package de.normalisiert.utils.graphs;

import gov.anl.gss.sdm.server.model.analysis.AnalysisManager;
import gov.anl.gss.sdm.server.model.analysis.loops.CyclesFile;
import gov.anl.gss.sdm.server.model.analysis.loops.GraphCyclesThread;
import gov.anl.gss.sdm.server.model.analysis.loops.LoopManager;
import gov.anl.gss.sdm.server.model.documentor.MessageProcessor;
import gov.anl.gss.sdm.server.model.documentor.SDMDocLanguageSpecific;
import gov.anl.gss.sdm.server.model.documentor.xmlwriter.NumberFormatter;
import gov.anl.gss.sdm.server.model.documentor.xmlwriter.StringConstants;
import gov.anl.gss.sdm.server.model.translator.InformationManagers;

import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Date;
import java.util.List;
import java.util.Vector;



/**
 * Searchs all elementary cycles in a given directed graph. The implementation
 * is independent from the concrete objects that represent the graphnodes, it
 * just needs an array of the objects representing the nodes the graph
 * and an adjacency-matrix of type boolean, representing the edges of the
 * graph. It then calculates based on the adjacency-matrix the elementary
 * cycles and returns a list, which contains lists itself with the objects of the 
 * concrete graphnodes-implementation. Each of these lists represents an
 * elementary cycle.<br><br>
 *
 * The implementation uses the algorithm of Donald B. Johnson for the search of
 * the elementary cycles. For a description of the algorithm see:<br>
 * Donald B. Johnson: Finding All the Elementary Circuits of a Directed Graph.
 * SIAM Journal on Computing. Volumne 4, Nr. 1 (1975), pp. 77-84.<br><br>
 *
 * The algorithm of Johnson is based on the search for strong connected
 * components in a graph. For a description of this part see:<br>
 * Robert Tarjan: Depth-first search and linear graph algorithms. In: SIAM
 * Journal on Computing. Volume 1, Nr. 2 (1972), pp. 146-160.<br>
 * 
 * @author Frank Meyer, web_at_normalisiert_dot_de
 * 
 * @author Eric Tatara - added cycle size limitation
 *
 */
public class ElementaryCyclesSearch {
	/** List of cycles */
	private List cycles = null;

	/** Adjacency-list of graph */
	private int[][] adjList = null;

	/** Graphnodes */
	private Object[] graphNodes = null;

	/** Blocked nodes, used by the algorithm of Johnson */
	private boolean[] blocked = null;

	/** B-Lists, used by the algorithm of Johnson */
	private Vector[] B = null;

	/** Stack for nodes, used by the algorithm of Johnson */
	private Vector stack = null;

	/** Cycle size limit restricts the size of cycles that are searched.  Longer
	 * lengths will find more cycles but search time can grow exponentially. */
	private static int DEFAULT_MAX_CYCLE_LENGTH = 30;
	private int maxCycleLength;
	
	private MessageProcessor mf;
	
	/**
	 * Number of cycles found;
	 */
	private long numCycles = 0;
	private int longest = 0;
	
	/**
	 * Determines if the actual cycle structure is stored, otherwise the algorithm
	 *   just counts the number of cycles.
	 */
	private boolean storeCycles = true; 
	private boolean writeCycles = false;
	private boolean processCycleWhenFound = false;
	private String cyclesFileName;
	private CyclesFile cyclesFile;
	private 
	GraphCyclesThread executingThread;
	
	int[] histo;
	
	private boolean withIVV;
	
	/**
	 * Constructor.
	 *
	 * @param matrix adjacency-matrix of the graph
	 * @param graphNodes array of the graphnodes of the graph; this is used to
	 * build sets of the elementary cycles containing the objects of the original
	 * graph-representation
	 */
	public ElementaryCyclesSearch(boolean[][] matrix, Object[] graphNodes, GraphCyclesThread executingThread) {
		this.graphNodes = graphNodes;
		this.adjList = AdjacencyList.getAdjacencyList(matrix);
		this.executingThread = executingThread;
		
		maxCycleLength = DEFAULT_MAX_CYCLE_LENGTH;
		histo = new int[maxCycleLength+1];
	}

	public ElementaryCyclesSearch(boolean[][] matrix, Object[] graphNodes, int maxCycleLength, GraphCyclesThread executingThread) {
		this.graphNodes = graphNodes;
		this.adjList = AdjacencyList.getAdjacencyList(matrix);
		this.maxCycleLength = maxCycleLength;
		this.executingThread = executingThread;
		histo = new int[maxCycleLength+1];
	}
	
	/**
	 * Returns List::List::Object with the Lists of nodes of all elementary
	 * cycles in the graph.
	 *
	 * @return List::List::Object with the Lists of the elementary cycles.
	 */
	public List getElementaryCycles(MessageProcessor mf) throws java.lang.OutOfMemoryError {
		this.mf = mf;
		this.cycles = new Vector();
		this.blocked = new boolean[this.adjList.length];
		this.B = new Vector[this.adjList.length];
		this.stack = new Vector();
		StrongConnectedComponents sccs = new StrongConnectedComponents(this.adjList);
		int s = 0;

		while (true) {
			if (executingThread != null && executingThread.isMyInterrupted("1"))
				return null;
			SCCResult sccResult = sccs.getAdjacencyList(s);
			if (sccResult != null && sccResult.getAdjList() != null) {
				Vector[] scc = sccResult.getAdjList();
				s = sccResult.getLowestNodeId();
				for (int j = 0; j < scc.length; j++) {
					if ((scc[j] != null) && (scc[j].size() > 0)) {
						this.blocked[j] = false;
						this.B[j] = new Vector();
					}
				}
				if (executingThread != null && executingThread.isMyInterrupted("2"))
					return null;
//				mf.addMessage(" findCycles(s, s, scc)\n");
				this.findCycles(s, s, scc);
				s++;
			} else {
				break;
			}
		}
		System.out.println(expand(histo));
		return this.cycles;
	}

	public long getNumCycles() {
		return numCycles;
	}
	
	public void processSingleCycles(int maxLen, boolean withIVV) {
		LoopManager lm = InformationManagers.getInstance().getAnalysisManager().getLoopManager();
		closeWrite();
		openForRead();
		Vector v = cyclesFile.read();
		while (v != null) {
			if (v.size() <= maxLen)
				lm.processCycleMinimumMemory(v, withIVV);
			v = cyclesFile.read();
		}
		
		closeRead();
	}

	/**
	 * Calculates the cycles containing a given node in a strongly connected
	 * component. The method calls itself recursivly.
	 *
	 * @param v
	 * @param s
	 * @param adjList adjacency-list with the subgraph of the strongly
	 * connected component s is part of.
	 * @return true, if cycle found; false otherwise
	 */
	private boolean findCycles(int v, int s, Vector[] adjList) {
		boolean f = false;
		if (executingThread != null && executingThread.isMyInterrupted("3"))
			return false;
		
		SimpleDateFormat timeStamp = new SimpleDateFormat("yyyy.MM.dd.   HH.mm.ss");
		SDMDocLanguageSpecific language = SDMDocLanguageSpecific.getInstance();
		AnalysisManager am = InformationManagers.getInstance().getAnalysisManager();
		LoopManager lm;
		if (am != null) {
			lm = InformationManagers.getInstance().getAnalysisManager().getLoopManager();
		} else {
			lm = InformationManagers.getInstance().getLoopManager();
			if (lm == null) {
			lm = new LoopManager();
			InformationManagers.getInstance().setLoopManager(lm);
			}
		}
		
		
		this.stack.add(new Integer(v));
		this.blocked[v] = true;

		if (stack.size() <= maxCycleLength){
			if (executingThread != null && executingThread.isMyInterrupted("4"))
				return false;
			for (int i = 0; i < adjList[v].size(); i++) {
				int w = ((Integer) adjList[v].get(i)).intValue();
				// found cycle
				if (w == s) {
					int l = this.stack.size();
					longest = Math.max(longest, l);
					histo[l]++;
					numCycles++;
					if ((numCycles % 500000) == 0) { // 1000000
						System.out.println("progress: "+numCycles+"  "+timeStamp.format(new Date()));
						System.out.println(expand(histo));
						if (mf != null) {
							mf.addMessage(language.getText(StringConstants.LOOP_SEARCH_PROGRESS)+". "+ 
									language.getText(StringConstants.FOUND)+": "+
									NumberFormatter.formatForCountNotConditional(numCycles)+" "+language.getText(StringConstants.LONGEST)+": "+longest);
						}
					}
					if (storeCycles || writeCycles || processCycleWhenFound){
						Vector cycle = new Vector();
						for (int j = 0; j < this.stack.size(); j++) {
							int index = ((Integer) this.stack.get(j)).intValue();
							cycle.add(this.graphNodes[index]);
//							System.out.println("<"+cycle+">");
						}
						if (storeCycles)
							this.cycles.add(cycle);
						if (writeCycles) {
							if (!cyclesFile.write(cycle)) {
								return false;
							}
						}
						if (processCycleWhenFound)
							lm.processCycleMinimumMemory(cycle, withIVV);
					} 
					f = true;
				} 
				else if (!this.blocked[w]) {
					if (this.findCycles(w, s, adjList)) {
						f = true;
					}
				}
			}
		}
		else{
			f = true;
		}

		if (f) {
			this.unblock(v);
		} else {
			for (int i = 0; i < adjList[v].size(); i++) {
				int w = ((Integer) adjList[v].get(i)).intValue();
				if (!this.B[w].contains(new Integer(v))) {
					this.B[w].add(new Integer(v));
				}
			}
		}

		this.stack.remove(new Integer(v));
		
		return f;
	}
	
	private String expand(int[] h) {
		StringBuilder sb = new StringBuilder();
		for (int i = 0; i < h.length; i++) {
			if (i>0)
				sb.append(",");
			sb.append(h[i]);
		}
		return sb.toString();
	}

	/**
	 * Unblocks recursively all blocked nodes, starting with a given node.
	 *
	 * @param node node to unblock
	 */
	private void unblock(int node) {
		this.blocked[node] = false;
		Vector Bnode = this.B[node];
		while (Bnode.size() > 0) {
			Integer w = (Integer) Bnode.get(0);
			Bnode.remove(0);
			if (this.blocked[w.intValue()]) {
				this.unblock(w.intValue());
			}
		}
	}

	public int getMaxCycleLength() {
		return maxCycleLength;
	}

	public void setMaxCycleLength(int maxCycleLength) {
		this.maxCycleLength = maxCycleLength;
	}
	
	public Vector loadSingleCycleFromFile() {
		
		Vector aCycle = cyclesFile.readSingle();
		
		return aCycle;
		
	}
	
	public List<List<String>> loadCyclesFromFile() {
		openForRead();
		List<List<String>> allCycles = new ArrayList<List<String>>();
		List<String> aCycle = cyclesFile.read();
		while(aCycle != null) {
			allCycles.add(aCycle);
			aCycle = cyclesFile.read();
		}
		return allCycles;
		
	}

	public void setStoreCycles(boolean storeCycles) {
		this.storeCycles = storeCycles;
	}

	public boolean isWriteCycles() {
		return writeCycles;
	}

	public void setWriteCycles(boolean writeCycles) {
		this.writeCycles = writeCycles;
	}

	public String getCyclesFileName() {
		return cyclesFileName;
	}

	public void setCyclesFileName(String cyclesFileName) {
		this.cyclesFileName = cyclesFileName;
//		cyclesFile = new CyclesFile(cyclesFileName);
	}

	public CyclesFile getCyclesFile() {
		return cyclesFile;
	}

	public void setCyclesFile(CyclesFile cyclesFile) {
		this.cyclesFile = cyclesFile;
	}
	
	public boolean openForWrite() {
		return cyclesFile.openForWrite();
	}
	
	public boolean openForRead() {
		return cyclesFile.openForRead();
	}
	
	public boolean closeWrite() {
		return cyclesFile.closeWrite();
	}
	
	public boolean closeRead() {
		return cyclesFile.closeRead();
	}

	public String getHisto() {
		return expand(histo);
	}

	public boolean isProcessCycleWhenFound() {
		return processCycleWhenFound;
	}

	public void setProcessCycleWhenFound(boolean processCycleWhenFound) {
		this.processCycleWhenFound = processCycleWhenFound;
	}

	public boolean isWithIVV() {
		return withIVV;
	}

	public void setWithIVV(boolean withIVV) {
		this.withIVV = withIVV;
	}
}