package gov.anl.gss.sdm.server.model.analysis.graphtheory;

import gov.anl.gss.sdm.server.model.analysis.polarity.PolarityResults;
import gov.anl.gss.sdm.server.model.structure.SDObjectManager;
import gov.anl.gss.sdm.server.model.support.Utilities;
import gov.anl.gss.sdm.server.model.translator.InformationManagers;

import java.io.BufferedWriter;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Collections;
import java.util.LinkedHashMap;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.apache.log4j.Logger;

public class GraphTheoryMatrix {
static Logger log = Logger.getLogger(GraphTheoryMatrix.class.getSimpleName());
		
		protected Map<String, Set<String>> successor;
		protected Map<String, Set<String>> predecessor;
		protected SDObjectManager objManager;
		protected PolarityResults polarityResults;
		
		protected Map<String, Set<String>> tmpsuccessor;
		
		public GraphTheoryMatrix(SDObjectManager objManager, PolarityResults polarityResults) {
			this.objManager = objManager;
			this.polarityResults = polarityResults;
			
			successor = new LinkedHashMap<String, Set<String>>();
			predecessor = new LinkedHashMap<String, Set<String>>();
			for (String var : InformationManagers.getInstance().getEquationManager().getVariablesNoSystem()) {
				successor.put(var, new LinkedHashSet<String>());
				predecessor.put(var, new LinkedHashSet<String>());
			}
		}
		
		public GraphTheoryMatrix(GraphTheoryMatrix gtm) {
			this.objManager = gtm.getObjManager();
			this.polarityResults = gtm.getPolarityResults();
			for (String row : gtm.keySet()) {
				successor.put(row, new LinkedHashSet<String>());
				predecessor.put(row, new LinkedHashSet<String>());
			}
			for (String row : gtm.keySet()) {
				Set<String> rowSet = gtm.getRowSet(row);
				for (String col : rowSet) {
					addSuccessor(row,col);
				}
			}
		}
		
		public GraphTheoryMatrix() {
			this.objManager = null;
			this.polarityResults = null;
			successor = new LinkedHashMap<String, Set<String>>();
			predecessor = new LinkedHashMap<String, Set<String>>();
		}
		

		
		public  Map<String, Set<String>> copyOfSuccessor() {
			 Map<String, Set<String>> tmpsuccessor = new LinkedHashMap<String, Set<String>>();
			 for (String key : successor.keySet()) {
				 tmpsuccessor.put(key, new LinkedHashSet<String>());
				 tmpsuccessor.get(key).addAll(successor.get(key));
			 }
			 return tmpsuccessor;
		}
		
		public Set<String> getPredecessorCopy(String s) {
			Set<String> set = predecessor.get(s);
			if (set == null)
				return new LinkedHashSet<String>();
			return new LinkedHashSet<String>(predecessor.get(s));
		}
		
		public Set<String> getSuccessorCopy(String s) {
			Set<String> set = successor.get(s);
			if (set == null)
				return new LinkedHashSet<String>();
			return new LinkedHashSet<String>(set);
		}
		public Set<String> getPredecessor(String s) {
			return predecessor.get(s);
		}
		
		public Set<String> getSuccessor(String s) {
			return successor.get(s);
		}
		
		public void initializeWith(GraphTheoryMatrix gtm) {
			for (String row : gtm.keySet()) {
				addSuccessor(row,row);
				Set<String> rowSet = gtm.getRowSet(row);
				for (String col : rowSet) {
					addSuccessor(row,col);
				}
			}
		}
		
		public void remove(String key) {
			successor.remove(key);
			predecessor.remove(key);
			
			for (Set<String> set : successor.values()) {
				set.remove(key);
			}
			
			for (Set<String> set : predecessor.values()) {
				set.remove(key);
			}
		}
		
		public void merge(String key, Set<String> set1, Set<String> set2) {
			set2.remove(key);
			for (String s : set2) {
				addSuccessor(key, s);
			}
		}
		
		public void add(GraphTheoryMatrix otherMatrix) {
			for (String otherKey : otherMatrix.keySet()) {
				Set<String> otherRowForKey = otherMatrix.getColSet(otherKey);
				for (String otherRowVal : otherRowForKey) {
					addSuccessor(otherKey, otherRowVal);
				}
			}
		}
		
		public void addSuccessor(String rowValue, String columnValue) {
			// rowset is sucessor
			// colset is predecessor
			
			Set<String> rowSet = getRowSet(rowValue);
			Set<String> colSet = getColSet(columnValue);
			boolean added = false;
			if (!rowSet.contains(columnValue)) {
				rowSet.add(columnValue);
				colSet.add(rowValue);
				added = true;
			}

		}
		
		public Set<String> getRowSet(String rowValue) {
			if (!successor.containsKey(rowValue))
				successor.put(rowValue, new LinkedHashSet<String>());
			return successor.get(rowValue);
		}
		
		public Set<String> getColSet(String columnValue) {
			if (!predecessor.containsKey(columnValue))
				predecessor.put(columnValue, new LinkedHashSet<String>());
			return predecessor.get(columnValue);
		}

		public Map<String, Set<String>> getSuccessor() {
			return successor;
		}

		public SDObjectManager getObjManager() {
			return objManager;
		}

		public PolarityResults getPolarityResults() {
			return polarityResults;
		}
		
		public Set<String> keySet() {
			return successor.keySet();
		}
		
		public Set<String> tmpKeySet() {
			return tmpsuccessor.keySet();
		}
		
		public Set<String> getSetForKey(String key) {
			return successor.get(key);
		}
		
		public Set<String> getTmpSetForKey(String key) {
			return tmpsuccessor.get(key);
		}
		
		public void printMatrix() {
			log.debug("### Matrix ###");
			List<String> al = new ArrayList<String>(successor.keySet());
			Collections.sort(al);
			for (String key : al) {
				log.debug("<"+key+"> -> "+successor.get(key).size());
//				List<String> al2 = new ArrayList<String>(matrix.get(key));
//				Collections.sort(al2);
//				log.debug("<"+key+"> ->");
//				for (String v : al2) {
//					log.debug(" <"+v+">");
//				}
//				log.debug("\n");
			}
//			log.debug("\n");
		}
		

		
		public void dumpMatrixToFile(String file, Map<String, Set<String>> matrix) {
			
			List<String> sorted = ordered(matrix.keySet());
			
			try {
				BufferedWriter bw = Utilities.openFileForWriting(
						InformationManagers.getInstance().getModelManager().getGeneratedOutputDirectoryName(), file);
				for (String var1 : sorted /*ordered(matrix.keySet()) */ ) {
					bw.append("\t");
					bw.append(var1);
				}
				bw.append("\n");
				
				for (String var1 : sorted /*ordered(matrix.keySet()) */) {
					bw.append(var1);
					Set<String> connected = matrix.get(var1);
					for (String var2 : sorted /*ordered(matrix.keySet()) */) {
						if (connected.contains(var2))
							bw.append("\t"+"1");
						else
						bw.append("\t");
					}
					bw.append("\n");
				}
				
				bw.close();
			} catch (IOException e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			}
		}
		
		private List<String> ordered(Set<String> aSet) {
			List<String> aList = new ArrayList<String>(aSet);
			Collections.sort(aList);
			return aList;
		}
		
		public Integer getRowCounts(String rowName) {
			return predecessor.get(rowName).size();
		}
		
		public Integer getColumnCounts(String columnName) {
			return successor.get(columnName).size();
		}
		
		public String getRowValue(String rowName) {
			StringBuilder sb = new StringBuilder();
			for (String s : ordered(successor.get(rowName))) {
				sb.append("|"+s);
			}
			
			return sb.toString();
			
		}
		
		public String getColValue(String columnName) {
			StringBuilder sb = new StringBuilder();
			for (String s : ordered(predecessor.get(columnName))) {
				sb.append("|"+s);
			}
			
			return sb.toString();
			
		}

		public void setSuccessor(Map<String, Set<String>> successor) {
			this.successor = successor;
		}

		public void setPredecessor(Map<String, Set<String>> predecessor) {
			this.predecessor = predecessor;
		}

		public Map<String, Set<String>> getPredecessor() {
			return predecessor;
		}
}
