package gov.anl.gss.sdm.server.model.analysis.micmac;

import gov.anl.gss.sdm.server.model.documentor.xmlwriter.NumberFormatter;
import gov.anl.gss.sdm.server.model.support.Utilities;
import gov.anl.gss.sdm.server.model.translator.InformationManagers;
import gov.anl.gss.sdm.server.model.translator.Parser;

import java.io.BufferedWriter;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;
import java.util.Set;

import org.apache.commons.math3.exception.OutOfRangeException;
import org.apache.commons.math3.linear.BlockRealMatrix;
import org.jfree.data.xy.XYSeries;
import org.jfree.data.xy.XYSeriesCollection;

public class MicMac {
	
	public static final int NTH_ORDER_ITERATIONS = 5;
	public static final int NTH_ORDER_RESULTS = NTH_ORDER_ITERATIONS + 1;
	
	public static final String[] tags = {"1st order", "2nd order", "4th order", "8th order", "16th order", "32nd order",
		"64th order", "128th order", "256th order", "512th order", "1024th order", "2048th order"};
	
	BlockRealMatrix matrix;
	int numEquations;
	BlockRealMatrix[] matrixArray = new BlockRealMatrix[NTH_ORDER_RESULTS];
	double[][] influence;
	double[][] exposure;
	double[][] influenceNormalized;
	double[][] exposureNormalized;
	
	public MicMac(int numEquations) {
		
		matrix = new BlockRealMatrix(numEquations, numEquations);
		influence = new double[numEquations][NTH_ORDER_RESULTS];
		exposure = new double[numEquations][NTH_ORDER_RESULTS];
		influenceNormalized = new double[numEquations][NTH_ORDER_RESULTS];
		exposureNormalized = new double[numEquations][NTH_ORDER_RESULTS];
		this.numEquations = numEquations;
	}
	
	public void run() {
		initialize();
		createFirstOrder();
		multiply();
		tabulateExposure();
		tabulateInfluence();
//		printExposureInfluence(influence, exposure, "Tabulated");
		normalizeInfluence();
		nomalizeExposure();
//		printExposureInfluence(influenceNormalized, exposureNormalized, "Normalized");
	}
	
	public XYSeriesCollection getScatterData() {
		XYSeriesCollection d = new XYSeriesCollection();
		for (int i = 0; i < numEquations; i++) {
			if (Parser.isSystemVariable(varName(i)) || Parser.isNaReplacement(varName(i)))
				continue;
			XYSeries series = new XYSeries(varName(i));
			series.add(influenceNormalized[i][influenceNormalized[i].length-1], exposureNormalized[i][exposureNormalized[i].length-1]);
			d.addSeries(series);
		}
	
		return d;

	}
	
	private void printExposureInfluence(double[][] influence, double[][] exposure, String tag) {
		System.out.println("\n##### "+tag+" #####\n");
		for (int row = 0; row < numEquations; row++) {
			
			if (Parser.isSystemVariable(varName(row)) || Parser.isNaReplacement(varName(row)))
					continue;
			
			StringBuilder sb = new StringBuilder();
			sb.append(row);
//			sb.append(","+varName(row));
			sb.append(varName(row));
			for (int order = 0; order < NTH_ORDER_ITERATIONS+1; order++) {
				sb.append(","+NumberFormatter.formatForRatio(influence[row][order]));
				sb.append(","+NumberFormatter.formatForRatio(exposure[row][order]));
			}
			System.out.println(sb.toString());
		}
	}
	
	private void tabulateExposure() {

		for (int order = 0; order < NTH_ORDER_RESULTS; order++) {
			double[][] data = matrixArray[order].getData();
			for (int col = 0; col < numEquations; col++) {
				for (int row = 0; row < numEquations; row++) {
					exposure[col][order] += data[row][col];
				}
			}
		}

	}
	
	private void tabulateInfluence() {
	
		for (int order = 0; order < NTH_ORDER_RESULTS; order++) {
			double[][] data = matrixArray[order].getData();
			for (int row = 0; row < numEquations; row++) {
				for (int col = 0; col < numEquations; col++) {
					influence[row][order] += data[row][col];
				}
			}
		}
	}
	
	private void normalizeInfluence() {
		for (int order = 0; order < NTH_ORDER_RESULTS; order++) {
			double max = Double.NEGATIVE_INFINITY;
			for (int row = 0; row < numEquations; row++) {
					max = Math.max(max, influence[row][order]);
			}
			for (int row = 0; row < numEquations; row++) {
				influenceNormalized[row][order] = max == 0.0 ? 0.0 : influence[row][order] / max;
			}
		}
	}
	
	private void nomalizeExposure() {
		for (int order = 0; order < NTH_ORDER_RESULTS; order++) {
			double max = Double.NEGATIVE_INFINITY;
			for (int row = 0; row < numEquations; row++) {
					max = Math.max(max, exposure[row][order]);
			}
			for (int row = 0; row < numEquations; row++) {
				exposureNormalized[row][order] = max == 0.0 ? 0.0 : exposure[row][order] / max;
			}
		}

	}
	
	private void printMatrix(BlockRealMatrix matrix, String tag) {
		System.out.println("\n######### "+ tag +" ##########\n");
		double[][] data = matrix.getData();
		for (int i = 0; i < data.length; i++) {
			StringBuilder sb = new StringBuilder();
			sb.append(i);
			sb.append(","+varName(i));
			for (int j = 0; j < data[i].length; j++) {
				sb.append(", "+NumberFormatter.formatForRatio(data[i][j]));
			}
			System.out.println(sb.toString());
			
		}
	}
	
	private void multiply() {
		
		matrixArray[0] = matrix;
//		printMatrix(matrixArray[0],tags[0]);
		for (int i = 0; i < NTH_ORDER_ITERATIONS; i++) {
			matrixArray[i+1] = matrixArray[i].multiply(matrixArray[i]);
//			printMatrix(matrixArray[i+1], tags[i+1]);
		}
	}
	
	private void initialize() {
		double[] zeros = new double[numEquations];
		Arrays.fill(zeros, 0.0);
		for (int row = 0; row < numEquations; row++) {
			matrix.setRow(row, zeros);
		}
		for (int i = 0; i < numEquations; i++) {
			Arrays.fill(influence[i], 0.0);
			Arrays.fill(exposure[i], 0.0);
		}
	}

	private void createFirstOrder() {
		
		String micmacTxt = InformationManagers.getInstance().getFileOrganizer().getMicmacTextFile();
		BufferedWriter bw = Utilities.openFileForWriting(micmacTxt);
		StringBuilder sb = new StringBuilder();

		try {
			for (String variable : InformationManagers.getInstance().getEquationManager().getVariables()) {
				sb.setLength(0);
				if (variable.startsWith("C_L_O_U_D"))
					continue;
					if (variable.startsWith("NoNamefor"))
					continue;
					if (Parser.isSystemVariable(variable))
						continue;
					if (Parser.isNaReplacement(variable))
						continue;
				int row = varID(variable);
				if (row < 0)
					continue;
				
				if (InformationManagers.getInstance().getEquationManager().getAnEquationFor(variable) != null) {
					List<String> usedBy = InformationManagers.getInstance().getEquationManager().getAnEquationFor(variable).getUsedBy();
					sb.append(Parser.removeQuotes(variable)); // variable.replace("\"", "")+";");
					if (usedBy != null) {
						for (String ub : usedBy) {
							if (ub.startsWith("C_L_O_U_D") || ub.startsWith("NoNamefor"))
								continue;
							int col = varID(ub);
							if (col < 0)
								continue;
							matrix.setEntry(row, col, 1.0);
							sb.append(Parser.removeQuotes(ub)); // ub.replace("\"", "")+";");
							
						}
					} else {
						System.out.println("No UsedBy for "+variable);
					}
					List<String> uses = new ArrayList<String>();
					Set<String> u = InformationManagers.getInstance().getAnalysisManager().getEquationManager().getUses().get(variable);
					if (u != null) {
							uses.addAll(InformationManagers.getInstance().getAnalysisManager().getEquationManager().getUses().get(variable));
					
						for (String ub : uses) {
							if (ub.startsWith("C_L_O_U_D") || ub.startsWith("NoNamefor"))
								continue;
							int col = varID(ub);

							if (col < 0)
								continue;
							matrix.setEntry(col, row, 1.0);
							
						} 
					} else {
						System.out.println("###No Uses for "+variable);
					}
				} else {
					System.out.println("No Equation for "+variable);
				}
				try {
					if (sb.toString().split(";").length > 1)
						bw.append(sb.toString()+"\n");
					
				} catch (IOException e) {
					// TODO Auto-generated catch block
					e.printStackTrace();
				}
			}
		} catch (OutOfRangeException e1) {
			// TODO Auto-generated catch block
			e1.printStackTrace();
		}
		try {
			bw.close();
		} catch (IOException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
	}
	
	private int asInt(String s) {
		return Integer.parseInt(s);
	}
	
	private int varID(String s) {
		return asInt(InformationManagers.getInstance().getEquationManager().getVariableID(s));
	}
	
	private String varName(int id) {
		return InformationManagers.getInstance().getEquationManager().getVariableNameForID(id);
	}

	public BlockRealMatrix getMatrix() {
		return matrix;
	}

	public int getNumEquations() {
		return numEquations;
	}

	public BlockRealMatrix[] getMatrixArray() {
		return matrixArray;
	}

	public double[][] getInfluence() {
		return influence;
	}

	public double[][] getExposure() {
		return exposure;
	}

	public double[][] getInfluenceNormalized() {
		return influenceNormalized;
	}

	public double[][] getExposureNormalized() {
		return exposureNormalized;
	}
}