package gov.anl.gss.sdm.server.model.analysis.micmac;

/* ---------------------
 * ScatterPlotDemo1.java
 * ---------------------
 * (C) Copyright 2002-2006, by Object Refinery Limited.
 *
 */

import gov.anl.gss.visualization.BoundingBox;
import gov.anl.gss.visualization.SFEXYPointerAnnotationMetaData;

import java.awt.BasicStroke;
import java.awt.Color;
import java.awt.Font;
import java.awt.Shape;
import java.awt.geom.Ellipse2D;
import java.text.DecimalFormat;
import java.util.ArrayList;
import java.util.Collections;
import java.util.List;

import org.jfree.chart.ChartFactory;
import org.jfree.chart.JFreeChart;
import org.jfree.chart.annotations.XYPointerAnnotation;
import org.jfree.chart.axis.NumberAxis;
import org.jfree.chart.axis.NumberTickUnit;
import org.jfree.chart.plot.PlotOrientation;
import org.jfree.chart.plot.XYPlot;
import org.jfree.chart.renderer.xy.XYLineAndShapeRenderer;
import org.jfree.data.xy.XYSeries;
import org.jfree.data.xy.XYSeriesCollection;
import org.jfree.ui.TextAnchor;

/**
 * A demo scatter plot.
 */
public class MicMacScatterPlot {
	
	public static final int PLOT_SIZE_BASLINE = 1200;
	public static final double PLOT_SIZE_SCALE_FACTOR = (double) MicMacPlot.PLOT_SIZE / (double) PLOT_SIZE_BASLINE;
	public static final int PLOT_DISTANCE_BASELINE = 900; 
	public  double DISTANCE_ADJUSTMENT = PLOT_DISTANCE_BASELINE;

	public static int FONT_SIZE = 11;
	
	
	int nonSqueezeCounts = 0;
	int squeezeCounts = 0;
	int failureCounts = 0;

	private static final Shape circle = new Ellipse2D.Double(-6.0, -6.0, 12.0, 12.0);

	private static final BasicStroke solidThick = new BasicStroke(3.0f);


	private static final BasicStroke dashThick = new BasicStroke(3.0f, 
			BasicStroke.CAP_SQUARE, 
			BasicStroke.JOIN_ROUND, 
			0.0f, new float[] {4.0f},  0.0f);

	private static final BasicStroke dashVeryThick = new BasicStroke(5.0f, 
			BasicStroke.CAP_SQUARE, 
			BasicStroke.JOIN_ROUND, 
			0.0f, new float[] {8.0f},  0.0f);

	private JFreeChart chart;

	private String title;
	private MicMacMinimumMemory micmac;
	private double minX;
	private double maxX;
	private double minY;
	private double maxY;
	private double tick;
	private double plotXMax;
	private double plotYMax;
	private double range;

	/**
	 * A demonstration application showing a scatter plot.
	 *
	 * @param title  the frame title.
	 */
	public MicMacScatterPlot(String title, MicMacMinimumMemory micmac, double minX, double maxX,
			double minY, double maxY) {
		this.title = title;
		this.micmac = micmac;
		this.minX = minX;
		this.maxX = maxX;
		this.minY = minY;
		this.maxY = maxY;
		range = maxX - minX;
		tick = range/5;
		plotXMax = maxX + tick;
		plotYMax = maxY + tick;
		DISTANCE_ADJUSTMENT = (PLOT_DISTANCE_BASELINE*PLOT_SIZE_SCALE_FACTOR)/range;
		
		System.out.println("\n$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$\n");
		System.out.println(title);
		System.out.println("X Range: "+minX+" -> "+maxX);
		System.out.println("Y Range: "+minY+" -> "+maxY);
		
		
		chart = createChart(micmac.getScatterData(minX, maxX, minY, maxY));
	}



	private JFreeChart createChart(XYSeriesCollection dataset) {
		JFreeChart chart = ChartFactory.createScatterPlot(title,
				"Influence", "Exposure", dataset, PlotOrientation.VERTICAL, false, false, false);

		XYPlot plot = (XYPlot) chart.getPlot();
		plot.setNoDataMessage("NO DATA");
		plot.setDomainZeroBaselineVisible(false);
		plot.setRangeZeroBaselineVisible(false);

		XYLineAndShapeRenderer renderer
		= (XYLineAndShapeRenderer) plot.getRenderer();


		for (int i = 0; i < dataset.getSeriesCount(); i++) {
			renderer.setSeriesOutlinePaint(i, Color.black);
			renderer.setSeriesStroke(i, solidThick);
			renderer.setSeriesShape(i, circle);
			renderer.setUseOutlinePaint(true);
			renderer.setSeriesItemLabelsVisible(i, true);
		}



		NumberAxis domainAxis = (NumberAxis) plot.getDomainAxis();
		domainAxis.setAutoRangeIncludesZero(true);
		domainAxis.setTickUnit(new NumberTickUnit(tick));
		domainAxis.setTickMarksVisible(true);
		domainAxis.setLowerBound(minX);
		domainAxis.setUpperBound(plotXMax);

		NumberAxis rangeAxis = (NumberAxis) plot.getRangeAxis();
		rangeAxis.setAutoRangeIncludesZero(true);
		rangeAxis.setTickUnit(new NumberTickUnit(tick));
		rangeAxis.setTickMarksVisible(true);


		if (tick < MicMacPlot.FORMAT_CHANGE) {
			rangeAxis.setNumberFormatOverride(new DecimalFormat("0.0E0"));
			domainAxis.setNumberFormatOverride(new DecimalFormat("0.0E0"));
		}

		for (int ds = 0; ds < dataset.getSeriesCount(); ds++) {

			renderer.setSeriesOutlinePaint(ds, Color.blue);
			renderer.setSeriesPaint(ds, Color.blue);
			renderer.setSeriesStroke(ds, dashThick);
			renderer.setSeriesItemLabelsVisible(ds, false);
			renderer.setSeriesLinesVisible(ds, false);
			renderer.setSeriesShapesVisible(ds, true);
			renderer.setSeriesVisibleInLegend(ds, false);
		}


		rangeAxis.setLowerBound(minY);
		rangeAxis.setUpperBound(plotYMax);

		annotateDataPoints(plot, dataset);

		// add line around data area

		// public XYSeries(Comparable key, boolean autoSort, boolean allowDuplicateXValues)
		XYSeries series = new XYSeries("DataArea", false, true);

		series.add(minX, minY);
		series.add(minX, maxY);
		series.add(maxX, maxY);
		series.add(maxX, minY);
		series.add(minX, minY);
		dataset.addSeries(series);

		int ds = dataset.getSeriesCount()-1;
		renderer.setSeriesOutlinePaint(ds, Color.black);
		renderer.setSeriesPaint(ds, Color.black);
		renderer.setSeriesStroke(ds, dashVeryThick);
		renderer.setSeriesItemLabelsVisible(ds, false);
		renderer.setSeriesLinesVisible(ds, true);
		renderer.setSeriesShapesVisible(ds, false);
		renderer.setSeriesVisibleInLegend(ds, false);

//		for (BoundingBox bb : micmac.getMeta().getUsedAnnotionSpace()) {
//
//			dataset.addSeries(bb.getXYSeries());
//			ds = dataset.getSeriesCount()-1;
//			renderer.setSeriesOutlinePaint(ds, Color.red);
//			renderer.setSeriesPaint(ds, Color.red);
//			renderer.setSeriesStroke(ds, solidThick);
//			renderer.setSeriesItemLabelsVisible(ds, false);
//			renderer.setSeriesLinesVisible(ds, true);
//			renderer.setSeriesShapesVisible(ds, false);
//			renderer.setSeriesVisibleInLegend(ds, false);
//		}

		return chart;
	}
	
	public boolean hasAnnotatedDataPoints() {
		return nonSqueezeCounts + squeezeCounts > 0;
	}

	private void annotateDataPoints(XYPlot plot, XYSeriesCollection dataset) {



		SFEXYPointerAnnotationMetaData meta = new SFEXYPointerAnnotationMetaData(dataset.getSeriesCount(), minX, maxX, minY, maxY, DISTANCE_ADJUSTMENT);
		micmac.setMeta(meta);

		int ptr = 0;

		for (int i = 0; i < dataset.getSeriesCount(); i++) {
			XYSeries s = dataset.getSeries(i);
			double x = s.getX(0).doubleValue();
			double y = s.getY(0).doubleValue();

			meta.addAnnotation((String) s.getKey(), x, y);

			ptr++;
		}

		meta.analyzeLayout();

		ptr = 0;

		List<DataPointProximitySummary> dppsList = new ArrayList<DataPointProximitySummary>();
		dppsList.addAll(meta.getProximitySummaryList().values());
		Collections.sort(dppsList);
		
		int[] hitCounts = new int[dppsList.size()];

//		for (DataPointProximitySummary dpps : dppsList) {
//			ptr = dpps.getPtr();
		for (DataPoint dp : meta.getOrderedPoints()) {
			ptr = dp.getPtr();
			//			for (int i = 0; i < dataset.getSeriesCount(); i++) {
			int i = ptr;
//			System.out.println("############### ptr: "+ptr+" ###############");
			XYSeries s = dataset.getSeries(i);

			String key = (String) s.getKey();
			Number x = s.getX(0);
			Number y = s.getY(0);
			
//			String tKey =  key +" <"+ptr+"><"+loc.getSlot()+">";
			String tKey = key;

			Annotation annot = null;

			List<CandidateLocation> locs = meta.getLocation(ptr);
			
			boolean foundLocation = false;
			
			int slotUsed = 0;
			
			if (locs == null) {
				//					System.out.println("NO LOCATION "+i);
			} else {
				StringBuilder prefs = new StringBuilder("ptr:  "+ptr);
				for (CandidateLocation cl : locs) {
					prefs.append("  "+cl.getSlot());
				}
//				System.out.println(prefs.toString());
				int loopCount = 0;
				for (CandidateLocation loc : locs) {
					
//					System.out.println("ptr: "+ptr+" trying slot: "+loc.getSlot()+" attempt: "+(loopCount));

					annot = new Annotation(ptr, tKey, 
							x.doubleValue(), y.doubleValue(),  
							loc.getAngle(),
							meta.getSlotAnchor(loc.getSlot()), 
							loc.getDistance(),loc.getSlot());
					

					if (meta.isAvailable(annot, true)) {
						hitCounts[loopCount]++;
						nonSqueezeCounts++;
						slotUsed = loc.getSlot();
						plot.addAnnotation(createXYPointerAnnotation(annot));
						foundLocation = true;
						break;
					} else {
						loopCount++;
//						System.out.println("NOT AVAILABLE");
//						meta.squeezeInAnnotation(loc, annot, true);
					}
					//						


				}
				if (annot != null && foundLocation) {
//					System.out.println("!!!!!!!!!!! Location Found Among !!!!!!!!!!!!!!!! slotUsed: "+slotUsed);
				} else {
					// one last try
					CandidateLocation loc = locs.get(0);
					annot = new Annotation(ptr, tKey  , 
							x.doubleValue(), y.doubleValue(),  
							loc.getAngle(),
							meta.getSlotAnchor(loc.getSlot()), 
							loc.getDistance(),loc.getSlot());
					if (meta.squeezeInAnnotation(annot, false /* true*/)) {
						squeezeCounts++;
//						System.out.println("@@@@@@@@@@@@ Able to squeeze it in @@@@@@@@@@@@ slotUsed: "+loc.getSlot());
						plot.addAnnotation(createXYPointerAnnotation(annot));
					} else {
						failureCounts++;
//						System.out.println("$$$$$$$$$$$$$$ could not find location $$$$$$$$$$$$$$");
					}
				}
			}
			ptr++;
		}
		
		System.out.println(title);
		System.out.println("X Range: "+minX+" -> "+maxX);
		System.out.println("Y Range: "+minY+" -> "+maxY);
		System.out.println("Num Annotations: "+meta.getNumAnnotations());
//		StringBuilder sb = new StringBuilder("Hit Counts by Ordered Position: ");
//		for (int h = 0 ; h < hitCounts.length; h++) {
//			sb.append("  "+hitCounts[h]);
//		}
//		System.out.println(sb.toString());
		System.out.println("Non Squeeze Count: "+nonSqueezeCounts);
		System.out.println("Squeeze Count: "+squeezeCounts);
		System.out.println("Failure Count: "+failureCounts);
	}

	private void printIntersectors(SFEXYPointerAnnotationMetaData meta, int ptr) {
		for (String key : meta.getIntersectors().keySet()) {
			if (key.startsWith(ptr+"#")) {
				System.out.println("   "+key+" "+meta.getIntersectors().get(key));
			}
		}
	}

	private XYPointerAnnotation createXYPointerAnnotation(Annotation annotation) {
		return createXYPointerAnnotation(annotation.getText(),
				annotation.getX(), annotation.getY(), annotation.getAngle(), annotation.getAnchor(), annotation.getDistanceFromXY());
	}

	private XYPointerAnnotation createXYPointerAnnotation(String text, double x, double y, double angle, TextAnchor anchor,
			double distanceFromXY) {
		XYPointerAnnotation annotation;
		Font font = new Font("Arial", Font.PLAIN, FONT_SIZE);
		annotation = new XYPointerAnnotation(text, x, y, angle);
		annotation.setFont(font);
		annotation.setTextAnchor(anchor);

		annotation.setTipRadius(10.0);
		annotation.setBaseRadius(distanceFromXY);
		annotation.setPaint(Color.black);
		annotation.setArrowLength(0.0); 

		annotation.setLabelOffset(5.0);

		return annotation;
	}


	private XYPointerAnnotation createAnnotation(String text, double x, double y, BoundingBox bb, TextAnchor anchor,
			double distanceFromXY) {
		XYPointerAnnotation annotation;
		Font font = new Font("Arial", Font.PLAIN, FONT_SIZE);
		annotation = new XYPointerAnnotation(text, x, y, computeAngle(x, y, bb, anchor));
//		System.out.println("### "+text+" x: "+x+" y: "+y+" bb: "+bb+" angle: "+computeAngle(x, y, bb, anchor));
		annotation.setFont(font);
		annotation.setTextAnchor(anchor);

		annotation.setTipRadius(10.0);
		annotation.setBaseRadius(distanceFromXY);
		annotation.setPaint(Color.black);
		annotation.setArrowLength(0.0); 

		annotation.setLabelOffset(5.0);

		return annotation;
	}

	private double computeAngle(double x, double y, BoundingBox bb, TextAnchor anchor) {


		if (anchor == TextAnchor.BOTTOM_LEFT) {
			return -Math.atan2(bb.getLowerLeft()[1]-y, bb.getLowerLeft()[0]-x);
			//	    		return 1.75 * Math.PI; 
		}
		if (anchor == TextAnchor.BOTTOM_CENTER) {
			return -Math.atan2(bb.getLowerLeft()[1]-y, (bb.getLowerLeft()[0]+bb.width()/2.0)-x);
			//	    		return 1.5 * Math.PI; 
		} 
		if (anchor == TextAnchor.BOTTOM_RIGHT) {
			return -Math.atan2(bb.getLowerRight()[1]-y, bb.getLowerRight()[0]-x);
			//	    		return 1.25 * Math.PI; 
		}
		if (anchor == TextAnchor.CENTER_LEFT) {
			return -Math.atan2((bb.getLowerLeft()[1]+bb.height()/2.0)-y, bb.getLowerLeft()[0]-x);
			//	    		return 0.0 * Math.PI; 
		}
		if (anchor == TextAnchor.CENTER) {
			return -Math.atan2((bb.getLowerLeft()[1]+bb.height()/2.0)-y, (bb.getLowerLeft()[0]+bb.width()/2.0)-x);
			//	    		return 1.75 * Math.PI;  
		}
		if (anchor == TextAnchor.CENTER_RIGHT) {
			return -Math.atan2((bb.getLowerLeft()[1]+bb.height()/2.0)-y, bb.getLowerRight()[0]-x);
			//	    		return Math.PI; 
		}
		if (anchor == TextAnchor.TOP_LEFT) {
			return -Math.atan2(bb.getUpperLeft()[1]-y, bb.getLowerLeft()[0]-x);
			//	    		return 0.25 * Math.PI; 
		}
		if (anchor == TextAnchor.TOP_CENTER) {
			return -Math.atan2(bb.getUpperLeft()[1]-y, bb.getLowerLeft()[0]+bb.width()/2.0-x);
			//	    		return 0.50 * Math.PI; 
		} 
		if (anchor == TextAnchor.TOP_RIGHT) {
			return -Math.atan2(bb.getUpperLeft()[1]-y, bb.getUpperRight()[0]-x);
			//	    		return 0.75 * Math.PI; 
		}
		return 0.0;
	}

	private double computeAngle(double x, double y, double bbx, double bby /*, TextAnchor anchor*/) {

		return -Math.atan2(bby-y, bbx-x); 

		//	    	if (anchor == TextAnchor.BOTTOM_LEFT) {
		//	    		return -Math.atan2(bby-y, bbx-x); 
		//	    	}
		//	    	if (anchor == TextAnchor.BOTTOM_CENTER) {
		//	    		return -Math.atan2(bb.getLowerLeft()[1]-y, (bb.getLowerLeft()[0]+bb.width()/2.0)-x);
		//	    	} 
		//	    	if (anchor == TextAnchor.BOTTOM_RIGHT) {
		//	    		return -Math.atan2(bb.getLowerRight()[1]-y, bb.getLowerRight()[0]-x);
		//	    	}
		//	    	if (anchor == TextAnchor.CENTER_LEFT) {
		//	    		return -Math.atan2((bb.getLowerLeft()[1]+bb.height()/2.0)-y, bb.getLowerLeft()[0]-x);
		//	    	}
		//	    	if (anchor == TextAnchor.CENTER) {
		//	    		return -Math.atan2((bb.getLowerLeft()[1]+bb.height()/2.0)-y, (bb.getLowerLeft()[0]+bb.width()/2.0)-x);
		//	    	}
		//	    	if (anchor == TextAnchor.CENTER_RIGHT) {
		//	    		return -Math.atan2((bb.getLowerLeft()[1]+bb.height()/2.0)-y, bb.getLowerRight()[0]-x); 
		//	    	}
		//	    	if (anchor == TextAnchor.TOP_LEFT) {
		//	    		return -Math.atan2(bb.getUpperLeft()[1]-y, bb.getLowerLeft()[0]-x); 
		//	    	}
		//	    	if (anchor == TextAnchor.TOP_CENTER) {
		//	    		return -Math.atan2(bb.getUpperLeft()[1]-y, bb.getLowerLeft()[0]+bb.width()/2.0-x);
		//	    	} 
		//	    	if (anchor == TextAnchor.TOP_RIGHT) {
		//	    		return -Math.atan2(bb.getUpperLeft()[1]-y, bb.getUpperRight()[0]-x);
		//	    	}
		//	    	return 0.0;
	}



	public JFreeChart getChart() {
		return chart;
	}

	public String getTitle() {
		return title;
	}

}