package gov.anl.gss.sdm.server.model.analysis.pathwaysparticipationmetrics;

	import gov.anl.gss.sdm.server.model.support.MutableInteger;
import gov.anl.gss.sdm.server.model.translator.InformationManagers;
import gov.anl.gss.sdm.server.model.translator.NativeDataTypeManager;
import gov.anl.gss.sdm.server.model.translator.Node;
import gov.anl.gss.sdm.server.model.translator.Parser;
import gov.anl.gss.sdm.server.model.translator.TreeTraversal;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.FileWriter;
import java.io.IOException;
import java.io.InputStreamReader;
import java.util.ArrayList;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;

import org.apache.log4j.Logger;

	public class Derivative {
static Logger log = Logger.getLogger(Derivative.class.getSimpleName());
		
		private List<String> symbols;
		private String equation;
		private List<String> withRespectTo;
		private BufferedWriter out;
		private BufferedReader in;
		private StringBuilder prg = new StringBuilder();
		private List<String> results;
		
		private Map<String, String> uniqueToOriginal;
		private Map<String, MutableInteger> nextID;
		
		private NativeDataTypeManager ndtm;
			
		
		public Derivative() {
			results = new ArrayList<String>();
			uniqueToOriginal = new LinkedHashMap<String, String>();
			nextID = new LinkedHashMap<String, MutableInteger>();
			ndtm = InformationManagers.getInstance().getNativeDataTypeManager();
		}
		
		public Derivative(Node root) {
			this();
			deriv(root);
		}
		
		public Derivative(String equation, List<String> symbols, List<String> withRespectTo) {
			this();
			this.symbols = new ArrayList<String>(symbols);
			this.equation = equation;
			this.withRespectTo = new ArrayList<String>(withRespectTo);
		}
		
		public List<String> getDeriv() {
			return results;
		}
		
		public void deriv() {
			
			printInformation();
			
			openScriptFile();
			importModules();
			defineSymbols();
			generateDiff();
			closeScriptFile();
			execute();
			results = gatherResults();
		}
		
		private void printInformation() {
			log.debug("\nSymbols:");
			for (String s : symbols)
				log.debug(" "+s);
			log.debug("\n");
			log.debug("Equation: "+equation);
			log.debug("WithRespectTo:");
			for (String s : withRespectTo)
				log.debug(" "+s);
			log.debug("\n");
			log.debug("\n");
		}
		
		private void deriv(Node aNode) {
			
			Node copy = TreeTraversal.copyStructure(aNode);
			makeNamesUnique(copy);
			equation = generateExpression(copy);
			log.debug("\nExp2': "+equation+"\n\n");
			symbols = Node.extractVariables(copy);  
		
			withRespectTo = new ArrayList<String>(symbols);
			
			deriv();
			
//			for (String d : getDeriv()) {
//				log.debug("Deriv: "+d);
//			}
		}
		
		private void makeNamesUnique(Node node) {
			if (node == null)
				return;

			Node aNode = node;

			// skip generated arguments
			if (Parser.isFunctionInvocation(node.getToken())) {
				aNode = TreeTraversal.getFunctionArgument(aNode, 1);
			}
			String token = ndtm.getOriginalNameForSubstitution(aNode.getToken());

			// process only variable names
			if (!Parser.isNumber(token) && !Parser.isOperator(token)) {
				aNode.setToken(toUnique(token));
			}

			makeNamesUnique(aNode.getChild());
			makeNamesUnique(aNode.getNext());
		}
		
		private String toUnique(String s) {
			if (!nextID.containsKey(s)) {
				nextID.put(s, new MutableInteger(0));
			}
			MutableInteger mi = nextID.get(s);
			return s + "__" + mi.valueAndInc();
		}
		
		private String fromUnique(String s) {
			return s.split("__")[0];
		}
		
		public String generateExpression(Node node) {
			StringBuilder sb = new StringBuilder();

			if (node != null) {
				

				// node is the root node of an expression
				// if an operator:
				// left child
				// operator
				// right child  
				if (Parser.isArithmeticOperator(node.getToken()) || Parser.isEqualSign(node.getToken())
						|| Parser.isRelationalOperator(node.getToken())) {
					if (!Parser.isEqualSign(node.getToken()))
							sb.append("(");
					if (!Parser.isUnaryOperator(node.getToken())) {
						sb.append(generateExpression(TreeTraversal.getLhs(node))); // node.getChild()
						sb.append(ndtm.getOriginalNameForSubstitution(node.getToken()));
						sb.append(generateExpression(TreeTraversal.getRhs(node))); // node.getChild().getNext()
					} else {
						sb.append(Parser.translateUnaryOperator(node.getToken()));
						sb.append(generateExpression(node.getChild())); // this is really a LHS location, but RHS of unary
					}
					if (!Parser.isEqualSign(node.getToken()))
							sb.append(")");
					return sb.toString();

				} else if (Parser.isFunctionInvocation(node.getToken())) {
//					sb.append(node.getToken());
					sb.append(generateFunctionCall(node));
				} else {
					sb.append(ndtm.getOriginalNameForSubstitution(node.getToken()));
				}

			}
			
			return sb.toString();
		}
		
		public String generateFunctionCall(Node node) {
			StringBuilder sb = new StringBuilder();
			
			if (node != null) {
				// create the entire invocation
				// node is the function name
				sb.append(ndtm.getOriginalNameForSubstitution(node.getToken()));
				sb.append("(");
				int i = 0;
				
				Node arg = TreeTraversal.getFunctionArgument(node, 1);
				while (arg != null) {
					if (i++ > 0)
						sb.append(",");
					sb.append(generateExpression(arg));
					arg = arg.getNext();
				}
				
				sb.append(")");
				
			}
			
			return sb.toString();
		}
		
		public void testIt() {

			
			List<String> syms = new ArrayList<String>();
			syms.add("a");
			syms.add("b");
			syms.add("c");
			syms.add("d");
			syms.add("e");
			syms.add("f");
			String eqn = "(a*b)*(c*((d+e)**-1)*f)";
			List<String> withRespectTo = new ArrayList<String>();
			withRespectTo.add("b");
			withRespectTo.add("c");
			withRespectTo.add("d");
			withRespectTo.add("e");
			
			Derivative d = new Derivative(eqn, syms, withRespectTo);
			
			for (String res : d.getDeriv()) {
				log.debug(res);
			}
		}
		
		public void testIt2() {
			
//			from sympy import *
//			Healthy = Symbol('Healthy')
//			Infected = Symbol('Infected')
//			g = diff((((Healthy*5)*(Infected/(Healthy+Infected)))*0.1), Healthy)
//			print(g)
//
//			g = diff((((Healthy*5)*(Infected/(Healthy+Infected)))*0.1), Infected)
//			print(g)
			
			List<String> syms = new ArrayList<String>();
			syms.add("Healthy");
			syms.add("Infected");
			
			String eqn = "(((Healthy*5)*(Infected/(Healthy+Infected)))*0.1)";
			List<String> withRespectTo = new ArrayList<String>();
			
			withRespectTo.add("Healthy");
			withRespectTo.add("Infected");
			
			
			Derivative d = new Derivative(eqn, syms, withRespectTo);
			
			for (String res : d.getDeriv()) {
				log.debug(res);
			}
		}
		
		private void openScriptFile() {
			try {
				out = new BufferedWriter(new FileWriter("test1.py"));
			} catch (IOException e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			}
		}
		
		private void closeScriptFile() {
			try {
				out.write(prg.toString());
				out.close();
			} catch (IOException e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			}
		}
		
		private void importModules() {
			prg.append("from sympy import *\n");
		}
		
		private void defineSymbols() {
			for (String symbol : symbols) {
				prg.append(symbol+" = Symbol('"+symbol+"')\n");
			}
		}
		
		private void generateDiff() {
			for (String var : withRespectTo) {
				prg.append("g = diff(" + equation + ", " + var + ")\n");
				prg.append("print(g)\n\n");
			}
		} 
		
		private void execute() {
			
			try {
				ProcessBuilder pb = new ProcessBuilder("python", "test1.py");
				Process p = pb.start();

				in = new BufferedReader(new InputStreamReader(
						p.getInputStream()));
			} catch (IOException e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			}
			
			
		}
		
		private List<String> gatherResults() {
			List<String> al = new ArrayList<String>();
			
			try {
				String line = in.readLine();
				int lnum = 1;
				while (line != null) {
//					log.debug("line number " + lnum + " <" + line + ">");
					al.add(line);
					line = in.readLine();
					lnum++;
				}
			} catch (IOException e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			}
			
			return al;
		}

		public static void main(String a[]) {
			Derivative t2 = new Derivative();
			t2.testIt();
			t2.testIt2();
//			t2.doit();
		}

		private void doit() {
			try {

				prg = new StringBuilder();
				String equation = "(a*b)*(c*((d+e)**-1)*f)";

				prg.append("from sympy import *\n");
				prg.append("a = Symbol('a')\n");
				prg.append("b = Symbol('b')\n");
				prg.append("c = Symbol('c')\n");
				prg.append("d = Symbol('d')\n");
				prg.append("e = Symbol('e')\n");
				prg.append("f = Symbol('f')\n");

				for (String var : new String[] { "b", "c", "d", "e" }) {
					prg.append("g = diff(" + equation + ", " + var + ")\n");
					prg.append("print(g)\n\n");
				}

				BufferedWriter out = new BufferedWriter(new FileWriter("test1.py"));
				out.write(prg.toString());
				out.close();

				ProcessBuilder pb = new ProcessBuilder("python", "test1.py");
				Process p = pb.start();

				BufferedReader in = new BufferedReader(new InputStreamReader(
						p.getInputStream()));
				String line = in.readLine();
				int lnum = 1;
				while (line != null) {
					log.debug("line number " + lnum + " <" + line + ">");
					line = in.readLine();
					lnum++;
				}

			} catch (Exception e) {
				log.debug(e);
			}
		}
}
