package gov.anl.gss.sdm.server.model.analysis.pathwaysparticipationmetrics;

import gov.anl.gss.sdm.server.model.analysis.AnalysisManager;
import gov.anl.gss.sdm.server.model.analysis.polarity.ArrowPolarity;
import gov.anl.gss.sdm.server.model.analysis.trees.CausalAnalyzer;
import gov.anl.gss.sdm.server.model.analysis.trees.VariableSubstitution;
import gov.anl.gss.sdm.server.model.support.ArrayReference;
import gov.anl.gss.sdm.server.model.support.Utilities;
import gov.anl.gss.sdm.server.model.translator.Equation;
import gov.anl.gss.sdm.server.model.translator.EquationComponents;
import gov.anl.gss.sdm.server.model.translator.ExpressionGenerator;
import gov.anl.gss.sdm.server.model.translator.InformationManagers;
import gov.anl.gss.sdm.server.model.translator.Node;
import gov.anl.gss.sdm.server.model.translator.SystemDynamicsObjectManager;
import gov.anl.gss.sdm.server.model.translator.TreeTraversal;

import java.util.ArrayList;
import java.util.LinkedHashMap;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.apache.log4j.Logger;

public class DerivativeData {
static Logger log = Logger.getLogger(DerivativeData.class.getSimpleName());
	
	private Equation eqn;
	private Map<String, Set<String>> referencedVariables;
	private Map<String, Equation> equations;
	private String currentLHS;
	private CausalAnalyzer causalAnalyzer;
	private SystemDynamicsObjectManager sdObjectManager;
	private AnalysisManager analysisManager;
	private Node causalRoot;
	private PPMCodeGenerator generator;
	
	private List<String> deltaEquations = new ArrayList<String>();
	private List<String> pathEquations;
	
	private Map<String, String> vensimToRSSD = new LinkedHashMap<String, String>();
	
	private String globalSubscript = "";
	private ArrayReference lhsArrayReference = null;
	
	private Map<String, Equation> equationsToAdd = new LinkedHashMap<String, Equation>();
	
	public DerivativeData(PPMCodeGenerator generator, Equation eqn, Map<String, Equation> equations, CausalAnalyzer causalAnalyzer,
			SystemDynamicsObjectManager sdObjectManager, AnalysisManager analysisManager) {
		this.generator = generator;
		this.eqn = eqn;
		referencedVariables = new LinkedHashMap<String, Set<String>>();
		this.equations = equations;
		this.causalAnalyzer = causalAnalyzer;
		this.sdObjectManager = sdObjectManager;
		this.analysisManager = analysisManager;
		initializeNameMapper();
		decompose();
		generateEquations();
		printEquations();
		log.debug("\n&&& Adding Generated Equations &&&\n");
		addGeneratedEquationsToModel();
		log.debug("\n&&& Done Adding Generated Equations &&&\n");
	}
	
	private void initializeNameMapper() {
		for (String lhs : equations.keySet()) {
			vensimToRSSD.put(InformationManagers.getInstance().getNativeDataTypeManager().getOriginalName(lhs), lhs);
		}
	}
	private void addGeneratedEquationsToModel() {
		addEquationToModel("delta=0.01"+"~"+"dmnl"+"~This is autogenerated for PPM calc");
		for (String eqn : getListOfEquations()) {
			log.debug("add Eqn: "+eqn);
			addEquationToModel(eqn);
		}
	}
	
	private void addEquationToModel(String e) {
		EquationComponents deriv = new EquationComponents(e);
		log.debug("Create New Equation: "+deriv.getVensimEquation());
		Equation newEquation = new Equation(deriv);
		newEquation.setAutoGenerated(true);
//		equations.put(newEquation.getLhs(), newEquation);
		equationsToAdd.put(newEquation.getLhs(), newEquation);
	}
	
	private void printEquations() {
		log.debug("\n\n&&& Printing Generated Equations &&&\n");
		for (String eqn : getListOfEquations()) {
			log.debug(eqn);
		}
		log.debug("\n\n&&& Done Printing Generated Equations &&&\n");
		
	}
	private void decompose() {
		
		// stock for which we want to generate PPM code
//		eqn.printTreeTerse();
		
		String lhs = eqn.getLhs();
		if (ArrayReference.isArrayReference(lhs)) {
			lhsArrayReference = new ArrayReference(lhs);
			globalSubscript = lhsArrayReference.getVensimSubscriptReference();
		}
		
		log.debug("\nExp1: "+ExpressionGenerator.generateExpression(eqn.getTreeRoot())+"\n");
		extractVariables();

	}
	
	private void extractVariables() {
		
		Node rootCopy = eqn.getCopyOfTree();
		currentLHS = TreeTraversal.getLhs(rootCopy).getToken();
		String screenName = InformationManagers.getInstance().getNativeDataTypeManager().getOriginalName(currentLHS);
		causalRoot = transformToArrays(causalAnalyzer.getCausalTree(transformFromArray(screenName)));
		log.debug("currentLHS = "+currentLHS);
		log.debug("screenName: "+screenName);
		
		if (causalRoot == null) {
			log.debug("Can't find causal tree for ");
		} else {
			Node.printTree(causalRoot);
		}
		
		// causal trees are generated from the drawing
		// we need to use equations as the source of causal variables so that we can ignore those not of interest
		// for this purpose. We can assume at this point that the causal tree info is good, but would need to
		// consider on the 1st arument to INTEG.
		
		// point to the first argument
		Node integ = TreeTraversal.getRhs(rootCopy);
		Node firstArg = TreeTraversal.getFunctionArgument(integ, 1);
		
		// get all variable children of this 1st arg
		List<String> level1 = asOrig(extractChildVariables(firstArg));
		
		log.debug("level1 = "+Utilities.asString(level1));
		String o = InformationManagers.getInstance().getNativeDataTypeManager().getOriginalName(causalRoot.getToken());
		referencedVariables.put(o, new LinkedHashSet<String>(level1));
		
		for (Node n : Node.extractChildNodes(causalRoot)) {
			String orig = InformationManagers.getInstance().getNativeDataTypeManager().getOriginalName(n.getToken());
			log.debug("causal node: "+orig);
			if (level1.contains(orig)) {
				log.debug("Extracting: "+orig);
				extractRHSVariables(n);
			}
		}
	}
	
	private String transformFromArray(String var) {
		if (ArrayReference.isArrayReference(var)) {
			return new ArrayReference(var).getArrayName();
		} else {
			return var;
		}
	}
	
	private String transformToVensimArray(String var) {
		if (ArrayReference.isArrayReference(var)) {
			return new ArrayReference(var).getVensimReference();
		} else {
			return var;
		}
	}
	
	private Node transformToArrays(Node n) {
		// causal trees are build with screen names
		// for PPM purposes, we must have array subscripts available for code generation
		
		generator.transformToArrays(n);
		
		return n;
	}

	
	private List<String> asOrig(List<String> al) {
		List<String> list = new ArrayList<String>();
		for (String s : al)
			list.add(InformationManagers.getInstance().getNativeDataTypeManager().getOriginalName(s));
		
		return list;
	}
	
	private List<String> extractChildVariables(Node node) {
		
		// we only want direct child of this tree
		// don't want to process node->next
		List<String> al = new ArrayList<String>();
		Node node2 = node;
		// if this is a terminal node, then it has no rhs vars to process
		if (node.isTerminal()) {
			al.add(node.getToken());
		} else {
			node2 = node.getChild();
			
			// note that this will process child and next
			al.addAll(Node.extractVariables(node2));
		}
		return al;
	}
	
	private void extractRHSVariables(Node node) {
		// node is a causual node
		if (node == null)
			return;
		if (referencedVariables.containsKey(node.getToken())) {
			log.debug("Already Extracted");
			return;
		}
		
		log.debug("ExRHSVar: "+node.getToken());
		Set<String> variables = new LinkedHashSet<String>();
		Node node2 = node;
		// if this is a terminal node, then it has no rhs vars to process
		if (node.isTerminal()) {
			referencedVariables.put(node.getToken(), variables);
		} else {
			// this gets the child names
			variables.addAll(Node.extractChildNames(node));
			referencedVariables.put(node.getToken(), variables);
		}
		
		for (Node childNode : Node.extractChildNodes(node)) {
			extractRHSVariables(childNode);
		}
		return;
	}
	
	public List<String> getListOfPathEquations() {
		if (pathEquations == null) {
			pathEquations = generatePathEquations();
		}
		return pathEquations;
	}
	public List<String> generatePathEquations() {
		List<String> al = new ArrayList<String>();
		int pathNum = 1;
		StringBuilder sb = new StringBuilder("// ");
		String units = "~dmnl";
		StringBuilder comment = new StringBuilder("~");
		for (List<String> path : causalAnalyzer.generatePaths(causalRoot)) {			
			sb.setLength(0);
			comment = new StringBuilder("~");
			sb.append(toPathNumberVariable(removeAnnotation(eqn.getLhs()), pathNum) + " = ");
			sb.append(getSignComponent(path.get(1), path.get(0)) + " * ") ;
			for (int j = 1; j < path.size()-1; j++) {
				if (j > 1) {
					sb.append(" * ");
					
				} else {
					comment.append(" {"+path.get(0)+" <-- "+path.get(1)+"}");
				}
				sb.append(toDerivativeVariable(path.get(j), path.get(j+1)));
				comment.append(" {"+path.get(j)+" <-- "+path.get(j+1)+"}");
			}
			
			// if the start of the causal path (i.e. the last in list) is a stock and different than
			// the variable of interest, need a partial derivative
			
			List<String> partialDerivatives = null;
			if (isMajor(path)) {
				partialDerivatives = new ArrayList<String>();
				log.debug("Found major path. Perform variable substitution for new equation");
				String sub1 = new VariableSubstitution(equations).substitute(eqn);
				String sub2 = new String(sub1);
				log.debug("Equation rewrite 1: "+sub1);
				String variableOfInterest = path.get(0);
				String otherStock = path.get(path.size()-1);
				String vOfIName1 = "effect1_p_"+pathNum+"_"+variableOfInterest;
				String vOfIName2 = "effect2_p_"+pathNum+"_"+variableOfInterest;
				sub1 = sub1.replaceFirst(variableOfInterest, vOfIName1);
				sub2 = sub2.replaceFirst(variableOfInterest, vOfIName2);
				
				log.debug("Equation rewrite 2(1): "+sub1);
				
				sub2 = addDelta(sub2, otherStock);
				log.debug("Equation rewrite 2(2): "+sub2);
				partialDerivatives.add(sub1+"~dmnl~Partial Derivative Equation");
				partialDerivatives.add(sub2+"~dmnl~Partial Derivative Equation");
				
				String partialLHS = "Partial"+toDerivativeVariable(variableOfInterest, otherStock);
				String partialEquation = partialLHS+	" = ("+vOfIName1+" - "+vOfIName2+")/delta";
				log.debug(partialEquation);
				partialDerivatives.add(partialEquation+units+comment);
				
				sb.append(" * "+partialLHS);
				
			}
			
			log.debug("GenPath: "+sb.toString());
			al.add(sb.toString()+units+comment);
			if (partialDerivatives != null) {
				al.addAll(partialDerivatives);
			}
			pathNum++;
		}
		
		// simple sum
		comment.setLength(0);
		comment.append("~Sum paths");
		sb.setLength(0);
		sb.append("Sum"+toPathNumberVariable(removeAnnotation(eqn.getLhs()), 0) + " = ");
		
		for (int i = 1; i < pathNum; i++) {
			if (i > 1)
				sb.append(" + ");
			sb.append(toPathNumberVariable(removeAnnotation(eqn.getLhs()), i));
		}
		log.debug("GenPath: "+sb.toString());
		al.add(sb.toString()+units+comment);
		
		// TPM calculation
		
		comment.setLength(0);
		comment.append("~TPM Calulation");
		sb.setLength(0);
		sb.append("TPM"+toPathNumberVariable(removeAnnotation(eqn.getLhs()), 0) + " = ");

		for (int i = 1; i < pathNum; i++) {
			if (i > 1) {
				sb.append(" + ");
			}
			String x = toPathNumberVariable(removeAnnotation(eqn.getLhs()), i);
//			sb.append("ABS("+toPathNumberVariable(removeAnnotation(eqn.getLhs()), i)+", "+x+", "+x+")");
			sb.append("ABS("+toPathNumberVariable(removeAnnotation(eqn.getLhs()), i)+")");
		}
		log.debug("TPMPath: "+sb.toString());
		al.add(sb.toString()+units+comment);

		// PPM calculation

		comment.setLength(0);
		comment.append("~PPM Calulation");
		


		for (int i = 1; i < pathNum; i++) {
			sb.setLength(0);
			sb.append("PPM"+toPathNumberVariable(removeAnnotation(eqn.getLhs()), i) + " = ");
			sb.append(toPathNumberVariable(removeAnnotation(eqn.getLhs()), i)+"/"+
					"TPM"+toPathNumberVariable(removeAnnotation(eqn.getLhs()), 0));
			log.debug("PPMPath: "+sb.toString());
			al.add(sb.toString()+units+comment);
		}
		
		// Dominant Path  calculation

		sb.setLength(0);
		comment.setLength(0);
		comment.append("~Dominant Path Calulation");
		
		sb.append("DomPath_"+toPathNumberVariable(removeAnnotation(eqn.getLhs()), 0));
		sb.append(" = DOMINANTPATH(");
		sb.append("Sum"+toPathNumberVariable(removeAnnotation(eqn.getLhs()), 0));
		for (int i = 1; i < pathNum; i++) {
			sb.append(", PPM"+toPathNumberVariable(removeAnnotation(eqn.getLhs()), i));
		}
		sb.append(")");
		al.add(sb.toString()+units+comment);
		log.debug("DomPath: "+sb.toString());
		return al;
	}
	
	private boolean isMajor(List<String> path) {
		String start = path.get(path.size()-1);
		String end = path.get(0);
		log.debug(asString(path));
		log.debug("isMajor? "+start+" -> "+end+" <"+vensimToRSSD.get(start)+">");
		if (!equations.get(vensimToRSSD.get(start)).isStock())
			return false;
		
		return !end.equals(start);
	}
	
	private String getSignComponent(String tail, String head) {
		String polarity = analysisManager.getPolarityResultsNoIVV().getPolarityForArrow(tail,  head);
		if (polarity.equals(ArrowPolarity.POSITIVE)) {
			return "(1.0)";
		} else if (polarity.equals(ArrowPolarity.NEGATIVE)) {
			return "(-1.0)";
		} else if (polarity.equals(ArrowPolarity.UNKNOWN)) {
			log.debug("UNKNOWN Polarity for "+tail+" --> "+head);
			return "(1.0)";
		} else {
			log.debug("????? Polarity for "+tail+" --> "+head);
			return "(1.0)";
		}
	}
	
	private void generateEquations() {
		deltaEquations.clear();
		for (String s : referencedVariables.keySet()) {
			Set<String> vars = referencedVariables.get(s);
			// we don't want to add equations for the stocks!
			if (equations.get(s).isStock())
				continue;
			deltaEquations.addAll(generateDeltaEquations(s, vars));
		}

		getListOfPathEquations();
		return;
	}
	
	public List<String> getListOfEquations() {
		List<String> al = new ArrayList<String>();
		al.addAll(deltaEquations);
		al.addAll(pathEquations);
		return al;
	}
	
	public List<String> generateDeltaEquations(String lhs, Set<String> vars) {
		List<String> al = new ArrayList<String>();
		if (vars.size() == 0 || eqn.getLhs().equals(lhs))
			return al;
		
		
		// first generate for lhs
		Equation eqn = equations.get(lhs);
		if (eqn == null) {
			log.debug("Could not find equation with LHS = <"+lhs+">");
		}
		String exp = ExpressionGenerator.generateExpression(eqn.getTreeRoot());
		log.debug("\n\ngenerateDeltaEquations // "+exp);
		int gener = 1;
		for (String var : getNonConstants(vars)) {
			String va = transformToVensimArray(lhs);
			String newLHS = toTempVariable(gener, va);
//			String newLHS = "Temp"+gener+"_"+va;
			String withDelta = addDelta(exp, var).replace(lhs, newLHS);
			al.add(withDelta+"~"+"dmnl"+"~Compute new value with delta");
			al.add(toDerivativeVariable(lhs, var)+"=("+newLHS+"-"+va+")/delta"+"~"+"dmnl"+"~Compute Derivative");
//			log.debug("lhs: "+lhs);
//			log.debug("var: "+var);
//			log.debug("va: "+va);
//			log.debug("newLHS: "+newLHS);
//			log.debug("\nwithDelta: "+withDelta);
//			log.debug("\nGENeqn: "+al.get(al.size()-1));
			gener++;
		}
		
		return al;
		
	}
	
	private String renameVariable(String equation, String origVar, String newVar) {
		return equation.replace(origVar, newVar);
		
	}
	
	private String removeAnnotation(String s ) {
		return s.replace("memory.", "").replace("lookup.", "").replace("array.", "").replace("sdFunctions.", "");
	}
	
	private String convertToArrayIfNecessary(String varName) {
		// if eqn lhs is not an array, we can just use this  varName
		if (globalSubscript.length() == 0) {
			return varName;
		}
		// if lhs is an array, and varname is not, simply add subscript
		if (!ArrayReference.isArrayReference(varName)) {
			return varName+globalSubscript;
		}
		// both are arrays. If they match return varName
		if (new ArrayReference(varName).getVensimSubscriptReference().equals(globalSubscript)) {
			return varName;
		} else {
			log.debug("Subscript mismatch: "+lhsArrayReference.getVensimReference()+" "+varName);
			return varName;
		}
		
	}
	
	private String toTempVariable(int gener, String va) {
		// note, both var and lhs can be arrays,
		// if var is an array, replace with arrayName
		return convertToArrayIfNecessary("Temp"+gener+"_"+va);
	}
	
	private String toDerivativeVariable(String lhs, String var) {
		// note, both var and lhs can be arrays,
		// if var is an array, replace with arrayName
		return convertToArrayIfNecessary("D_"+transformFromArray(var)+"_"+lhs);
	}
	
	private String toPathNumberVariable(String lhs, int pathNum) {
		return convertToArrayIfNecessary("Path_"+pathNum+"_"+removeAnnotation(eqn.getLhs()));
	}
	
	public String asString(Set<String> s) {
		StringBuilder sb = new StringBuilder();
		for (String str : s) {
			if (sb.length() > 0)
				sb.append(", ");
			sb.append(str);
		}
		return "{"+sb.toString()+"}";
	}
	
	public String asString(List<String> s) {
		StringBuilder sb = new StringBuilder();
		for (String str : s) {
			if (sb.length() > 0)
				sb.append(", ");
			sb.append(str);
		}
		return "{"+sb.toString()+"}";
	}
	
	private String addDelta(String equation, String variable) {
		return equation.replace(variable, "("+variable+"+delta)");
	}
	
	private Set<String> getNonConstants(Set<String> allVars) {
		Set<String> set = new LinkedHashSet<String>();
		for (String var : allVars) {
			if (referencedVariables.get(var).size() > 0 ||
					equations.get(var).isStock())
				set.add(var);
		}
		
		return set;
	}

	public Map<String, Equation> getEquationsToAdd() {
		return equationsToAdd;
	}

	public Equation getEqn() {
		return eqn;
	}
}
