package gov.anl.gss.sdm.server.model.analysis.pathwaysparticipationmetrics;

import gov.anl.gss.sdm.server.model.analysis.AnalysisManager;
import gov.anl.gss.sdm.server.model.analysis.trees.CausalAnalyzer;
import gov.anl.gss.sdm.server.model.analysis.trees.VariableSubstitution;
import gov.anl.gss.sdm.server.model.support.ArrayReference;
import gov.anl.gss.sdm.server.model.translator.Equation;
import gov.anl.gss.sdm.server.model.translator.EquationComponents;
import gov.anl.gss.sdm.server.model.translator.EquationGraphicValidation;
import gov.anl.gss.sdm.server.model.translator.InformationManagers;
import gov.anl.gss.sdm.server.model.translator.Node;
import gov.anl.gss.sdm.server.model.translator.SystemDynamicsObjectManager;
import gov.anl.gss.sdm.server.model.translator.TreeTraversal;

import java.util.ArrayList;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;

import org.apache.log4j.Logger;

public class PPMCodeGenerator extends VariableSubstitution {
static Logger log = Logger.getLogger(PPMCodeGenerator.class.getSimpleName());

	private CausalAnalyzer causalAnalyzer;
	private Map<String, DerivativeData> derivativeData;
	private SystemDynamicsObjectManager sdObjectManager;
	private Map<String, String> arrayTransform = new LinkedHashMap<String, String>();
	private AnalysisManager analysisManager;

	public PPMCodeGenerator(Map<String, Equation> equations, CausalAnalyzer causalAnalyzer, 
			SystemDynamicsObjectManager sdObjectManager, AnalysisManager analysisManager) {
		super(equations);
		this.causalAnalyzer = causalAnalyzer;
		this.sdObjectManager = sdObjectManager;
		this.analysisManager = analysisManager;
		derivativeData = new LinkedHashMap<String, DerivativeData>();
		buildArrayTransform(equations);
		linkEquationLHS(equations);
	}
	
	private void linkEquationLHS(Map<String, Equation> equations) {
		// in various parts of the tool, we need to be able to access eqn
		// by screen name, annotated array or non-annotated array
		// provide linkage for all
		Map<String, Equation> equationsAdded = new LinkedHashMap<String, Equation>();
		for (String key : equations.keySet()) {
			Equation eqn = equations.get(key);
			if (ArrayReference.isArrayReference(eqn.getLhs())) {
				ArrayReference ar = new ArrayReference(eqn.getLhs());
				// need to create two new lhs values
				equationsAdded.put(ar.getVensimReference(), eqn);
				equationsAdded.put(ar.getArrayName(), eqn);
			}
		}
		for (String key : equationsAdded.keySet()) {
			equations.put(key, equationsAdded.get(key));
		}
	}
	
	private void buildArrayTransform(Map<String, Equation> equations) {
		
		// provides mapping of screen names to a fully specified lhs (i.e. array subscripts)
		for (String key : equations.keySet()) {
			Equation eqn = equations.get(key);
			String lhs = eqn.getLhs();
			if (ArrayReference.isArrayReference(lhs)) {
				ArrayReference ar = new ArrayReference(lhs);
				arrayTransform.put(ar.getArrayName(), ar.getVensimReference());
			} else {
				arrayTransform.put(lhs, lhs);
			}
		}
		
	}

	public void processAllStocks() {

		log.debug("@@@ PPM Code Generator @@@");

		for (Equation eqn : equations.values()) {
			if (eqn.isStock()) {
				Node rootCopy = eqn.getCopyOfTree();
				currentLHS = TreeTraversal.getLhs(rootCopy).getToken();
				Node rhs = TreeTraversal.getRhs(rootCopy);
				derivativeData.put(currentLHS, extractDerivativeData(eqn));
			}
		}
		
		// now that we've generated code for PPM calcuations, add to the model
		addEquationToModel("delta=0.01"+"~"+"dmnl"+"~This is autogenerated for PPM calc");
		for (DerivativeData dd : derivativeData.values()) {
			log.debug("DD Eqn: "+dd.getEqn().getVensimEquationOnly());
			for (Equation eqnToAdd : dd.getEquationsToAdd().values()) {
				log.debug("toModel: "+eqnToAdd.getVensimEquationOnly());
				equations.put(eqnToAdd.getLhs(), eqnToAdd);
			}
		}
		
		// since we've added equations that may contain arrays, we must process subscript space again
		InformationManagers.getInstance().getArrayManager().populateArraySubscriptSpace();
		List<EquationGraphicValidation> addedScreenNames = sdObjectManager.validate(equations);
		
		sdObjectManager.createSystemDynamicsObjectForNonGraphic(addedScreenNames, equations);
		
	}
	
	public Node transformToArrays(Node n) {
		if (n == null) {
			log.debug("transformToArrays Node is null");
			return n;
		}
		for (Node x : TreeTraversal.getTreeAsList(n)) {
			if (arrayTransform.containsKey(x.getToken()))
				x.setToken(arrayTransform.get(x.getToken()));
			else
				if (!arrayTransform.containsValue(x.getToken()))
					log.debug("!!! NO ARRAY TRANSFORM FOR "+n.getToken());
		}
		return n;
	}
	

	private void addEquationToModel(String e) {
		EquationComponents deriv = new EquationComponents(e);
		Equation newEquation = new Equation(deriv);
		newEquation.setAutoGenerated(true);
//		equations.put(newEquation.getLhs(), newEquation);
		equations.put(newEquation.getLhs(), newEquation);
	}
	

	public DerivativeData extractDerivativeData(Equation eqn) {

		log.debug("\n\nextractDerivativeData: "+eqn.getVensimEquationOnly());

		return new DerivativeData(this, eqn, equations, causalAnalyzer, sdObjectManager, analysisManager);
	}
	
	public List<String> getPPMCode(String variable) {
		List<String> al = new ArrayList<String>();
		
		return al;
	}

}
