package gov.anl.gss.sdm.server.model.analysis.trees;

import gov.anl.gss.sdm.server.model.translator.Arrow;
import gov.anl.gss.sdm.server.model.translator.Node;
import gov.anl.gss.sdm.server.model.translator.SystemDynamicsObject;
import gov.anl.gss.sdm.server.model.translator.SystemDynamicsObjectManager;
import gov.anl.gss.sdm.server.model.translator.TreeTraversal;

import java.util.ArrayList;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;

import org.apache.log4j.Logger;


public class CausalAnalyzer {
static Logger log = Logger.getLogger(CausalAnalyzer.class.getSimpleName());
	
	private Map<String, Node> causalTrees;
	private SystemDynamicsObjectManager sdObjectManager;
	
	private boolean debugPrint = false;
	int extendCount;

	public CausalAnalyzer(SystemDynamicsObjectManager sdObjectManager) {
		causalTrees = new LinkedHashMap<String, Node>();
		this.sdObjectManager = sdObjectManager;
	}

	public void generateCausalTrees() {
//		log.debug("@@@@@@ GenerateCausalTrees @@@@@@");
		Node root = null;
		for (SystemDynamicsObject stock : sdObjectManager.stockVariables()) {
			//    		log.debug(stock.getScreenName());
			root = new Node(stock.getScreenName());
			if (debugPrint)
				log.debug("######### top level extend "+root.getToken());
			extendCount = 0;
			extend(root);
//			log.debug("####################### Causal Tree for "+stock.getScreenName()+"############################\n");
//			printTreeCausal(root, 0, sdObjectManager);
			causalTrees.put(root.getToken(), root);
		}
//		log.debug("@@@@@@ End GenerateCausalTrees @@@@@@");
	}
	
	public Node getCausalTree(String token) {
		return causalTrees.get(token);
	}

	private void extend(Node node) {
		if (node == null) {
			if (debugPrint)
				log.debug("DONE Extending: NULL Node");
			return;
		}
		
		if (extendCount > 100) {
			if (debugPrint)
				log.debug("DONE extendCount > 100");
			return;
		}
		
		extendCount++;
		if (debugPrint)
			log.debug("Entering Extending: "+node.getToken()+" count = "+extendCount);

		Node currentNode = null;
		if (sdObjectManager.getIncomingArrows(node.getToken()).size() == 0) {
			if (debugPrint)
				log.debug(node.getToken()+" No input links. Done");
			return;
		} else {
			if (debugPrint)
				log.debug(node.getToken()+" Num incoming arrows = "+
						sdObjectManager.getIncomingArrows(node.getToken()).size());
		}
		for (Arrow in : sdObjectManager.getIncomingArrows(node.getToken())) {
			Node n = new Node(in.getOtherEnd());
			if (currentNode == null) {
				node.setChild(n);
				n.setParent(node);
			} else {
				currentNode.setNext(n);
				n.setPrevious(currentNode);
				n.setParent(node);
			}
			currentNode = n;

			SystemDynamicsObject obj = sdObjectManager.getObjectWithName(in.getOtherEnd());
			if (obj != null && !obj.isStock()) {
				if (debugPrint)
					log.debug("Will extend "+obj.getScreenName());
				extend(n);
			} else {
				if (debugPrint)
					log.debug("Skip Extending: "+n.getToken()+" obj = "+(obj == null ? "NULL" : "Stock"));
			}

		}
		if (debugPrint)
			log.debug("DONE Looking at arrows for "+node.getToken()+" count = "+extendCount);
	}

	public static void printTreeCausal(Node node, int level, SystemDynamicsObjectManager sdObjectManager) {
		if (node == null)
			return;
		String blanks="";
		for (int i = 0; i < level; i++)
			blanks += "    ";

		String stock = "";
		if (sdObjectManager.getObjectWithName(node.getToken()).isStock())
			stock = " <STOCK>";

		log.debug(blanks+"("+level + ") "+node.getToken()+stock);

		printTreeCausal(node.getChild(), level+1, sdObjectManager);
		printTreeCausal(node.getNext(), level, sdObjectManager);
	}
	
	public List<List<String>> generatePaths(Node node) {
		List<List<String>> paths = new ArrayList<List<String>>();
		for (Node n : getStockEndPoints(node)) {
			paths.add(generatePathFrom(n));
		}
		
		return paths;
	}
	
	private List<String> generatePathFrom(Node n) {
		List<String> backwards = new ArrayList<String>();
		Node b = n;
		while (b != null) {
			backwards.add(b.getToken());
			b = b.getParent();
		}
		return reverse(backwards);
	}
	
	private List<String> reverse(List<String> list) {
		List<String> rev = new ArrayList<String>();
		for (int i = list.size()-1; i >=0; i--) {
			rev.add(list.get(i));
		}
		return rev;
	}
	
	public int getNumberOfPaths(Node node) {
		int num = 0;
		for (Node n : TreeTraversal.getTreeAsList(node)) {
			if (sdObjectManager.getObjectWithName(n.getToken()).isStock())
				num++;
		}
		return num-1;
	}
	
	public List<Node> getStockEndPoints(Node node) {
		List<Node> list = new ArrayList<Node>();
		int num = 0;
		for (Node n : TreeTraversal.getTreeAsList(node)) {
			if (sdObjectManager.getObjectWithName(n.getToken()).isStock() && num++ > 0)
				list.add(n);
		}
		return list;
	}


}
