package gov.anl.gss.sdm.server.model.analysis.variableclass;

import gov.anl.gss.sdm.server.model.analysis.graphtheory.ReachabilityMatrix;
import gov.anl.gss.sdm.server.model.analysis.loops.LoopManager;
import gov.anl.gss.sdm.server.model.structure.SDObjectManager;
import gov.anl.gss.sdm.server.model.support.Utilities;
import gov.anl.gss.sdm.server.model.translator.Equation;
import gov.anl.gss.sdm.server.model.translator.InformationManagers;
import gov.anl.gss.sdm.server.model.translator.SDM_Doc;

import java.io.BufferedWriter;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Collections;
import java.util.LinkedHashMap;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.apache.log4j.Logger;

public class VariableClassAnalyzer {
static Logger log = Logger.getLogger(VariableClassAnalyzer.class.getSimpleName());
	
	private SDObjectManager objManager;
	private LoopManager loopManager;
	private ReachabilityMatrix reachabilityMatrix;
	
	private Map<String, ExogenousEndogenous> variableTypes = new LinkedHashMap<String, ExogenousEndogenous>();
	
	private Set<String> remainingVariables = new LinkedHashSet<String>();
	
	public VariableClassAnalyzer(SDObjectManager objManager, LoopManager loopManager, ReachabilityMatrix reachabilityMatrix) {
		this.objManager = objManager;
		this.loopManager = loopManager;
		this.reachabilityMatrix = reachabilityMatrix;
		remainingVariables.addAll(InformationManagers.getInstance().getEquationManager().getVariables());
		analyze();
		if (SDM_Doc.DUMP_TABULAR_DATA) {
			dumpToFile(Utilities.openFileForWriting(
					InformationManagers.getInstance().getModelManager().getGeneratedOutputDirectoryName(), "variableClass.tab"));
			dumpWithDistanceToFile(Utilities.openFileForWriting(
					InformationManagers.getInstance().getModelManager().getGeneratedOutputDirectoryName(), "variableClassDistance.tab"));
		}
	}
	
	private void analyze() {

		// find exogenous
		for (String var : InformationManagers.getInstance().getEquationManager().getVariables()) {
			Equation eqn = InformationManagers.getInstance().getEquationManager().getAnEquationFor(var);
			if (eqn.isData() || eqn.isConstant() || eqn.isArrayInitialization() || eqn.isDefinesLookup()) {
				variableTypes.put(var, new ExogenousEndogenous(ExogenousEndogenous.EXOGENOUS));
				remainingVariables.remove(var);
			} 
		}

		// find endogenous
		for (String var : InformationManagers.getInstance().getEquationManager().getVariables()) {
			if (loopManager.getNumberOfLoopsWith(var) > 0) {
				variableTypes.put(var, new ExogenousEndogenous(ExogenousEndogenous.ENDOGENOUS));
				remainingVariables.remove(var);
			} 
		}

		// X = A + B
		// usedBy(A) = X   A -> X  A usedBy X
		// usedBy(B) = X   B -> X  B usedBy X
		// uses(X) = {A, B} {A -> X, B -> X} {X uses A, X uses B}
		
		// We define type of X based on it's inputs

		// find endogenous transformed
		// find exogenous transformed
		// ENDOGENOUS_TRANSFORMED has priority over EXOGENOUS_TRANSFORMED
		

		while (remainingVariables.size() > 0) {
			Set<String> toRemove = new LinkedHashSet<String>();
			for (String head : remainingVariables) {
				
				// if not used by anyone, we need to get to it via uses
					for (String tail : objManager.getUses(head)) {
						ExogenousEndogenous tailExEn = variableTypes.get(tail);
						// do we know the type of the variable pointing to us?
						if (tailExEn != null) {
							ExogenousEndogenous t = variableTypes.get(head);
							if (t == null && (tailExEn.getType().equals(ExogenousEndogenous.EXOGENOUS) || tailExEn.getType().equals(ExogenousEndogenous.EXOGENOUS_TRANSFORMED))) {
								variableTypes.put(head, new ExogenousEndogenous(ExogenousEndogenous.EXOGENOUS_TRANSFORMED));
								toRemove.add(head);
							} else if (tailExEn.getType().equals(ExogenousEndogenous.ENDOGENOUS) || tailExEn.getType().equals(ExogenousEndogenous.ENDOGENOUS_TRANSFORMED)) {
								variableTypes.put(head, new ExogenousEndogenous(ExogenousEndogenous.ENDOGENOUS_TRANSFORMED));
								toRemove.add(head);
							}
						}
					}
			}
			if (toRemove.size() == 0) {
				log.debug("Stuck defining variable types, abort");
				log.debug("Remaining Variables:");
				for (String head : remainingVariables) {
					log.debug("  "+head);
				}
				break;
			}
			remainingVariables.removeAll(toRemove);
		}


	}

public void dumpToFile(BufferedWriter bw) {
	try {
		bw.append("Variable\tType\n");
		for (String s : variableTypes.keySet()) {
			bw.append(s+"\t"+variableTypes.get(s).getType()+"\n");
		}
		bw.close();
	} catch (IOException e) {
		// TODO Auto-generated catch block
		e.printStackTrace();
	}
}

public void dumpWithDistanceToFile(BufferedWriter bw) {
	
//	ReachabilityMatrix reachabilityMatrix;
//	Map<String, ExogenousEndogenous> variableTypes = new LinkedHashMap<String, ExogenousEndogenous>();
	
	try {
		bw.append("Type\t");
		for (String var1 : ordered(variableTypes.keySet())) {
			bw.append("Variable\t");
			bw.append(var1+" <"+variableTypes.get(var1).getType()+">");
		}
		bw.append("\n");
		
		for (String var1 : ordered(variableTypes.keySet())) {
			bw.append(variableTypes.get(var1).getType()+"\t");
			bw.append(var1);
			for (String var2 : ordered(variableTypes.keySet())) {
				int dist = reachabilityMatrix.getDistance(var1, var2);
				if (var1.equals(var2) || dist != -1)
					bw.append("\t"+dist);
				else
				bw.append("\t");
			}
			bw.append("\n");
		}
		
		bw.close();
	} catch (IOException e) {
		// TODO Auto-generated catch block
		e.printStackTrace();
	}
}

private List<String> ordered(Set<String> aSet) {
	List<String> aList = new ArrayList<String>(aSet);
	Collections.sort(aList);
	return aList;
}
	

}
