package gov.anl.gss.sdm.server.model.bridge;

import gov.anl.gss.sdm.shared.CustomGraphObject;
import gov.anl.gss.sdm.shared.Lever;
import gov.anl.gss.sdm.shared.LoopObject;
import gov.anl.gss.sdm.shared.Node;
import gov.anl.gss.sdm.shared.Output;
import gov.anl.gss.sdm.shared.RunObject;
import gov.anl.gss.sdm.shared.ViewObject;

import java.util.List;
import java.util.Map;
import java.util.Set;

public interface Bridge {
	
	public void addCustomGraph(CustomGraphObject cgo);
	
	/**
	 * backset time one timeStep
	 * @return
	 */
	public Boolean backstepModel();
	
	public Boolean backstepModelToTime(double time);
	
	public Boolean closeSelectedModel();
	
	/**
	 * create a new run for the selected model 
	 * @param runName
	 * @return
	 */
	public Boolean createNewRun(String runName);
	
	/**
	 * Create a new vignette with the given name 
	 * @param vignette
	 * @return
	 */
	public Boolean createNewVignette(String vignette);
	
	public Boolean done();
	
	public double[][] getAllValues(String runName, String[] vars);
	
	/**
	 * Returns for all variables all values generated in an existing run
	 * @param run
	 * @return
	 */
	public Map<String, List<Double>> getAllVariableValues(String run);
	
	/**
	 * returns the values for all variables for a specific run at simulation time
	 * @param run
	 * @param time
	 * @return
	 */

	public Map<String, Double> getAllVariableValues(String run, double time, Set<String> selectedVariables);
	
	/**
	 * get the list of arrows that have the specified variable as
	 * @param variable
	 * @return
	 */
	public List<BridgeArrow> getArrowsWithHead(String variable);
	
	/**
	 * get the list of arrows that have the specified variable as tail
	 * @param variable
	 * @return
	 */
	
	public List<BridgeArrow> getArrowsWithTail(String variable);
	
	/**
	 * Returns the alphabetically sorted list of available models
	 * @return
	 */
	public List<String> getAvailableModels();
	

	/**
	 * Returns a list of all the view names in the SD model
	 * @return
	 */
	public List<ViewObject> getAvailableViewNames();
	
	/**
	 * return causes tree for variable to specified levels
	 * @param variable
	 * @param numLevels
	 * @return
	 */
	public Node getCausesTree(String variable, int numLevels);
	
	/**
	 * get the current time of the running model
	 * @return
	 */
	public Double getCurrentTime();
	
	/**
	 * Returns the current value for each of the variables specified
	 * @param variables
	 * @return
	 */
	public double[] getCurrentValues(String[] variables);
	
	public List<CustomGraphObject> getCustomGraphs();
	
	/**
	 * get the dt used in running the model
	 * @return
	 */
	public Double getDT();

	/**
	 * get the end time for the in-core model
	 * @return
	 */
	public Double getEndTime();

	/**
	 * get the list of all know loops
	 * @return
	 */
	public List<LoopObject> getLoops();
	/**
	 * 
	 * @param length
	 * @return
	 */
	public List<LoopObject> getLoopsOfLength(int length);

	/**
	 * get the list loops that contain the variable
	 * @param variable
	 * @return
	 */
	public List<LoopObject> getLoopsWithVariable(String variable);

	/**
	 * get the set of variables from the model that are considered "levers"
	 * @return
	 */
	public Set<Lever> getModelLeverVariables();

	/**
	 * get the set of variables from the model that are considered "outputs"
	 * @return
	 */
	public Set<Output> getModelOutputVariables();
	
	public Map<String, Map<String, String>> getPolarity();
	
	/**
	 * Returns the alphabetically sorted list of previous runs that the user
	 * can select for inclusion on graphs and in tables
	 * @return
	 */
	public List<RunObject> getRuns();

	/**
	 * get the save period for the model
	 * @return
	 */
	public Double getSavePer();
	
	/**
	 * get the html file name with SDM Doc
	 * @return
	 */
	public String getSdmdocHtmlFile();
	
	/**
	 * get the start time for the in-core model
	 * @return
	 */
	public Double getStartTime();
	
	/**
	 * get the Time Step for the in-core model
	 * @return
	 */
	public Double getTimeStep();
	
	/**
	 * get the units for time
	 * @return
	 */
	public String getTimeUnits();
		
	/**
	 * get the list of time steps for which data will be captured
	 * @return
	 */
	public List<Double> getTimeVector();
	
	/**
	 * time vector for existing run
	 * @param run
	 * @return
	 */
	public List<Double> getTimeVector(String run);


	/**
	 * return uses tree for variable to specified levels
	 * @param variable
	 * @param numLevels
	 * @return
	 */
	public Node getUsesTree(String variable, int numLevels);
	
	/**
	 * retrieve variable values from a previous run
	 * @param runName
	 * @param variable
	 * @param time
	 * @return
	 */
	public double[] getValues(String runName, String[] variable, double time);
	
	public double[][] getValues(String[] variables);
	
	/**
	 * Returns the value for each of the variables at a specific time
	 * @param variable
	 * @param time
	 * @return
	 */
	public double[] getValues(String[] variable, double time);
	
	public double[][] getValuesForCurrentTime(String[] vars);
	
	public double[][] getValuesUpToCurrentTime(String[] vars);
	
	public double[][] getValuesUpToTime(String[] vars, double time);
	
	/**
	 * returns the value of the variable at a specific simulation time from a specific run
	 * @param run
	 * @param variable
	 * @param time
	 * @return
	 */
	public Double getVariableValue(String run, String variable, double time);
	
	/**
	 * returns the list of values available for the specified variable in the specified run  
	 * @param run
	 * @param variable
	 * @return
	 */
	public List<Double> getVariableValues(String run, String variable);
	
//	/**
//	 * return all Variables in the currently selected vignette
//	 * @return
//	 */
//	
//	public Set<BridgeVariable> getVignetteVariables();
	
	/**
	 * returns the list of graphic object definitions for the specified view
	 * @param name
	 * @return
	 */
	public List<String> getView(String name);
	
	/**
	 * returns the alphabetically sorted list of vignettes that have been
	 * defined for the selected
	 * @return
	 */
	public List<String> getVignettes();
	
	/**
	 * specifies whether model is complete or just a map
	 * @return
	 */
	public Boolean modelIsMap();
	
	/**
	 * specifies whether model is complete or just a map
	 * @return
	 */
	public Boolean modelIsRunnable();
	
	/**
	 * Begin execution of the model -- performs all model initialization and starts execution
	 * of the model returns when ready to step through time
	 * @return
	 */
	public Boolean run();
	
	public Boolean runTo(double time);

	public void save(BridgeVignetteObject bvo);
	public void save(Lever lever);
	public void save(Output output);
	public void saveCustomGraphs();
	/**
	 * save the selected vignette to disk
	 * @return
	 */
	public Boolean saveSelectedVignette();
	public void saveVignette();
	/**
	 * Use the specified model for all processing
	 * @param model
	 * @return success on initial processing
	 */
	public Boolean selectModel(String model);
	/**
	 * Specifies one or more runs that will be included in this session 
	 * @param runss
	 * @return
	 */
	public Boolean selectRuns(Set<String> runs);	
	/**
	 * The vignette the user selects
	 * @param vignette
	 * @return
	 */
	public Boolean selectVignette(String vignette);
	/**
	 * Sets the value of a variable in the model
	 * @param variableName
	 * @param value
	 * @return
	 */
	public Boolean setValue(String variableName, double value);
	public Boolean startExecution();
	/**
	 * run the model forward one step
	 * @return
	 */
	public Boolean stepModel();
	
	public Boolean validateRun(String runName);
	
}
