package gov.anl.gss.sdm.server.model.bridge;

import gov.anl.gss.sdm.server.model.support.Utilities;
import gov.anl.gss.sdm.server.model.support.database.SDModelRunDatabase;

import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;

public class BridgeModelResultsDatabaseImplementation {
	
	private final static String UNKNOWN = "Unknown";
	
	private String filename;
	private String scenario;
	private double startTime;
	private double endTime;
	private double timeStep;
	private double dt;
	
	private SDModelRunDatabase db;
	
	private String fileFormat = UNKNOWN;
	
	private boolean cleanDB = false;
	private boolean completeRun = false;
	
	public BridgeModelResultsDatabaseImplementation() {
	}
	
	/**
	 * For existing model runs on disk
	 * @param filename
	 */
	public BridgeModelResultsDatabaseImplementation(String filename) {
		this();
		this.filename = filename;
		setDb(new SDModelRunDatabase(filename));
		cleanDB = validate();
		completeRun = completedRun();
		if (cleanDB) {
			getDb().load();
			getDb().loadRunParameters();


			// if this is the new run database, this information is not yet available in its entirety

			startTime = getDb().getStartTime();
			double[] tv = getDb().getTimeVector();
			if (tv.length > 1) {
				endTime = getDb().getEndTime();
				timeStep = getDb().getTimeStep();
				dt = getDb().getDt();
			}
		}
	}
	
	/**
	 * For a new model run
	 * @param filename
	 * @param scenario
	 * @param startTime
	 * @param endTime
	 * @param timeStep
	 * @param dt
	 */
	public BridgeModelResultsDatabaseImplementation(
			String filename,
			String mapFile,
			String scenario,
			double startTime,
			double endTime,
			double timeStep,
			double dt
			) {
		this();
		this.filename = filename;
		this.scenario = scenario;
		this.startTime = startTime;
		this.endTime = endTime;
		this.timeStep = timeStep;
		this.dt = dt;
		setDb(new SDModelRunDatabase(filename));
		getDb().initializeNewRunDatabase(mapFile);
		
		cleanDB = true;
	}
	
	public boolean completedRun() {
		return getDb().isCompletedRun();
	}
	
	private boolean validate() {
		
		cleanDB = getDb().validate();
		return cleanDB;
	}
	
//	private int mapTimeToIndex(double time) {
//		for (int i = 0; i < timeVector.size(); i++)
//			if (timeVector.get(i).doubleValue() == time)
//				return i;
//		return 0;
//	}
	
	public List<Double> getValues(String variable) {
		return Utilities.asList(getDb().getVarTimeVector(variable));
	}
	
	public double getValue(String variable, double time) {
			
		return getDb().getVarValue(variable, time);
	}
	
	public double[][] getValues(String[] variable) {
		
		return getDb().getVarsValues(variable);
	}
	
	public double[] getValue(String[] variable, double time) {
		
		return getDb().getVarsValue(variable, time);
	}
	
	public Map<String, Double> getAllVariableValues(double time, Set<String> selected) {
		
		Map<String, Double> map = new LinkedHashMap<String, Double>();
		
		for (String var : selected) {
				map.put(var, getDb().getVarValue(var, time));
		}

		return map;
	}
	
	public double[] getValuesAsArray(String variable) {
		return getDb().getVarTimeVector(variable);
		
	}
	
	public Map<String, List<Double>> getAllVariableValues() {
		return getDb().getAllVariableValuesAsMapList();
	}

	/**
	 * @return the filename
	 */
	public String getFilename() {
		return filename;
	}

	/**
	 * @param filename the filename to set
	 */
	public void setFilename(String filename) {
		this.filename = filename;
	}

	/**
	 * @return the scenario
	 */
	public String getScenario() {
		return scenario;
	}

	/**
	 * @param scenario the scenario to set
	 */
	public void setScenario(String scenario) {
		this.scenario = scenario;
	}

	/**
	 * @return the startTime
	 */
	public double getStartTime() {
		return startTime;
	}

	/**
	 * @param startTime the startTime to set
	 */
	public void setStartTime(double startTime) {
		this.startTime = startTime;
	}

	/**
	 * @return the endTime
	 */
	public double getEndTime() {
		return endTime;
	}

	/**
	 * @param endTime the endTime to set
	 */
	public void setEndTime(double endTime) {
		this.endTime = endTime;
	}

	/**
	 * @return the timeStep
	 */
	public double getTimeStep() {
		return timeStep;
	}

	/**
	 * @param timeStep the timeStep to set
	 */
	public void setTimeStep(double timeStep) {
		this.timeStep = timeStep;
	}

	/**
	 * @return the dt
	 */
	public double getDt() {
		return dt;
	}

	/**
	 * @param dt the dt to set
	 */
	public void setDt(double dt) {
		this.dt = dt;
	}

	/**
	 * @return the cleanLoad
	 */
	public boolean isOk() {
		return cleanDB;
	}

	/**
	 * @return the timeVector
	 */
	public List<Double> getTimeVector() {
		return Utilities.asList(getDb().getTimeVector());
	}

	public List<String> getVariables() {
		return Utilities.asList(getDb().getVariables());
	}

	/**
	 * @return the fileFormat
	 */
	public String getFileFormat() {
		return fileFormat;
	}
	
	public void shutdowndatabase() {
		getDb().shutdown();
	}

	public boolean isCompleteRun() {
		return completeRun;
	}

	public SDModelRunDatabase getDb() {
		return db;
	}

	public void setDb(SDModelRunDatabase db) {
		this.db = db;
	}
}
