package gov.anl.gss.sdm.server.model.bridge;

import gov.anl.gss.sdm.server.model.support.Utilities;
import gov.anl.gss.sdm.server.model.translator.Parser;

import java.util.ArrayList;
import java.util.LinkedHashMap;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

public class BridgeRunValidator {

	private BridgeModelMeta meta;

	private List<Double> timeVector = new ArrayList<Double>();

	private Map<String, List<Double>> values = new LinkedHashMap<String, List<Double>>();
	
	private boolean valid;
	private String runFile;

	public boolean isValid() {
		return valid;
	}

	public BridgeRunValidator(BridgeModelMeta meta) {
		this.meta = meta;
	}

	public boolean validate(String runName) {

		System.out.println("Validating Run: "+runName);

		runFile = meta.getModellRunFileName(runName);
		String varFile = meta.getModelVariablesFileName();

		double initialTime = Double.parseDouble(meta.getInitialTime());
		double finalTime = Double.parseDouble(meta.getFinalTime());
		double saveper = 1.0;
		
		try {
			saveper= Double.parseDouble(meta.getSaveper());
		} catch (NumberFormatException e) {
			if (meta.getSaveper().contains("TIME"))
				try {
					saveper = Double.parseDouble(meta.getDt());
				} catch (NumberFormatException e1) {
					
				}
		}

		Set<String> modelVariables = new LinkedHashSet<String>();
		Set<String> runVariables = new LinkedHashSet<String>();

		List<String> runFileContents = Utilities.getAsList(runFile); // var, time, val
		List<String> varFileContents = Utilities.getAsList(varFile); // only 1st column

		for (String line : varFileContents) {
			modelVariables.add(line.split("\t")[0]);
		}

		int i = 0;
		for (String line : runFileContents) {
			if (i++ == 0) {
				if (line.equals("Variable\tTime\tValue"))
					continue;
			}
			runVariables.add(line.split("\t")[0]);
		}

		// verify consistency of variables

		boolean variableConsistency = true;

		System.out.println("ModelVariablesSize: "+modelVariables.size());
		System.out.println("runVariablesSize: "+runVariables.size());

		boolean sameSize = modelVariables.size() == runVariables.size();
		Set<String> modelMinusRun = minus(modelVariables, runVariables);
		Set<String> runMinusModel = minus(runVariables, modelVariables);

		System.out.println("modelMinusRun: "+modelMinusRun);
		System.out.println("runMinusModel: "+runMinusModel);

		if (!sameSize || modelMinusRun.size() > 0 || runMinusModel.size() > 0) {
			variableConsistency = false;
			System.out.println("variableConsistency "+sameSize+" "+(modelMinusRun.size() > 0)+
					" "+(runMinusModel.size() > 0));
		}

		System.out.println("");

		boolean valueConsistency = true;
		boolean timeConsistency = true;

//		load(runFileContents);
		
		BridgeModelResultsLoader bmrl = new BridgeModelResultsLoader();
		if (!bmrl.load(runFile))
			return false;
		timeVector = bmrl.getTimeVector();
		values = bmrl.getValues();
		

		int timeValues = 1+ (int) ((finalTime - initialTime)/saveper);

		for (String var : values.keySet()) {
			List<Double> vals = values.get(var);
			if (vals.size() != timeValues) {
				System.out.println("vals.size() = "+vals.size()+ " timeValues = "+timeValues);
				timeConsistency = false;
			}
			for (Double val : vals) {
				
				if (Double.isInfinite(val) ||
						Double.isNaN(val))
					valueConsistency = false;
			}
		}

		System.out.println(variableConsistency+" "+ valueConsistency +" "+timeConsistency);

		valid =  variableConsistency && valueConsistency && timeConsistency;
		return valid;
	}

	public BridgeModelMeta getMeta() {
		return meta;
	}

	public List<Double> getTimeVector() {
		return timeVector;
	}

	public String getRunFile() {
		return runFile;
	}

	public Map<String, List<Double>> getValues() {
		return values;
	}

	private Set<String> minus(Set<String> set1, Set<String> set2) {
		Set<String> set = new LinkedHashSet<String>(set1);
		set.removeAll(set2);
		return set;
	}

	private void load(List<String> runFileContents) {

		String previousTime = "";

		int i = 0;
		for (String line : runFileContents) {
			if (i++ == 0)
				continue;
			String[] v = line.split("\t");
			if (v.length == 3) {
				if (!values.containsKey(v[0])) {
					values.put(v[0], new ArrayList<Double>());
				}
				if (!previousTime.equals(v[1])) {
					if (Parser.isNumber(v[1])) {
					timeVector.add(Double.parseDouble(v[1]));
					previousTime = v[1];
					}
				}
				if (Parser.isNumber(v[2]))
					values.get(v[0]).add(Double.parseDouble(v[2]));
				else
					values.get(v[0]).add(Double.NaN);
			}
		}

		return;
	}

}
