package gov.anl.gss.sdm.server.model.bridge;

import gov.anl.gss.sdm.server.model.support.Utilities;
import gov.anl.gss.sdm.server.model.translator.Parser;
import gov.anl.gss.sdm.shared.CustomGraphObject;
import gov.anl.gss.sdm.shared.Lever;
import gov.anl.gss.sdm.shared.Output;

import java.io.BufferedWriter;
import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashSet;
import java.util.LinkedHashMap;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.apache.commons.io.FileUtils;

public class BridgeVignetteObject {

	private String name;
	private String model;
	private ModelStructure modelStructure;

	private BridgeCustomGraphManager graphManager;

	private Map<String, BridgeVariable> levers;
	private Map<String,BridgeVariable> outputs;
	
	private Set<String> requiredForCustomGraphs;

	private String dir;

	public BridgeVignetteObject() {
		levers = new LinkedHashMap<String, BridgeVariable>();
		outputs = new LinkedHashMap<String, BridgeVariable>();
		requiredForCustomGraphs = new LinkedHashSet<String>();
	}

	public BridgeVignetteObject(String model, ModelStructure modelStructure, String name, String rootDirectory) {
		this();
		this.name = name;
		this.model = model;
		this.modelStructure = modelStructure;
		dir = rootDirectory +"/vignettes/";
		graphManager = new BridgeCustomGraphManager(modelStructure);
	}
	
	public boolean copyHelpFile() {
		
		String template = dir+"../../../ScenarioDocumentationTemplate.docx";
		String helpFile = dir+name+".docx";
		try {
			FileUtils.copyFile(new File(template), new File(helpFile));
			
		} catch (IOException e1) {
			// TODO Auto-generated catch block
			e1.printStackTrace();
			return false;
		}
		return true;
	}

	/**
	 * A new vignette is being created. Read the model variables and initialize...
	 * @param file
	 * @return
	 */
	public boolean create() {
		saveCustomGraphsOnCreate();
		copyHelpFile();
		for (BridgeModelVariable bv : modelStructure.getVariables().values()) {
			BridgeVariable vv = new BridgeVariable(bv);
			if (vv.isLever())
				levers.put(vv.getName(), vv);
			else
				outputs.put(vv.getName(), vv);
			
			if (requiredForCustomGraphs.contains(vv.getName()) || requiredForCustomGraphs.contains(Parser.quoted(vv.getName()))) // "\""+vv.getName()+"\""))
				vv.setIncluded(true);
		}
		
		List<String> vars = new ArrayList<String>();
		vars.addAll(levers.keySet());
		vars.addAll(outputs.keySet());
		Collections.sort(vars);
		int pos = 0;
		for (String v : vars) {
			if (levers.containsKey(v)) {
				levers.get(v).setPosition(pos);
			} else {
				outputs.get(v).setPosition(pos);
			}
			pos++;
		}
		
		return true;
	}

	/**
	 * load an existing vignette into the object
	 * @param file
	 * @return
	 */
	public boolean load() {	
		if (Utilities.fileExists(dir+"/"+name+".tab")) {
		List<String> vars = Utilities.getAsList(dir+"/"+name+".tab");
		for (String var : vars) {
			String[] p = var.split("\t");
			BridgeVariable vv = new BridgeVariable(p[0], p[1], p[2], p[3], p[4], p[5], p[6],
					p[7], p[8], p[9], p[10], p[11], p[12]);
			if (vv.isLever())
				levers.put(vv.getName(), vv);
			else
				outputs.put(vv.getName(), vv);
			
		
		}

		loadCustomGraphs(dir + "/"+name+ "_customGraphs.tab");
		requiredForCustomGraphs = graphManager.getRequiredForCustomGraphs();
		} else {
			create();
			
			save();
		}

		return true;
	}



	public boolean save() {
		if (name == null)
			return false;
		BufferedWriter bw = Utilities.openFileForWriting(dir+name+".tab");
		for (BridgeVariable v : levers.values()) {
			if (requiredForCustomGraphs.contains(v.getName()) || 
					requiredForCustomGraphs.contains(Parser.quoted(v.getName()))) // "\""+v.getName()+"\""))
				v.setIncluded(true);
			v.save(bw);
		}
		for (BridgeVariable v : outputs.values()) {
			if (requiredForCustomGraphs.contains(v.getName()) || 
					requiredForCustomGraphs.contains(Parser.quoted(v.getName()))) // "\""+v.getName()+"\""))
				v.setIncluded(true);
			v.save(bw);
		}
		try {
			bw.close();
		} catch (IOException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
			return false;
		}
		
		return true;
	}
	
	public boolean saveCustomGraphsOnCreate() {
		String cgFile = dir+"../structure/customGraphs.tab";
		File cg = new File(cgFile);
		if (!cg.exists())
			return false;
		try {
			BufferedWriter bw = Utilities.openFileForWriting(dir+name+"_customGraphs.tab");
			for (String g : Utilities.getAsList(cgFile)) {
				String[] fields = g.split("\t");
				for (int i = 1; i < fields.length; i++)
					requiredForCustomGraphs.add(fields[i]);
				bw.append(g);
				bw.append("\n");
			}
			bw.close();
		} catch (IOException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
			return false;
		}
		return true;
	}

	/**
	 * @return the name
	 */
	public String getName() {
		return name;
	}

	/**
	 * @param name the name to set
	 */
	public void setName(String name) {
		this.name = name;
	}

	/**
	 * @return the levers
	 */
	public Set<Lever> getLevers() {

		HashSet<Lever> asLevers = new LinkedHashSet<Lever>();
		for (BridgeVariable bv : levers.values()) {
			asLevers.add(bv.asLever());
		}
		return asLevers;
	}

	/**
	 * @return the outputs
	 */
	public Set<Output> getOutputs() {

		HashSet<Output> asOutputs = new LinkedHashSet<Output>();

		for (BridgeVariable bv : outputs.values()) {
			asOutputs.add(bv.asOutput());
		}
		return asOutputs;
	}

	/**
	 * @return the model
	 */
	public String getModel() {
		return model;
	}

	public void save(Lever lever) {
		levers.put(lever.getName(), new BridgeVariable(lever));	
	}

	public void save(Output output) {
		outputs.put(output.getName(), new BridgeVariable(output));	
	}




	public void loadCustomGraphs(String file) {
		graphManager.loadCustomGraphs(file);
	}

	public List<CustomGraphObject> getCustomGraphs() {
		return graphManager.getCustomGraphs();
	}

	public void saveCustomGraphs() {
		graphManager.save(dir + "/"+name+ "_customGraphs.tab");

	}
	public void addCustomGraph(CustomGraphObject cgo) {
		graphManager.addCustomGraph(cgo);
	}
}
