package gov.anl.gss.sdm.server.model.bridge;

import gov.anl.gss.sdm.server.model.support.Utilities;
import gov.anl.gss.sdm.server.model.translator.Parser;
import gov.anl.gss.sdm.shared.Lever;
import gov.anl.gss.sdm.shared.LoopObject;
import gov.anl.gss.sdm.shared.Output;

import java.io.BufferedWriter;
import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.LinkedHashMap;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

public class ModelStructure {

	private Map<String, Map<Integer, List<String>>> allocatedIndicies = new LinkedHashMap<String, Map<Integer, List<String>>>();
	private Map<String, Set<String>> tailHeadMap;
	private Map<String, Set<String>> headTailMap;
	private Map<String, BridgeModelVariable> variables;

	private BridgeLoopManager loopManager = new BridgeLoopManager();

	private Map<String, List<BridgeArrow>> headArrows = new LinkedHashMap<String, List<BridgeArrow>>();
	private Map<String, List<BridgeArrow>> tailArrows = new LinkedHashMap<String, List<BridgeArrow>>();

	private Map<String, Map<String, String>> polarity = new LinkedHashMap<String, Map<String, String>>();
	
	private Map<String, Map<String, String>> arrowThickness = new LinkedHashMap<String, Map<String, String>>();
	private Map<String,String> extendedLoopNames = new LinkedHashMap<String, String>();

	private String modelName;
	private String structureDir;

	public ModelStructure(String modelName, String rootDirectory) {
		this.modelName = modelName;
		structureDir = rootDirectory +  "/structure/";

		loadSubscriptMap(structureDir + "subscriptSpace.tab");
		loadBridgeArrows(structureDir + "arrowPolarityFromEquations.tab");
		
		loadExtendedLoopNames(structureDir + "extendedLoopNames.tab");
		loadArrowThickness(structureDir + "arrowThickness.tab");
		
		loadLoops(structureDir + "loopsNoIVV.tab");
		headTailMap = loadMap(structureDir + "headTailMap.tab");
		tailHeadMap = loadMap(structureDir + "tailHeadMap.tab");
		variables = loadVariables(structureDir + "variables.tab");
	}
	
	public void saveModifiedLoopData(LoopObject lo) {
		// update Structure and save
		// first was name extended?
		String name = lo.getName();
		if (lo.getExtendedName() != null) {
			String extendedName = lo.getExtendedName();
			String loopID = name;
			extendedLoopNames.put(loopID, extendedName);
			saveExtendedLoopNames(structureDir + "extendedLoopNames.tab");
		}
		
		boolean saveThick = false;
		// first link -- this is from the variable to the first intermediate
		String origThick = arrowThickness.get(lo.getVariable()).get(lo.getIntermediates().get(0));
		String newThick = lo.getLinkThickness().get(0);
		if (!newThick.equals(origThick)) {
			saveThick = true;
			arrowThickness.get(lo.getVariable()).put(lo.getIntermediates().get(0), newThick);
		}
		// these are the links from intermediate to intermediate
		for (int i = 0; i < lo.getIntermediates().size()-1; i++) {
			origThick = arrowThickness.get(lo.getIntermediates().get(i)).get(lo.getIntermediates().get(i+1));
			newThick = lo.getLinkThickness().get(i+1);  // needs to be offset by one to account for the 
			if (!newThick.equals(origThick)) {
				saveThick = true;
				arrowThickness.get(lo.getIntermediates().get(i)).put(lo.getIntermediates().get(i+1), newThick);
			}
			
		}
		
		// this is the last intermediate to the variable
		String lastIntermediate = lo.getIntermediates().get(lo.getIntermediates().size()-1);
		
		origThick = arrowThickness.get(lastIntermediate).get(lo.getVariable());
		newThick = lo.getLinkThickness().get(lo.getLinkThickness().size()-2);  // for some reason, we have
		if (!newThick.equals(origThick)) {
			saveThick = true;
			arrowThickness.get(lastIntermediate).put(lo.getVariable(), newThick);
		}
		if (saveThick) {
			saveArrowThickness(structureDir + "arrowThickness.tab");
		}
	}
	
	public void saveExtendedLoopNames(String file) {
		BufferedWriter bw = Utilities.openFileForWriting(file);

		try {
			for (String v1 : extendedLoopNames.keySet()) {					
				bw.append(v1+"\t"+extendedLoopNames.get(v1)+"\n");
			}
			bw.close();
		} catch (IOException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}

	}
	
	public void loadExtendedLoopNames(String file) {
		
		if (Utilities.fileExists(file)) {
			for (String line : Utilities.getAsList(file)) {
				String[] fields = line.split("\t", 2);
				extendedLoopNames.put(fields[0], fields[1]);
			}
			
		}
		loopManager.setExtendedLoopNames(extendedLoopNames);
	}
	
	public void saveArrowThickness(String file) {
		BufferedWriter bw = Utilities.openFileForWriting(file);
		try {
			for (String v1 : arrowThickness.keySet()) {
				Map<String, String> map = arrowThickness.get(v1);
					for (String v2 : map.keySet()) {
						bw.append(v1+"\t"+v2+"\t"+map.get(v2)+"\n");
					}
				}
			bw.close();
		} catch (IOException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
	}
	
	public void loadArrowThickness(String file) {
		if (Utilities.fileExists(file)) {
			for (String line : Utilities.getAsList(file)) {
				String[] fields = line.split("\t");
				if (!arrowThickness.containsKey(fields[0])) {
					arrowThickness.put(fields[0], new LinkedHashMap<String, String>());
				}
				arrowThickness.get(fields[0]).put(fields[1], fields[2]);
			}
			
		}
		loopManager.setArrowThickness(arrowThickness);
	}

	public void loadLoops(String file) {
		loopManager.loadLoops(file, polarity);
	}


	public void loadBridgeArrows(String file) {
		List<String> mapInfo = Utilities.getAsList(file);
		for (String line : mapInfo) {
			String[] p = line.split("\t");
			BridgeArrow ba = new BridgeArrow(p[0], p[1], p[2]);
			if (!polarity.containsKey(p[0])) {
				polarity.put(p[0], new LinkedHashMap<String, String>());
			}
			polarity.get(p[0]).put(p[1], p[2]);

			addArrowToMap(ba.getTail(), ba, tailArrows);
			addArrowToMap(ba.getHead(), ba, headArrows);
		}
	}

	public List<BridgeArrow> getArrowsWithTail(String variable) {
		return tailArrows.get(variable);
	}

	public List<BridgeArrow> getArrowsWithHead(String variable) {
		return headArrows.get(variable);
	}

	public List<LoopObject> getLoops() {

		List<LoopObject> lo = new ArrayList<LoopObject>();
		for (BridgeLoop bl : loopManager.getLoops()) {
			lo.add(bl.asLoopObject());
		}
		return lo;
	}

	public List<LoopObject> getLoopsWithVariable(String variable) {
		List<LoopObject> al = new ArrayList<LoopObject>();
		for (BridgeLoop bl : loopManager.getLoopsWithVariable(variable)) {
			al.add(bl.getLoopStartingAt(variable).asLoopObject());
		}

		return al;
	}

	public List<LoopObject> getLoopsOfLength(int length) {
		List<LoopObject> al = new ArrayList<LoopObject>();
		for (BridgeLoop bl : loopManager.getLoopsOfLength(length)) {
			al.add(bl.asLoopObject());
		}
		return al;
	}

	public void addArrowToMap(String key, BridgeArrow ba, Map<String, List<BridgeArrow>> map) {
		if (!map.containsKey(key))
			map.put(key,  new ArrayList<BridgeArrow>());
		map.get(key).add(ba);
	}

	public Map<String, BridgeModelVariable> loadVariables(String file) {
		Map<String, BridgeModelVariable> map = new LinkedHashMap<String, BridgeModelVariable>();
		List<String> mapInfo = Utilities.getAsList(file);

		for (String line : mapInfo) {
			String[] p = line.split("\t");
			map.put(p[0], new BridgeModelVariable(p[0], p[1], p[2], p[3], p[4], p[5], p[6]));
		}

		return map;
	}

	public void loadSubscriptMap(String file) {
		File f = new File(file);
		if (!f.exists())
			return;
		List<String> mapInfo = Utilities.getAsList(file);
		mapInfo.remove(0);
		for (String line : mapInfo) {
			String[] p = line.split("\t");
			if (!allocatedIndicies.containsKey(p[0])) {
				allocatedIndicies.put(p[0], new LinkedHashMap<Integer, List<String>>());
			}
			Map<Integer, List<String>> array = allocatedIndicies.get(p[0]);
			int subscriptIndex = Integer.parseInt(p[1]);
			if (!array.containsKey(subscriptIndex)) {
				array.put(subscriptIndex, new ArrayList<String>());
			}
			List<String> subs = array.get(subscriptIndex);
			subs.add(p[2]);
		}
	}

	public Map<String, Set<String>> loadMap(String file) {
		Map<String, Set<String>> map = new LinkedHashMap<String, Set<String>>();
		List<String> mapInfo = Utilities.getAsList(file);

		for (String line : mapInfo) {
			String[] p = line.split("\t");
			map.put(p[0], new LinkedHashSet<String>());
			for (int i = 1; i < p.length; i++)
				map.get(p[0]).add(p[i]);
		}

		return map;
	}
	
	public Set<Output> getModelOutputVariables() {
		int position = 0;
		Set<Output> outputs = new LinkedHashSet<Output>();
		for (BridgeModelVariable bmv : getVariables().values()) {
			if (!bmv.isLever()) {
				outputs.add(bmv.asOutput(position++));
			}
		}
		
		return outputs;
		
	}
	
	public Set<Lever> getModelLeverVariables() {
		int position = 0;
		Set<Lever> levers = new LinkedHashSet<Lever>();
		for (BridgeModelVariable bmv : getVariables().values()) {
			if (bmv.isLever()) {
				levers.add(bmv.asLever(position++));
			}
		}
		
		return levers;
		
	}

	/**
	 * @return the allocatedIndicies
	 */
	public Map<String, Map<Integer, List<String>>> getAllocatedIndicies() {
		return allocatedIndicies;
	}

	/**
	 * @return the tailHeadMap
	 */
	public Map<String, Set<String>> getTailHeadMap() {
		return tailHeadMap;
	}

	/**
	 * @return the headTailMap
	 */
	public Map<String, Set<String>> getHeadTailMap() {
		return headTailMap;
	}

	/**
	 * @return the variables
	 */
	public Map<String, BridgeModelVariable> getVariables() {
		return variables;
	}

	/**
	 * @return the modelName
	 */
	public String getModelName() {
		return modelName;
	}

	/**
	 * @return the polarity
	 */
	public Map<String, Map<String, String>> getPolarity() {
		return polarity;
	}
	
	public String getPolarity(String tail, String head) {
		return polarity.get(tail).get(head);
	}

	public BridgeLoopManager getLoopManager() {
		return loopManager;
	}

	public Map<String, List<BridgeArrow>> getHeadArrows() {
		return headArrows;
	}

	public Map<String, List<BridgeArrow>> getTailArrows() {
		return tailArrows;
	}

	public Map<String, Map<String, String>> getArrowThickness() {
		return arrowThickness;
	}

	public Map<String, String> getExtendedLoopNames() {
		return extendedLoopNames;
	}

	public String getStructureDir() {
		return structureDir;
	}
	
	public boolean isKnownVariable(String var) {
		if (variables.containsKey(var)) {
			return true;
		} else {
			if (var.startsWith("\"") || var.endsWith("\"")) {
				if (variables.containsKey(Parser.removeQuotes(var)))// var.replace("\"", "")))
					return true;
				else 
					return false;
			} else {

				if (variables.containsKey(Parser.quoted(var))) // "\""+var+"\""))
					return true;
				else 
					return false;
			}
		}

	}
}
