package gov.anl.gss.sdm.server.model.bridge;

import gov.anl.gss.sdm.server.model.documentor.gui.ImportSDMDoc;
import gov.anl.gss.sdm.server.model.documentor.gui.StandaloneGUI;
import gov.anl.gss.sdm.server.model.documentor.xmlwriter.Anchors;
import gov.anl.gss.sdm.server.model.support.MutableInteger;
import gov.anl.gss.sdm.server.model.support.Utilities;
import gov.anl.gss.sdm.server.model.support.database.ModelRunDatabaseVerifier;
import gov.anl.gss.sdm.server.model.tools.FileOrganizer;
import gov.anl.gss.sdm.server.model.tools.MDL;
import gov.anl.gss.sdm.server.model.tools.MDLProcessor;
import gov.anl.gss.sdm.server.model.translator.InformationManagers;
import gov.anl.gss.sdm.shared.AnnotatedModelName;
import gov.anl.gss.sdm.shared.CustomGraphObject;
import gov.anl.gss.sdm.shared.Lever;
import gov.anl.gss.sdm.shared.LoopObject;
import gov.anl.gss.sdm.shared.Node;
import gov.anl.gss.sdm.shared.Output;
import gov.anl.gss.sdm.shared.RunObject;
import gov.anl.gss.sdm.shared.ViewObject;

import java.awt.Desktop;
import java.io.BufferedReader;
import java.io.File;
import java.io.IOException;
import java.lang.reflect.Constructor;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.net.MalformedURLException;
import java.net.URI;
import java.net.URISyntaxException;
import java.net.URL;
import java.net.URLClassLoader;
import java.util.ArrayList;
import java.util.LinkedHashMap;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.apache.commons.io.filefilter.DirectoryFileFilter;
import org.apache.commons.io.filefilter.SuffixFileFilter;

public class SDMBridge implements Bridge {

	public static final String MDL_EXTENSION = ".mdl";
	public static final String VIGNETTES_DIR = "vignettes";
	public static final String VIGNETTE_EXTENSION = ".tab";
	public static final String RUNS_DIR = "runs";
	public static final String RUN_EXTENSION = "HSQLDB.data"; // ".tab";
	public static final String HSQLDB_EXTENSION = "HSQLDB";
	public static final String STRUCTURE = "structure";

	public static final String CUSTOM_GRAPH_TAG = "_customGraphs.tab";
	
	private String currentRunName = "";

	/**
	 * All results that are available to the SDM-UI
	 */
	private Map<String, BridgeModelResultsDatabaseImplementation> resultMap = new LinkedHashMap<String, BridgeModelResultsDatabaseImplementation>();

	private Map<String, BridgeRunValidator> validatedRuns = new LinkedHashMap<String, BridgeRunValidator>();

	private ModelStructure modelStructure = null;
	private BridgeModelState modelState;
	private BridgeModelMeta meta;
	private boolean checkedRuns = false;

	/**
	 * This is the executable form of the SD Model
	 */

	private SDMUIExecutable sdmuiExecutable;

	private MDL theSelectedModel;

	private BridgeVignetteObject theSelectedVignette;

	private String rootDirectory;
	private String modelDirectory;
	private String outputDirectory;

	private FileOrganizer fileOrganizer;

	private Set<String> selectedVariables = new LinkedHashSet<String>();

	private Map<String, Map<Integer, List<String>>> allocatedIndicies = new LinkedHashMap<String, Map<Integer, List<String>>>();

	public SDMBridge(String rootDirectory) {
		// this is an SDM-UI only constructory

		this(new FileOrganizer(true, rootDirectory, ""));
		System.out.println("New Bridge SDM-UI Only " + rootDirectory);
		System.out.println("Using SDM-Doc Version "+StandaloneGUI.version);

	}

	public SDMBridge(FileOrganizer fileOrganizer) {

		this.fileOrganizer = fileOrganizer;
		resultMap = new LinkedHashMap<String, BridgeModelResultsDatabaseImplementation>();
		this.rootDirectory = fileOrganizer.getRootDirectory();
		System.out.println("NEW BRIDGE: " + this.rootDirectory);
		System.out.println("Using SDM-Doc Version "+StandaloneGUI.version);
		modelState = new BridgeModelState(this);
	}
	
	public boolean displaySDMDocInBrowserAtVariable(String variable) {
		String file = fileOrganizer.getModelDirectory()+"/"+meta.getModelName()+"_English.html#"+Anchors.createAnchorName(variable);
		return displayInBrowser(file);
	}
	
	public boolean displaySDMDocInBrowser() {
		String file = fileOrganizer.getModelDirectory()+"/"+meta.getModelName()+"_English.html";
		return displayInBrowser(file);
	}
	
	public boolean runSDMDoc() {
		String file = fileOrganizer.getModelDirectory()+"/"+meta.getMdlFile();
		StandaloneGUI.viaSDMUI(file);
		return true;
	}
	
	public boolean importSDMDoc() {
		String file = fileOrganizer.getModelDirectory()+"/"+meta.getMdlFile();
		ImportSDMDoc.viaSDMUI(meta.getModelName(), file);
		return true;
	}
	
	public boolean displayInBrowser(String file) {
		if (Desktop.isDesktopSupported()) {
			try {
				
				URI uri;
				
				String[] parts = file.split("#");
				if (parts.length > 1) {
					uri = new URI("file", parts[0], parts[1]);
				} else {
					uri = new URI("file", parts[0], null);
				}
				System.out.println(uri.getScheme()+" "+uri.getSchemeSpecificPart()+" "+uri.getFragment()+" "+uri.toASCIIString());
				Desktop.getDesktop().browse(uri);
				// Desktop.getDesktop().browse(new
				// URI("file:///"+out));
			} catch (IOException e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			} catch (URISyntaxException e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			}
		}
		return true;
	}

	@Override
	public Boolean done() {
		
		// set up a lock file for the database shutdown activity
		
		File closeLock = new File(fileOrganizer.getRunsDirectory()+"/CLOSELOCK.lck");
		try {
			boolean canLock = closeLock.createNewFile();
			if (!canLock) {
				System.out.println("close lock troubles");
			}
		} catch (IOException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
		
		for (BridgeModelResultsDatabaseImplementation bmrdi : resultMap.values()) {
			bmrdi.shutdowndatabase();
		}
		if (modelState.canDone()) {
			if (sdmuiExecutable != null) {
				if (!meta.isMap()) {
					sdmuiExecutable.done();
				}
			}
			if (modelState != null) {
				modelState.setModelRunComplete(true);
			}
			boolean canDelete = closeLock.delete();
			if (!canDelete)
				System.out.println("delete close lock troubles");
			return true;
		} else {
			return false;
		}
	}
	
	public void ping() {
		System.out.println("SDMBridge: ping!");
	}

	public void importModel(String mdlFile, String objectName, String systemRoot) {
		
		try {
			InformationManagers.getInstance().getModelManager().setFileOrganizer(fileOrganizer);
			String className = Utilities.convertModelNameToClassName(mdlFile);
			MDLProcessor mdlProc = new MDLProcessor(mdlFile, mdlFile, className, systemRoot);
			mdlProc.process(true);
		} catch (Exception e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
	}

	public String makeLegal(String name) {
		return name.replace(" ", "_").replace("/", "_").replace(":", "_")
				.replace("-", "_").replace("&", "_").replace(".", "_")
				.replace("(", "_").replace(")", "_").replace("(", "_")
				.replace(")", "_").replace(".", "_").replace(",", "_")
				.replace("'", "").replace("*", "_");
	}

	@Override
	public Boolean stepModel() {
		modelState.setModelRunning(true);
		if (modelState.canStepModel()) {
			
			if (sdmuiExecutable.isReinitialize())
				sdmuiExecutable.reinitializeRun();

			sdmuiExecutable.stepMap();
			modelState.stepTime();
			return true;
		} else {
			return false;
		}
	}

	@Override
	public Boolean backstepModel() {

		if (modelState.canBackstepModel()) {
			sdmuiExecutable.backstepMap();
			modelState.backstepTime();
			return true;
		} else {
			return false;
		}
	}

	@Override
	public Boolean backstepModelToTime(double time) {
		sdmuiExecutable.backstepExtendedMemory(time);
		modelState.setCurrentTime(time);
		return true;
//		if (time > 0.0) {
//			while (modelState.getCurrentTime() > time) {
//				backstepModel();
//			}
//			return true;
//		} else {
//			return false;
//		}
	}

	@Override
	public Set<Lever> getModelLeverVariables() {
		if (modelState.canGetModelLeverVariables()) {
			// if no selected vignette (i.e. a map) get
			// variables from structure
			if (theSelectedVignette.getName() == null) {
				return modelStructure.getModelLeverVariables();
			} else {
				return theSelectedVignette.getLevers();
			}
		} else {
			return new LinkedHashSet<Lever>();
		}
	}

	@Override
	public Set<Output> getModelOutputVariables() {
		if (modelState.canGetModelOutputVariables()) {
			if (theSelectedVignette.getName() == null) {
				return modelStructure.getModelOutputVariables();
			} else {
				return theSelectedVignette.getOutputs();
			}
		} else {
			return new LinkedHashSet<Output>();
		}
	}

	@Override
	public Double getTimeStep() {
		if (modelState.canGetTimeStep())
			return Double.parseDouble(meta.getDt());
		else
			return null;

	}

	@Override
	public Double getDT() {
		if (modelState.canGetDT())
			return Double.parseDouble(meta.getDt());
		else
			return null;

	}

	@Override
	public Double getCurrentTime() {
		if (modelState.canGetCurrentTime())
			return modelState.getCurrentTime();
		else
			return null;
	}

	@Override
	public String getTimeUnits() {
		if (modelState.canGetTimeUnits())
			return meta.getTimeUnits();
		else
			return null;

	}

	@Override
	public List<Double> getVariableValues(String scenario, String variable) {
		if (modelState.canGetVariableValues(scenario, variable)) {
			if (resultMap.containsKey(scenario)) {
				return resultMap.get(scenario).getValues(variable);
			} else {
				return new ArrayList<Double>();
			}
		} else {
			return null;
		}
	}

	@Override
	public Double getVariableValue(String scenario, String variable, double time) {
		if (modelState.canGetVariableValue(scenario, variable, time)) {
			if (resultMap.containsKey(scenario)) {
				return resultMap.get(scenario).getValue(variable, time);
			} else {
				return null;
			}
		} else {
			return null;
		}
	}

	@Override
	public Map<String, Double> getAllVariableValues(String scenario,
			double time, Set<String> selectedVariables) {
		if (modelState.canGetAllVariableValues(scenario, time,
				selectedVariables)) {
			if (resultMap.containsKey(scenario)) {
				return resultMap.get(scenario).getAllVariableValues(time,
						selectedVariables);
			} else {
				return null;
			}
		} else {
			return null;
		}
	}

	@Override
	public Double getStartTime() {
		if (modelState.canGetStartTime()) {
			return new Double(meta.getInitialTime());
		} else {
			return null;
		}
	}

	@Override
	public Double getEndTime() {
		if (modelState.canGetEndTime()) {
			return new Double(meta.getFinalTime());
		} else {
			return null;
		}
	}

	@Override
	public List<String> getAvailableModels() {

		// InformationManagers.getInstance().getModelManager().getGeneratedOutputDirectoryName();
		String d = fileOrganizer.getModelsDirectory();

		System.out.println("FO: models dir: " + d);

		File dir = new File(d);
		System.out.println("Available models: " + dir.getAbsolutePath());
		// if (modelState.canGetAvailableModels(d)) {

		// String[] annotated = modelAnnotation(rootDirectory,
		// dir.list(DirectoryFileFilter.INSTANCE));
		String[] annotated = modelAnnotation(InformationManagers.getInstance()
				.getModelManager().getGeneratedOutputDirectoryName(),
				dir.list(DirectoryFileFilter.INSTANCE));

		System.out.println(annotated);

		return asList(annotated);
		// } else {
		// return new ArrayList<String>();
		// }
	}

	private String[] modelAnnotation(String rootDirectory, String[] models) {
		String[] annotated = new String[models.length];
		int i = 0;
		for (String model : models)
			annotated[i++] = modelAnnotation(rootDirectory, model);
		return annotated;
	}

	private String modelAnnotation(String rootDirectory, String model) {
		System.out.println("modelAnnotation");
		fileOrganizer.updateForModel(model);
		fileOrganizer.print();
		BridgeModelMeta meta = new BridgeModelMeta(
				fileOrganizer.getStructureDirectory(), model);
		meta.load();

		return model
				+ " "
				+ (meta.isMap() ? AnnotatedModelName.MAP
						: AnnotatedModelName.MODEL);
	}

	@Override
	public Boolean selectModel(String modelName) {
		String model = AnnotatedModelName.removeAnnotation(modelName);
		if (modelState.canSelectModel()) {

			fileOrganizer.updateForModel(model);
			modelState = new BridgeModelState(this);
			modelState.setSelectedModel(model);
			modelDirectory = fileOrganizer.getModelDirectory(); // InformationManagers.getInstance().getModelManager().getModelDirectory();
			outputDirectory = fileOrganizer.getDocGraphDirectory(); // InformationManagers.getInstance().getModelManager().getGeneratedOutputDirectoryName();

			System.out.println("SelectModel");
			fileOrganizer.print();

			meta = new BridgeModelMeta(fileOrganizer.getStructureDirectory(),
					model);
			meta.load();
			modelState.extractFromMeta(meta);

			theSelectedModel = new MDL(fileOrganizer.getStructureDirectory()
					+ "/" + meta.getMdlFile());
			resultMap.clear();
			validatedRuns.clear();
			selectedVariables.clear();
			checkedRuns = false;

//			modelStructure = new ModelStructure(model, outputDirectory);

			if (modelStructure != null) {
				System.out.println("MJB modelStructure initialized "
						+ modelStructure.toString());
			} else {
				System.out.println("modelStructure NULL! ");
			}

			modelState.setModelSelected(true);
			modelState.setMetaLoaded(true);
			modelState.setStructureLoaded(true);

			modelState
					.setCurrentTime(Double.parseDouble(meta.getInitialTime()));
			modelState
					.setInitialTime(Double.parseDouble(meta.getInitialTime()));
			modelState.setFinalTime(Double.parseDouble(meta.getFinalTime()));
			modelState.setDt(Double.parseDouble(meta.getDt()));

			modelState.setModelRunnable(meta.isRunnable());
			
			ModelRunDatabaseVerifier verifier = new ModelRunDatabaseVerifier(fileOrganizer.getRunsDirectory());
			if (!verifier.isClean()) {
				verifier.restore();
				System.out.println("Need to restore runs database directory");
			}

			return true;
		} else {
			return null;
		}
	}

	@Override
	public List<String> getVignettes() {
		// String dir =
		// rootDirectory+"/"+modelState.getSelectedModel()+"/"+VIGNETTES_DIR;

		System.out.println("Vignettes");
		fileOrganizer.print();
		String dir = fileOrganizer.getVignettesDirectory();
		if (modelState.canGetVignettes(dir)) {

			File vig = new File(dir);
			String[] files = vig.list(new SuffixFileFilter(VIGNETTE_EXTENSION));
			List<String> list = new ArrayList<String>();
			for (String file : files) {
				if (file.endsWith(CUSTOM_GRAPH_TAG))
					continue;
				list.add(file.replace(VIGNETTE_EXTENSION, ""));
			}
			return list;
		} else {
			return new ArrayList<String>();
		}
	}
	
	public List<String> getVignettes(String forModel) {
		// String dir =
		// rootDirectory+"/"+modelState.getSelectedModel()+"/"+VIGNETTES_DIR;
		
		String currentFOModel = fileOrganizer.getModelname();
		fileOrganizer.updateForModel(forModel);

		
		String dir = fileOrganizer.getVignettesDirectory();
		if (modelState.canGetVignettes(dir)) {

			File vig = new File(dir);
			String[] files = vig.list(new SuffixFileFilter(VIGNETTE_EXTENSION));
			List<String> list = new ArrayList<String>();
			for (String file : files) {
				if (file.endsWith(CUSTOM_GRAPH_TAG))
					continue;
				list.add(file.replace(VIGNETTE_EXTENSION, ""));
			}
			fileOrganizer.updateForModel(currentFOModel);
			return list;
		} else {
			fileOrganizer.updateForModel(currentFOModel);
			return new ArrayList<String>();
		}
	}

	@Override
	public List<RunObject> getRuns() {
		List<RunObject> list = new ArrayList<RunObject>();
		
		String dir = fileOrganizer.getRunsDirectory();
		if (modelState.canGetRuns(dir)) {
			File vig = new File(dir);
			String[] files = vig.list(/* new SuffixFileFilter(RUN_EXTENSION) */);

			for (String file : files) {
				if (!file.endsWith(RUN_EXTENSION))
					continue;
				String runName = file.replace(RUN_EXTENSION, "");
				System.out.println("Checking for complete run "+runName);
				 if (validateRun(runName)) {
					list.add(new RunObject(runName, false));
				 } else {
					 
					 boolean success = ModelRunDatabaseVerifier.removeDB(fileOrganizer.getRunsDirectory(), runName);
					 System.out.println("Run invalid -- "+runName+" successful removal = "+success);
				 }
			}
			checkedRuns = true;
			return list;
		} else {
			return list;
		}
		// }
	}

	@Override
	public Boolean selectRuns(Set<String> runs) {
		if (modelState.canSelectRun()) {
			modelState.setSelectedRuns(runs);
			loadRunResults();
			return true;
		} else {
			return false;
		}
	}

	public void addSelectedRun(String run) {
		modelState.getSelectedRuns().add(run);
		loadRunResults(run);
	}

	public void removeSelectedRun(String run) {
		modelState.getSelectedRuns().remove(run);
		resultMap.remove(run);
	}

	@Override
	public Boolean createNewRun(String runName) {

		System.out.println("Creating New Run: " + runName);

		if (modelState.canCreateRun()) {
			modelState.setModelRunName(runName);
			modelState.newRun();
			currentRunName = runName;
			return run();
		} else {
			return false;
		}
	}

	@Override
	public Boolean run() {
		if (!modelState.canRun())
			return false;
		// this is where we instantiate
		// sdModelClass = new SystemDynamicsModel(selectedModel);
		// return false;

		// set local currentTime to our start time
		modelState.setCurrentTime(getStartTime());

		ClassLoader cl = ClassLoader.getSystemClassLoader();
		URL[] urls = ((URLClassLoader) cl).getURLs();
//		for (URL url : urls) {
//			System.out.println(url.getFile());
//		}

		try {

			try {
				File f = new File(fileOrganizer.getDynamicDirectory()
						+ "/"+FileOrganizer.getSrcClassDirectory());
				URL u = f.toURL();
				URLClassLoader urlClassLoader = (URLClassLoader) ClassLoader
						.getSystemClassLoader();
				Class urlClass = URLClassLoader.class;
				Method method = urlClass.getDeclaredMethod("addURL",
						new Class[] { URL.class });
				method.setAccessible(true);
				method.invoke(urlClassLoader, new Object[] { u });
			} catch (MalformedURLException e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			}

			// System.out.println("Class: "+"gov.anl.gss.sdm.server.simulation.models."+modelState.getSelectedModel());
			System.out.println("Class: "
					+ "gov.anl.gss.sdm.server.simulation.models."
					+ meta.getClassName());
			URLClassLoader loader = (URLClassLoader) ClassLoader
					.getSystemClassLoader();
			System.out.println(System.getProperty("user.dir"));
//			for (URL url : loader.getURLs()) {
//				System.out.println("URL: " + url.getPath());
//			}

			String[] args = new String[3];
			args[0] = fileOrganizer.getMapperFile();
			args[1] = fileOrganizer.getRunsDirectory() + "/"
					+ modelState.getModelRunName() + "HSQLDB";
			args[2] = meta.getDt();

			Class<?> clazz = Class
					.forName("gov.anl.gss.sdm.server.simulation.models."
							+ meta.getClassName());
			Constructor<?> constructor = clazz.getConstructor(String.class,
					String[].class);
			sdmuiExecutable = (SDMUIExecutable) constructor.newInstance(
					modelState.getSelectedModel(), args);
		} catch (ClassNotFoundException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
			return false;
		} catch (NoSuchMethodException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
			return false;
		} catch (SecurityException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		} catch (InstantiationException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
			return false;
		} catch (IllegalAccessException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
			return false;
		} catch (IllegalArgumentException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
			return false;
		} catch (InvocationTargetException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
			return false;
		}
		sdmuiExecutable.sdmuiExecuteStep1(modelState.getModelRunName(),
				fileOrganizer.getRunsDirectory());

		modelState.setModelInstantiated(true);

		return true;

	}

	@Override
	public List<ViewObject> getAvailableViewNames() {
		if (modelState.canGetAvailableViewNames()) {
			List<ViewObject> views = new ArrayList<ViewObject>();
			for (String v : theSelectedModel.getViewNames())
				views.add(new ViewObject(v, false));
			return views;
		} else {
			return new ArrayList<ViewObject>();
		}
	}

	@Override
	public List<String> getView(String name) {
		if (modelState.canGetView(name)) {
			return theSelectedModel.getView(name);
		} else {
			return null;
		}
	}

	public String[] getViewAsArray(String name) {
		List<String> al = getView(name);
		if (al == null)
			return null;
		String[] a = new String[al.size()];
		for (int i = 0; i < a.length; i++)
			a[i] = al.get(i);
		return a;
	}

	@Override
	public Boolean setValue(String variableName, double value) {
		if (modelState.canSetValue(variableName, value)) {
			sdmuiExecutable.setViaSDMUI(variableName, value);
			return true;
		} else {
			return false;
		}
	}

	public double getCurrentValue(String variable) {

		double cv = sdmuiExecutable.get(variable);
		return cv;
	}

	@Override
	public double[] getCurrentValues(String[] variables) {
		if (modelState.canGetCurrentValues(variables)) {
			double[] cv = new double[variables.length];
			for (int i = 0; i < variables.length; i++)
				cv[i] = sdmuiExecutable.get(variables[i]);
			return cv;
		} else {
			return null;
		}
	}

	@Override
	public double[][] getAllValues(String runName, String[] vars) {
		double[][] cv = new double[vars.length][(int) (1 / (Double
				.parseDouble(meta.getSaveper())))];
		if (resultMap.containsKey(runName)) {
			BridgeModelResultsDatabaseImplementation bmr = resultMap
					.get(runName);
			for (int i = 0; i < vars.length; i++) {
				cv[i] = bmr.getValuesAsArray(vars[i]);
			}
		}
		return cv;
	}

	@Override
	public double[][] getValues(String[] variables) {
		if (modelState.canGetValues(variables)) {
			double[][] cv = new double[variables.length][(int) (1 / (Double
					.parseDouble(meta.getSaveper())))];
			for (int i = 0; i < variables.length; i++)
				cv[i] = sdmuiExecutable.getForTimeStep(variables[i]);
			return cv;
		} else {
			return null;
		}
	}

	@Override
	public double[] getValues(String[] variables, double time) {
		
		if (modelState.canGetValues(variables, time)) {
			return sdmuiExecutable.get(variables, time);
//			double[] cv = new double[variables.length];
//			for (int i = 0; i < variables.length; i++)
//				cv[i] = sdmuiExecutable.get(variables[i], time);
//			return cv;
		} else {
			return null;
		}
	}
	
	public double[][] getValues(String runName, String[] variables) {

		BridgeModelResultsDatabaseImplementation bmrdi = new BridgeModelResultsDatabaseImplementation(
				buildRunOutputFileName(runName));
		if (bmrdi.isOk()) {
			resultMap.put(runName,bmrdi);
		}

		BridgeModelResultsDatabaseImplementation bmr = resultMap
				.get(runName);

		double[][] v = bmr.getValues(variables);

		return v;
	}

	@Override
	public double[] getValues(String runName, String[] variable, double time) {
		if (modelState.canGetValues(runName, variable, time)) {
			BridgeModelResultsDatabaseImplementation bmr = resultMap
					.get(runName);
			double[] v = new double[variable.length];
//			for (int i = 0; i < variable.length; i++)
//				v[i] = bmr.getValue(variable[i], time);
			
			v = bmr.getValue(variable, time);
			
			return v;
		} else {
			return null;
		}
	}

	// @Override
	// public Set<BridgeVariable> getVignetteVariables() {
	// if (modelState.canGetVignetteVariables()) {
	// Set<Lever> s = theSelectedVignette.getLevers();
	// s.addAll(theSelectedVignette.getOutputs());
	// return s;
	// } else {
	// return null;
	// }
	// }

	private List<String> asList(String[] array) {
		List<String> list = new ArrayList<String>();
		for (String element : array)
			list.add(element);
		return list;
	}

	@Override
	public Boolean createNewVignette(String vignette) {
		if (modelState.canCreateNewVignette(vignette)) {
			theSelectedVignette = new BridgeVignetteObject(
					modelState.getSelectedModel(), modelStructure, vignette,
					outputDirectory);
			theSelectedVignette.create();
			theSelectedVignette.save();
			selectedVariables.clear();
			modelState.setSelectedVignette(vignette);
			return true;
		} else {
			return false;
		}
	}

	@Override
	public Boolean selectVignette(String vignette) {
		if (vignette == null)
			return false;
		if (modelState.canSelectVignette(vignette)) {
			modelState.setSelectedVignette(vignette);
			theSelectedVignette = new BridgeVignetteObject(
					modelState.getSelectedModel(), modelStructure, vignette,
					outputDirectory);
			theSelectedVignette.load();
			selectedVariables.clear();
			return true;
		} else {
			return false;
		}
	}

	@Override
	public Map<String, List<Double>> getAllVariableValues(String run) {
		if (modelState.canGetAllVariableValues(run)) {
			BridgeModelResultsDatabaseImplementation bmr = resultMap
					.get(modelState.getModelRunName());
			return bmr.getAllVariableValues();
		} else {
			return null;
		}
	}

	// private boolean validateRuns() {
	// boolean valid = true;
	// for (String run : modelState.getSelectedRuns()) {
	// valid = valid && validateRun(run);
	// }
	// return valid;
	// }

	private void loadRunResults() {

		int timeout = 0;
		// this will open databases. We must make sure that we are not trying to close these
		// at this time
		File closeLock = new File(fileOrganizer.getRunsDirectory()+"/CLOSELOCK.lck");
		boolean okToProceed = false;
		while(timeout < 5 && !okToProceed) {
			if (closeLock.exists()) {
				timeout++;
				try {
					Thread.sleep(1000);
				} catch (InterruptedException e) {
					// TODO Auto-generated catch block
					e.printStackTrace();
				}
			} else {
				okToProceed = true;
			}
		}
		if (!okToProceed) {
			System.out.println("load run results troubles lock file");
		}

		for (String run : modelState.getSelectedRuns()) {
			if (!resultMap.containsKey(run)) {
				BridgeModelResultsDatabaseImplementation bmrdi = new BridgeModelResultsDatabaseImplementation(
						buildRunOutputFileName(run));
				if (bmrdi.isOk()) {
					resultMap.put(run,bmrdi);
				} else {
					System.out.println("Run "+run+" is being discarded");
					bmrdi.shutdowndatabase();
					boolean success = ModelRunDatabaseVerifier.removeDB(fileOrganizer.getRunsDirectory(), run);
					if (!success) {
						System.out.println("Couldn't remove database "+run);
					}
				}
			}
		}
	}

	public void loadRunResults(String run) {
		if (!resultMap.containsKey(run))
			resultMap.put(run, new BridgeModelResultsDatabaseImplementation(
					buildRunOutputFileName(run)));
	}

	public String buildRunOutputFileName(String run) {
		StringBuilder sb = new StringBuilder();

		sb.append(fileOrganizer.getRunsDirectory() + "/");
		sb.append(run);
		sb.append(HSQLDB_EXTENSION);
		return sb.toString();
	}

	/**
	 * @return the resultMap
	 */
	public Map<String, BridgeModelResultsDatabaseImplementation> getResultMap() {
		return resultMap;
	}

	/**
	 * @return the sdModelClass
	 */
	public SDMUIExecutable getSdModelClass() {
		return sdmuiExecutable;
	}

	/**
	 * @return the selectedModel
	 */
	public String getSelectedModel() {
		return modelState.getSelectedModel();
	}

	/**
	 * @return the theSelectedModel
	 */
	public MDL getTheSelectedModel() {
		return theSelectedModel;
	}

	/**
	 * @return the selectedVignette
	 */
	public String getSelectedVignette() {
		return modelState.getSelectedVignette();
	}

	/**
	 * @return the theSelectedVignette
	 */
	public BridgeVignetteObject getTheSelectedVignette() {
		return theSelectedVignette;
	}

	/**
	 * @return the runName
	 */
	public String getRunName() {
		return modelState.getModelRunName();
	}

	/**
	 * @return the selectedRuns
	 */
	public Set<String> getSelectedRuns() {
		return modelState.getSelectedRuns();
	}

	/**
	 * @return the rootDirectory
	 */
	public String getRootDirectory() {
		return rootDirectory;
	}

	/**
	 * @return the selectedVariables
	 */
	public Set<String> getSelectedVariables() {
		return selectedVariables;
	}

	public void loadSubscriptMap(BufferedReader br) {
		List<String> mapInfo = Utilities.getAsList(br);
		mapInfo.remove(0);
		for (String line : mapInfo) {
			String[] p = line.split("\t");
			if (!allocatedIndicies.containsKey(p[0])) {
				allocatedIndicies.put(p[0],
						new LinkedHashMap<Integer, List<String>>());
			}
			Map<Integer, List<String>> array = allocatedIndicies.get(p[0]);
			int subscriptIndex = Integer.parseInt(p[1]);
			if (!array.containsKey(subscriptIndex)) {
				array.put(subscriptIndex, new ArrayList<String>());
			}
			List<String> subs = array.get(subscriptIndex);
			subs.add(p[2]);
		}
	}

	public Node getUsesTree(String variable, int numLevels) {

		// specified variable is at the head of the arrow
		// get entries in headTailMap for this variable for a single level

		Node root = new Node(variable, 0);
		Map<String, MutableInteger> processed = new LinkedHashMap<String, MutableInteger>();
		processed.put(variable, new MutableInteger(1));
		if (modelStructure == null) {
			System.out.println("WTF! modelStructure " + variable);
		} else {
			extendTree(modelStructure.getTailHeadMap(), root, 1, numLevels,
					processed, false);
		}
		return root;

	}

	public Node getCausesTree(String variable, int numLevels) {

		// specified variable is at the head of the arrow
		// get entries in headTailMap for this variable for a single level

		Node root = new Node(variable, 0);
		Map<String, MutableInteger> processed = new LinkedHashMap<String, MutableInteger>();
		processed.put(variable, new MutableInteger(1));
		if (modelStructure == null) {
			System.out.println("WTF! modelStructure " + variable);
		} else {
			extendTree(modelStructure.getHeadTailMap(), root, 1, numLevels,
					processed, true);
		}
		return root;

	}

	public void extendTree(Map<String, Set<String>> map, Node node, int level,
			int numLevels, Map<String, MutableInteger> processed,
			boolean reverseDirectionForPolarity) {
		if (level > numLevels)
			return;
		if (!map.containsKey(node.getVariable())) {
			return;
		}

		// extend one level
		for (String var : map.get(node.getVariable())) {
			String polarity = reverseDirectionForPolarity ? getPolarity(var,
					node.getVariable()) : getPolarity(node.getVariable(), var);
			node.addVertex(new Node(var, level), polarity);
			if (!processed.containsKey(var))
				processed.put(var, new MutableInteger(0));
			processed.get(var).add(1);
		}

		for (Node aNode : node.getVertices()) {
			if (processed.get(aNode.getVariable()).value() < 2) {
				extendTree(map, aNode, level + 1, numLevels, processed,
						reverseDirectionForPolarity);
			}
		}

	}

	@Override
	public Boolean saveSelectedVignette() {
		if (modelState.canSaveSelectedVignette()) {
			theSelectedVignette.save();
			return true;
		} else {
			return false;
		}
	}

	@Override
	public Boolean modelIsRunnable() {
		if (modelState.canModelIsRunnable()) {
			return meta.isRunnable();
		} else {
			return false;
		}
	}

	@Override
	public Boolean modelIsMap() {
		if (modelState.canModelIsMap()) {
			return meta.isMap();
		} else {
			return false;
		}
	}

	@Override
	public List<Double> getTimeVector() {
		if (modelState.canGetTimeVector()) {
			List<Double> timeVector = new ArrayList<Double>();
			double st = getStartTime();
			double ft = getEndTime();
			double sav = getSavePer();

			for (double t = st; t <= ft; t += sav) {
				int tm = (int) (t / sav);
				double remainder = (t - (((double) tm) * sav));
				if (remainder == 0.0) {
					timeVector.add(t);
				}
			}

			return timeVector;
		} else {
			return null;
		}
	}

	@Override
	public List<Double> getTimeVector(String runName) {
		if (modelState.canGetTimeVector(runName)) {
			return resultMap.get(runName).getTimeVector();
		} else {
			return null;
		}
	}

	@Override
	public Double getSavePer() {
		if (modelState.canGetSavePer()) {
			return Double.parseDouble(meta.getSaveper());
		} else {
			return null;
		}
	}

	@Override
	public Boolean runTo(double time) {
		// sdmuiExecutable.runTo(time);
		// return true;
		if (modelState.canRunTo(time)) {
			while (modelState.getCurrentTime() < time) {
				stepModel();
			}
			return true;
		} else {
			return false;
		}
	}

	@Override
	public Boolean startExecution() {
		if (modelState.canStartExecution()) {
			sdmuiExecutable.sdmuiExecuteStep2();
			modelState.setModelInitialized(true);
			return true;
		} else {
			return null;
		}
	}

	@Override
	public List<BridgeArrow> getArrowsWithTail(String variable) {
		if (modelState.canGetArrowsWithTail(variable)) {
			return modelStructure.getArrowsWithTail(variable);
		} else {
			return null;
		}
	}

	@Override
	public List<BridgeArrow> getArrowsWithHead(String variable) {
		if (modelState.canGetArrowsWithHead(variable)) {
			return modelStructure.getArrowsWithHead(variable);
		} else {
			return null;
		}
	}

	@Override
	public List<LoopObject> getLoops() {
		if (modelState.canGetLoops()) {
			return modelStructure.getLoops();
		} else {
			return new ArrayList<LoopObject>();
		}
	}

	@Override
	public List<LoopObject> getLoopsWithVariable(String variable) {
		if (modelState.canGetLoopsWithVariable(variable)) {
			return modelStructure.getLoopsWithVariable(variable);
		} else {
			return new ArrayList<LoopObject>();
		}
	}

	@Override
	public List<LoopObject> getLoopsOfLength(int length) {
		if (modelState.canGetLoopsOfLength(length)) {
			return modelStructure.getLoopsOfLength(length);
		} else {
			return new ArrayList<LoopObject>();
		}
	}

	/**
	 * @return the modelStructure
	 */
	public ModelStructure getModelStructure() {
		return modelStructure;
	}

	/**
	 * @param modelStructure
	 *            the modelStructure to set
	 */
	public void setModelStructure(ModelStructure modelStructure) {
		this.modelStructure = modelStructure;
	}

	/**
	 * @return the modelState
	 */
	public BridgeModelState getModelState() {
		return modelState;
	}

	/**
	 * @param modelState
	 *            the modelState to set
	 */
	public void setModelState(BridgeModelState modelState) {
		this.modelState = modelState;
	}

	/**
	 * @return the meta
	 */
	public BridgeModelMeta getMeta() {
		return meta;
	}

	/**
	 * @param meta
	 *            the meta to set
	 */
	public void setMeta(BridgeModelMeta meta) {
		this.meta = meta;
	}

	/**
	 * @param resultMap
	 *            the resultMap to set
	 */
	public void setResultMap(
			Map<String, BridgeModelResultsDatabaseImplementation> resultMap) {
		this.resultMap = resultMap;
	}

	/**
	 * @param theSelectedModel
	 *            the theSelectedModel to set
	 */
	public void setTheSelectedModel(MDL theSelectedModel) {
		this.theSelectedModel = theSelectedModel;
	}

	/**
	 * @param theSelectedVignette
	 *            the theSelectedVignette to set
	 */
	public void setTheSelectedVignette(BridgeVignetteObject theSelectedVignette) {
		this.theSelectedVignette = theSelectedVignette;
	}

	/**
	 * @param rootDirectory
	 *            the rootDirectory to set
	 */
	public void setRootDirectory(String rootDirectory) {
		this.rootDirectory = rootDirectory;
	}

	/**
	 * @param selectedVariables
	 *            the selectedVariables to set
	 */
	public void setSelectedVariables(Set<String> selectedVariables) {
		this.selectedVariables = selectedVariables;
	}

	@Override
	public Boolean closeSelectedModel() {

		if (modelState.canCloseSelectedModel()) {

			resultMap.clear();
			validatedRuns.clear();
			modelStructure = null;
			;
			modelState = null;
			meta = null;
			sdmuiExecutable = null;
			theSelectedModel = null;
			theSelectedVignette = null;
			selectedVariables = new LinkedHashSet<String>();
			allocatedIndicies = new LinkedHashMap<String, Map<Integer, List<String>>>();
			modelState = new BridgeModelState(this);

			return true;
		} else {
			return false;
		}
	}

	@Override
	public String getSdmdocHtmlFile() {
		if (modelState.isStructureLoaded()) {
			String dir = fileOrganizer.getSdmdocDirectory();
			String html = dir + "/" + meta.getModelName() + "_English.html";
			if (modelState.canGetSdmdocHtmlFile(html))
				return html;
			else
				return "";
		} else {
			return null;
		}
	}

	public String getSdmdocGraphicsDirectory() {
		return InformationManagers.getInstance().getModelManager()
				.getGeneratedOutputDirectoryName();
	}

	@Override
	public Boolean validateRun(String runName) {
		if (currentRunName.equals(runName))
			return true;
		BridgeModelResultsDatabaseImplementation dmrdi = new BridgeModelResultsDatabaseImplementation(buildRunOutputFileName(runName));
		boolean complete = dmrdi.isCompleteRun();
		System.out.println("Showdown "+runName);
		dmrdi.shutdowndatabase();
		return complete;
//		return true;
	}

	@Override
	public double[][] getValuesUpToTime(String[] variables, double time) {
		if (modelState.canGetValues(variables)) {
			double[][] cv = new double[variables.length][];
			for (int i = 0; i < variables.length; i++) {
				cv[i] = sdmuiExecutable.getValuesUpToTime(variables[i], time);
				if (badComputation(cv[i]))
					System.out.println("Bad Comp "+variables[i]);
			}
			return cv;
		} else {
			return null;
		}
	}
	
	public boolean badComputation(double[] values) {
		boolean bad = false;
		for (int i = 0; i < values.length; i++) {
			if (Double.isNaN(values[i]) || Double.isInfinite(values[i]))
				return true;
		}
		return bad;
	}

	@Override
	public double[][] getValuesForCurrentTime(String[] vars) {
		return getValues(vars);
	}

	@Override
	public double[][] getValuesUpToCurrentTime(String[] vars) {
		return getValuesUpToTime(vars, modelState.getCurrentTime());
	}

	@Override
	public Map<String, Map<String, String>> getPolarity() {
		return modelStructure.getPolarity();
	}

	public String getPolarity(String tail, String head) {
		if (modelStructure.getPolarity().get(tail) == null
				|| modelStructure.getPolarity().get(tail).get(head) == null)
			return "?";
		return modelStructure.getPolarity().get(tail).get(head);
	}

	@Override
	public void save(BridgeVignetteObject bvo) {
		bvo.save();

	}

	@Override
	public void saveVignette() {
		this.theSelectedVignette.save();
	}

	@Override
	public void save(Lever lever) {
		this.theSelectedVignette.save(lever);

	}

	@Override
	public void save(Output output) {
		this.theSelectedVignette.save(output);

	}

	@Override
	public List<CustomGraphObject> getCustomGraphs() {
		return theSelectedVignette.getCustomGraphs();
	}

	@Override
	public void addCustomGraph(CustomGraphObject cgo) {
		theSelectedVignette.addCustomGraph(cgo);

	}

	@Override
	public void saveCustomGraphs() {
		theSelectedVignette.saveCustomGraphs();

	}

	public String getModelDirectory() {
		return modelDirectory;
	}

	public void setModelDirectory(String modelDirectory) {
		this.modelDirectory = modelDirectory;
	}

	public String getOutputDirectory() {
		return outputDirectory;
	}

	public void setOutputDirectory(String outputDirectory) {
		this.outputDirectory = outputDirectory;
	}

	public void releaseExecutable() {
		sdmuiExecutable = null;
		System.gc();
		try {
			Thread.sleep(2000l);
		} catch (InterruptedException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
	}
	
	public void saveModifiedLoopData(LoopObject lo) {
		modelStructure.saveModifiedLoopData(lo);
	}
	
	public boolean isKnownVariable(String var) {
		return modelStructure.isKnownVariable(var);
	}

}
