package gov.anl.gss.sdm.server.model.documentor;

import gov.anl.gss.sdm.server.model.documentor.docx.Table;
import gov.anl.gss.sdm.server.model.documentor.docx.writer.ReportWriter;
import gov.anl.gss.sdm.server.model.documentor.docx.writer.Reportable;
import gov.anl.gss.sdm.server.model.documentor.xmlwriter.StringConstants;
import gov.anl.gss.sdm.server.model.support.Utilities;

import java.util.ArrayList;
import java.util.Collections;
import java.util.List;

public class LanguageAdministration {

	/**
	 * This class performs the behind the scenes processing
	 * to generate the LanguageKey file. This is a docx file 
	 * that contains all the strings used in the SDM-doc tool.
	 */
	
	public static final String KEY_FILE = "data/LanguageKey.docx";
	
	
	public LanguageAdministration() {
		
	}
	
	public boolean generateKeyFile() {
		return generateKeyFile(KEY_FILE);
	}
	
	public boolean generateKeyFile(String keyFile) {
		/*
		 * Get the list of all strings used in SDM-Doc.
		 * These are the keys. For each of the keys, get the
		 * English words. Generate a two column table with English
		 * in the left column and the keys in the right column.
		 */
		
		List<String> keys = Utilities.asList(StringConstants.ALL_STRING_CONSTANTS);
		Collections.sort(keys);
		
		boolean setAsDefault = false;
		SDMDocLanguageSpecific english = new SDMDocLanguageSpecific(
				SDMDocLanguageSpecific.ENGLISH, setAsDefault, null);
		List<Reportable> reportables = new ArrayList<Reportable>();
		List<List<String>> tableData = new ArrayList<List<String>>();
		
		for (String k : keys) {
			List<String> row = new ArrayList<String>();
			if (english.hasTranslation(k)) {
				row.add(english.getText(k));
			} else {
				row.add("");
			}
			row.add(k);
			tableData.add(row);
		}
		
		
		String title = "SDM-Doc Language Keys";
		String[] columnHeaders = {"English", "Key"};
		reportables.add(new Table(title, columnHeaders, tableData));
		ReportWriter rw = new ReportWriter(reportables, keyFile, false);
		rw.generateDocx();
		
		return true;
	}
	
	public boolean generateTranslationFile(String fullName, String language, String file) {
	
		/*
		 * Get the list of all strings used in SDM-Doc.
		 * These are the keys. For each of the keys, get the
		 * English words. Generate a two column table with English
		 * in the left column and either the language specific translation if
		 * it has already been defined or a blank field into which the translator
		 * will store the translation.
		 */
		
		List<String> keys = Utilities.asList(StringConstants.ALL_STRING_CONSTANTS);
		Collections.sort(keys);
		
		boolean setAsDefault = false;
		
		SDMDocLanguageSpecific english = new SDMDocLanguageSpecific(SDMDocLanguageSpecific.ENGLISH, setAsDefault, null);
		SDMDocLanguageSpecific otherLanguage = new SDMDocLanguageSpecific(language, setAsDefault, null);
		List<Reportable> reportables = new ArrayList<Reportable>();
		List<List<String>> tableData = new ArrayList<List<String>>();
		
		for (String k : keys) {
			List<String> row = new ArrayList<String>();
			if (english.hasTranslation(k)) {
				row.add(english.getText(k));
			} else {
				row.add(k);
			}
			if (otherLanguage.hasTranslation(k)) {
				row.add(otherLanguage.getText(k));
			} else {
				row.add("");
			}
			tableData.add(row);
		}
		
		
		String title = "SDM-Doc Translation File";
		String[] columnHeaders = {"English", fullName};
		reportables.add(new Table(title, columnHeaders, tableData));
		ReportWriter rw = new ReportWriter(reportables, file, false);
		rw.generateDocx();
		
		return true;
	}
	
	public static void main(String[] args) {
	
		SDMDocLanguageSpecific.initialize();
		LanguageAdministration la = new LanguageAdministration();
		la.generateKeyFile();
		la.generateTranslationFile("Japanese", "ja", "data/LanguageTranslationJapanese.docx");
		la.generateTranslationFile("Spanish", "es", "data/LanguageTranslationSpanish.docx");
		la.generateTranslationFile("German", "de", "data/LanguageTranslationGerman.docx");
		la.generateTranslationFile("Italian", "it", "data/LanguageTranslationItalian.docx");
		la.generateTranslationFile("Portuguese (Brazil)", "pt,BR", "data/LanguageTranslationPortugueseBrazil.docx");
		la.generateTranslationFile("French", "fr", "data/LanguageTranslationFrench.docx");
		la.generateTranslationFile("English", "en,US", "data/LanguageTranslationEnglishUS.docx");
	}
}
