package gov.anl.gss.sdm.server.model.documentor;

import gov.anl.gss.sdm.server.model.analysis.loops.Loop;
import gov.anl.gss.sdm.server.model.analysis.loops.LoopManager;
import gov.anl.gss.sdm.server.model.analysis.polarity.ArrowPolarity;
import gov.anl.gss.sdm.server.model.analysis.polarity.PolarityResults;
import gov.anl.gss.sdm.server.model.documentor.xmlwriter.FlatFileWriter;
import gov.anl.gss.sdm.server.model.documentor.xmlwriter.NumberFormatter;
import gov.anl.gss.sdm.server.model.documentor.xmlwriter.StringConstants;
import gov.anl.gss.sdm.server.model.documentor.xmlwriter.UserSelectedOptions;
import gov.anl.gss.sdm.server.model.documentor.xmlwriter.variableselector.VariableSelector;
import gov.anl.gss.sdm.server.model.documentor.xmlwriter.variableselector.VariableSelectorDataLookup;
import gov.anl.gss.sdm.server.model.documentor.xmlwriter.variableselector.VariableSelectorDmnlUnits;
import gov.anl.gss.sdm.server.model.documentor.xmlwriter.variableselector.VariableSelectorEndogenous;
import gov.anl.gss.sdm.server.model.documentor.xmlwriter.variableselector.VariableSelectorExogenous;
import gov.anl.gss.sdm.server.model.documentor.xmlwriter.variableselector.VariableSelectorFunctionSensitivityParameter;
import gov.anl.gss.sdm.server.model.documentor.xmlwriter.variableselector.VariableSelectorLevel;
import gov.anl.gss.sdm.server.model.documentor.xmlwriter.variableselector.VariableSelectorSourceInformation;
import gov.anl.gss.sdm.server.model.documentor.xmlwriter.variableselector.VariableSelectorState;
import gov.anl.gss.sdm.server.model.documentor.xmlwriter.variableselector.VariableSelectorUsingMacros;
import gov.anl.gss.sdm.server.model.documentor.xmlwriter.variableselector.VariableSelectorValueRangeSpecified;
import gov.anl.gss.sdm.server.model.translator.Equation;
import gov.anl.gss.sdm.server.model.translator.InformationManagers;
import gov.anl.gss.sdm.server.model.translator.Macro;
import gov.anl.gss.sdm.server.model.translator.MacroManager;
import gov.anl.gss.sdm.server.model.translator.Parser;
import gov.anl.gss.sdm.server.model.translator.RateToRateLinkReport;
import gov.anl.gss.sdm.server.model.translator.TranslatorConstants;
import gov.anl.gss.sdm.server.model.translator.UnitsManager;

import java.util.ArrayList;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.apache.commons.lang3.StringEscapeUtils;
import org.apache.log4j.Logger;

public class ModelInformation extends TableData implements TableRow, TableDataLoad {
static Logger log = Logger.getLogger(ModelInformation.class.getSimpleName());

	
	private List<Equation> variables;
	private long effectiveVariableCount;
	
	private List<Equation> stateVariables;
	private long effectiveStateVariableCount;
	
	private long stocks;
	private long effectiveStockVariableCount;
	
	private long exogenous;
	private long effectiveExogenousCount;
	private long withLookupCount;
	private long effectiveWithLookupCount;
	
	private long endogenous;
	private long effectiveEndogenousCount;
	
	private List<Macro> macros;
	private long effectiveMacrosCount;
	private List<Equation> usingMacros;
	private long effectiveUsingMacrosCount;
	
	private long rateToRateLinks;
	
	private long feedbackLoops;
	private long effectiveFeedbackCount;
	
	private long positiveCausalLinks;
	private long effectivePositiveCausalLinkCount;
	
	private long negativeCausalLinks;
	private long effectiveNegativeCausalLinkCount;
	
	private long unknownCausalLinks;
	private long effectiveUnknownCausalLinkCount;

	private List<Equation> sourceInformation;
	private long effectiveSourceInformationCount;
	
	private List<Equation> dmnl;
	private long effectiveDmnlCount;
	
	private List<Equation> valueRange;
	private long effectiveValueRangeCount;
	
	private List<Equation> sensitivityParameters;
	private long effectiveSensitivityParametersCount;
	private List<Equation> lookups;
	private long effectiveLookupsCount;

	private String timeUnit;
	private String initialTime;
	private String finalTime;
	private String saveper;
	private String timeStep;

	private boolean fullyFormed;
	private boolean definedGroups;
	private boolean vpmFile = false;
	
	private boolean subs = false;

	private Set<String> uniqueVariables = new LinkedHashSet<String>();

	public ModelInformation(Documentor documentor) {
		super(documentor);
		load();
	}
	
	public void document() {
		log.debug("### Model Information ###\n");
		log.debug("Variables: "+getVariablesCount());
		log.debug("State Variables: "+getStateVariablesCount());
		log.debug("Stocks: "+getStocksCount());
		log.debug("Feedback Loops: "+getFeedbackLoopsCount());
		log.debug("Macros: "+getMacrosCount());
		log.debug("Using Macros: "+getUsingMacrosCount());
		log.debug("SourceInformation: "+getSourceInformationCount());
		log.debug("Sensitivity Parameters: "+getSensitivityParametersCount());
		log.debug("Lookups: "+getLookupsCount());
		
		log.debug("\nTime Unit: "+timeUnit);
		log.debug("Initial Time: "+initialTime);
		log.debug("Final Time: "+finalTime);
		log.debug("Saveper: "+saveper);
		log.debug("Time Step: "+timeStep);
		
		log.debug("\nFully Formed: "+fullyFormed);
		log.debug("Defined Groups: "+definedGroups);
		log.debug("VPM File: "+vpmFile);
	}

	public void load() {
		loadVariables();
		loadStateVariables();
		loadStocks();
		
		loadExogenous();
		loadEndogenous();
		
		loadFeedbackLoops();
		loadPositiveCausalLinks();
		loadNegativeCausalLinks();
		loadUnknownCausalLinks();
		loadMacros();
		loadUsingMacros();
		
		loadRateToRateLinks();

		loadSourceInformation();
		loadDmnl();
		loadValueRange();
		loadSensitivityParameters();
		loadLookups();

		loadStrings();
		loadBooleans();
		
		subs = variables.size() != effectiveVariableCount;
		documentor.setSubs(subs);
	}

	private void loadStrings() {
		
		Map<String, Equation> equations = documentor.getEquations();
		
		Equation e = equations.get(TranslatorConstants.TIME_STEP);
		if (e == null) {
			e = equations.get(TranslatorConstants.TIME_STEP.replace(" ", "_"));
		}
		timeUnit = e.getUnits().split("\\[")[0];
		
		 e = equations.get(TranslatorConstants.INITIAL_TIME);
		if (e == null) {
			e = equations.get(TranslatorConstants.INITIAL_TIME.replace(" ", "_"));
		}
		initialTime = e.getRhs();
		
		 e = equations.get(TranslatorConstants.FINAL_TIME);
		if (e == null) {
			e = equations.get(TranslatorConstants.FINAL_TIME.replace(" ", "_"));
		}
		finalTime = e.getRhs();
		
		 e = equations.get(TranslatorConstants.SAVEPER);
		if (e == null) {
			e = equations.get(TranslatorConstants.SAVEPER.replace(" ", "_"));
		}
		saveper = e.getRhs();
		
		 e = equations.get(TranslatorConstants.TIME_STEP);
		if (e == null) {
			e = equations.get(TranslatorConstants.TIME_STEP.replace(" ", "_"));
		}
		timeStep = e.getRhs();
		
	}

	private void loadBooleans() {
		
		if (documentor.getAnalysisManager().getModelComponents().isComplete())
			fullyFormed = true;
		else
			fullyFormed = false;
		if (documentor.getGroupsCounter().isUserDefinedGroups())
			definedGroups = true;
		else
			definedGroups = false;
		vpmFile = false;
	}

	private void loadVariables() {
		clearUnique();
		VariableSelector selector = new VariableSelector /*Variable */ (StringConstants.ALL_VARIABLES, StringConstants.ALL_VARIABLES);
		variables = new ArrayList<Equation>();
		effectiveVariableCount = 0;
		
		Map<String, Equation> equations = documentor.getEquations();
		
		loadSelectedVariables(selector, variables);
		for (Equation eqn : equations.values()) {
			if (selector.isSelectMultipleEquation(eqn)  && eqn.isSyntacticallyCorrect())
				effectiveVariableCount += eqn.getTotalEquationCount();
		}
		
	}

	private void loadStateVariables() {

		clearUnique();
		Map<String, Equation> equations = documentor.getEquations();
		
		VariableSelector variableSelector = new VariableSelectorState("State Variables", "State");
		stateVariables = new ArrayList<Equation>();
		for (Equation eqn : equations.values()) {
			if ((variableSelector.isSelectSingleEquation(eqn)) &&
					eqn.isSyntacticallyCorrect()) {
					stateVariables.add(eqn);
			}
			if ((variableSelector.isSelectMultipleEquation(eqn)) &&
					eqn.isSyntacticallyCorrect()) {
				effectiveStateVariableCount += eqn.getTotalEquationCount();

			}
		}
	}
	
	private void clearUnique() {
		uniqueVariables.clear();
	}
	
	private boolean isUnique(String vn) {
		if (uniqueVariables.contains(vn))
			return false;
		uniqueVariables.add(vn);
		return true;
	}

	private void loadStocks() {
		Map<String, Equation> equations = documentor.getEquations();
		stocks = 0;
		
		VariableSelector vs = new VariableSelectorLevel("Level Variables", "Level");
		
		for (Equation eqn : equations.values()) {
			if (vs.isSelectSingleEquation(eqn)
					 && eqn.isSyntacticallyCorrect()) {
				stocks++;
			}
			if (vs.isSelectMultipleEquation(eqn)
					 && eqn.isSyntacticallyCorrect()) {
				effectiveStockVariableCount += eqn.getTotalEquationCount();
			}
		}
	}
	
	private void loadExogenous() {
		Map<String, Equation> equations = documentor.getEquations();
		exogenous = 0;
		effectiveExogenousCount = 0;
		withLookupCount = 0;
		VariableSelector vs = new VariableSelectorExogenous(StringConstants.EXOGENOUS,StringConstants.EXOGENOUS);
		

		for (Equation eqn : equations.values()) {
			if (vs.isSelectSingleEquation(eqn)
					 && eqn.isSyntacticallyCorrect()) {
				exogenous++;
				if (Parser.isWithLookup(eqn.getLhs())) {
					withLookupCount++;
				}
			}
			if (vs.isSelectMultipleEquation(eqn)
					 && eqn.isSyntacticallyCorrect()) {
				effectiveExogenousCount += eqn.getTotalEquationCount();
				if (Parser.isWithLookup(eqn.getLhs())) {
						effectiveWithLookupCount += eqn.getTotalEquationCount();
				}
						
			}
		}
	}
	
	private void loadEndogenous() {
		Map<String, Equation> equations = documentor.getEquations();
		endogenous = 0;
		effectiveEndogenousCount = 0;
		VariableSelector vs = new VariableSelectorEndogenous(StringConstants.ENDOGENOUS,StringConstants.ENDOGENOUS);
		
		for (Equation eqn : equations.values()) {
			if (vs.isSelectSingleEquation(eqn)
					 && eqn.isSyntacticallyCorrect()) {
				endogenous++;
				
			}
			if (vs.isSelectMultipleEquation(eqn)
					 && eqn.isSyntacticallyCorrect()) {
				effectiveEndogenousCount += eqn.getTotalEquationCount();						
			}
		}
	}
	
	private void loadRateToRateLinks() {
		RateToRateLinkReport rtr = new RateToRateLinkReport();
		rateToRateLinks = rtr.getNumRateToRate();
	}
	
	private void loadFeedbackLoops() {
		if (documentor.getAnalysisManager().getLoopManager() == null) {
			feedbackLoops = 0;
		} else {
			feedbackLoops = documentor.getAnalysisManager().getLoopManager().getNumberOfLoopsNoIVV();
		}
	}
	
	private void loadPositiveCausalLinks() {
		List<ArrowPolarity> al = documentor.getAnalysisManager().getPolarityResultsNoIVV().getPositivePolarityArrows();
		int size = al.size();
		for (ArrowPolarity ap : al) {
			if (ap.isGeneratedCloudOrRate()) {
				size--;
			} else {
				String lhs = ap.getHead();
				if (InformationManagers.getInstance().getEquationManager().getAllEquationsFor(lhs) != null) {
				for (Equation eqn : InformationManagers.getInstance().getEquationManager().getAllEquationsFor(lhs))
					if (eqn.isSyntacticallyCorrect())
					effectivePositiveCausalLinkCount += eqn.getTotalEquationCount();
				}
			}
		}
		positiveCausalLinks = size;
	}
	
	private void loadNegativeCausalLinks() {

		List<ArrowPolarity> al = documentor.getAnalysisManager().getPolarityResultsNoIVV().getNegativePolarityArrows();
		int size = al.size();
		for (ArrowPolarity ap : al) {
			if (ap.isGeneratedCloudOrRate()) {
				size--;
			} else {
				String lhs = ap.getHead();
				if (InformationManagers.getInstance().getEquationManager().getAllEquationsFor(lhs) != null) {
					for (Equation eqn : InformationManagers.getInstance().getEquationManager().getAllEquationsFor(lhs))
						if (eqn.isSyntacticallyCorrect())
							effectiveNegativeCausalLinkCount += eqn.getTotalEquationCount();
				}
			}
		}
		negativeCausalLinks = size;
	}
	
	private void loadUnknownCausalLinks() {

		List<ArrowPolarity> al = documentor.getAnalysisManager().getPolarityResultsNoIVV().getUnknownPolarityArrows();
		int size = al.size();
		for (ArrowPolarity ap : al) {
			if (ap.isGeneratedCloudOrRate()){
				size--;
			} else {
				String lhs = ap.getHead();
				if (InformationManagers.getInstance().getEquationManager().getAllEquationsFor(lhs) != null) {
				for (Equation eqn : InformationManagers.getInstance().getEquationManager().getAllEquationsFor(lhs))
					if (eqn.isSyntacticallyCorrect())
					effectiveUnknownCausalLinkCount += eqn.getTotalEquationCount();
				}
			}
		}
		unknownCausalLinks = size;
	}

	private void loadMacros() {
		MacroManager mm = InformationManagers.getInstance().getMacroManager();
		Map<String, Macro> map = mm.getMacros();
		macros = new ArrayList<Macro>();
		for (Macro macro : map.values()) {
			macros.add(macro);
		}
	}
	
	private void loadUsingMacros() {
		VariableSelector selector = new VariableSelectorUsingMacros(StringConstants.USING_MACROS, StringConstants.USING_MACROS);
		usingMacros = new ArrayList<Equation>();
		loadSelectedVariables(selector, usingMacros);
		
		Map<String, Equation> equations = documentor.getEquations();
		for (Equation eqn : equations.values()) {
			if (selector.isSelectMultipleEquation(eqn) && eqn.isSyntacticallyCorrect())
				effectiveUsingMacrosCount += eqn.getTotalEquationCount();
		}
	}

	private void loadSourceInformation() {
		VariableSelector selector = new VariableSelectorSourceInformation(StringConstants.VARIABLES_WITH_SOURCE_INFORMATION, StringConstants.VARIABLES_WITH_SOURCE_INFORMATION);
		sourceInformation = new ArrayList<Equation>();
		loadSelectedVariables(selector, sourceInformation);
		
		Map<String, Equation> equations = documentor.getEquations();
		for (Equation eqn : equations.values()) {
			if (selector.isSelectMultipleEquation(eqn) && eqn.isSyntacticallyCorrect())
				effectiveSourceInformationCount += eqn.getTotalEquationCount();
		}
	}
	
	private void loadDmnl() {
		VariableSelector selector = new VariableSelectorDmnlUnits(StringConstants.DIMENSIONLESS_UNIT_VARIABLES, StringConstants.DIMENSIONLESS_UNIT_VARIABLES);
		dmnl = new ArrayList<Equation>();
		loadSelectedVariables(selector, dmnl);
		Map<String, Equation> equations = documentor.getEquations();
		for (Equation eqn : equations.values()) {
			if (selector.isSelectMultipleEquation(eqn) && eqn.isSyntacticallyCorrect())
				effectiveDmnlCount += eqn.getTotalEquationCount();
		}
	}
	
	private void loadValueRange() {
		VariableSelector selector = new VariableSelectorValueRangeSpecified(StringConstants.VALUE_RANGE_SPECIFIED, StringConstants.VALUE_RANGE_SPECIFIED);
		valueRange = new ArrayList<Equation>();
		loadSelectedVariables(selector, valueRange);
		Map<String, Equation> equations = documentor.getEquations();
		for (Equation eqn : equations.values()) {
			if (selector.isSelectMultipleEquation(eqn) && eqn.isSyntacticallyCorrect())
				effectiveValueRangeCount += eqn.getTotalEquationCount();
		}
	}

	private void loadSensitivityParameters() {
		VariableSelector selector = new VariableSelectorFunctionSensitivityParameter(StringConstants.FUNCTION_SENSITIVITY_PARAMETERS, StringConstants.FUNCTION_SENSITIVITY_PARAMETERS);
		sensitivityParameters = new ArrayList<Equation>();
		loadSelectedVariables(selector, sensitivityParameters);
		
		Map<String, Equation> equations = documentor.getEquations();
		for (Equation eqn : equations.values()) {
			if (selector.isSelectMultipleEquation(eqn) && eqn.isSyntacticallyCorrect())
				effectiveSensitivityParametersCount += eqn.getTotalEquationCount();
		}
	}

	private void loadLookups() {
		VariableSelector selector = new VariableSelectorDataLookup(StringConstants.DATA_LOOKUP_TABLES, StringConstants.DATA_LOOKUP_TABLES);
		lookups = new ArrayList<Equation>();
		loadSelectedVariables(selector, lookups);
		
		Map<String, Equation> equations = documentor.getEquations();
		for (Equation eqn : equations.values()) {
			if (selector.isSelectMultipleEquation(eqn) && eqn.isSyntacticallyCorrect())
				effectiveLookupsCount += eqn.getTotalEquationCount();
		}
	}
	
	private void loadSelectedVariables(VariableSelector selector, List<Equation> selectedVariables) {
		
		// this loads variable names based on equation characteristics, but placed in list only once.
		clearUnique();
		Map<String, Equation> equations = documentor.getEquations();
		for (Equation eqn : equations.values()) {
			if (selector.isSelectSingleEquation(eqn) && eqn.isSyntacticallyCorrect()) {
				if (isUnique(eqn.getVariableName()))
					selectedVariables.add(eqn);
				
			} 
		}
	}
	
	public String formatInt(long i) {
		String s = NumberFormatter.formatForCountNotConditional(i);
		return s;
	}

	
	public String formatIntNoHTML(int i) {
		String s = NumberFormatter.formatForCountNotConditional(i);
		return s;
	}
	
	public String formatIntNoHTML(long i) {
		String s = NumberFormatter.formatForCountNotConditional(i);
		return s;
	}

	
	public String formatInt(int i) {
		String s = NumberFormatter.formatForCountNotConditional(i);
		return s;
	}

	public int getVariablesCount() {
		return variables.size();
	}
	

	public String getVariablesCountString() {
		return formatInt(variables.size())+ (subs ? "|"+formatInt(effectiveVariableCount) : "");
	}

	public int getStateVariablesCount() {
		return stateVariables.size();
	}
	
	public String getExogenousVariablesCountWithPctNoHTML() {
		return  NumberFormatter.getAsCountAndPercentTotalNoHTML(exogenous, getVariablesCount()) + (subs ? "|"+
		NumberFormatter.getAsCountAndPercentTotalNoHTML(effectiveExogenousCount, effectiveVariableCount)  : "");
	}
	
	public String getEndogenousVariablesCountWithPctNoHTML() {
		return  NumberFormatter.getAsCountAndPercentTotalNoHTML(endogenous, getVariablesCount()) + (subs ? "|"+
		NumberFormatter.getAsCountAndPercentTotalNoHTML(effectiveEndogenousCount, effectiveVariableCount)  : "");
	}
	
	public String getExogenousVariablesCountWithPct() {
		return  NumberFormatter.getAsCountAndPercentTotal(exogenous, getVariablesCount()) + (subs ? "|"+
		NumberFormatter.getAsCountAndPercentTotal(effectiveExogenousCount, effectiveVariableCount)  : "");
	}
	
	public String getEndogenousVariablesCountWithPct() {
		return  NumberFormatter.getAsCountAndPercentTotal(endogenous, getVariablesCount()) + (subs ? "|"+
		NumberFormatter.getAsCountAndPercentTotal(effectiveEndogenousCount, effectiveVariableCount)  : "");
	}
	
	public String getStateVariablesCountWithPct() {
		if (NumberFormatter.map)
			return "NA";
		return  NumberFormatter.getAsCountAndPercentTotal(stateVariables.size(), getVariablesCount()) + (subs ? "|"+
		NumberFormatter.getAsCountAndPercentTotal(effectiveStateVariableCount, effectiveVariableCount)  : "");
	}
	
	
	public String getStateVariablesCountWithPctNoHTML() {
		if (NumberFormatter.map)
			return "NA";
		return  NumberFormatter.getAsCountAndPercentTotalNoHTML(stateVariables.size(), getVariablesCount()) + (subs ? "|"+
		NumberFormatter.getAsCountAndPercentTotalNoHTML(effectiveStateVariableCount, effectiveVariableCount)  : "");
	}

	public long getStocksCount() {
		return stocks;
	}
	
	public String getStocksCountWithPctNoHTML() {
		if (NumberFormatter.map)
			return "NA";
		return  NumberFormatter.getAsCountAndPercentTotalNoHTML(stocks, getVariablesCount()) + (subs ? "|"+
				NumberFormatter.getAsCountAndPercentTotalNoHTML(effectiveStockVariableCount, effectiveVariableCount) : "");
	}
	
	public String getStocksCountWithPct() {
		if (NumberFormatter.map)
			return "NA";
		return  NumberFormatter.getAsCountAndPercentTotal(stocks, getVariablesCount()) + (subs ? "|"+
				NumberFormatter.getAsCountAndPercentTotal(effectiveStockVariableCount, effectiveVariableCount) : "");
	}
	
	public String getExogenousCount() {
		
		return  formatInt(exogenous) + (subs ? "|"+
				formatInt(effectiveExogenousCount): "");
	}
	
	public String getEndogenousCount() {
		return  formatInt(endogenous)+ (subs ? "|"+	formatInt(effectiveEndogenousCount) : "");
	}
	
	public long getFeedbackLoopsCount() {
		return feedbackLoops;
	}
	
	public String getUnitsUsed() {
		StringBuilder sb = new StringBuilder();
		UnitsManager um = InformationManagers.getInstance().getUnitsManager();
		sb.append(um.getBasicCount());
		sb.append("/");
		sb.append(um.getCombinedCount());
		return sb.toString();
	}
	
	
	
	public String getCausalLinkResults() {
		StringBuilder sb = new StringBuilder();
		
		PolarityResults pr = documentor.getAnalysisManager().getPolarityResultsNoIVV();

		int total = pr.getArrows().size();
		int pos = pr.getPositivePolarityArrows().size();
		int neg = pr.getNegativePolarityArrows().size();
		int unk = pr.getUnknownPolarityArrows().size();
		
		long totalEffective = effectivePositiveCausalLinkCount + 
				effectiveNegativeCausalLinkCount + 
				effectiveUnknownCausalLinkCount;

		sb.append(HTMLConstants.BLACK_START);
		sb.append(formatInt(total)+HTMLConstants.SPACE+"("+HTMLConstants.COLOR_END);
		sb.append(HTMLConstants.BLUE_START+formatInt(pos)+HTMLConstants.COLOR_END);
		sb.append(HTMLConstants.BLACK_START+"|"+HTMLConstants.COLOR_END);
		sb.append(HTMLConstants.RED_START+formatInt(neg)+HTMLConstants.COLOR_END);
		sb.append(HTMLConstants.BLACK_START+"|"+HTMLConstants.COLOR_END);
		sb.append(HTMLConstants.BLACK_START+formatInt(unk)+HTMLConstants.COLOR_END);
		sb.append(HTMLConstants.BLACK_START+")"+HTMLConstants.COLOR_END);
		if (subs) {
			sb.append("|");
			sb.append(HTMLConstants.BLACK_START);
			sb.append(formatInt(totalEffective)+HTMLConstants.SPACE+"("+HTMLConstants.COLOR_END);
			sb.append(HTMLConstants.BLUE_START+formatInt(effectivePositiveCausalLinkCount)+HTMLConstants.COLOR_END);
			sb.append(HTMLConstants.BLACK_START+"|"+HTMLConstants.COLOR_END);
			sb.append(HTMLConstants.RED_START+formatInt(effectiveNegativeCausalLinkCount)+HTMLConstants.COLOR_END);
			sb.append(HTMLConstants.BLACK_START+"|"+HTMLConstants.COLOR_END);
			sb.append(HTMLConstants.BLACK_START+formatInt(effectiveUnknownCausalLinkCount)+HTMLConstants.COLOR_END);
			sb.append(HTMLConstants.BLACK_START+")"+HTMLConstants.COLOR_END);
		}

		return sb.toString();
	}
	
	public String getCausalLinkResultsNoHTML() {
		StringBuilder sb = new StringBuilder();
		
		PolarityResults pr = documentor.getAnalysisManager().getPolarityResultsNoIVV();

		int total = pr.getArrows().size();
		int pos = pr.getPositivePolarityArrows().size();
		int neg = pr.getNegativePolarityArrows().size();
		int unk = pr.getUnknownPolarityArrows().size();
		
		long totalEffective = effectivePositiveCausalLinkCount + 
				effectiveNegativeCausalLinkCount + 
				effectiveUnknownCausalLinkCount;

		sb.append(formatInt(total)+" (");
		sb.append(formatInt(pos));
		sb.append("|");
		sb.append(formatInt(neg));
		sb.append("|");
		sb.append(formatInt(unk));
		sb.append(")");
		if (subs) {
			sb.append("|");
			sb.append(formatInt(totalEffective)+" (");
			sb.append(formatInt(effectivePositiveCausalLinkCount));
			sb.append("|");
			sb.append(formatInt(effectiveNegativeCausalLinkCount));
			sb.append("|");
			sb.append(formatInt(effectiveUnknownCausalLinkCount));
			
		}

		return sb.toString();
	}
	
	public String getFeedbackResultsWithIVVNoHTML() {
		return getFeedbackResultsWithIVVNoHTML(documentor.getAnalysisManager().getLoopManager().getAllLoopsWithIVV());
	}

	public String getFeedbackResultsWithIVV() {
		return getFeedbackResultsWithIVV(documentor.getAnalysisManager().getLoopManager().getAllLoopsWithIVV());
	}
	
	public String getFeedbackResultsNoIVVNoHTML() {
		return getFeedbackResultsNoHTML(documentor.getAnalysisManager().getLoopManager().getAllLoopsNoIVV());
	}

	public String getFeedbackResultsNoIVV() {
		return getFeedbackResults(documentor.getAnalysisManager().getLoopManager().getAllLoopsNoIVV());
	}
	
	public String getFeedbackResultsWithIVVNoHTML(Set<Loop> loops) {
		StringBuilder sb = new StringBuilder();
		
		LoopManager lm = documentor.getAnalysisManager().getLoopManager();
		
		if (!lm.isUsingMinimumMemory()) {

		int total = loops.size();
		int pos = lm.getNumberOfPositivePolarityLoops(loops);
		int neg = lm.getNumberOfNegativePolarityLoops(loops);
		int unk = lm.getNumberOfProblemPolarityLoops(loops);

		sb.append(formatInt(total)+" (");
		sb.append(formatInt(pos));
		sb.append("|");
		sb.append(formatInt(neg));
		sb.append("|");
		sb.append(formatInt(unk));
		sb.append(")");
		
//		pos = lm.getEffectiveNumberOfPositivePolarityLoops(loops);
//		neg = lm.getEffectiveNumberOfNegativePolarityLoops(loops);
//		unk = lm.getEffectiveNumberOfProblemPolarityLoops(loops);
//		
//		total = pos+neg+unk;
		} else {
			long total = lm.getNumberOfLoopsWithIVV();
			int pos = lm.getNumberOfPositivePolarityLoopsWithIVV();
			int neg = lm.getNumberOfNegativePolarityLoopsWithIVV();
			int unk = lm.getNumberOfProblemPolarityLoopsWithIVV();
			
			sb.append(formatInt(total)+" (");
			sb.append(formatInt(pos));
			sb.append("|");
			sb.append(formatInt(neg));
			sb.append("|");
			sb.append(formatInt(unk));
			sb.append(")");
		}

		return sb.toString();
	}
	
	public String getFeedbackResultsWithIVV(Set<Loop> loops) {
		StringBuilder sb = new StringBuilder();
		
		LoopManager lm = documentor.getAnalysisManager().getLoopManager();
		
		if (!lm.isUsingMinimumMemory()) {

		int total = loops.size();
		int pos = lm.getNumberOfPositivePolarityLoops(loops);
		int neg = lm.getNumberOfNegativePolarityLoops(loops);
		int unk = lm.getNumberOfProblemPolarityLoops(loops);

		sb.append(HTMLConstants.BLACK_START);
		sb.append(formatInt(total)+HTMLConstants.SPACE+"("+HTMLConstants.COLOR_END);
		sb.append(HTMLConstants.BLUE_START+formatInt(pos)+HTMLConstants.COLOR_END);
		sb.append(HTMLConstants.BLACK_START+"|"+HTMLConstants.COLOR_END);
		sb.append(HTMLConstants.RED_START+formatInt(neg)+HTMLConstants.COLOR_END);
		sb.append(HTMLConstants.BLACK_START+"|"+HTMLConstants.COLOR_END);
		sb.append(HTMLConstants.BLACK_START+formatInt(unk)+HTMLConstants.COLOR_END);
		sb.append(HTMLConstants.BLACK_START+")"+HTMLConstants.COLOR_END);
		
//		pos = lm.getEffectiveNumberOfPositivePolarityLoops(loops);
//		neg = lm.getEffectiveNumberOfNegativePolarityLoops(loops);
//		unk = lm.getEffectiveNumberOfProblemPolarityLoops(loops);
//		
//		total = pos+neg+unk;
		} else {
			long total = lm.getNumberOfLoopsWithIVV();
			int pos = lm.getNumberOfPositivePolarityLoopsWithIVV();
			int neg = lm.getNumberOfNegativePolarityLoopsWithIVV();
			int unk = lm.getNumberOfProblemPolarityLoopsWithIVV();
			
			sb.append(HTMLConstants.BLACK_START);
			sb.append(formatInt(total)+HTMLConstants.SPACE+"("+HTMLConstants.COLOR_END);
			sb.append(HTMLConstants.BLUE_START+formatInt(pos)+HTMLConstants.COLOR_END);
			sb.append(HTMLConstants.BLACK_START+"|"+HTMLConstants.COLOR_END);
			sb.append(HTMLConstants.RED_START+formatInt(neg)+HTMLConstants.COLOR_END);
			sb.append(HTMLConstants.BLACK_START+"|"+HTMLConstants.COLOR_END);
			sb.append(HTMLConstants.BLACK_START+formatInt(unk)+HTMLConstants.COLOR_END);
			sb.append(HTMLConstants.BLACK_START+")"+HTMLConstants.COLOR_END);
		}

		return sb.toString();
	}
	
	public String getFeedbackResultsNoHTML(Set<Loop> loops) {
		StringBuilder sb = new StringBuilder();
		
		LoopManager lm = documentor.getAnalysisManager().getLoopManager();
		
		if (!lm.isUsingMinimumMemory()) {

			int total = loops.size();
			int pos = lm.getNumberOfPositivePolarityLoops(loops);
			int neg = lm.getNumberOfNegativePolarityLoops(loops);
			int unk = lm.getNumberOfProblemPolarityLoops(loops);

			sb.append(formatInt(total)+" (");
			sb.append(formatInt(pos));
			sb.append("|");
			sb.append(formatInt(neg));
			sb.append("|");
			sb.append(formatInt(unk));
			sb.append(")");

		} else {
			long total = lm.getNumberOfLoopsNoIVV();
			long pos = lm.getNumberOfPositivePolarityLoops();
			long neg = lm.getNumberOfNegativePolarityLoops();
			long unk = lm.getNumberOfProblemPolarityLoops();

			sb.append(formatInt(total)+" (");
			sb.append(formatInt(pos));
			sb.append("|");
			sb.append(formatInt(neg));
			sb.append("|");
			sb.append(formatInt(unk));
			sb.append(")");
		}

		return sb.toString();
	}

	public String getFeedbackResults(Set<Loop> loops) {
		StringBuilder sb = new StringBuilder();
		
		LoopManager lm = documentor.getAnalysisManager().getLoopManager();
		
		if (!lm.isUsingMinimumMemory()) {

			int total = loops.size();
			int pos = lm.getNumberOfPositivePolarityLoops(loops);
			int neg = lm.getNumberOfNegativePolarityLoops(loops);
			int unk = lm.getNumberOfProblemPolarityLoops(loops);

			sb.append(HTMLConstants.BLACK_START);
			sb.append(formatInt(total)+HTMLConstants.SPACE+"("+HTMLConstants.COLOR_END);
			sb.append(HTMLConstants.BLUE_START+formatInt(pos)+HTMLConstants.COLOR_END);
			sb.append(HTMLConstants.BLACK_START+"|"+HTMLConstants.COLOR_END);
			sb.append(HTMLConstants.RED_START+formatInt(neg)+HTMLConstants.COLOR_END);
			sb.append(HTMLConstants.BLACK_START+"|"+HTMLConstants.COLOR_END);
			sb.append(HTMLConstants.BLACK_START+formatInt(unk)+HTMLConstants.COLOR_END);
			sb.append(HTMLConstants.BLACK_START+")"+HTMLConstants.COLOR_END);

		} else {
			long total = lm.getNumberOfLoopsNoIVV();
			long pos = lm.getNumberOfPositivePolarityLoops();
			long neg = lm.getNumberOfNegativePolarityLoops();
			long unk = lm.getNumberOfProblemPolarityLoops();

			sb.append(HTMLConstants.BLACK_START);
			sb.append(formatInt(total)+HTMLConstants.SPACE+"("+HTMLConstants.COLOR_END);
			sb.append(HTMLConstants.BLUE_START+formatInt(pos)+HTMLConstants.COLOR_END);
			sb.append(HTMLConstants.BLACK_START+"|"+HTMLConstants.COLOR_END);
			sb.append(HTMLConstants.RED_START+formatInt(neg)+HTMLConstants.COLOR_END);
			sb.append(HTMLConstants.BLACK_START+"|"+HTMLConstants.COLOR_END);
			sb.append(HTMLConstants.BLACK_START+formatInt(unk)+HTMLConstants.COLOR_END);
			sb.append(HTMLConstants.BLACK_START+")"+HTMLConstants.COLOR_END);
		}

		return sb.toString();
	}


	public int getMacrosCount() {
		return macros.size();
	}
	
	public String getMacrosCountWithPct() {
		if (NumberFormatter.map)
			return "NA";
		return NumberFormatter.getAsCountAndPercentTotal(macros.size(), getVariablesCount());
	}
	
	public int getUsingMacrosCount() {
		return usingMacros.size();
	}
	
	public String getUsingMacrosCountWithPct() {
		if (NumberFormatter.map)
			return "NA";
		return NumberFormatter.getAsCountAndPercentTotal(usingMacros.size(), getVariablesCount()) + (subs ? "|" + 
		NumberFormatter.getAsCountAndPercentTotal(effectiveUsingMacrosCount, getEffectiveVariableCount()) : "");
	}
	
	public String getUsingMacrosCountWithPctNoHTML() {
		if (NumberFormatter.map)
			return "NA";
		return NumberFormatter.getAsCountAndPercentTotalNoHTML(usingMacros.size(), getVariablesCount()) + (subs ? "|" + 
		NumberFormatter.getAsCountAndPercentTotalNoHTML(effectiveUsingMacrosCount, getEffectiveVariableCount()) : "");
	}


	public List<Equation> getVariables() {
		return variables;
	}

	public List<Equation> getStateVariables() {
		return stateVariables;
	}

//	public List<Equation> getStocks() {
//		return stocks;
//	}

	public List<Macro> getMacros() {
		return macros;
	}

	public List<Equation> getSourceInformation() {
		return sourceInformation;
	}

	public List<Equation> getSensitivityParameters() {
		return sensitivityParameters;
	}

	public List<Equation> getLookups() {
		return lookups;
	}

	public int getSourceInformationCount() {
		return sourceInformation.size();
	}
	
	public String getSourceInformationCountWithPct() {
		if (NumberFormatter.map)
			return "NA";
		return NumberFormatter.getAsCountAndPercentTotal(sourceInformation.size(), getVariablesCount()) + (subs ? "|" +
		NumberFormatter.getAsCountAndPercentTotal(effectiveSourceInformationCount, getEffectiveVariableCount()) : "");
	}
	
	public String getSourceInformationCountWithPctNoHTML() {
		if (NumberFormatter.map)
			return "NA";
		return NumberFormatter.getAsCountAndPercentTotalNoHTML(sourceInformation.size(), getVariablesCount()) + (subs ? "|" +
		NumberFormatter.getAsCountAndPercentTotalNoHTML(effectiveSourceInformationCount, getEffectiveVariableCount()) : "");
	}
	
	public String getDmnlCountAsWithPct() {

		if (NumberFormatter.map)
			return "NA";

		return NumberFormatter.getAsCountAndPercentTotal(dmnl.size(), getVariablesCount()) + (subs ? "|" + 
		NumberFormatter.getAsCountAndPercentTotal(effectiveDmnlCount, getEffectiveVariableCount()) : "");
	}
	
	public String getDmnlCountAsWithPctNoHTML() {

		if (NumberFormatter.map)
			return "NA";

		return NumberFormatter.getAsCountAndPercentTotalNoHTML(dmnl.size(), getVariablesCount()) + (subs ? "|" + 
		NumberFormatter.getAsCountAndPercentTotalNoHTML(effectiveDmnlCount, getEffectiveVariableCount()) : "");
	}
	
	public String getValueRangeCountAsWithPct() {

		return NumberFormatter.getAsCountAndPercentTotal(valueRange.size(), getVariablesCount()) + (subs ? "|" + 
		NumberFormatter.getAsCountAndPercentTotal(effectiveValueRangeCount, getEffectiveVariableCount()) : "");
	}
	
	public String getValueRangeCountAsWithPctNoHTML() {

		return NumberFormatter.getAsCountAndPercentTotalNoHTML(valueRange.size(), getVariablesCount()) + (subs ? "|" + 
		NumberFormatter.getAsCountAndPercentTotalNoHTML(effectiveValueRangeCount, getEffectiveVariableCount()) : "");
	}

	public int getSensitivityParametersCount() {
		return sensitivityParameters.size();
	}
	
	public String getSensitivityParametersCountWithPct() {
		if (NumberFormatter.map)
			return "NA";
		return NumberFormatter.getAsCountAndPercentTotal(sensitivityParameters.size(), getVariablesCount()) + (subs ? "|" +
		NumberFormatter.getAsCountAndPercentTotal(effectiveSensitivityParametersCount, getEffectiveVariableCount()) : "");
	}
	
	public String getSensitivityParametersCountWithPctNoHTML() {
		if (NumberFormatter.map)
			return "NA";
		return NumberFormatter.getAsCountAndPercentTotalNoHTML(sensitivityParameters.size(), getVariablesCount()) + (subs ? "|" +
		NumberFormatter.getAsCountAndPercentTotalNoHTML(effectiveSensitivityParametersCount, getEffectiveVariableCount()) : "");
	}

	public int getLookupsCount() {
		return lookups.size();
	}
	
	public String getLookupsCountWithPct() {
		if (NumberFormatter.map)
			return "NA";
		return NumberFormatter.getAsCountAndPercentTotal(lookups.size(), getVariablesCount()) + (subs ? "|" +
		NumberFormatter.getAsCountAndPercentTotal(effectiveLookupsCount, getEffectiveVariableCount()) : "");
	}
	
	public String getLookupsCountWithPctNoHTML() {
		if (NumberFormatter.map)
			return "NA";
		return NumberFormatter.getAsCountAndPercentTotalNoHTML(lookups.size(), getVariablesCount()) + (subs ? "|" +
		NumberFormatter.getAsCountAndPercentTotalNoHTML(effectiveLookupsCount, getEffectiveVariableCount()) : "");
	}

	public String getTimeUnit() {
		return timeUnit;
	}

	public String getInitialTime() {
		return initialTime;
	}

	public String getFinalTime() {
		return finalTime;
	}

	public String getSaveper() {
		return saveper;
	}

	public String getTimeStep() {
		return timeStep;
	}

	public boolean isFullyFormed() {
		return fullyFormed;
	}

	public boolean isDefinedGroups() {
		return definedGroups;
	}

	public boolean isVpmFile() {
		return vpmFile;
	}
	
	private List<List<List<String>>> createFlatFileTable() {
		SDMDocLanguageSpecific language = SDMDocLanguageSpecific.getInstance();
		List<List<List<String>>> table = createTable();
		
		addTitle(table,language.getTextHtml(StringConstants.MODEL_ASSESSMENT_RESULTS));
		List<String> headers = new ArrayList<String>();
		headers.add(language.getTextHtml(StringConstants.MODEL_INFORMATION));
		headers.add(language.getTextHtml(StringConstants.RESULT));
		addColumnHeaders(table, headers);
		
		return table;
	}
	
//	@Override
	public List<List<List<String>>> getTableData() {

		
		UserSelectedOptions options = InformationManagers.getInstance().getUserSelectedOptions();
		SDMDocLanguageSpecific language = SDMDocLanguageSpecific.getInstance();
		String t = UserSelectedOptions.TRUE;

		List<List<List<String>>> table = createFlatFileTable();
		
			rowAddNoSplit(HTMLConstants.WITHOUT_HTML, table, StringConstants.TOTAL_NUMBER_OF_VARIABLES, /*StringConstants.ALL_VARIABLES, */
					language.getTextHtml(StringConstants.TOTAL_NUMBER_OF_VARIABLES),getVariablesCountString());
			
			rowAddNoSplit(HTMLConstants.WITHOUT_HTML, table, StringConstants.TOTAL_NUMBER_OF_STATE_VARIABLES, /*StringConstants.STATE_VARIABLES,   */
					language.getTextHtml(StringConstants.TOTAL_NUMBER_OF_STATE_VARIABLES),getStateVariablesCountWithPctNoHTML());
			
			rowAddNoSplit(HTMLConstants.WITHOUT_HTML, table, StringConstants.TOTAL_NUMBER_OF_STOCKS, /*StringConstants.STOCK_VARIABLES, */
					language.getTextHtml(StringConstants.TOTAL_NUMBER_OF_STOCKS),getStocksCountWithPctNoHTML());
			
			rowAddNoSplit(HTMLConstants.WITHOUT_HTML, table, StringConstants.TOTAL_NUMBER_OF_EXOGENOUS_VARIABLES, /*StringConstants.EXOGENOUS_VARIABLE_ANALYSIS, */
					language.getTextHtml(StringConstants.TOTAL_NUMBER_OF_EXOGENOUS_VARIABLES),getExogenousVariablesCountWithPctNoHTML());
			
			rowAddNoSplit(HTMLConstants.WITHOUT_HTML, table, StringConstants.TOTAL_NUMBER_OF_ENDOGENOUS_VARIABLES, /* StringConstants.ENDOGENOUS_VARIABLE_ANALYSIS,  */
					language.getTextHtml(StringConstants.TOTAL_NUMBER_OF_ENDOGENOUS_VARIABLES),getEndogenousVariablesCountWithPctNoHTML());
			
			
			String len =  documentor.getAnalysisManager().getLoopManager().getMaximumLoopLengthAsString();
			
			String extraStuff = " ("+language.getTextHtml(StringConstants.MAXIMUM_LOOP_LENGTH)+": "+len+") ["+
					documentor.getAnalysisManager().getLoopManager().getShortestLoopLengthNoIVV()+", "+
					documentor.getAnalysisManager().getLoopManager().getLongestLoopLengthNoIVV()+"]";
			
			rowAddNoSplit(HTMLConstants.WITHOUT_HTML, table, StringConstants.TOTAL_NUMBER_OF_FEEDBACK_LOOPS_NO_IVV, /* StringConstants.FEEDBACK_LOOPS, */
					language.getTextHtml(StringConstants.TOTAL_NUMBER_OF_FEEDBACK_LOOPS_NO_IVV)+ extraStuff, getFeedbackResultsNoIVVNoHTML());
			
			extraStuff = " ("+language.getTextHtml(StringConstants.MAXIMUM_LOOP_LENGTH)+": "+len+") ["+
					documentor.getAnalysisManager().getLoopManager().getShortestLoopLengthWithIVV()+", "+
					documentor.getAnalysisManager().getLoopManager().getLongestLoopLengthWithIVV()+"]";
			rowAddNoSplit(HTMLConstants.WITHOUT_HTML, table, StringConstants.TOTAL_NUMBER_OF_FEEDBACK_LOOPS_WITH_IVV, /* StringConstants.FEEDBACK_LOOPS_WITH_IVV, */
					language.getTextHtml(StringConstants.TOTAL_NUMBER_OF_FEEDBACK_LOOPS_WITH_IVV)+ extraStuff, getFeedbackResultsWithIVVNoHTML());
			
			rowAddNoSplit(HTMLConstants.WITHOUT_HTML, table, StringConstants.TOTAL_NUMBER_OF_CAUSAL_LINKS, /* StringConstants.CAUSAL_LINKS,  */
					language.getTextHtml(StringConstants.TOTAL_NUMBER_OF_CAUSAL_LINKS),getCausalLinkResultsNoHTML());
			
			rowAddNoSplit(HTMLConstants.WITHOUT_HTML, table, StringConstants.RATE_TO_RATE_LINKS, /* StringConstants.RATE_TO_RATE_LINKS, */
					language.getTextHtml(StringConstants.NUM_RATE_TO_RATE_LINKS),getRateToRateLinks());

			rowAddSeparator(table);
			
			rowAddNoSplit(HTMLConstants.WITHOUT_HTML, table, StringConstants.UNITS_USED, /*StringConstants.UNITS_NAME, */
					language.getTextHtml(StringConstants.UNITS_USED),getUnitsUsed());
			
			
			
			
			rowAddNoSplit(HTMLConstants.WITHOUT_HTML, table, StringConstants.TOTAL_NUMBER_OF_EQUATIONS_USING_MACROS, /* StringConstants.USING_MACROS, */
					language.getTextHtml(StringConstants.TOTAL_NUMBER_OF_EQUATIONS_USING_MACROS),getUsingMacrosCountWithPctNoHTML());
			
			rowAddNoSplit(HTMLConstants.WITHOUT_HTML, table, StringConstants.VARIABLES_WITH_SOURCE_INFORMATION,  /* StringConstants.VARIABLES_WITH_SOURCE_INFORMATION,  */
					language.getTextHtml(StringConstants.VARIABLES_WITH_SOURCE_INFORMATION), getSourceInformationCountWithPctNoHTML());
			
			rowAddNoSplit(HTMLConstants.WITHOUT_HTML, table, StringConstants.DIMENSIONLESS_UNIT_VARIABLES,  /*StringConstants.DIMENSIONLESS_UNIT_VARIABLES,  */
					language.getTextHtml(StringConstants.DIMENSIONLESS_UNIT_VARIABLES), getDmnlCountAsWithPctNoHTML());
			
			rowAddNoSplit(HTMLConstants.WITHOUT_HTML, table, StringConstants.VALUE_RANGE_SPECIFIED,  /*StringConstants.VALUE_RANGE_SPECIFIED,  */
					language.getTextHtml(StringConstants.VALUE_RANGE_SPECIFIED), getValueRangeCountAsWithPctNoHTML());
			
			
			rowAddNoSplit(HTMLConstants.WITHOUT_HTML, table, StringConstants.FUNCTION_SENSITIVITY_PARAMETERS, /* StringConstants.FUNCTION_SENSITIVITY_PARAMETERS, */
					language.getTextHtml(StringConstants.FUNCTION_SENSITIVITY_PARAMETERS),getSensitivityParametersCountWithPctNoHTML());
			
			rowAddNoSplit(HTMLConstants.WITHOUT_HTML, table, StringConstants.DATA_LOOKUP_TABLES, /* StringConstants.DATA_LOOKUP_TABLES, */
					language.getTextHtml(StringConstants.DATA_LOOKUP_TABLES),getLookupsCountWithPctNoHTML());
			
			rowAddSeparator(table);
			
			rowAddNoSplit(HTMLConstants.WITHOUT_HTML, table, StringConstants.TIME_UNIT, 
					language.getTextHtml(StringConstants.TIME_UNIT),timeUnit);
			rowAddNoSplit(HTMLConstants.WITHOUT_HTML, table, StringConstants.INITIAL_TIME, 
					language.getTextHtml(StringConstants.INITIAL_TIME),initialTime);
			rowAddNoSplit(HTMLConstants.WITHOUT_HTML, table, StringConstants.FINAL_TIME,
					language.getTextHtml(StringConstants.FINAL_TIME),finalTime);
			rowAddNoSplit(HTMLConstants.WITHOUT_HTML, table, StringConstants.REPORTED_TIME_INTERVAL, 
					language.getTextHtml(StringConstants.REPORTED_TIME_INTERVAL),saveper);
			rowAddNoSplit(HTMLConstants.WITHOUT_HTML, table, StringConstants.TIME_STEP, 
					language.getTextHtml(StringConstants.TIME_STEP),StringEscapeUtils.escapeHtml4(timeStep));
			
			rowAddSeparator(table);
			
			rowAdd(HTMLConstants.WITHOUT_HTML, table, StringConstants.MODEL_IS_FULLY_FORMULATED, 
					language.getTextHtml(StringConstants.MODEL_IS_FULLY_FORMULATED),fullyFormed);
			rowAdd(HTMLConstants.WITHOUT_HTML, table, StringConstants.MODELER_DEFINED_GROUPS, 
					language.getTextHtml(StringConstants.MODELER_DEFINED_GROUPS),definedGroups);


		return table;
	}


	@Override
	public List<List<List<String>>> getRows() {

		
		UserSelectedOptions options = InformationManagers.getInstance().getUserSelectedOptions();
		SDMDocLanguageSpecific language = SDMDocLanguageSpecific.getInstance();
		String t = UserSelectedOptions.TRUE;
		

		List<List<List<String>>> rows = new ArrayList<List<List<String>>>();
		
			rowAdd(HTMLConstants.WITH_HTML, rows, StringConstants.TOTAL_NUMBER_OF_VARIABLES, StringConstants.ALL_VARIABLES, 
					language.getTextHtml(StringConstants.TOTAL_NUMBER_OF_VARIABLES),getVariablesCountString());
			
			rowAdd(HTMLConstants.WITH_HTML, rows, StringConstants.TOTAL_NUMBER_OF_STATE_VARIABLES, StringConstants.STATE_VARIABLES,  
					language.getTextHtml(StringConstants.TOTAL_NUMBER_OF_STATE_VARIABLES),getStateVariablesCountWithPct());
			
			rowAdd(HTMLConstants.WITH_HTML, rows, StringConstants.TOTAL_NUMBER_OF_STOCKS, StringConstants.STOCK_VARIABLES, 
					language.getTextHtml(StringConstants.TOTAL_NUMBER_OF_STOCKS),getStocksCountWithPct());
			
			rowAdd(HTMLConstants.WITH_HTML, rows, StringConstants.TOTAL_NUMBER_OF_EXOGENOUS_VARIABLES, StringConstants.EXOGENOUS_VARIABLE_ANALYSIS, 
					language.getTextHtml(StringConstants.TOTAL_NUMBER_OF_EXOGENOUS_VARIABLES),getExogenousVariablesCountWithPct() /*getExogenousCount() */);
			rowAdd(HTMLConstants.WITH_HTML, rows, StringConstants.TOTAL_NUMBER_OF_ENDOGENOUS_VARIABLES, StringConstants.ENDOGENOUS_VARIABLE_ANALYSIS, 
					language.getTextHtml(StringConstants.TOTAL_NUMBER_OF_ENDOGENOUS_VARIABLES),getEndogenousVariablesCountWithPct() /*getEndogenousCount() */);
			
			
			String len =  documentor.getAnalysisManager().getLoopManager().getMaximumLoopLengthAsString();
			
			String extraStuff = " ("+language.getTextHtml(StringConstants.MAXIMUM_LOOP_LENGTH)+": "+len+") ["+
					documentor.getAnalysisManager().getLoopManager().getShortestLoopLengthNoIVV()+", "+
					documentor.getAnalysisManager().getLoopManager().getLongestLoopLengthNoIVV()+"]";
			
			rowAdd(HTMLConstants.WITH_HTML, rows, StringConstants.TOTAL_NUMBER_OF_FEEDBACK_LOOPS_NO_IVV, StringConstants.FEEDBACK_LOOPS, 
					language.getTextHtml(StringConstants.TOTAL_NUMBER_OF_FEEDBACK_LOOPS_NO_IVV),extraStuff, getFeedbackResultsNoIVV());
			
			extraStuff = " ("+language.getTextHtml(StringConstants.MAXIMUM_LOOP_LENGTH)+": "+len+") ["+
					documentor.getAnalysisManager().getLoopManager().getShortestLoopLengthWithIVV()+", "+
					documentor.getAnalysisManager().getLoopManager().getLongestLoopLengthWithIVV()+"]";
			rowAdd(HTMLConstants.WITH_HTML, rows, StringConstants.TOTAL_NUMBER_OF_FEEDBACK_LOOPS_WITH_IVV, StringConstants.FEEDBACK_LOOPS_WITH_IVV, 
					language.getTextHtml(StringConstants.TOTAL_NUMBER_OF_FEEDBACK_LOOPS_WITH_IVV), extraStuff, getFeedbackResultsWithIVV());
			
			rowAdd(HTMLConstants.WITH_HTML, rows, StringConstants.TOTAL_NUMBER_OF_CAUSAL_LINKS, StringConstants.CAUSAL_LINKS, 
					language.getTextHtml(StringConstants.TOTAL_NUMBER_OF_CAUSAL_LINKS),getCausalLinkResults());
			
			rowAdd(HTMLConstants.WITH_HTML, rows, StringConstants.RATE_TO_RATE_LINKS, StringConstants.RATE_TO_RATE_LINKS, 
					language.getTextHtml(StringConstants.NUM_RATE_TO_RATE_LINKS),getRateToRateLinks());

			rowAddSeparator(rows);
			
			rowAdd(HTMLConstants.WITH_HTML, rows, StringConstants.UNITS_USED, StringConstants.UNITS_NAME, 
					language.getTextHtml(StringConstants.UNITS_USED),getUnitsUsed());
			
			
			
			
			rowAdd(HTMLConstants.WITH_HTML, rows, StringConstants.TOTAL_NUMBER_OF_EQUATIONS_USING_MACROS, StringConstants.USING_MACROS, 
					language.getTextHtml(StringConstants.TOTAL_NUMBER_OF_EQUATIONS_USING_MACROS),getUsingMacrosCountWithPct());
			rowAdd(HTMLConstants.WITH_HTML, rows, StringConstants.VARIABLES_WITH_SOURCE_INFORMATION,  StringConstants.VARIABLES_WITH_SOURCE_INFORMATION,  
					language.getTextHtml(StringConstants.VARIABLES_WITH_SOURCE_INFORMATION), getSourceInformationCountWithPct());
			rowAdd(HTMLConstants.WITH_HTML, rows, StringConstants.DIMENSIONLESS_UNIT_VARIABLES,  StringConstants.DIMENSIONLESS_UNIT_VARIABLES,  
					language.getTextHtml(StringConstants.DIMENSIONLESS_UNIT_VARIABLES), getDmnlCountAsWithPct());
			
			rowAdd(HTMLConstants.WITH_HTML, rows, StringConstants.VALUE_RANGE_SPECIFIED,  StringConstants.VALUE_RANGE_SPECIFIED,  
					language.getTextHtml(StringConstants.VALUE_RANGE_SPECIFIED), getValueRangeCountAsWithPct());
			
			
			rowAdd(HTMLConstants.WITH_HTML, rows, StringConstants.FUNCTION_SENSITIVITY_PARAMETERS, StringConstants.FUNCTION_SENSITIVITY_PARAMETERS, 
					language.getTextHtml(StringConstants.FUNCTION_SENSITIVITY_PARAMETERS),getSensitivityParametersCountWithPct());
			rowAdd(HTMLConstants.WITH_HTML, rows, StringConstants.DATA_LOOKUP_TABLES, StringConstants.DATA_LOOKUP_TABLES, 
					language.getTextHtml(StringConstants.DATA_LOOKUP_TABLES),getLookupsCountWithPct());
			
			rowAddSeparator(rows);
			
			rowAdd(HTMLConstants.WITH_HTML, rows, StringConstants.TIME_UNIT, "", 
					language.getTextHtml(StringConstants.TIME_UNIT),timeUnit);
			rowAdd(HTMLConstants.WITH_HTML, rows, StringConstants.INITIAL_TIME, "", 
					language.getTextHtml(StringConstants.INITIAL_TIME),initialTime);
			rowAdd(HTMLConstants.WITH_HTML, rows, StringConstants.FINAL_TIME, "", 
					language.getTextHtml(StringConstants.FINAL_TIME),finalTime);
			rowAdd(HTMLConstants.WITH_HTML, rows, StringConstants.REPORTED_TIME_INTERVAL, "", 
					language.getTextHtml(StringConstants.REPORTED_TIME_INTERVAL),saveper);
			rowAdd(HTMLConstants.WITH_HTML, rows, StringConstants.TIME_STEP, "", 
					language.getTextHtml(StringConstants.TIME_STEP),StringEscapeUtils.escapeHtml4(timeStep));
			
			rowAddSeparator(rows);
			
			rowAdd(HTMLConstants.WITH_HTML, rows, StringConstants.MODEL_IS_FULLY_FORMULATED, "", 
					language.getTextHtml(StringConstants.MODEL_IS_FULLY_FORMULATED),fullyFormed);
			rowAdd(HTMLConstants.WITH_HTML, rows, StringConstants.MODELER_DEFINED_GROUPS, "", 
					language.getTextHtml(StringConstants.MODELER_DEFINED_GROUPS),definedGroups);
			
			if (options.isSelected(StringConstants.CREATE_FLAT_FILES)) {

				List<List<List<String>>> table = getTableData();
				//			System.out.println(table);
				String fn = InformationManagers.getInstance().getFileOrganizer().getFlatFileTablesDirectory()+"/"+"ModelInformation"+".tab";
				FlatFileWriter.writeFile(fn, table, "\t");
			}
		return rows;
	}

	public long getPositiveCausalLinks() {
		return positiveCausalLinks;
	}

	public long getNegativeCausalLinks() {
		return negativeCausalLinks;
	}
	
	public long getUnknownCausalLinks() {
		return unknownCausalLinks;
	}

	public long getEffectiveVariableCount() {
		return effectiveVariableCount;
	}

	public long getRateToRateLinks() {
		return rateToRateLinks;
	}

	@Override
	public List<List<List<String>>> getDataTableRows() {
		// TODO Auto-generated method stub
		return null;
	}


}
