package gov.anl.gss.sdm.server.model.documentor;

import gov.anl.gss.sdm.server.model.documentor.xmlwriter.NumberFormatter;
import gov.anl.gss.sdm.server.model.documentor.xmlwriter.QuickLinkAlphabet;
import gov.anl.gss.sdm.server.model.documentor.xmlwriter.StringConstants;
import gov.anl.gss.sdm.server.model.documentor.xmlwriter.UserSelectedOptions;
import gov.anl.gss.sdm.server.model.documentor.xmlwriter.variableselector.VariableSelector;
import gov.anl.gss.sdm.server.model.translator.Equation;
import gov.anl.gss.sdm.server.model.translator.InformationManagers;

import java.util.ArrayList;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;

import org.apache.commons.lang3.StringEscapeUtils;

public class ModelVariablesInformation extends TableData implements TableRow, TableDataLoad {
	
	private static final String HEADER = 
					"$$$ADEF$$$ ($$$TYPE$$$) $$$NAME$$$ ($$$NUM$$$ ##variables##)";
	
	private static final String TYPE_MARKER = "$$$TYPE$$$";
	private static final String NAME_MARKER = "$$$NAME$$$";
	private static final String NUM_MARKER = "$$$NUM$$$";

	
	private Map<String, VariableInformation> variableInformation;
	private String groupingName;
	private String groupingType;
	private VariableSelector selector;
	private String internalName;
	
	private QuickLinkAlphabet quickLinkAlphabet;

	public ModelVariablesInformation(Documentor documentor, String internalName, String groupingType, String grouping, VariableSelector selector) {
		super(documentor);
		this.groupingType = new String(groupingType);
		this.groupingName = new String(grouping);
		this.selector = selector;
		variableInformation = new LinkedHashMap<String, VariableInformation>();
		this.internalName = internalName;
		
		load();
//		quickLinkAlphabet = new QuickLinkAlphabet(variableInformation, "Detail"+groupingType+"-"+grouping);
	}

	@Override
	public void load() {
		for (String variableName : InformationManagers.getInstance().getEquationManager().getVariables()) {
			Equation eqn = InformationManagers.getInstance().getEquationManager().getAnEquationFor(variableName);
			if (!selector.isSelectSingleEquation(eqn))
				continue;

			//			if (groupingName.equals("Auxiliary")) {
			//				System.out.println("SELECTAUX: "+variableName);
			//			}
			VariableInformation vi = InformationManagers.getInstance().getVariableInformationFactory().
					getVariableInformation(documentor, documentor.getAnalysisManager(), variableName);
			variableInformation.put(variableName, vi);
		}

	}

	
	@Override
	public List<List<List<String>>> getDataTableRows() {
		// need to get this information from individual VariableInformation
		List<List<List<String>>> rows = new ArrayList<List<List<String>>>();
		// work only with this set of variables
		for (String key : orderedVariables()) {
			VariableInformation vi = variableInformation.get(key);
			if (vi == null || vi.isSkipMe())
				continue;
//			System.out.println("     >>> "+key);
			rows.add(vi.createDataTableRow());
		}
		
		
		return rows;
	}

	@Override
	public List<List<List<String>>> getRows() {
		// need to get this information from individual VariableInformation
		List<List<List<String>>> rows = new ArrayList<List<List<String>>>();
		// work only with this set of variables
		for (String key : orderedVariables()) {
			VariableInformation vi = variableInformation.get(key);
			if (vi == null || vi.isSkipMe())
				continue;

			rows.add(vi.createRow());
		}
		
		
		return rows;
	}

	public Map<String, VariableInformation> getVariableInformation() {
		return variableInformation;
	}
	

	private List<String> orderedVariables() {
		List<String> elist = InformationManagers.getInstance().getEquationLookup().orderedVariables();
		return elist;
	}
	
	
	public int size() {
		return variableInformation.size();
	}

	public String getGroupingName() {
		return groupingName;
	}

	public String getGroupingType() {
		return groupingType;
	}

	public QuickLinkAlphabet getQuickLinkAlphabet() {
		return quickLinkAlphabet;
	}

	public String getInternalName() {
		return internalName;
	}

	@Override
	public List<List<List<String>>> getTableData() {

		SDMDocLanguageSpecific language = SDMDocLanguageSpecific.getInstance();

		UserSelectedOptions options = InformationManagers.getInstance().getUserSelectedOptions();
		List<List<List<String>>> table = TableUtilities.createFlatFileTable();
		String titlel;


		//		($$$TYPE$$$) $$$NAME$$$ ($$$NUM$$$ ##variables##)
		titlel = (HEADER
				.replace(NAME_MARKER, StringEscapeUtils.escapeHtml4(groupingName))
				.replace(NUM_MARKER, NumberFormatter.formatForCountNotConditional(23))
				.replace(TYPE_MARKER, groupingType)
				.replace("##variables##", language.getTextHtml(StringConstants.VARIABLES)));
		TableUtilities.addTitle(table, titlel);

		List<List<String>> row = TableUtilities.createRow();
		TableUtilities.addRow(table, row);
		if (documentor.displayModule()) {
			TableUtilities.addColumn(row, 
					TableUtilities.createColumn(language.getTextHtml(StringConstants.MODULE)));
		}

		TableUtilities.addColumn(row, TableUtilities.createColumn(language.getTextHtml(StringConstants.GROUP)));
		TableUtilities.addColumn(row, TableUtilities.createColumn(language.getTextHtml(StringConstants.UNITS_TYPE)));
		TableUtilities.addColumn(row, TableUtilities.createColumn(language.getTextHtml(StringConstants.VARIABLE)));


		if (options.isSelected(StringConstants.DESCRIPTION)) {
			TableUtilities.addColumn(row, TableUtilities.createColumn(language.getTextHtml(StringConstants.DESCRIPTION)));
}
		if (options.isSelected(StringConstants.PRESENT_IN_VIEWS)) {
			TableUtilities.addColumn(row, TableUtilities.createColumn(language.getTextHtml(StringConstants.PRESENT_IN_VIEWS)));
		}
		if (options.isSelected(StringConstants.USED_BY)) {
			TableUtilities.addColumn(row, TableUtilities.createColumn(language.getTextHtml(StringConstants.USED_BY)));
		}
		if (options.isSelected(StringConstants.FEEDBACK_LOOPS)) {
			TableUtilities.addColumn(row, TableUtilities.createColumn(language.getTextHtml(StringConstants.NUMBER_OF_FEEDBACK_LOOPS)));
			TableUtilities.addColumn(row, TableUtilities.createColumn(language.getTextHtml(StringConstants.PCT_OF_FEEDBACK_LOOPS)));

			TableUtilities.addColumn(row, TableUtilities.createColumn(language.getTextHtml(StringConstants.POSITIVE_POLARITY_FEEDBACK_LOOPS)));
			TableUtilities.addColumn(row, TableUtilities.createColumn(language.getTextHtml(StringConstants.POSITIVE_POLARITY_MIN_LENGTH_FEEDBACK_LOOPS)));
			TableUtilities.addColumn(row, TableUtilities.createColumn(language.getTextHtml(StringConstants.POSITIVE_POLARITY_MAX_LENGTH_FEEDBACK_LOOPS)));

			TableUtilities.addColumn(row, TableUtilities.createColumn(language.getTextHtml(StringConstants.NEGATIVE_POLARITY_FEEDBACK_LOOPS)));
			TableUtilities.addColumn(row, TableUtilities.createColumn(language.getTextHtml(StringConstants.NEGATIVE_POLARITY_MIN_LENGTH_FEEDBACK_LOOPS)));
			TableUtilities.addColumn(row, TableUtilities.createColumn(language.getTextHtml(StringConstants.NEGATIVE_POLARITY_MAX_LENGTH_FEEDBACK_LOOPS)));
		}

		return table;
	}


}
