package gov.anl.gss.sdm.server.model.documentor;

import gov.anl.gss.sdm.server.model.documentor.xmlwriter.Anchors;
import gov.anl.gss.sdm.server.model.documentor.xmlwriter.NumberFormatter;
import gov.anl.gss.sdm.server.model.documentor.xmlwriter.StringConstants;

import java.util.ArrayList;
import java.util.Collections;
import java.util.LinkedHashMap;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

public class ViewVariables {
	
	private static final String TITLE = "<a name='byView' />$$$TOP$$$<h3>##List of## $$$NUM_VIEWS$$$ ##Views And Their## $$$NUM_VARIABLES$$$ ##Variables##</h3>\n";
	private static final String TABLE_HEADER = "<table class='sample'>\n";
	private static final String TABLE_FOOTER = "</table>\n";
	private static final String BLANK_COLUMN_TITLE = "<td>&nbsp</td>\n";
	private static final String COLUMN_TITLE = "<td style='vertical-align:bottom; color:Blue'><center>$$$VIEW_ANCHOR$$$</center></td>\n";
	private static final String BLANK_COUNT_HEADER = "<td style='vertical-align:bottom; color:Blue; text-align:right'><b>$$$TOTAL$$$:</b></td>\n";
	private static final String BLANK_COUNT_FOOTER = "<td style='vertical-align:bottom; color:Blue;'><b>$$$TOTAL$$$:</b></td>\n";
	private static final String COUNT_HEADER = "<td style='vertical-align:bottom; color:Blue; text-align:center;'><b>$$$NUM_VARIABLES$$$</b></td>\n";
	private static final String COLUMN_HEADER = "<td style='text-align:center;color:White;'><b><font style='background-color:$$$COLOR$$$'>$$$CHARACTER$$$</font></b></td>\n";
	private static final String VARIABLE_HEADER = "<td style='vertical-align:bottom; color:Blue; text-align:right'>$$$VARIABLE_ANCHOR$$$ (##in## $$$NUM_VIEWS$$$ ##view##)</td>\n";
	
	private static final String TOP_MARKER = "$$$TOP$$$";
	private static final String NUM_VIEWS_MARKER = "$$$NUM_VIEWS$$$";
	private static final String NUM_VARIABLES_MARKER = "$$$NUM_VARIABLES$$$";
	
	private static final String VIEW_ANCHOR_MARKER = "$$$VIEW_ANCHOR$$$";
	private static final String VIEW_NAME_MARKER = "$$$VIEW_NAME$$$";
	
	private static final String VARIABLE_ANCHOR_MARKER = "$$$VARIABLE_ANCHOR$$$";
	private static final String VARIABLE_NAME_MARKER = "$$$VARIABLE_NAME$$$";
	
	private static final String COLOR_MARKER = "$$$COLOR$$$";
	private static final String CHARACTER_MARKER = "$$$CHARACTER$$$";
	
	private static final String NOT_IN_VIEW_CHARACTER = "&nbsp;";
	private static final String IN_VIEW_CHARACTER = "&nbsp;&nbsp;&nbsp;";
	
	private static final String NOT_IN_VIEW_COLOR = "White;";
	private static final String IN_VIEW_COLOR = "Blue";
	
	private static final String ROW_START = "<tr>\n";
	private static final String ROW_END = "</tr>\n";
	
	private static final String COLUMN_START = "<td>\n";
	private static final String COLUMN_END = "</td>\n";
	
	
	
	private Map<String, Set<String>> viewVariables;
	private Map<String, Set<String>> variableViews;
	private Map<String, Integer> viewOrder;
	private List<String> views;
	
	public ViewVariables(Map<String, Set<String>> viewVariables) {
		this.viewVariables = viewVariables;
		variableViews = new LinkedHashMap<String, Set<String>>();
		viewOrder = new LinkedHashMap<String, Integer>();
		loadVariableViews();
		loadViewOrder();
	}

	private void loadVariableViews() {
		for (String view : viewVariables.keySet()) {
			for (String variable : viewVariables.get(view)) {
				if (!variableViews.containsKey(variable))
					variableViews.put(variable, new LinkedHashSet<String>());
				variableViews.get(variable).add(view);
			}
		}
	}
	
	private void loadViewOrder() {
		views = new ArrayList<String>(viewVariables.keySet());
		Collections.sort(views);
		int vNum = 0;
		for (String view : views) {
			viewOrder.put(view, vNum++);
		}
	}
	
	public String getXML() {
		SDMDocLanguageSpecific language = SDMDocLanguageSpecific.getInstance();
		
		StringBuilder sb = new StringBuilder();

		sb.append(TITLE.replace(NUM_VIEWS_MARKER, NumberFormatter.formatForCountNotConditional(viewVariables.size()))
				.replace(NUM_VARIABLES_MARKER, NumberFormatter.formatForCountNotConditional(variableViews.size()))
				.replace(TOP_MARKER, Anchors.getAnchorReference("top", language.getTextHtml(StringConstants.TOP)))
				.replace("##List of##", language.getTextHtml(StringConstants.LIST_OF))
				.replace("##Views And Their##", language.getTextHtml(StringConstants.VIEWS_AND_THEIR))
				.replace("##Variables##", language.getTextHtml(StringConstants.VARIABLES)));
		sb.append(TABLE_HEADER);
		
		// VIEW NAMES
		sb.append(ROW_START);
		sb.append(BLANK_COLUMN_TITLE);
		for (String view : views) {
			sb.append(COLUMN_TITLE.replace(VIEW_ANCHOR_MARKER, Anchors.getAnchorReference("-VIEW-"+view)));
		}
		sb.append(BLANK_COLUMN_TITLE);
		sb.append(ROW_END);

		// VIEW Counts
		sb.append(ROW_START);
		sb.append(BLANK_COUNT_HEADER.replace("$$$TOTAL$$$", language.getTextHtml(StringConstants.TOTAL)));
		for (String view : views) {
			sb.append(COUNT_HEADER.replace(NUM_VARIABLES_MARKER, NumberFormatter.formatForCountNotConditional(viewVariables.get(view).size())));
		}
		sb.append(BLANK_COUNT_FOOTER.replace("$$$TOTAL$$$", language.getTextHtml(StringConstants.TOTAL)));
		sb.append(ROW_END);

		// VARIABLES
		for (String variable : variableViews.keySet()) {
			sb.append(ROW_START);
			
			String singularOrPlural = "View";
			if (variableViews.get(variable).size() > 1)
				singularOrPlural = "Views";
			sb.append(VARIABLE_HEADER.replace(VARIABLE_ANCHOR_MARKER, Anchors.getAnchorReference(variable)).replace(NUM_VIEWS_MARKER, NumberFormatter.formatForCountNotConditional(variableViews.get(variable).size()))
					.replace("##in##", language.getTextHtml(StringConstants.IN)).replace("##view##", language.getTextHtml(singularOrPlural)));
			sb.append(rowForVariable(variable));
			sb.append(VARIABLE_HEADER.replace(VARIABLE_ANCHOR_MARKER, Anchors.getAnchorReference(variable)).replace(NUM_VIEWS_MARKER, NumberFormatter.formatForCountNotConditional(variableViews.get(variable).size()))
					.replace("##in##", language.getTextHtml(StringConstants.IN)).replace("##view##", language.getTextHtml(singularOrPlural)));
			sb.append(ROW_END);
		}

		// VIEW Counts
		sb.append(ROW_START);
		sb.append(BLANK_COUNT_HEADER.replace("$$$TOTAL$$$", language.getTextHtml(StringConstants.TOTAL)));
		for (String view : views) {
			sb.append(COUNT_HEADER.replace(NUM_VARIABLES_MARKER, NumberFormatter.formatForCountNotConditional(viewVariables.get(view).size())));
		}
		sb.append(BLANK_COUNT_FOOTER.replace("$$$TOTAL$$$", language.getTextHtml(StringConstants.TOTAL)));
		sb.append(ROW_END);

		// VIEW NAMES
		sb.append(ROW_START);
		sb.append(BLANK_COLUMN_TITLE);
		for (String view : views) {
			sb.append(COLUMN_TITLE.replace(VIEW_ANCHOR_MARKER, Anchors.getAnchorReference("-VIEW-"+view)).replace(VIEW_NAME_MARKER, view));
		}
		sb.append(BLANK_COLUMN_TITLE);
		sb.append(ROW_END);
		sb.append(TABLE_FOOTER);

		return sb.toString();
	}
	
	private String rowForVariable(String variable) {
		StringBuilder sb = new StringBuilder();
		String[] col = new String[viewVariables.keySet().size()];
		for (int i = 0; i < viewVariables.keySet().size(); i++) {
			col[i] = new String(COLUMN_HEADER);
		}
		for (String view : variableViews.get(variable)) {
			col[viewOrder.get(view)] = col[viewOrder.get(view)].replace(COLOR_MARKER, IN_VIEW_COLOR).replace(CHARACTER_MARKER, IN_VIEW_CHARACTER);
		}
		
		for (int i = 0; i < viewVariables.keySet().size(); i++) {
		col[i] = col[i].replace(COLOR_MARKER, NOT_IN_VIEW_COLOR).replace(CHARACTER_MARKER, NOT_IN_VIEW_CHARACTER);
		}
		
		for (String xml : col)
			sb.append(xml);
		
		return sb.toString();
		
	}

	public Map<String, Set<String>> getViewVariables() {
		return viewVariables;
	}

	public Map<String, Set<String>> getVariableViews() {
		return variableViews;
	}

	public Map<String, Integer> getViewOrder() {
		return viewOrder;
	}

	public List<String> getViews() {
		return views;
	}
}
