package gov.anl.gss.sdm.server.model.support.database;

import java.io.File;
import java.io.IOException;
import java.util.LinkedHashSet;
import java.util.Set;

import org.apache.commons.io.FileUtils;

public class ModelRunDatabaseVerifier {
	
	private String directory;
	
	private static String   dbIdentifier = "HSQLDB";
	private static String[] closedFileExtensions = {".data", ".properties", ".script"};
	private static String[] openFileExtensions = {".tmp", ".lck", ".log"};
	
	private Set<String> openDatabases;
	private Set<String> closedDatabases;
	private Set<String> allDatabases;
	
	private Set<String> incompleteDatabases;
	
	private boolean clean = true;
	
	public ModelRunDatabaseVerifier(String directory) {
		
		this.directory = directory;
		openDatabases = new LinkedHashSet<String>();
		closedDatabases = new LinkedHashSet<String>();
		allDatabases = new LinkedHashSet<String>();
		incompleteDatabases = new LinkedHashSet<String>();
		analyze();
		
	}
	
	private void analyze() {
		File d = new File(directory);
		for (String file : d.list()) {
			String dbName = extractDBName(file);
			if (dbName != null) {
				allDatabases.add(dbName);
			}
		}
		for (String db : allDatabases) {
			if (isComplete(db)) {
				if (isOpen(db))
					openDatabases.add(db);
				else
					closedDatabases.add(db);
			} else {
				incompleteDatabases.add(db);
			}
		}
		
		if (incompleteDatabases.size() > 0 || openDatabases.size() > 0)
			clean = false;
	}
	
	public boolean cleanup() {
		boolean success = true;
		for (String db : openDatabases) {
			success = success && forceClose(db);
		}
		
		for (String db : incompleteDatabases) {
			success = success && removeDB(db);
		}
			
		return success;
	}
	
	private boolean removeDB(String db) {
		boolean success = true;
		for (String ext : closedFileExtensions) {
			String file = directory+"/"+db+dbIdentifier+ext;
			File f = new File(file);
			if (f.exists())
				success = success && f.delete();
		}
		
		for (String ext : openFileExtensions) {
			String file = directory+"/"+db+dbIdentifier+ext;
			File f = new File(file);
			if (f.exists())
				success = success && f.delete();
		}
		
		return success;
	}
	
	public static boolean removeDB(String directory, String db) {
		boolean success = true;
		for (String ext : closedFileExtensions) {
			String file = directory+"/"+db+dbIdentifier+ext;
			File f = new File(file);
			if (f.exists())
				success = success && f.delete();
		}
		
		for (String ext : openFileExtensions) {
			String file = directory+"/"+db+dbIdentifier+ext;
			File f = new File(file);
			if (f.exists())
				success = success && f.delete();
		}
		
		return success;
	}
	
	public boolean exists(String db) {
		boolean exist = true;
		// exist -- true if closedFileExtensions exist
		// won't know if opne or closed
		for (String ext : closedFileExtensions) {
			
			String file = directory+"/"+db+dbIdentifier+ext;
			if (!new File(file).exists())
				exist = false;
			
		}
		return exist;
	}
	
	public boolean isOpen(String db) {
		
		if (!exists(db))
			return false;
		
		boolean open = true;
		for (String ext : openFileExtensions) {
			
			String file = directory+"/"+db+dbIdentifier+ext;
			if (!new File(file).exists())
				open = false;
			
		}
		return open;
	}
	
	public boolean isClosed(String db) {
		boolean closed = true;
		for (String ext : openFileExtensions) {
			
			String file = directory+"/"+db+dbIdentifier+ext;
			if (new File(file).exists())
				closed = false;
			
		}
		return closed;
	}
	
	public boolean isComplete(String db) {
		
		int numDBFiles = 0;
		for (String ext : closedFileExtensions) {
			if(new File(directory+"/"+db+dbIdentifier+ext).exists())
				numDBFiles++;
		}
		
		return numDBFiles == 3;
	}
	
	public boolean restore() {
		return forceClose();
	}
	
	public static boolean purge(String directory) {
		
		// delete everything in this directory
		boolean success = true;
		boolean frc = true;
		File d = new File(directory);
		for (String f : d.list()) {
			
				File fil = new File(directory+"/"+f);
				if (fil.isDirectory()) {
					try {
						FileUtils.deleteDirectory(fil);
					} catch (IOException e) {
						// TODO Auto-generated catch block
						e.printStackTrace();
					}
				} else {
					try {
						fil.setWritable(true);
						frc = fil.delete();
					} catch (Exception e) {
						// TODO Auto-generated catch block
						e.printStackTrace();
					}
				}
			
		}
		return success;
	}
	
	public boolean forceClose() {
		// force close all database in directory
		// ignore non-db files
		boolean success = true;
		boolean frc = true;
		File d = new File(directory);
		for (String f : d.list()) {
			if (isOpenDBFile(f)) {
				File fil = new File(directory+"/"+f);
				if (fil.isDirectory()) {
					try {
						FileUtils.deleteDirectory(fil);
					} catch (IOException e) {
						// TODO Auto-generated catch block
						e.printStackTrace();
					}
				} else {
					try {
						fil.setWritable(true);
						frc = fil.delete();
					} catch (Exception e) {
						// TODO Auto-generated catch block
						e.printStackTrace();
					}
				}
			}
		}
		return success;
	}
	
	public boolean forceClose(String db) {
		boolean force = true;
		boolean frc = true;
		
		for (String ext : closedFileExtensions) {
			
			String file = directory+"/"+db+dbIdentifier+ext;
			File f = new File(file);
			if (f.isDirectory()) {
				try {
					FileUtils.deleteDirectory(f);
				} catch (IOException e) {
					// TODO Auto-generated catch block
					e.printStackTrace();
				}
			} else {
				try {
					frc = f.delete();
				} catch (Exception e) {
					// TODO Auto-generated catch block
					e.printStackTrace();
				}
			}
			
		}
		return force;
	}
	
	public boolean isClosedDBFile(String file) {
		
		for (String s : closedFileExtensions) {
			String tail = dbIdentifier + s;
			if (file.endsWith(tail))
				return true;
		}
		
		return false;
	}
	
	public boolean isOpenDBFile(String file) {
		
		for (String s : openFileExtensions) {
			String tail = dbIdentifier + s;
			if (file.endsWith(tail))
				return true;
		}
		
		return false;
	}
	
	public String extractDBName(String file) {
		
		for (String s : openFileExtensions) {
			String tail = dbIdentifier + s;
			if (file.endsWith(tail))
				return file.replace(tail, "");
		}
		
		for (String s : closedFileExtensions) {
			String tail = dbIdentifier + s;
			if (file.endsWith(tail))
				return file.replace(tail, "");
		}
		
		return null;
	}

	public boolean isClean() {
		return clean;
	}

}
