package gov.anl.gss.sdm.server.model.support.database;

import gov.anl.gss.sdm.server.model.documentor.MessageGobbler;
import gov.anl.gss.sdm.server.model.support.MemoryMapper;
import gov.anl.gss.sdm.server.model.support.MemoryMapperInterface;
import gov.anl.gss.sdm.server.model.support.StreamlinedMemoryMapper;
import gov.anl.gss.sdm.server.model.support.Utilities;

import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Statement;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;

import org.apache.commons.lang.ArrayUtils;
import org.hsqldb.jdbc.JDBCArrayBasic;

public class SDModelRunDatabase {

	/**
	 * This is the gateway to the SDM-Doc/SDM-UI results database
	 * 
	 * Supported Task:
	 * 
	 * database creation based on .map file
	 * database update
	 * database access
	 */

	private String dbFile;
	private String mapFile;
	private DBConnection dbConn;

	private PreparedStatement psConstants;
	private PreparedStatement psTimeSeries;
	
//	private PreparedStatement psLookups;
//	private PreparedStatement psLookupsValues;
	
	private double[] timeVector = null;

	private Map<String, ModelVariable> modelVariables = new LinkedHashMap<String, ModelVariable>();
	
	private double startTime;
	private double endTime;
	private double timeStep;
	private double dt;

	public SDModelRunDatabase(String dbFile) {
		this.dbFile = dbFile;
		try {
			dbConn = new DBConnection(dbFile);
			
			// verify that database is readable
		} catch (Exception e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
	}

	public void shutdown() {
		try {
			
			dbConn.shutdown();
		} catch (SQLException e) {
			// TODO Auto-generated catch block
//			e.printStackTrace();
		}
	}
	
	public void reinitialize() {
		exec("delete from constants");
		exec("delete from timeSeries");
	}
	
	public void setRunParameters(double initialTime, double finalTime, double dt, double savper) {
		startTime = initialTime;
		endTime = finalTime;
		timeStep = dt;
		this.dt = dt;
		exec("drop table RunParameters If Exists");
		exec("CREATE Cached TABLE RunParameters( id INTEGER IDENTITY, InitialTime DOUBLE,"
				+ " FinalTime DOUBLE, DT DOUBLE, SAVPER DOUBLE)");
		exec("insert into RunParameters(InitialTime, FinalTime, DT, SAVPER) VALUES"
				+ " ("+initialTime+","+finalTime+","+dt+","+savper+")");
	}
	
	// HERE
	public double[][] getVarsValues(String[] variables) {
		// return the contents of the constants and timeseries tables
		double[][] varValues = new double[variables.length][];
		
		// note that some variables will be constants and some will be
		StringBuilder constantsSql = new StringBuilder("Select modelTime, ");
		StringBuilder dynamicsSql = new StringBuilder("Select ");
		createColumnSelect(variables, constantsSql, dynamicsSql);
		constantsSql.append(" from Constants order by modelTime");
		dynamicsSql.append(" from TimeSeries order by modelTime");
		double[][] cValues = null;
		
		if (constantsSql.toString().contains("DataValue")) {
			cValues = DBUtilities.getExpandedConstantsDoubles(dbConn, constantsSql.toString(), timeVector);
		}
		
//		double[][] cValues = DBUtilities.getExpandedConstantsDoubles(dbConn, constantsSql.toString(), timeVector);
		double[][] dValues = DBUtilities.getTableDoubles(dbConn, dynamicsSql.toString());
		int vIndex = 0;
		int cIndex = 0;
		int dIndex = 0;
		for (String var : variables) {
			ModelVariable mv = modelVariables.get(var);
			String type = mv.getType();
			int index = mv.getIndex()+1;
			if (type.equals(MemoryMapper.CONSTANT)) {
				varValues[vIndex] = cValues[cIndex++];
			} else {
				varValues[vIndex] = dValues[dIndex++];
			}
			vIndex++;
		}
		return varValues;
		
	}
	
	private String buildSelectOfLength(int len) {
		StringBuilder sb = new StringBuilder();
		for (int i = 0; i < len; i++) {
			if (i > 0)
				sb.append(",");
			sb.append("DataValues["+(i+1)+"]");
		}
		return sb.toString();
	}
	
	private void resetConstantMemory(double[] mem, String sql, double time) {
		double[] tmp = DBUtilities.getConstantDoubles(dbConn, sql, time);
		for (int i = 0; i < tmp.length; i++) {
			mem[i] = tmp[i];
		}
	}
	
	private void resetTimeSeriesMemory(double[] mem, String sql) {
		double[] tmp = DBUtilities.getDoubles(dbConn, sql);
		for (int i = 0; i < tmp.length; i++) {
			mem[i] = tmp[i];
		}
	}
	
	public void backstepToTime(double time, double[] cMem, double vMem[]) {
		
		double priorToTime = time;//  - dt;
		// reload memory
		
		StringBuilder constantsSQL = new StringBuilder("select modelTime, ");
		StringBuilder timeseriesSQL = new StringBuilder("select ");
		
		constantsSQL.append(buildSelectOfLength(cMem.length));
		timeseriesSQL.append(buildSelectOfLength(vMem.length));
		
		constantsSQL.append(" from constants order by modelTime");
		timeseriesSQL.append("from timeseries where modelTime = "+priorToTime);
		
		resetConstantMemory(cMem, constantsSQL.toString(), time);
		resetTimeSeriesMemory(vMem, timeseriesSQL.toString());
		
		// delete from db
		
		exec("delete from constants where modelTime >= "+time);
		exec("delete from timeseries where modelTime > "+time);
		
		
		
	}

	public void printVariables() {
		try {
			dbConn.query("select * from variables");
		} catch (SQLException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
	}


	public void initializeNewRunDatabase(String mapFile) {
		this.mapFile = mapFile;
		createRunStructureTables(mapFile);
		makePrepared();
	}
	
	public boolean dbUsable() {
		Statement st = null;
		ResultSet rs = null;

		//        exec("CREATE Cached TABLE Variables ( id INTEGER IDENTITY, VariableName VARCHAR(256), "
		//				+ "VariableType VARCHAR(2), VariableIndex INTEGER, numElements INTEGER)");

		st = dbConn.createStatement(ResultSet.TYPE_SCROLL_INSENSITIVE, ResultSet.CONCUR_READ_ONLY);
		try {
			rs = st.executeQuery("Select VariableName, VariableType, VariableIndex, numElements from Variables");
			st.close();
			return true;
		} catch (SQLException e) {
			// TODO Auto-generated catch block
//			e.printStackTrace();
			return false;
		}
	}

	public void load() {
		Statement st = null;
		ResultSet rs = null;

		//        exec("CREATE Cached TABLE Variables ( id INTEGER IDENTITY, VariableName VARCHAR(256), "
		//				+ "VariableType VARCHAR(2), VariableIndex INTEGER, numElements INTEGER)");

		st = dbConn.createStatement(ResultSet.TYPE_SCROLL_INSENSITIVE, ResultSet.CONCUR_READ_ONLY);
		try {
			rs = st.executeQuery("Select VariableName, VariableType, VariableIndex, numElements from Variables");
			loadVariables(rs);
			st.close();
		} catch (SQLException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
	}
	public void loadRunParameters() {
		
		startTime = getStartTime();
		endTime = getEndTime();
		timeStep = getTimeStep();
		dt = getDt();
	}

	public void loadVariables(ResultSet rs) throws SQLException {
		for (; rs.next(); ) {
			String name = rs.getString(1);
			String type = rs.getString(2);
			int index = rs.getInt(3);
			int numElements = rs.getInt(4);
			ModelVariable mv = new ModelVariable(name, type, index, numElements);
			modelVariables.put(mv.getName(), mv);
		}
	}

	public double[] getTimeVector() {
		if (timeVector == null)
			timeVector = getTimeVector("Select distinct modelTime from TimeSeries order by modelTime");
		return timeVector;
	}
	

	public double[][] getVarTimeVector(String[] vars) {
		double[][] varValues = new double[vars.length][];
		int vIndex = 0;
		for (String var : vars) {
			ModelVariable mv = modelVariables.get(var);
			String type = mv.getType();
			int index = mv.getIndex()+1;
			if (type.equals(MemoryMapper.CONSTANT))
				varValues[vIndex] =  getConstantAsTimeVector("Select modelTime, DataValues["+index+"] from Constants order by modelTime");
			else
				varValues[vIndex] =   getTimeVector("Select DataValues["+index+"], modelTime from TimeSeries order by modelTime");
			vIndex++;
		}
		return varValues;
	}
	
	public double[] getVarTimeVector(String var, double time) {
		ModelVariable mv = modelVariables.get(var);
		String type = mv.getType();
		int index = mv.getIndex()+1;
		if (type.equals(MemoryMapper.CONSTANT))
			return getConstantAsTimeVector("Select modelTime, DataValues["+index+"] from Constants order by modelTime", time);
		else
			return getTimeVector("Select DataValues["+index+"], modelTime from TimeSeries where modelTime <= "+time+" order by modelTime");
	}
	
	public double getVarAtTime(String var, double time) {
		ModelVariable mv = modelVariables.get(var);
		String type = mv.getType();
		int index = mv.getIndex()+1;
		if (type.equals(MemoryMapper.CONSTANT))
			return getSingleDoubleValue("Select modelTime, DataValues["+index+"] from Constants order by modelTime", time);
		else
			return getSingleDoubleValue("Select DataValues["+index+"], modelTime from TimeSeries where modelTime = "+time+" order by modelTime");
	}
	
	private boolean doesDbExist() {
		return DBUtilities.tableExists(dbConn, "TIMESERIES");
	}
	
	public boolean isCompletedRun() {
		boolean completed = false;
		double lastTime = 0.0;
		boolean dbExists = doesDbExist();
		if (!dbExists)
			return false;
		try {
			lastTime = getSingleDoubleValue("Select MAX(modelTime) from TimeSeries");
		} catch (Exception e) {
			// TODO Auto-generated catch block
//			e.printStackTrace();
			System.out.println("Run Not Generated");
			return false;
		}
		if (lastTime == getEndTime())
			completed = true;
		return completed;

	}
	
	public boolean validate() {
		boolean cleanDB = true;
		
		String[] requiredTables = {"Constants", "TimeSeries", "Variables", "Lookups", "LookupsValues"};
		
		for (String table : requiredTables) {
			if (!DBUtilities.tableExists(dbConn, table))
				cleanDB = false;
		}
		return cleanDB;
		
	}
	
	public double[] getVarTimeVectorByIndex(String type, int index) {
		if (type.equals(MemoryMapper.CONSTANT))
			return getConstantAsTimeVector("Select modelTime, DataValues["+index+"] from Constants order by modelTime");
		else
			return getTimeVector("Select DataValues["+index+"], modelTime from TimeSeries order by modelTime");
	}
	
	public double[] getVarTimeVector(String var) {
		

		ModelVariable mv = modelVariables.get(var);
		
		if (mv == null) {
			System.out.println("No modelvariable for "+var);
			return null;
		}

		String type = mv.getType();
		int index = mv.getIndex()+1;
		if (type.equals(MemoryMapper.CONSTANT))
			return getConstantAsTimeVector("Select modelTime, DataValues["+index+"] from Constants order by modelTime");
		else
			return getTimeVector("Select DataValues["+index+"], modelTime from TimeSeries order by modelTime");
	}
	
	private void createColumnSelect(String[] vars, StringBuilder constantsList, StringBuilder dynamicsList) {
		int cCnt = 0;
		int dCnt = 0;
		for (String var : vars) {
			ModelVariable mv = modelVariables.get(var);
			
			if (mv == null) {
				continue;
			}
			
			String type = mv.getType();
			int index = mv.getIndex()+1;
			if (type.equals(MemoryMapper.CONSTANT)) {
				if (cCnt++ > 0)
					constantsList.append(",");
				constantsList.append("DataValues["+index+"]");
			} else {
				if (dCnt++ > 0)
					dynamicsList.append(",");
				dynamicsList.append("DataValues["+index+"]");
			}
		}
	}
	
	// this should only be used for non-memory resident data
	// multiple variables, single time point
	public double[] getVarsValue(String[] vars, double time) {
		
		double[] varValues = new double[vars.length];
		
		// note that some variables will be constants and some will be
		StringBuilder constantsSql = new StringBuilder("Select modelTime, ");
		StringBuilder dynamicsSql = new StringBuilder("Select ");
		createColumnSelect(vars, constantsSql, dynamicsSql);
		constantsSql.append(" from Constants order by modelTime");
		dynamicsSql.append(" from TimeSeries where modelTime = "+time);
		
		double[] cValues = DBUtilities.getConstantDoubles(dbConn, constantsSql.toString(), time);
		double[] dValues = DBUtilities.getDoubles(dbConn, dynamicsSql.toString());
		
		int vIndex = 0;
		int cIndex = 0;
		int dIndex = 0;
		for (String var : vars) {
			ModelVariable mv = modelVariables.get(var);
			String type = mv.getType();
			int index = mv.getIndex()+1;
			if (type.equals(MemoryMapper.CONSTANT)) {
				varValues[vIndex] = cValues[cIndex++];
			} else {
				varValues[vIndex] = dValues[dIndex++];
			}
			vIndex++;
		}
		return varValues;
	}
	
	public double[] getVarValue(String[] vars, double time) {
		return getVarsValue(vars, time);
//		double[] varValues = new double[vars.length];
//		int vIndex = 0;
//		for (String var : vars) {
//			ModelVariable mv = modelVariables.get(var);
//			String type = mv.getType();
//			int index = mv.getIndex()+1;
//			if (type.equals(MemoryMapper.CONSTANT))
//				varValues[vIndex] = getSingleDoubleValue("Select DataValues["+index+"] from Constants");
//			else
//				varValues[vIndex] = getSingleDoubleValue("Select DataValues["+index+"] from TimeSeries where modelTime = "+time);
//			vIndex++;
//		}
//		return varValues;
	}
	
	public double getVarValue(String var, double time) {
		ModelVariable mv = modelVariables.get(var);
		String type = mv.getType();
		int index = mv.getIndex()+1;
		if (type.equals(MemoryMapper.CONSTANT))
			return getSingleDoubleValue("Select DataValues["+index+"] from Constants order by modelTime");
		else
			return getSingleDoubleValue("Select DataValues["+index+"] from TimeSeries where modelTime = "+time);
	}
	
	public double[] getConstantAsTimeVector(String sql, double time) {
		
		Statement st = null;
		ResultSet rs = null;
		int len = (int) ((time-startTime)/dt) + 1;
		double[] tv = new double[len];
		
		st = dbConn.createStatement();
		try {
			rs = st.executeQuery(sql);
			for (; rs.next(); ) {
				double tt = rs.getDouble(1);
				double value = rs.getDouble(2);
				int index = (int) ((tt-startTime)/dt);
				for (int i = index; i < tv.length; i++) {
					tv[i] = value;
				}
			}
			st.close();
		} catch (SQLException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
		
		
	
		return tv;
		
	}
	public double[] getConstantAsTimeVector(String sql) {

		Statement st = null;
		ResultSet rs = null;
		if (timeVector == null) {
			getTimeVector();
		}
		double[] tv = new double[timeVector.length];

		st = dbConn.createStatement();
		try {
			rs = st.executeQuery(sql);

			for (; rs.next(); ) {
				double t = rs.getDouble(1);
				int index = (int) ((t-startTime)/dt);
				double value = rs.getDouble(2);
				for (int i = index; i < timeVector.length; i++) {
					tv[i] = value;
				}
			}
			st.close();
		} catch (SQLException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}

		return tv;
	}
	
	public double getSingleDoubleValue(String sql) {
		return DBUtilities.getSingleDoubleValue(dbConn, sql);
	}
	
	public double getSingleDoubleValue(String sql, double time) {
		return DBUtilities.getSingleDoubleValue(dbConn, sql, time);
	}
	
	public double[][] getLookupValues(String variable) {
//		public static double[][] getTableDoubles(DBConnection dbConn, String sql) {
		int lookupIndex = DBUtilities.getSingleIntegerValue(dbConn, "select LookupIndex from Lookups where LookupName = '"+variable+"'");
		double[][] l = DBUtilities.getTableDoubles(dbConn, "select x, y from LookupsValues where LookupIndex = "+lookupIndex+" order by PairOrder");
		return l;
	}
	
	public double[] getTimeVector(String sql) {

		Statement st = null;
		ResultSet rs = null;
		double[] tv = null;

		st = dbConn.createStatement(ResultSet.TYPE_SCROLL_INSENSITIVE, ResultSet.CONCUR_READ_ONLY);
		try {
			rs = st.executeQuery(sql);
			tv = new double[DBUtilities.numRows(rs)];
			int i = 0;
			for (; rs.next(); ) {
				double t = rs.getDouble(1);
				tv[i++] = t;
			}
			st.close();
		} catch (SQLException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
		
	
		return tv;
	}
	
	public String[] getStringValues(String sql) {
		return DBUtilities.getStringValues(dbConn, sql);
	}

	public void createRunStructureTables(String mapFile) {

		MemoryMapperInterface mapper = new StreamlinedMemoryMapper(mapFile, new MessageGobbler());
		createConstants(mapper);
		createTimeSeries(mapper);
		createLookups(mapper); 
		createLookupsValues(mapper); 
		createRunVariables(mapper);
		
//		createAndLoadMapperTables(mapper);

	}

	public void exec(String sql) {
		DBUtilities.exec(dbConn, sql);
	}
	
	public void createAndLoadMapperTables(MemoryMapperInterface mapper) {
		
		exec("drop table InitializationExecutionOrder If Exists");
		exec("CREATE Cached TABLE InitializationExecutionOrder ( id INTEGER IDENTITY, ExpandedLHS VARCHAR(256))");
		
		exec("drop table RepeatedExecutionOrder If Exists");
		exec("CREATE Cached TABLE RepeatedExecutionOrder ( id INTEGER IDENTITY, ExpandedLHS VARCHAR(256))");
		
		exec("drop table EquationLHS If Exists");
		exec("CREATE Cached TABLE EquationLHS ( id INTEGER IDENTITY, Variable VARCHAR(256))");
		
		exec("drop table ExpandedLHS If Exists");
		exec("CREATE Cached TABLE ExpandedLHS ( id INTEGER IDENTITY, Variable VARCHAR(256))");
		
		loadEquationLHS(mapper);
		loadExpandedLHS(mapper);
		loadLHSInitializationExecutionOrder(mapper);
		loadRepeatedExecutionOrder(mapper);

	}
	
	private void loadEquationLHS(MemoryMapperInterface mapper) {
		for (String s : mapper.getSortedLHS()) {
			exec("insert into EquationLHS(Variable) VALUES ('"
					+ specialCharacterProcess(s)+"')");
		}
	}
	
	private void loadExpandedLHS(MemoryMapperInterface mapper){
		for (String s : mapper.getExpandedSortedLHS()) {
			exec("insert into ExpandedLHS(Variable) VALUES ('"
					+ specialCharacterProcess(s)+"')");
		}
	}
	

	private void loadLHSInitializationExecutionOrder(MemoryMapperInterface mapper){

		for (String expLHS : mapper.getEvaluationOrderOneTime()) {
			exec("insert into InitializationExecutionOrder(ExpandedLHS) VALUES ('"
					+ specialCharacterProcess(expLHS)+"')");
		}
		
		
	}
	private void loadRepeatedExecutionOrder(MemoryMapperInterface mapper){
		
		for (String expLHS : mapper.getEvaluationOrderRepeated()) {
			exec("insert into RepeatedExecutionOrder(ExpandedLHS) VALUES ('"
					+ specialCharacterProcess(expLHS)+"')");
		}
		
	}

	public void createConstants(MemoryMapperInterface mapper) {

		int numConstants = mapper.getNumElementsForType(MemoryMapper.CONSTANT);
		exec("drop table Constants If Exists");
		exec("CREATE Cached TABLE Constants ( id INTEGER IDENTITY, modelTime DOUBLE, DataValues DOUBLE ARRAY["+numConstants+"])");

	}
	public void createTimeSeries(MemoryMapperInterface mapper) {

		int numDynamic = mapper.getNumElementsForType(MemoryMapper.DYNAMIC);
		exec("drop table TimeSeries If Exists");
		exec("CREATE Cached TABLE TimeSeries ( id INTEGER IDENTITY, modelTime DOUBLE, DataValues DOUBLE ARRAY["+numDynamic+"])");
	}
	
	public void createLookupsValues(MemoryMapperInterface mapper) {

		exec("drop table LookupsValues If Exists");
		exec("CREATE Cached TABLE LookupsValues ( id INTEGER IDENTITY, LookupIndex INTEGER, PairOrder INTEGER, x DOUBLE, y DOUBLE)");
	}
	public void createLookups(MemoryMapperInterface mapper) {
		exec("drop table Lookups If Exists");
		exec("CREATE Cached TABLE Lookups ( id INTEGER IDENTITY, LookupName VARCHAR(256), "
				+ "LookupIndex INTEGER, "
				+ "numElements INTEGER)");

		List<List<String>> records = mapper.getLookupMapRecords();

		int lookupIndex = 0;
		for (List<String> record : records) {
			if (record.get(1).equals(StreamlinedMemoryMapper.LOOKUP)) {
				exec("insert into Lookups(LookupName, LookupIndex, NumElements) VALUES ('"
						+ specialCharacterProcess(record.get(0))+"',"+lookupIndex+","+record.get(2)+")");
				lookupIndex++;
			}
		}
	}
	public void createRunVariables(MemoryMapperInterface mapper) {
		exec("drop table Variables If Exists");
		exec("CREATE Cached TABLE Variables ( id INTEGER IDENTITY, VariableName VARCHAR(256), "
				+ "VariableType VARCHAR(2), VariableIndex INTEGER, numElements INTEGER)");

		List<List<String>> records = mapper.getVariableMapRecords();

		for (List<String> record : records) {
			exec("insert into Variables(VariableName, VariableType, VariableIndex, NumElements) VALUES ('"
					+ specialCharacterProcess(record.get(0))+"','"+record.get(1)+"',"+record.get(2)+","+record.get(3)+")");
		}
	}

	public String specialCharacterProcess(String s) {
		return s.replace("'", "''");
	}
	
	public Map<String, List<Double>> getAllVariableValuesAsMapList() {
		Map<String, List<Double>> map = new LinkedHashMap<String, List<Double>>();
		
		for (String var : getVariables()) {
			map.put(var, Utilities.asList(getVarTimeVector(var)));
		}
		
		return map;
	}
	
	public String[] getVariables() {
		return getStringValues("select Variable From EquationLHS order by Variable");
	}
	
	public void writeLookupValues(String var, double[][] lookupValues){

		String sql = "select LookupIndex from Lookups where LookupName = '"+var+"'";
		int li = DBUtilities.getSingleIntegerValue(dbConn, sql);

		int order = 0;
		for (int i = 0; i < lookupValues[0].length; i++) {
			exec("insert into LookupsValues(LookupIndex, PairOrder, x, y) VALUES ("
					+ li+","+order+","+lookupValues[0][i]+","+lookupValues[1][i]+")");
			order++;
		}
//		for (double[] record : lookupValues) {
//			exec("insert into LookupsValues(LookupIndex, PairOrder, x, y) VALUES ("
//					+ li+","+order+","+record[0]+","+record[1]+")");
//			order++;
//		}		
	}

	public void writeCMem(double time, double[] cMem){
		org.hsqldb.types.Type type = org.hsqldb.types.Type.SQL_DOUBLE;
		Double[] cMemD = ArrayUtils.toObject(cMem);
		JDBCArrayBasic array = new JDBCArrayBasic(cMemD, type);

		try {
			psConstants.setArray(2, array);
			psConstants.setDouble(1, time);
			psConstants.executeUpdate();
		} catch (SQLException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
	}
	public void writeVMem(double time, double[] vMem){
		org.hsqldb.types.Type type = org.hsqldb.types.Type.SQL_DOUBLE;
		Double[] vMemD = ArrayUtils.toObject(vMem);
		JDBCArrayBasic array = new JDBCArrayBasic(vMemD, type);
		
		try {
			psTimeSeries.setArray(2, array);
			psTimeSeries.setDouble(1, time);
			psTimeSeries.executeUpdate();
		} catch (SQLException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
	}

	private void makePrepared() {
		String sql;
		sql = "insert into timeSeries(modelTime, dataValues) values( ?, ?)";
		psTimeSeries = dbConn.prepareStatement(sql);

		sql = "insert into constants(modelTime, dataValues) values(?, ?)";
		psConstants = dbConn.prepareStatement(sql);
	}

	public String getMapFile() {
		return mapFile;
	}

	public void setMapFile(String mapFile) {
		this.mapFile = mapFile;
	}

	public String getDbFile() {
		return dbFile;
	}

	public double getStartTime() {
		return DBUtilities.getSingleDoubleValue(dbConn, "Select initialTime from RunParameters");
	}

	public double getEndTime() {
		return DBUtilities.getSingleDoubleValue(dbConn, "Select finalTime from RunParameters");
	}

	public double getTimeStep() {
		return DBUtilities.getSingleDoubleValue(dbConn, "Select dt from RunParameters");
	}

	public double getDt() {
		return DBUtilities.getSingleDoubleValue(dbConn, "Select dt from RunParameters");
	} 
	
	public double getSavper() {
		return DBUtilities.getSingleDoubleValue(dbConn, "Select savper from RunParameters");
	}

}
