package gov.anl.gss.sdm.server.model.tools;

import gov.anl.gss.sdm.server.model.documentor.MessageProcessor;
import gov.anl.gss.sdm.server.model.documentor.SDMDocLanguageSpecific;
import gov.anl.gss.sdm.server.model.documentor.xmlwriter.StringConstants;
import gov.anl.gss.sdm.server.model.structure.SDObjectManager;
import gov.anl.gss.sdm.server.model.support.ArrayReference;
import gov.anl.gss.sdm.server.model.support.MemoryMapperInterface;
import gov.anl.gss.sdm.server.model.support.StreamlinedMemoryMapper;
import gov.anl.gss.sdm.server.model.support.SubscriptCombination;
import gov.anl.gss.sdm.server.model.support.Utilities;
import gov.anl.gss.sdm.server.model.translator.ActiveInitialProcessor;
import gov.anl.gss.sdm.server.model.translator.ArrayManager;
import gov.anl.gss.sdm.server.model.translator.CodeGenerator;
import gov.anl.gss.sdm.server.model.translator.Equation;
import gov.anl.gss.sdm.server.model.translator.EquationProcessor;
import gov.anl.gss.sdm.server.model.translator.FunctionDescription;
import gov.anl.gss.sdm.server.model.translator.GraphicsProcessor;
import gov.anl.gss.sdm.server.model.translator.InformationManagers;
import gov.anl.gss.sdm.server.model.translator.Parser;
import gov.anl.gss.sdm.server.model.translator.StaticSwitches;
import gov.anl.gss.sdm.server.model.translator.SystemDynamicsObjectManager;
import gov.anl.gss.sdm.server.model.translator.VariableDependencyProcessor;
import gov.anl.gss.sdm.server.model.translator.View;

import java.io.BufferedWriter;
import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedHashMap;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.apache.log4j.Logger;

public class SDModelCodeGenerator extends SDModelTool {
	
	static Logger log = Logger.getLogger(SDModelCodeGenerator.class.getSimpleName());
	
	private static final boolean DISPLAY_CATEGORY = false;

	private SDModel sdModel;
	
	private boolean cleanEquationOrdering = true;

	private EquationProcessor equationProcessor;
	private Map<String, Equation> equations;
	private GraphicsProcessor graphicsProcessor;
	private Map<String, View> graphics;

	protected boolean requiresActiveInitialProcessing;
	
//	boolean newMemoryDefinition;

	protected SystemDynamicsObjectManager sdObjectManager;

	// configuration for generated code
//	protected boolean generateC = false;
	protected boolean generateJava = true;
//	private String target; // one of "Java" or "C"
//	private String memoryImplementation; // one of "Native" or "HashMap" boolean
											// nativeMemory()
//	private boolean repastSimphony = false; // isRepastSimphony()
//	private boolean rsHybrid = false; // isRsHybrid()

	private String miscDirectory = "misc";
	protected String destinationDirectory;

	private VariableDependencyProcessor variableDependencyProcessor;
	
	private FileOrganizer fileOrganizer;

	private String packageName;
	private String objectName;
	
	private MessageProcessor info;

	private String javaFilePath = null;

	private boolean codeGenerationFailure = false;
	
	private static final String STOCK = "S";
	private static final String AUXILIARY = "A";
	private static final String INITIALIZER = "I";
	private static final String CONSTANT_DATA = "C";
		
		
	private Map<String, List<String>> byVariableType = new LinkedHashMap<String, List<String>>();

	public SDModelCodeGenerator(SDModel sdModel, String packageName,
			String objectName, String destinationDirectory,
			/* String memoryImplementation, */ String javaFilePath, MessageProcessor info,
			/* boolean newMemoryDefinition,*/ FileOrganizer fileOrganizer) {
		this(sdModel, packageName, objectName, destinationDirectory,
				/* memoryImplementation,*/ info, /*newMemoryDefinition,*/ fileOrganizer);
		this.javaFilePath = javaFilePath;
		
		if (InformationManagers.getInstance().getFileOrganizer() != null && 
				InformationManagers.getInstance().getFileOrganizer().getMiscDirectory().length() > 0) {
			miscDirectory = InformationManagers.getInstance().getFileOrganizer().getMiscDirectory();
		} else {
			miscDirectory = "misc";
		}

	}

	public SDModelCodeGenerator(SDModel sdModel, String packageName,
			String objectName, String destinationDirectory,
			/* String memoryImplementation, */ MessageProcessor info, /* boolean newMemoryDefinition,*/  FileOrganizer fileOrganizer) {
		this.sdModel = sdModel;
		this.info = info;
		
//		this.newMemoryDefinition = newMemoryDefinition;

		equationProcessor = sdModel.getEquationProcessor();
		equations = sdModel.getEquations();
		graphicsProcessor = sdModel.getGraphicsProcessor();
		graphics = sdModel.getGraphics();
		sdObjectManager = sdModel.getSdObjectManager();
		this.fileOrganizer = fileOrganizer;
		

		this.objectName = objectName;
		this.packageName = packageName;
		this.destinationDirectory = destinationDirectory;
//		this.memoryImplementation = memoryImplementation;

//		for (Equation eqn : equations.values())
//			eqn.setNativeMemoryImplementation(nativeMemory());
	}

	public boolean isValid(List<String> messages) {

		boolean errors = equationProcessor.getDuplicateLHS().size() > 0;
		if (errors) {
			messages.add("+++ Duplicate LHS +++");
			for (String lhs : equationProcessor.getDuplicateLHS()) {
				messages.add(lhs);
			}
			return false;
		}
		
		// Should these be done here?

		processSubscriptDefinition(equations);
		processExponentiaion(equations);

		Map<String, Equation> fatalErrors = equationProcessor
				.getFatalErrors(equations);
		errors = fatalErrors.size() > 0;
		if (errors) {
			sdModel.printErrors("Fatal", fatalErrors);
			sdModel.generateErrorReport("Fatal Errors", fatalErrors, messages);
			return false;
		}

		InformationManagers.getInstance().getArrayManager().populateArraySubscriptSpace();
		info.addMessage(SDMDocLanguageSpecific.getInstance().
				getText(StringConstants.DETERMINING_EQUATION_VARIABLE_CHARACTERISTICS));
		
		String mapperFile = InformationManagers.getInstance().getModelManager().getMapperFile();
		MemoryMapperInterface mapper = null;
//		if (newMemoryDefinition) {
			mapper = new StreamlinedMemoryMapper(equations, mapperFile, info);
//		}
		InformationManagers.getInstance().setMemoryMapper(mapper);

		equationProcessor.checkUsage(equations);


		generateRPN(equations);
		generateTrees(equations);

		Map<String, Equation> unitsErrors = InformationManagers
				.getInstance()
				.getUnitsManager()
				.performUnitsConsistencyCheck(equations, InformationManagers.getInstance().
						getFileOrganizer().getMiscDirectory()+
						"/unitsConsistencyCheckResultsFile.txt");
		errors = unitsErrors.size() > 0;
//		if (errors) {
//			sdModel.printErrors("Units", unitsErrors);
//			sdModel.generateErrorReport("Units Errors", unitsErrors, messages);
//			// return false;
//		}

		return true;

	}

	protected boolean process(/*
							 * Map<String, Equation> equations, Map<String,
							 * View> graphics
							 */) {
		CodeGenerator cg = null;

		ActiveInitialProcessor aip = null;
		SDMDocLanguageSpecific language = SDMDocLanguageSpecific.getInstance();
		

		if (equations != null) {
			requiresActiveInitialProcessing = ActiveInitialProcessor
					.requiresProcessing(equations);
			generateMemory(aip, requiresActiveInitialProcessing);
			List<String> evaluationOrderOneTime = new ArrayList<String>();
			List<String> evaluationOrderRepeated = new ArrayList<String>();

			info.addMessage(language.getText(StringConstants.DETERMINING_EVALUATION_ORDER));
			
//			determineEvaluationOrder(equations, evaluationOrderOneTime,	evaluationOrderRepeated);
			
			determineEvaluationOrder(evaluationOrderOneTime,evaluationOrderRepeated);
			System.out.println("evaluationOrderOneTime: "+evaluationOrderOneTime.size());
			System.out.println("evaluationOrderRepeated: "+evaluationOrderRepeated.size());
			
			boolean badEvalOrder = false;
			if (!isCleanEquationOrdering()) {
				info.addMessage(language.getText(StringConstants.EVALUATION_ORDERING_FAILED));
				InformationManagers.getInstance().getModelManager().setRunModel(false);
				ok = false;
				badEvalOrder = true;
//				return;
			}
			
			info.addMessage(language.getText(StringConstants.ADDING_EVALUATION_ORDER_TO_MAPPER_FILE));
			
			MemoryMapperInterface memoryMapper = InformationManagers.getInstance().getMemoryMapper();
			
			List<String> mapperEvaluationOrderOneTime = new ArrayList<String>(evaluationOrderOneTime);
			List<String> mapperEvaluationOrderRepeated = new ArrayList<String>(evaluationOrderRepeated);
			
			List<String> toRemove = new ArrayList<String>();
			
			// we do not want any lookups or other data stuff in the mapper
			for (String lhs : evaluationOrderOneTime) {
				Equation eqn = InformationManagers.getInstance().getEquationManager().getAnEquationFor(lhs); // equations.get(lhs);
				if (eqn != null) {
					if (eqn.isDefinesLookup() || eqn.isDefinesLookupGetXls() || eqn.isDefinesSubscript() || eqn.isData())
						toRemove.add(lhs);
				}
			}
			mapperEvaluationOrderOneTime.removeAll(toRemove);
			
			for (String lhs : evaluationOrderRepeated) {
				Equation eqn = InformationManagers.getInstance().getEquationManager().getAnEquationFor(lhs); // equations.get(lhs);
				if (eqn != null) {
					if (eqn.isDefinesLookup() || eqn.isDefinesLookupGetXls() || eqn.isDefinesSubscript() || eqn.isData())
						toRemove.add(lhs);
				}
			}
			mapperEvaluationOrderRepeated.removeAll(toRemove);
			
			memoryMapper.setEvaluationOrderOneTime(mapperEvaluationOrderOneTime);
			memoryMapper.setEvaluationOrderRepeated(mapperEvaluationOrderRepeated);
			memoryMapper.writeMapper(badEvalOrder);
			
//			if (badEvalOrder)
//				return badEvalOrder;

			// writeEquationInitializationSequence(evaluationOrder);

			// if (requiresActiveInitialProcessing) {
			// aip = new ActiveInitialProcessor(equations, evaluationOrder,
			// isGenerateJava() ? "Java" : "C");
			// aip.analyze(variableDependencyProcessor);
			// generateMemory(aip, requiresActiveInitialProcessing);
			// }

			// if (evaluationOrder != null) {
			if (javaFilePath == null) {
				new File(destinationDirectory + "/" + objectName + "/src" + "/"
						+ asDirectoryPath(packageName) + "/").mkdirs();
				cg = new CodeGenerator(destinationDirectory + "/" + objectName
						+ "/" + "src" + "/" + asDirectoryPath(packageName)
						+ "/", evaluationOrderOneTime, evaluationOrderRepeated,
						equations, objectName, packageName,
						/*isGenerateJava() ? "Java" : "C", */ /* memoryImplementation, */ /*newMemoryDefinition,*/
								fileOrganizer);
			} else {
				File jsrc = new File(javaFilePath);
				jsrc.getParentFile().mkdirs();
				cg = new CodeGenerator(jsrc.getParent() + "/",
						evaluationOrderOneTime, evaluationOrderRepeated,
						equations, objectName, packageName,
						/* isGenerateJava() ? "Java" : "C",*/  /* memoryImplementation, */ /* newMemoryDefinition, */
								fileOrganizer);
			}

			info.addMessage(language.getText(StringConstants.WRITING_CODE));
			
			codeGenerationFailure = cg.generateCode(aip) || codeGenerationFailure;
			ok = !codeGenerationFailure;
			// }
		}
		return codeGenerationFailure;
	}

	private void writeEquationInitializationSequence(
			List<String> evaluationOrder) {
		BufferedWriter bw = Utilities
				.openFileForWriting(miscDirectory+"/initializationSequence.txt");
		try {

			bw.append("\n##### One Time ##### (Constants)\n");
			for (String lhs : evaluationOrder) {
				Equation eqn = equations.get(lhs);
				if (!eqn.isOneTime())
					continue;
				bw.append("<" + eqn.getLhs() + "> "
						+ eqn.getVensimEquationOnly() + "\n");
			}

			Set<String> required = new LinkedHashSet<String>();
			for (String lhs : evaluationOrder) {
				Equation eqn = equations.get(lhs);
				if (eqn.containsInitialValueFunction()) {
					for (String s : eqn.getFunctionInitialVariables()) {
						required.add(InformationManagers.getInstance()
								.getNativeDataTypeManager().getOriginalName(s));
					}
				}
			}

			bw.append("\n##### Stock Initialization #####\n");
			for (String lhs : evaluationOrder) {
				Equation eqn = equations.get(lhs);
				if (!eqn.isOneTime() && required.contains(lhs))
					bw.append("<" + eqn.getLhs() + "> "
							+ eqn.getVensimEquationOnly() + "\n");
			}

			for (String lhs : evaluationOrder) {
				Equation eqn = equations.get(lhs);
				if (eqn.isStock())
					bw.append("<" + eqn.getLhs() + "> "
							+ eqn.getVensimEquationOnly() + "\n");
			}

			bw.append("\n##### Repeated Non-Stocks #####\n");
			for (String lhs : evaluationOrder) {
				Equation eqn = equations.get(lhs);
				if (eqn.isStock() || eqn.isOneTime())
					continue;
				bw.append("<" + eqn.getLhs() + "> "
						+ eqn.getVensimEquationOnly() + "\n");
			}

			bw.append("\n##### Repeated Stocks #####\n");
			for (String lhs : evaluationOrder) {
				Equation eqn = equations.get(lhs);
				if (eqn.isOneTime() || !eqn.isStock())
					continue;
				bw.append("<" + eqn.getLhs() + "> "
						+ eqn.getVensimEquationOnly() + "\n");
			}

			bw.close();
		} catch (IOException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
	}

	protected void generateMemory(ActiveInitialProcessor aip,
			boolean requiresActiveInitialProcessing) {
		InformationManagers.getInstance().getMappedSubscriptManager()
				.makeConsistent();

		if (/* isGenerateC() || */ isGenerateJava()) {
			// ############## WRITE_DEBUG_FILES
			if (StaticSwitches.WRITE_DEBUG_FILES) {
			InformationManagers
					.getInstance()
					.getMappedSubscriptManager()
					.dumpMappings(
							Utilities.openFileForWriting(miscDirectory + "/"
									+ "SubscriptMappings_" + objectName
									+ ".csv"));
			InformationManagers
					.getInstance()
					.getNamedSubscriptManager()
					.dumpMappings(
							Utilities.openFileForWriting(miscDirectory + "/"
									+ "SubscriptNameValues_" + objectName
									+ ".csv"), ",");

			InformationManagers
					.getInstance()
					.getUnitsManager()
					.dumpLhsUnits(
							Utilities.openFileForWriting(miscDirectory + "/"
									+ "Units_" + objectName + ".csv"));
			InformationManagers
					.getInstance()
					.getUnitsManager()
					.dumpLhsUnitsRaw(
							Utilities.openFileForWriting(miscDirectory + "/"
									+ "UnitsRaw_" + objectName + ".csv"));
			}

			if (isGenerateJava()) {
				
				String SourceDirectory = fileOrganizer.getDynamicDirectory()+"/"+ 
				FileOrganizer.getSrcClassDirectory()+ "/"
						+ asDirectoryPath(packageName) + "/";
				
				// ############## WRITE_DEBUG_FILES
				if (StaticSwitches.WRITE_DEBUG_FILES) {
				InformationManagers
						.getInstance()
						.getNativeDataTypeManager()
						.dumpLegalNames(
								Utilities.openFileForWriting(miscDirectory
										+ "/" + "LegalNames_" + objectName
										+ ".csv"));
				}

//				if (repastSimphony) {
//					RepastSimphonyEnvironment.generateContextBuilder(
//							Utilities.openFileForWriting(SourceDirectory
//									+ "ContextBuilder" + objectName + ".java"),
//							objectName, packageName);
//
//					String ScenarioDirectory = destinationDirectory + "/"
//							+ objectName + "/" + objectName + ".rs/";
//
//					RepastSimphonyEnvironment.generateScenarioXml(
//							Utilities.openFileForWriting(ScenarioDirectory
//									+ "scenario.xml"), objectName);
//					RepastSimphonyEnvironment.generateUserPathXml(
//							Utilities.openFileForWriting(ScenarioDirectory
//									+ "user_path.xml"), objectName);
//					RepastSimphonyEnvironment
//							.generateClassLoaderXml(
//									Utilities
//											.openFileForWriting(ScenarioDirectory
//													+ "repast.simphony.dataLoader.engine.ClassNameDataLoaderAction_1.xml"),
//									objectName, packageName);
//					RepastSimphonyEnvironment.generateContextXml(
//							Utilities.openFileForWriting(ScenarioDirectory
//									+ "context.xml"), objectName);
//				}
//				if (nativeMemory()) {
					// MJB write memory
					InformationManagers
							.getInstance()
							.getNativeDataTypeManager()
							.generateMemoryJava(
									Utilities.openFileForWritingCharset(SourceDirectory
											+ "Memory" + objectName + ".java", 
											SDMDocLanguageSpecific.getInstance().getFileCharSet()),
									objectName, packageName /*, newMemoryDefinition*/ );
//				}
			}

//			} else if (isGenerateC()) {
//				String SourceDirectory = destinationDirectory + "/"
//						+ objectName + "/" + "src" + "/"
//						+ asDirectoryPath(packageName) + "/";
//				// ############## WRITE_DEBUG_FILES
//				if (StaticSwitches.WRITE_DEBUG_FILES) {
//				InformationManagers
//						.getInstance()
//						.getNativeDataTypeManager()
//						.dumpLegalNames(
//								Utilities.openFileForWriting(miscDirectory
//										+ "/" + "LegalNames_" + objectName
//										+ ".csv"));
//				}
//				InformationManagers
//						.getInstance()
//						.getNativeDataTypeManager()
//						.generateMemoryC(
//								Utilities.openFileForWriting(SourceDirectory
//										+ "memory" + objectName + ".h"),
//								objectName, aip,
//								requiresActiveInitialProcessing);
//
//			}
			
			String dir = InformationManagers.getInstance().getModelManager().getGeneratedOutputDirectoryName();

			// ############## WRITE_DEBUG_FILES
			if (StaticSwitches.WRITE_DEBUG_FILES) {
			InformationManagers
					.getInstance()
					.getArrayManager()
					.dumpSubscriptSpace(
							Utilities.openFileForWriting(dir + "/structure/"
									+ "subscriptSpace.tab"));
			InformationManagers
			.getInstance()
			.getNamedSubscriptManager()
			.dumpMappings(
					Utilities.openFileForWriting(dir + "/structure/"
							+ "subscriptDefinition.tab"), "\t");
			}

		}

	}
	
	protected void determineEvaluationOrder(List<String> oneTime, List<String> repeated) {
		
		VariableDependencyProcessor vdp = new VariableDependencyProcessor(
				equations, objectName);
		String lastInitial = vdp.loadForOrdering();
		
		variableDependencyProcessor = vdp;
		
		if (vdp.isCleanOrderOfAvailableVariables()) {
			categorizeVariables(vdp.getOrderedListOfAvailableVariables());
			
			Equation lastStock = null;
			for (String var : vdp.getOrderedListOfAvailableVariables()) {
				Equation eqn = InformationManagers.getInstance().getEquationManager().getAnEquationFor(var);
				if (eqn == null)
					continue;
				if (eqn.isStock())
					lastStock = eqn;
			}
			
//			the problem here is that we hit last stock before the others -- this needs to change!
			
			boolean locatedLastStock = false;
			boolean locatedLastInitial = false;
			
			for (String var : vdp.getOrderedListOfAvailableVariables()) {
				Equation eqn = InformationManagers.getInstance().getEquationManager().getAnEquationFor(var);
				if (eqn == null)
					continue;
				if (eqn.isOneTime() || eqn.isStock() || byVariableType.get(INITIALIZER).contains(var) || byVariableType.get(AUXILIARY).contains(var)) {
					oneTime.add(var);
				}
				if (var.equals(lastInitial))
					locatedLastInitial = true;
				if (eqn == lastStock)
					locatedLastStock = true;
				if (locatedLastInitial && locatedLastStock)
					break;
			}
			
//			oneTime.addAll(byVariableType.get(CONSTANT_DATA));
//			oneTime.addAll(byVariableType.get(INITIALIZER));
//			oneTime.addAll(byVariableType.get(STOCK));
//			oneTime.addAll(stocks);
			
			// Sept 22 2015 -- stocks need to be available for one time
			
			for (String stock : byVariableType.get(STOCK)) {
				if (!oneTime.contains(stock))
					oneTime.add(stock);
			}
			
			repeated.addAll(byVariableType.get(AUXILIARY));
			repeated.addAll(byVariableType.get(STOCK));
			
			List<String> evaluationOrder = new ArrayList<String>();
			SDMDocLanguageSpecific language = SDMDocLanguageSpecific.getInstance();
			
			evaluationOrder.addAll(oneTime);
			evaluationOrder.addAll(repeated);
			info.addMessage(language.getText(StringConstants.ANALYZE_EVALUATION_ORDER));
			
			// this needs to be rewritten with new data structures
//			analyzeEvaluationOrder(evaluationOrder, equations);

		} else {
			cleanEquationOrdering = false;
		}
	}

	public void categorizeVariables(List<String> variables) {

		//		private Map<String, Set<String>> byVariableType = new LinkedHashMap<String, Set<String>>();

		byVariableType.put(STOCK, new ArrayList<String>());
		byVariableType.put(AUXILIARY, new ArrayList<String>());
		byVariableType.put(INITIALIZER, new ArrayList<String>());
		byVariableType.put(CONSTANT_DATA, new ArrayList<String>());

		for (String var : variables) {
			Equation eqn = InformationManagers.getInstance().getEquationManager().getAnEquationFor(var);
			if (eqn == null) {
				System.out.println("NO EQUATION: "+var);
				continue;
			}

			if (eqn.isStock()) {
				if (!byVariableType.get(STOCK).contains(var)) {
					byVariableType.get(STOCK).add(var);
					if (DISPLAY_CATEGORY)
						System.out.println(var+" -> STOCK");
				}
			} else if (eqn.isAuxiliary()) {
				if (!byVariableType.get(AUXILIARY).contains(var)) {
					byVariableType.get(AUXILIARY).add(var);
					if (DISPLAY_CATEGORY)
						System.out.println(var+" -> AUXILIARY");
				}
			}
			if (eqn.isConstant() || eqn.isData()) {
				if (!byVariableType.get(CONSTANT_DATA).contains(var)) {
					byVariableType.get(CONSTANT_DATA).add(var);
					if (DISPLAY_CATEGORY)
						System.out.println(var+" -> CONSTANT");
				}
			}
			if (eqn.isStock() || eqn.isActiveInitial()) {
				Set<String> fiv = eqn.getFunctionInitialVariables();

				for (String original : InformationManagers.getInstance()
						.getNativeDataTypeManager().getOriginalNames(fiv)) {
					// some constants already processed may also qualify as initializers
					// do not put them in twice
					if (!byVariableType.get(CONSTANT_DATA).contains(original)) {
						if (!byVariableType.get(INITIALIZER).contains(original)) {
							byVariableType.get(INITIALIZER).add(original);
							if (DISPLAY_CATEGORY)
								System.out.println(var+" -> INITIALIZER");
						}
					}

				}
			}
		}
		// since INITIALIZERS are not ordered at this point -- order them

		List<String> nonordered = byVariableType.get(INITIALIZER);
		byVariableType.put(INITIALIZER, new ArrayList<String>());

		for (String var : variables) {
			if (nonordered.contains(var)) {
				if (!byVariableType.get(INITIALIZER).contains(var)) {
					byVariableType.get(INITIALIZER).add(var);
//					System.out.println(var+" -> INITIALIZER ordered");
				}
			}
		}

	}
	protected void determineEvaluationOrder(Map<String, Equation> equations,
			List<String> oneTime, List<String> repeated) {

		variableDependencyProcessor = new VariableDependencyProcessor(
				equations, objectName);
		String lastInitial = variableDependencyProcessor.loadForOrdering();

		List<String> evaluationOrder = new ArrayList<String>();
		List<String> constantsAndData = new ArrayList<String>();
		List<String> stockInitializers = new ArrayList<String>();
		List<String> stocks = new ArrayList<String>();
		List<String> auxiliaries = new ArrayList<String>();
		
		String dir = InformationManagers.getInstance().getModelManager().getGeneratedOutputDirectoryName();

//		Map<String, Set<String>> headTailMap = variableDependencyProcessor
//				.getHeadTailMap();
//		// ############## WRITE_DEBUG_FILES
//				if (StaticSwitches.WRITE_DEBUG_FILES) {
//		writeMapToFile(dir
//				+ "/structure/headTailMap.tab", headTailMap);
//				}
//		Map<String, Set<String>> tailHeadMap = variableDependencyProcessor
//				.getTailHeadMap();
//		// ############## WRITE_DEBUG_FILES
//				if (StaticSwitches.WRITE_DEBUG_FILES) {
//		writeMapToFile(dir
//				+ "/structure/tailHeadMap.tab", tailHeadMap);
//				}
				
		SDMDocLanguageSpecific language = SDMDocLanguageSpecific.getInstance();
		
		Set<String> v = validate(variableDependencyProcessor
				.getRequiresExpandedLhsExpandedRHS()); 
		if (v.size() > 0) {
			info.addMessage(language.getText(StringConstants.INVALID));
		}
		
		
		// C O N S T A N T S

		info.addMessage(language.getText(StringConstants.ORDER_CONSTANTS));
		Set<String> removedConstants = orderConstants(equations, constantsAndData);
		
		print("constantsAndData", constantsAndData);
		
		v = validate(variableDependencyProcessor
				.getRequiresExpandedLhsExpandedRHS()); 
		if (v.size() > 0) {
			info.addMessage(language.getText(StringConstants.INVALID));
		}
		
		// S T O C K   I N I T I A L I Z E R S
		
		info.addMessage(language.getText(StringConstants.ORDER_INITIALIZERS));
		orderInitializers(equations, stockInitializers, constantsAndData);
		print("stockInitializers", stockInitializers);
		
		v = validate(variableDependencyProcessor
				.getRequiresExpandedLhsExpandedRHS()); 
		if (v.size() > 0) {
			info.addMessage(language.getText(StringConstants.INVALID));
		}
		info.addMessage(language.getText(StringConstants.ORDER_STOCKS));
		
		// S T O C K S 
		
		orderStocksAndActiveInitial(equations, stocks);
		print("stocks", stocks);
		
		v = validate(variableDependencyProcessor
				.getRequiresExpandedLhsExpandedRHS()); 
		if (v.size() > 0) {
			info.addMessage(language.getText(StringConstants.INVALID));
		}

		Map<String, HashSet<String>> requiresExpandedLhsExpandedRHS = variableDependencyProcessor
				.getRequiresExpandedLhsExpandedRHS();
		boolean startChain = false; // want this to be false for production
		// ############## WRITE_DEBUG_FILES
				if (StaticSwitches.WRITE_DEBUG_FILES) {
		dumpRequiresExpandedWithDepend("postStock.txt",
				requiresExpandedLhsExpandedRHS, false);
		dumpRequiresExpanded("postStockWithRHS.txt",
				requiresExpandedLhsExpandedRHS, false);
				}
		
		List<String> alreadyDefined = new ArrayList<String>();
		
		
		alreadyDefined.addAll(constantsAndData);
		alreadyDefined.addAll(stockInitializers);
		alreadyDefined.addAll(stocks);
		
		// A U X I L I A R I E S

		info.addMessage(language.getText(StringConstants.ORDER_AUXILIARIES));
		orderAuxiliaries(equations, auxiliaries, alreadyDefined);
//		orderAuxiliaries(equations, auxiliaries, new ArrayList<String>());
		
		v = validate(variableDependencyProcessor.getRequiresExpandedLhsExpandedRHS()); 
		if (v.size() > 0) {
			info.addMessage(language.getText(StringConstants.INVALID));
		}
		
		// ############## WRITE_DEBUG_FILES
				if (StaticSwitches.WRITE_DEBUG_FILES) {
		dumpRequiresExpandedWithDepend("postAuxiliaries.txt",
				requiresExpandedLhsExpandedRHS, false);
				}
		
		oneTime.addAll(constantsAndData);
		oneTime.addAll(stockInitializers);
		
//		oneTime.addAll(stocks);
		
		// Sept 22 2015 -- stocks need to be available for one time
		
		for (String stock : stocks) {
			if (!oneTime.contains(stock))
				oneTime.add(stock);
		}
		
		repeated.addAll(auxiliaries);
		repeated.addAll(stocks);
		
		evaluationOrder.addAll(oneTime);
		evaluationOrder.addAll(repeated);
		info.addMessage(language.getText(StringConstants.ANALYZE_EVALUATION_ORDER));
		analyzeEvaluationOrder(evaluationOrder, equations);

		return;
	}
	
	public String getVariableOrArrayName(String lhs) {
		String aVar = lhs;
		if (ArrayReference.isArrayReference(lhs)) {
			aVar = ArrayReference.extractArrayName(aVar);
		}
		aVar = ArrayReference.deannotate(aVar);
		return aVar;
	}
	
	protected Set<String> validate(Map<String, HashSet<String>> requiresExpandedLhsExpandedRHS) {
		Set<String> badOnes = new LinkedHashSet<String>();
		SDObjectManager equationManager = InformationManagers.getInstance().getSdObjectManager();

		// verify that each variable appearing on the rhs has a lhs in data

		for (Set<String> rhsVars : requiresExpandedLhsExpandedRHS.values()) {
			for (String rhsVar0 : rhsVars) {
				String rhsVar = rhsVar0.replace("!]", "]").replace("!,", ",");
				
				
				if (!requiresExpandedLhsExpandedRHS.containsKey(rhsVar)) {
					
					// no equation registered for this variable
					// if this is not an array reference, then it is definitely bad
					// if it is an array reference, it still may be ok based on terminal vs
					// named subscript usage
					if (ArrayReference.isArrayReference(rhsVar)) {
						
						// get the variable name without subscript information
						String aVar = getVariableOrArrayName(rhsVar);
						
						// get both lhs and rhs references for this variable
						Map<String, Set<String>> refs = InformationManagers.getInstance().getSdObjectManager().getVariableReferences(aVar);
						
						boolean bad = true;
						
						// if the rhs reference does not have a terminal subscript, check if there are
						// lhs references that use terminal subscripts to specifiy data for the array
						if (!containsTerminalSubscript(rhsVar)) {
							for (String lhs : refs.get(SDObjectManager.LHS)) {
								if (containsTerminalSubscript(lhs)) {
									bad = false;
								}
							}
							if (bad)
								badOnes.add(rhsVar);
						} else {
							// if the rhsvar has a terminal subscript, then there may be a lsh with named
							// subscripts
							for (String lhs : refs.get(SDObjectManager.LHS)) {
								if (!containsTerminalSubscript(lhs)) {
									bad = false;
								}
							}
							if (bad)
								badOnes.add(rhsVar);
						}
					} else {
						badOnes.add(rhsVar);
					}
				}
			}
		}

		return badOnes;
	}
	
	private boolean containsTerminalSubscript(String var) {

		ArrayManager amgr = InformationManagers.getInstance().getArrayManager();
		ArrayReference ar = new ArrayReference(var);
		List<String> subs = ar.getSubscripts();
		List<Integer> terminalSubs = new ArrayList<Integer>();
		int dim = 0;
		for (String sub : subs) {
			if (amgr.isTerminalSubscript(ar.getArrayName(), sub, dim)) {
				return true;
			}
			dim++;
		}
		return false;
	}

	protected Set<String> orderConstants(Map<String, Equation> equations,
			List<String> evaluationOrder) {
		// only equations that have no requirements or get data from outside
		// sources
		Map<String, HashSet<String>> requiresExpandedLhsExpandedRHS = variableDependencyProcessor
				.getRequiresExpandedLhsExpandedRHS();
		Map<String, HashSet<String>> usesExpandedRhsExpandedLhs = variableDependencyProcessor
				.getUsesExpandedRhsExpandedLhs();
		Map<String, HashSet<String>> realLhsToExpandedLhsMap = variableDependencyProcessor
				.getRealLhsToExpandedLhsMap();
		Map<String, String> expandedLhsToRealLhsMap = variableDependencyProcessor
				.getExpandedLhsToRealLhsMap();
		Map<String, HashSet<String>> realLhsToExpandedLhsNotInitialized = variableDependencyProcessor
				.getRealLhsToExpandedLhsNotInitialized();
		
		Set<String> allRemoved = new LinkedHashSet<String>();

		for (Equation eqn : equations.values()) {
			if (eqn.isConstant() || eqn.isData()) {
				// verify based on requires

				String origLHS = eqn.getOriginalLhs();
//				if (ArrayReference.isArrayReference(origLHS))
//					origLHS = ArrayReference.annotate(origLHS);
				
//				if (origLHS == null || realLhsToExpandedLhsMap.get(origLHS) == null)
//					System.out.println("something wrong");

				for (String expandedLHS : realLhsToExpandedLhsMap.get(origLHS)) {
					// if this expanded LHS have no requirements,
					// remove it from realToExpNotInit -- when remove all, add
					// real to list
//					if (requiresExpandedLhsExpandedRHS.containsKey(expandedLHS)) {
//						printSet(expandedLHS, requiresExpandedLhsExpandedRHS.get(expandedLHS));
//					}
					if (requiresExpandedLhsExpandedRHS.containsKey(expandedLHS)
							&& requiresExpandedLhsExpandedRHS.get(expandedLHS)
									.size() == 0) {
						// remove this requirement for other equations
						removeFromRHS(expandedLHS,
								requiresExpandedLhsExpandedRHS,
								usesExpandedRhsExpandedLhs);

						// see if we are ready to place this into equation order
						// list
						String realLHS = expandedLhsToRealLhsMap.get(expandedLHS);
//						printSet(realLHS, realLhsToExpandedLhsNotInitialized.get(realLHS));
						realLhsToExpandedLhsNotInitialized.get(realLHS).remove(expandedLHS);
						if (realLhsToExpandedLhsNotInitialized.get(realLHS).size() == 0) {
							
							if (!evaluationOrder.contains(realLHS)) 
								evaluationOrder.add(realLHS);
							allRemoved.add(realLHS);
						}
					}
					// }
				}
			}

		}
		return allRemoved;

	}
	
//	private void printSet(String key, HashSet<String> set) {
//		System.out.println("key: "+key+"  set: "+set);
//	}

	protected void orderInitializers(Map<String, Equation> equations,
			List<String> stockInitializers, List<String> constantsAndData) {
		Map<String, HashSet<String>> requiresExpandedLhsExpandedRHS = variableDependencyProcessor
				.getRequiresExpandedLhsExpandedRHS();
		Map<String, HashSet<String>> realLhsToExpandedLhsMap = variableDependencyProcessor
				.getRealLhsToExpandedLhsMap();

		// build the set of variables that are used to initialize our stocks
		Set<String> initializers = new LinkedHashSet<String>();

		for (Equation eqn : equations.values()) {
			if (eqn.isStock() || eqn.isActiveInitial()) {
				// need to find out what is still required
				String realLHS = eqn.getOriginalLhs();
//				if (ArrayReference.isArrayReference(realLHS))
//					realLHS = ArrayReference.annotate(realLHS);

				Set<String> fiv = eqn.getFunctionInitialVariables();
				
				for (String original : InformationManagers.getInstance()
						.getNativeDataTypeManager().getOriginalNames(fiv)) {
					if (!initializers.contains(original))
						initializers.add(original);
					
				}
			}
		}

		initializers = expandInitializers(initializers, constantsAndData);
//		for (String i : initializers) {
//			if (constantsAndData.contains(i))
//				System.out.println("orderStockInitializers DEBUG DUP "+i);
//		}
//		stockInitializers
//				.addAll(orderVariables(ArrayReference.annotate(initializers), constantsAndData));
		stockInitializers
		.addAll(orderVariables(initializers, constantsAndData));
	}

	public Set<String> expandInitializers(String realLhs) {

		// need to expand this set to all variables required to define these
		Set<String> expanded = new LinkedHashSet<String>();

		expanded.add(realLhs);
		expand(realLhs, expanded);

		return expanded;
	}

	public Set<String> expandInitializers(Set<String> initializers, List<String> constantsAndData) {

		// need to expand this set to all variables required to define these
		Set<String> expanded = new LinkedHashSet<String>();
		for (String realLhs : initializers) {
			
			if (constantsAndData.contains(realLhs))
				continue;
			
			if (ArrayReference.isArrayReference(realLhs) && realLhs.contains("!")) {
				System.out.println("initializer with !");
			}
			
			expanded.add(realLhs.replace("!", ""));
			
			expand(realLhs.replace("!", ""), expanded);
		}
		return expanded;
	}

	private void expand(String realLHS, Set<String> expanded) {

		Map<String, HashSet<String>> requiresExpandedLhsExpandedRHS = variableDependencyProcessor
				.getRequiresExpandedLhsExpandedRHS();
		Map<String, HashSet<String>> realLhsToExpandedLhsMap = variableDependencyProcessor
				.getRealLhsToExpandedLhsMap();
		Map<String, String> expandedLhsToRealLhsMap = variableDependencyProcessor
				.getExpandedLhsToRealLhsMap();
		
		Map<String, HashSet<String>> usesExpandedRhsExpandedLhs = variableDependencyProcessor
				.getUsesExpandedRhsExpandedLhs();

		// given a real LHS, need to loop over the expanded LHS to see if
		// anything is required
		// i.e. take this rhs var, treat as a lhs, and get a list of its
		// required lhs
		// repeat this process recursively to generate a list of all vars
		// required for initial value variables
		// this list will be of "real" values

		String realLhs = realLHS;
//		if (ArrayReference.isArrayReference(realLHS))
//			realLhs = ArrayReference.annotate(realLHS);
		if (!realLhsToExpandedLhsMap.containsKey(realLhs)) {
			log.debug("DEBUG expand");
			return;
		}
		for (String expandedLHS : realLhsToExpandedLhsMap.get(realLhs)) {
			for (String expandedRHS : requiresExpandedLhsExpandedRHS
					.get(expandedLHS)) {
				String realRhs = expandedLhsToRealLhsMap.get(expandedRHS);
				if (realRhs == null)
					continue;
				if (!expanded.contains(realRhs)) {
					expanded.add(realRhs);
					expand(realRhs, expanded);
				} else {
					// log.debug("Already Expanded: "+realRhs);
				}
			}
		}

		return;
	}

	protected List<String> orderVariables(Set<String> variablesIn, List<String> alreadyDefined) {

		Map<String, HashSet<String>> requiresExpandedLhsExpandedRHS = variableDependencyProcessor
				.getRequiresExpandedLhsExpandedRHS();
		Map<String, HashSet<String>> realLhsToExpandedLhsMap = variableDependencyProcessor
				.getRealLhsToExpandedLhsMap();
		Map<String, HashSet<String>> realLhsToExpandedLhsNotInitialized = variableDependencyProcessor
				.getRealLhsToExpandedLhsNotInitialized();
		Map<String, HashSet<String>> usesExpandedRhsExpandedLhs = variableDependencyProcessor
				.getUsesExpandedRhsExpandedLhs();

		Set<String> variables = new LinkedHashSet<String>(variablesIn);
		int stuckCount = 0;
		
		// never use time or timeStep for ordering as we know they will be defined
		variables.remove("time");
		variables.remove("timeStep");
		variables.remove("Time");
		
		// first make sure that we clean out all vars already defined
		


		List<String> vars = new ArrayList<String>();
		boolean done = false;
		while (!done) {
//			System.out.println("############## while not done ##################");
			List<String> added = new ArrayList<String>();

			// variables are "real" LHS with or without annotation -- why?

//			for (String realLhs :  realLhsToExpandedLhsMap.keySet() /*requiresExpandedLhsExpandedRHS.keySet() */  /* variables */ ) {
			for (String annotatedLHS :  equations.keySet() /*requiresExpandedLhsExpandedRHS.keySet() */  /* variables */ ) {
				String originalLHS = equations.get(annotatedLHS).getOriginalLhs();
				
//				System.out.println("          "+originalLHS);
				
				// make sure that it is annotated

//				if (ArrayReference.isArrayReference(realLhs)) {
//					realLHS = ArrayReference.annotate(realLHS);
//				}
				
				if (!realLhsToExpandedLhsMap.containsKey(originalLHS)) {
					if (!equations.get(originalLHS).isDefinesSubscript()) {
						log.debug("huh? "+originalLHS);
					}
					continue;
				}

				// get each of its expanded LHS
				for (String expandedLHS : realLhsToExpandedLhsMap.get(originalLHS)) {
					
					// is this expanded LHS not waiting for a value?
					// one of three conditions:
					// 1) requiresExpandedLhsExpandedRHS has Xlhs as key, but
					// there are no Xrhs vars waiting for data
					// 2) requiresExpandedLhsExpandedRHS does not hax Xlhs as a
					// key
					// 3) this realLHS has already been defined
					
					// is this necessary?
					
					if (requiresExpandedLhsExpandedRHS.get(expandedLHS) != null &&
							requiresExpandedLhsExpandedRHS.get(expandedLHS).size() > 0) {
						Set<String> s = requiresExpandedLhsExpandedRHS.get(expandedLHS);
						s.remove("time");
						s.remove("timeStep");
						s.remove("TIME");
						s.remove("Time");
						s.remove("TIMESTEP");
						
					}
					
					boolean condition0 = false;
					if (equations.get(annotatedLHS).isHasInitialValue()) {
//						log.debug("Condition 0 "+equations.get(realLHS).getVensimEquationOnly()+
//								(requiresExpandedLhsExpandedRHS.containsKey(expandedLHS) ? requiresExpandedLhsExpandedRHS.get(expandedLHS).size() : "Nothing"));
						if (requiresExpandedLhsExpandedRHS.get(expandedLHS).size() > 0) {
							Set<String> dependencies = new LinkedHashSet<String>(requiresExpandedLhsExpandedRHS.get(expandedLHS));
							int count = requiresExpandedLhsExpandedRHS.get(expandedLHS).size();
							for (String s : dependencies) {
								if (alreadyDefined.contains(s))
									count--;
							}
							if (count == 0)
								condition0 = true;
							
						} else {
							condition0 = true;
						}
//						log.debug("Condition 0 "+realLHS);
					}

					boolean condition1 = requiresExpandedLhsExpandedRHS.containsKey(expandedLHS)
							&& requiresExpandedLhsExpandedRHS.get(expandedLHS).size() == 0;
					
					boolean condition1a = requiresExpandedLhsExpandedRHS.containsKey(expandedLHS)
							&& requiresExpandedLhsExpandedRHS.get(expandedLHS).size() == 1 && 
							requiresExpandedLhsExpandedRHS.get(expandedLHS).contains(expandedLHS);
					
					boolean condition2 = !requiresExpandedLhsExpandedRHS.containsKey(expandedLHS);
					
					boolean condition3 = alreadyDefined.contains(originalLHS);

					if (condition0 || condition1 || condition1a || condition2 || condition3) {
						removeFromRHS(expandedLHS, requiresExpandedLhsExpandedRHS, usesExpandedRhsExpandedLhs);
						realLhsToExpandedLhsNotInitialized.get(originalLHS).remove(expandedLHS);
						if (realLhsToExpandedLhsNotInitialized.get(originalLHS).size() == 0) {
							
							if (variables.contains(originalLHS) && !added.contains(originalLHS))
								added.add(originalLHS);
							if (variablesIn.contains(originalLHS)) {
								if (!alreadyDefined.contains(originalLHS) && !vars.contains(originalLHS))
									vars.add(originalLHS);
//								if (!alreadyDefined.contains(realLHS))
//									alreadyDefined.add(realLHS);
							} else {
								// log.debug("Already Initialized: "+realLHS);
							}
						}
					}
				}
			}

			System.out.println("Added: " + added.size() + " vars left = "
					+ variables.size());

			if (added.size() == 0) {
				stuckCount++;
				// check if the remaining equation does not have an equation
				for (String remaining : variables) {
					
					if (InformationManagers.getInstance().getEquationManager().getAnEquationFor(remaining) == null) {
//					if (!equations.containsKey(remaining))
						added.add(remaining);
					}
				}
				if (stuckCount > 3) {
					cleanEquationOrdering = false;
					System.out.println("######################### Stuck Variables");
					printStuckWithFunctions(variables, realLhsToExpandedLhsMap, requiresExpandedLhsExpandedRHS);
					System.out.println("######################### Stuck Variables");
					return vars;
				}
			}

			variables.removeAll(added);
			if (variables.size() == 0)
				done = true;
		}

		System.out.println("########NOT STUCK!########");
		return vars;

	}

	protected void orderStocksAndActiveInitial(Map<String, Equation> equations,
			List<String> evaluationOrder) {
		Map<String, HashSet<String>> requiresExpandedLhsExpandedRHS = variableDependencyProcessor
				.getRequiresExpandedLhsExpandedRHS();
		Map<String, HashSet<String>> realLhsToExpandedLhsMap = variableDependencyProcessor
				.getRealLhsToExpandedLhsMap();
		Map<String, String> expandedLhsToRealLhsMap = variableDependencyProcessor
				.getExpandedLhsToRealLhsMap();
		Map<String, HashSet<String>> realLhsToExpandedLhsNotInitialized = variableDependencyProcessor
				.getRealLhsToExpandedLhsNotInitialized();
		Map<String, HashSet<String>> usesExpandedRhsExpandedLhs = variableDependencyProcessor
				.getUsesExpandedRhsExpandedLhs();

		// since we have forced all initial values variables to be initialized,
		// order is not important
		for (Equation eqn : equations.values()) {
			if (eqn.isStock() || eqn.isActiveInitial()) {
				String realLHS = eqn.getOriginalLhs();
//				if (ArrayReference.isArrayReference(realLHS))
//					realLHS = ArrayReference.annotate(realLHS);
				evaluationOrder.add(realLHS);
				for (String expandedLHS : realLhsToExpandedLhsMap.get(realLHS)) {
					removeFromRHS(expandedLHS, requiresExpandedLhsExpandedRHS, usesExpandedRhsExpandedLhs);
					requiresExpandedLhsExpandedRHS.get(expandedLHS).clear();
				}
				realLhsToExpandedLhsNotInitialized.get(realLHS).clear();
			}
		}
	}

	protected void orderAuxiliaries(Map<String, Equation> equations,
			List<String> evaluationOrder, List<String> alreadyDefined) {
		variableDependencyProcessor.loadForActiveInitial();  // what does thiis do?
		
		Map<String, HashSet<String>> requiresExpandedLhsExpandedRHS = variableDependencyProcessor
				.getRequiresExpandedLhsExpandedRHS();
		Map<String, HashSet<String>> realLhsToExpandedLhsMap = variableDependencyProcessor
				.getRealLhsToExpandedLhsMap();
		Map<String, HashSet<String>> realLhsToExpandedLhsNotInitialized = variableDependencyProcessor
				.getRealLhsToExpandedLhsNotInitialized();
		Map<String, HashSet<String>> usesExpandedRhsExpandedLhs = variableDependencyProcessor
				.getUsesExpandedRhsExpandedLhs();
		
		
		
		for (String realLhs :  alreadyDefined ) {
			String realLHS = realLhs;
//			if (ArrayReference.isArrayReference(realLhs)) {
//				realLHS = ArrayReference.annotate(realLHS);
//			}
			for (String expandedLHS : realLhsToExpandedLhsMap.get(realLHS)) {
				removeFromRHS(expandedLHS, requiresExpandedLhsExpandedRHS, usesExpandedRhsExpandedLhs);
				realLhsToExpandedLhsNotInitialized.get(realLHS).remove(expandedLHS);
			}
		}
		
		Set<String> variables = new LinkedHashSet<String>();

		for (Equation eqn : equations.values()) {
			
			if (!eqn.isAssignment())
				continue;
			if (eqn.isStock() || eqn.isData() || eqn.isConstant())
				continue;
			if (eqn.isOneTime())
				continue;
			
			String lhs = eqn.getOriginalLhs();
			variables.add(lhs);
		}
		
		
		evaluationOrder.addAll(orderVariables(variables,
				alreadyDefined));
//		evaluationOrder.addAll(orderVariables(ArrayReference.annotate(variables),
//				alreadyDefined));

	}





	protected void dumpRequires(String fileName,
			Map<String, HashSet<String>> requires) {

		BufferedWriter bw = Utilities.openFileForWriting(miscDirectory + "/"
				+ fileName);
		ArrayList<String> allLHS = new ArrayList<String>();

		for (String lhs : requires.keySet()) {
			if (lhs == null) {
				log.debug("Skipping null in requires");
			} else {
				allLHS.add(lhs);
			}
		}

		Collections.sort(allLHS);

		try {
			for (String lhs : allLHS) {
				bw.append("LHS: <" + lhs + ">\n");
				for (String r : sortSet(requires.get(lhs))) {
					bw.append("     RHS <" + r + ">\n");
				}
			}
			bw.flush();
			bw.close();
		} catch (IOException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
	}

	protected void dumpRequires2(String fileName, Map<String, String> requires) {

		BufferedWriter bw = Utilities.openFileForWriting(miscDirectory + "/"
				+ fileName);
		ArrayList<String> allLHS = new ArrayList<String>();

		for (String lhs : requires.keySet()) {
			if (lhs == null) {
				log.debug("Skipping null in requires");
			} else {
				allLHS.add(lhs);
			}
		}

		Collections.sort(allLHS);

		try {
			for (String lhs : allLHS) {
				bw.append("LHS: <" + lhs + ">\n");
				String r = requires.get(lhs);
				bw.append("     RHS <" + r + ">\n");
			}

			bw.flush();
			bw.close();
		} catch (IOException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
	}
	
	protected void analyzeDependancies(
			Map<String, HashSet<String>> requires) {

		ArrayList<String> allLHS = new ArrayList<String>();

		for (String lhs : requires.keySet()) {
			if (lhs == null) {
//				log.debug("Skipping null in requires");
			} else {
				allLHS.add(lhs);
			}
		}

		Collections.sort(allLHS);

	
			for (String lhs : allLHS) {
				if (requires.get(lhs).size() == 0)
					continue;
				
				int level = 1;
				// want to find the loop
			}
	
	}
	
	protected void dumpRequiresExpandedWithDepend(String fileName,
			Map<String, HashSet<String>> requires, boolean printChain) {

//		 boolean printChain = true;

		BufferedWriter bw = Utilities.openFileForWriting(miscDirectory + "/"
				+ fileName);
		ArrayList<String> allLHS = new ArrayList<String>();

		for (String lhs : requires.keySet()) {
			if (lhs == null) {
				log.debug("Skipping null in requires");
			} else {
				allLHS.add(lhs);
			}
		}

		Collections.sort(allLHS);
		
		int totalCount = 0;

		try {
			for (String lhs : allLHS) {
				if (InformationManagers.getInstance().getEquationManager().getAnEquationFor(lhs) == null) {
					continue;
				}
//				if (requires.get(lhs).size() == 0)
//					continue;
				totalCount++;
				if (totalCount > 10)
					continue;
				bw.append("LHS: <" + lhs + "> "+InformationManagers.getInstance().getEquationManager().getAnEquationFor(lhs).getVensimEquationOnly()+"\n");
				int level = 1;
				if (printChain) {
					printChain(bw, level, lhs, requires, new ArrayList<String>());
				} else {
					for (String r : sortSet(requires.get(lhs))) {
						bw.append("     RHS <" + r + "> "+InformationManagers.getInstance().getEquationManager().getAnEquationFor(r).getVensimEquationOnly()+"\n");
					}
				}

			}
			bw.flush();
			bw.close();
		} catch (IOException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
	}

	protected void dumpRequiresExpanded(String fileName,
			Map<String, HashSet<String>> requires, boolean printChain) {

		// boolean printChain = true;

		BufferedWriter bw = Utilities.openFileForWriting(miscDirectory + "/"
				+ fileName);
		ArrayList<String> allLHS = new ArrayList<String>();

		for (String lhs : requires.keySet()) {
			if (lhs == null) {
				log.debug("Skipping null in requires");
			} else {
				allLHS.add(lhs);
			}
		}

		Collections.sort(allLHS);

		try {
			for (String lhs : allLHS) {
				if (requires.get(lhs).size() == 0)
					continue;
				bw.append("LHS: <" + lhs + ">\n");
				int level = 1;
				if (printChain) {
					printChain(bw, level, lhs, requires,
							new ArrayList<String>());
				} else {
					for (String r : sortSet(requires.get(lhs))) {
						bw.append("     RHS <" + r + ">\n");
					}
				}

			}
			bw.flush();
			bw.close();
		} catch (IOException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
	}

	// perform any misc transformations that are required
	public void processSubscriptDefinition(Map<String, Equation> equations) {
		for (String lhs : equations.keySet()) {
			Equation eqn = equations.get(lhs);
			eqn.processSubscriptDefinition();

		}
	}

	// perform any misc transformations that are required
	public void processExponentiaion(Map<String, Equation> equations) {
		for (String lhs : equations.keySet()) {
			Equation eqn = equations.get(lhs);
			eqn.processExponentiation();

		}
	}

	// perform any misc transformations that are required
	public void generateRPN(Map<String, Equation> equations) {
		for (String lhs : equations.keySet()) {

			Equation eqn = equations.get(lhs);

			try {
				eqn.generateRpn();
			} catch (Exception e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			}
		}
	}

	protected void analyzeEvaluationOrder(List<String> evaluationOrder,
			Map<String, Equation> equations) {
		
		info.addMessage(SDMDocLanguageSpecific.getInstance().getText(StringConstants.DETERMINING_EVALUATION_ORDER));

		Map<String, HashSet<String>> lhsExpandedMap = variableDependencyProcessor
				.getRealLhsToExpandedLhsMap();
		Map<String, String> expandedLhsMap = variableDependencyProcessor
				.getExpandedLhsToRealLhsMap();

		boolean orderProblem = false;
		BufferedWriter bw = Utilities.openFileForWriting(miscDirectory + "/"
				+ "evaluationOrder.txt");
		BufferedWriter bw2 = Utilities.openFileForWriting(miscDirectory + "/"
				+ "evaluationOrder.tab");
		HashMap<String, List<Integer>> initializationLine = new LinkedHashMap<String, List<Integer>>();

		Set<String> printed = new LinkedHashSet<String>();
		for (int eqnOrder = 0; eqnOrder < evaluationOrder.size(); eqnOrder++) {
			if (!printed.contains(evaluationOrder.get(eqnOrder))) {
				if (InformationManagers.getInstance().getEquationManager().getAnEquationFor(evaluationOrder.get(eqnOrder)).isAssignment()) {
//				if (equations.get(evaluationOrder.get(eqnOrder)).isAssignment()) {
					try {
						bw2.append(eqnOrder + "\t" + evaluationOrder.get(eqnOrder) + "\n");
						printed.add(evaluationOrder.get(eqnOrder));
					} catch (IOException e) {
						// TODO Auto-generated catch block
						e.printStackTrace();
					}
				}
			}

			if (!initializationLine.containsKey(evaluationOrder.get(eqnOrder))) {
				// allow for an equation to be in the order twice
				// once for initialization and once for repeated execution
				initializationLine.put(evaluationOrder.get(eqnOrder), new ArrayList<Integer>());
				initializationLine.get(evaluationOrder.get(eqnOrder)).add(eqnOrder);

				String lhs = evaluationOrder.get(eqnOrder);

				if (expandedLhsMap.containsKey(lhs)) {
					String real = expandedLhsMap.get(lhs);
					initializationLine.put(real,  new ArrayList<Integer>());
					initializationLine.get(real).add(eqnOrder);
				}

				if (lhsExpandedMap.containsKey(lhs)) {
					for (String exp : lhsExpandedMap.get(lhs)) {
						initializationLine.put(exp, new ArrayList<Integer>());
						initializationLine.get(exp).add(eqnOrder);
					}
				}
			}
		}

		try {
			bw2.close();
		} catch (IOException e1) {
			// TODO Auto-generated catch block
			e1.printStackTrace();
		}

		try {
			for (int lhsInitializationLine = 0; lhsInitializationLine < evaluationOrder
					.size(); lhsInitializationLine++) {
				String lhs = evaluationOrder.get(lhsInitializationLine);
				Equation eqn = InformationManagers.getInstance().getEquationManager().getAnEquationFor(lhs);
				
				// some lhs don't actually exist in the model but are created in
				// code

				if (lhs.equalsIgnoreCase("time")
						|| lhs.equalsIgnoreCase("timestep"))
					continue;

				bw.append(lhsInitializationLine + " ");
				bw.append("EQN:<" + lhs + "> " 
						+ (InformationManagers.getInstance().getEquationManager().getAnEquationFor(lhs).isOneTime() ? "Once" : "Repeated") + " "
						+ InformationManagers.getInstance().getEquationManager().getAnEquationFor(lhs).getVensimEquationOnly() + "\n");
//				bw.flush();

				Set<String> rhsVars = null;
				
				boolean useExpandedSubscriptOrdering = InformationManagers.getInstance().getModelManager().isRequiresExpandedSubscriptOrdering();

				if (eqn.isRepeated()) {
					if (useExpandedSubscriptOrdering) {
						rhsVars = eqn.getRHSVariablesForOrderingExpanded();
					} else {
						rhsVars = eqn.getRHSVariablesForOrdering();
					}
				}

				// if we are working with a stock that uses a constant as an
				// initializer, we are ok without checking
				if (eqn.isStock()) {
					String initialVal = eqn.getInitialValue();
					if (Parser.isNumber(initialVal) && rhsVars != null)
						rhsVars.clear();
				}

				boolean problem = false;
				boolean orderingProblem = false;
				boolean missingProblem = false;

				if (rhsVars == null || rhsVars.size() == 0) {
					problem = false;
					orderingProblem = false;
					missingProblem = false;
				} else {
					int rhsInitializationLine = 0;
					for (String rhst : rhsVars) {
						String rhs = InformationManagers.getInstance()
								.getNativeDataTypeManager()
								.getOriginalName(rhst);
						// log.debug("DEBUG analyze RHS: "+rhs);

						if (rhs.equals("TIME STEP"))
							continue;

						// need to contend with array references 30Sep

						if (initializationLine.get(rhs) != null) {
							rhsInitializationLine = initializationLine.get(rhs).get(0);
							eqn = InformationManagers.getInstance().getEquationManager().getAnEquationFor(lhs);

							if (lhsInitializationLine < rhsInitializationLine) {
								// if (containsInitialValueFunction(eqn)) {
								// Set<String> ivars =
								// eqn.getFunctionInitialVariables();
								// if (ivars.contains(rhs)) {
								// orderingProblem = true;
								// }
								// } else {
								// orderingProblem = true;
								// }
								orderingProblem = true;
								if (orderingProblem) {
									bw.append("     <"
											+ rhs
											+ "> "
											+ rhsInitializationLine
											+ (lhsInitializationLine < rhsInitializationLine ? " <<<<<<<<"
													: "") + "\n");
								} else {
									bw.append("     <"
											+ rhs
											+ "> "
											+ rhsInitializationLine
											+ " not part of initialization calculation\n");
								}
							} else {
								bw.append("     <" + rhs + "> "
										+ rhsInitializationLine + "\n");
							}
						} else {
							// this might be a array assignment and we alooking
							// at an expanded LHS
							// convert back to the realLHS
							String rLHS = expandedLhsMap.get(rhs);
							if (initializationLine.containsKey(rLHS)) {
								rhsInitializationLine = initializationLine
										.get(rLHS).get(0);
								if (lhsInitializationLine < rhsInitializationLine) {
									orderingProblem = true;
									bw.append("     <"
											+ rLHS
											+ "> "
											+ rhsInitializationLine
											+ (lhsInitializationLine < rhsInitializationLine ? " <<<<<<<<"
													: "") + "\n");
								} else {
									bw.append("     <"
											+ rLHS
											+ "> "
											+ rhsInitializationLine
											+ (lhsInitializationLine < rhsInitializationLine ? " <<<<<<<<"
													: "") + "\n");
								}
								// if (lhsInitializationLine <
								// rhsInitializationLine ) {
								// if (containsInitialValueFunction(eqn)) {
								// Set<String> ivars =
								// eqn.getFunctionInitialVariables();
								// if (ivars.contains(rhs)) {
								// orderingProblem = true;
								// }
								// }
								// }
							} else {
								if (useExpandedSubscriptOrdering) {
									bw.append(" >>>> initializtion line missing for <"
											+ rhs + "> try individuals\n");
									missingProblem = false;
									// if the rhs is an array reference to a "group"
									// see if the individual array refers have
									// initialization lines
									if (rhs.contains("!")
											|| !ArrayReference
											.isArrayReference(lhs)) {
										if (!ArrayReference.isArrayReference(rhs)) {
											System.out
											.println("about to check for array reference when none is there! lhs: "
													+ lhs + " rhs: " + rhs);
											continue;
										}

										ArrayReference rhsArrayReference = new ArrayReference(
												rhs.replace("!", ""));

										for (SubscriptCombination subrCombo : InformationManagers
												.getInstance()
												.getArrayManager()
												.getSubscriptValueCombinations(
														rhsArrayReference
														.getSubscriptsAsArray())) {
											// Build an array reference based on the
											// subscript value
											String expandedRHS = "array."
													+ rhsArrayReference
													.getArrayName() + "[";
											int subn = 0;
											for (String subscript : rhsArrayReference
													.getSubscriptsAsArray()) {
												if (subn++ > 0) {
													expandedRHS += ",";
												}
												expandedRHS += subrCombo
														.getSubscriptValue(subscript);
											}
											expandedRHS += "]";
											if (!initializationLine
													.containsKey(expandedRHS)) {
												bw.append(" >>>> initializtion line missing for individual <"
														+ expandedRHS + ">\n");
												missingProblem = true;
											} else {
												rhsInitializationLine = initializationLine
														.get(expandedRHS).get(0);
												bw.append(" >>>> initializtion line FOUND for individual <"
														+ expandedRHS
														+ "> "
														+ rhsInitializationLine
														+ "\n");
											}
										}

									} else if (ArrayReference.isArrayReference(lhs)) {
										if (!ArrayReference.isArrayReference(rhs)) {
											// log.debug("Why do I think that this should be an AR");
											continue;
										}
										ArrayReference rhsArrayReference = new ArrayReference(
												rhs.replace("!", ""));
										ArrayReference lhsArrayReference = new ArrayReference(
												lhs);
										for (SubscriptCombination subrCombo : InformationManagers
												.getInstance()
												.getArrayManager()
												.getSubscriptValueCombinations(
														lhsArrayReference
														.getSubscriptsAsArray())) {
											// Build an array reference based on the
											// subscript value
											String expandedRHS = "array."
													+ rhsArrayReference
													.getArrayName() + "[";
											int subn = 0;
											for (String subscript : rhsArrayReference
													.getSubscriptsAsArray()) {
												if (subn++ > 0) {
													expandedRHS += ",";
												}
												expandedRHS += subrCombo
														.getSubscriptValue(subscript);
											}
											expandedRHS += "]";
											if (!initializationLine
													.containsKey(expandedRHS)) {
												bw.append(" >>>> initializtion line missing for individual <"
														+ expandedRHS + ">\n");
												missingProblem = true;
											} else {
												rhsInitializationLine = initializationLine
														.get(expandedRHS).get(0);
												bw.append(" >>>> initializtion line FOUND for individual <"
														+ expandedRHS
														+ "> "
														+ rhsInitializationLine
														+ "\n");
											}
										}
									} else {
									}
								}
							}
						}
					}
				}
				if (problem) {
					bw.append("######### P R O B L E M ######### \n");
					orderProblem = true;
				}

				if (orderingProblem) {
					bw.append("######### ORDERING P R O B L E M #########\n");
					orderProblem = true;
				}

				if (missingProblem) {
					bw.append("######### MISSING P R O B L E M #########\n");
					orderProblem = true;
				}
			}
//			bw.flush();
			bw.close();

			if (orderProblem) {
				info.addMessage(SDMDocLanguageSpecific.getInstance().getText(StringConstants.INVALID));
				info.addMessage(SDMDocLanguageSpecific.getInstance().getText(StringConstants.SEE)+" evaluationOrder.txt");
			} else {
				info.addMessage(SDMDocLanguageSpecific.getInstance().getText(StringConstants.CLEAN_EVALUATION_ORDER));
			}

		} catch (IOException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
	}

	protected void removeRequirement(ArrayList<String> toRemoveFromRequires,
			Map<String, HashSet<String>> requiresExpandedLhsExpandedRHS,
			Map<String, HashSet<String>> realLhsToExpandedLhsNotInitialized,
			Map<String, String> expandedLhsToRealLhsMap,
			List<String> evaluationOrder) {
		
		Map<String, HashSet<String>> usesExpandedRhsExpandedLhs = variableDependencyProcessor
				.getUsesExpandedRhsExpandedLhs();

		System.out
				.println("removeRequirement # " + toRemoveFromRequires.size());

		for (String toRemove : toRemoveFromRequires) {

			if (ArrayReference.isArrayReference(toRemove)) {

				// we've determined that this array reference can take place
				// remove it from requiresExpanded
				// this does not added to evaluation order until we know that
				// all other references to this array have been processed

				if (requiresExpandedLhsExpandedRHS.remove(toRemove) == null) { // "LHS"
					// of
					// equation
					log.debug("Bad Remove from requiresExpanded: "
							+ toRemove);

				}

				if (realLhsToExpandedLhsNotInitialized.remove(toRemove) == null) { // "LHS"
					// of
					// equation
					// log.debug("Bad Remove from lhsExpandedNotInitialized: "+toRemove);

				}
				removeFromRHS(toRemove, requiresExpandedLhsExpandedRHS, usesExpandedRhsExpandedLhs); // "RHS"
																			// of
																			// equation

				// at this point, we need to see if all array subscripts have
				// been achieved. Then we can add to evaluationOrder

				String realLHS = expandedLhsToRealLhsMap.get(toRemove);
				HashSet<String> ehs = realLhsToExpandedLhsNotInitialized
						.get(realLHS);
				if (ehs != null && ehs.contains(toRemove)) {
					ehs.remove(toRemove);

					if (ehs.size() == 0) {
						evaluationOrder.add(realLHS);
						realLhsToExpandedLhsNotInitialized.remove(realLHS);
						removeFromRHS(realLHS, requiresExpandedLhsExpandedRHS, usesExpandedRhsExpandedLhs);

					} else {

					}
				} else if (ehs == null) {
					if (!evaluationOrder.contains(realLHS))
						evaluationOrder.add(realLHS);
					else
						log.debug("Already in Order: " + realLHS);
					realLhsToExpandedLhsNotInitialized.remove(realLHS);
					removeFromRHS(realLHS, requiresExpandedLhsExpandedRHS, usesExpandedRhsExpandedLhs);
				}

			} else {
				// we remove from requires and then check requires and
				// requiresExpanded
				// this will always be a non array LHS

				// test this 3/28/2015
				if (realLhsToExpandedLhsNotInitialized.containsKey(toRemove)) {
					if ((realLhsToExpandedLhsNotInitialized.get(toRemove)
							.size() == 1
							&& realLhsToExpandedLhsNotInitialized.get(toRemove)
									.contains(toRemove) || realLhsToExpandedLhsNotInitialized
							.get(toRemove).size() == 0))
						realLhsToExpandedLhsNotInitialized.remove(toRemove);
				}
				evaluationOrder.add(toRemove);
				removeFromRHS(toRemove, requiresExpandedLhsExpandedRHS, usesExpandedRhsExpandedLhs);
			}
		}
	}

	protected void removeFromRHS(String lhs,
			Map<String, HashSet<String>> requiresExpandedLhsExpandedRHS,
			Map<String, HashSet<String>> usesExpandedRhsExpandedLhs) {
		removeFromRHS(lhs, requiresExpandedLhsExpandedRHS, usesExpandedRhsExpandedLhs, "requiresExpanded");
	}

	protected void removeFromRHS(String key,
			Map<String, HashSet<String>> requires, 
			Map<String, HashSet<String>> usesExpandedRhsExpandedLhs, String tag) {

		if (usesExpandedRhsExpandedLhs.containsKey(key)) {
			HashSet<String> hs = usesExpandedRhsExpandedLhs.get(key);
			for (String use : hs) {
				if (requires.containsKey(use))
					requires.get(use).remove(key);
			}
		}
	}

	public void generateTrees(Map<String, Equation> equations) {
		for (String lhs : equations.keySet()) {
			Equation eqn = equations.get(lhs);
			eqn.generateTree();
		}
	}

	/**
	 * @return the sdModel
	 */
	public SDModel getSdModel() {
		return sdModel;
	}

	/**
	 * @return the generateC
	 */
//	public boolean isGenerateC() {
//		return generateC;
//	}

	/**
	 * @return the generateJava
	 */
	public boolean isGenerateJava() {
		return generateJava;
	}

	protected String getProjectLocation() {
		return "output";
	}

	public String getSourceDirectory() {
		return destinationDirectory + "/" + "src" + "/";
	}

	protected String asDirectoryPath(String packageName) {
		return packageName.replace(".", "/");
	}

	protected boolean hasInitialValue(String lhs,
			Map<String, HashSet<String>> requires,
			Map<String, Equation> equations) {
		boolean initialized = false;

		Equation eqn = equations.get(lhs);

		// functions that support initial values
		if (containsInitialValueFunction(equations.get(lhs))) {

			// HERE -- get a list of the initialization variables in the
			// statement and check each of them

			// if (Equation.isNumber(var))
			// return true;
			// if (requires.containsKey(var))
			// return false;

			initialized = true;
			Set<String> initializationVariables = equations.get(lhs)
					.getFunctionInitialVariables();

			for (String var : initializationVariables) {

				String v = removeValueOf(var).replaceAll("\"", "");
				// the requires
				v = InformationManagers.getInstance()
						.getNativeDataTypeManager().getOriginalName(v);

				if (Parser.isNumber(v))
					continue;
				else if (requires.containsKey(v)) {
					initialized = false;
				} else {
					// log.debug("Not in requires: "+v+" "+equations.get(lhs).getVensimEquationOnly());
				}
			}
		}
		if (!equations.get(lhs).isOrderedWithInitialValue()) {
			equations.get(lhs).setOrderedWithInitialValue(initialized);
		}
		return initialized;
	}

	protected boolean hasInitialValue(String lhs, String expandedLHS,
			Map<String, HashSet<String>> requiresExpanded,
			Map<String, HashSet<String>> lhsExpandedNotInitialized,
			Map<String, Equation> equations) {
		boolean initialized = false;

		Equation eqn = equations.get(lhs);
		// functions that support initial values
		if (containsInitialValueFunction(equations.get(lhs))) {

			initialized = true;
			Set<String> initializationVariables = equations.get(lhs)
					.getFunctionInitialVariables();

			for (String var : initializationVariables) {

				String v = removeValueOf(var).replaceAll("\"", "");
				// the requires
				v = InformationManagers.getInstance()
						.getNativeDataTypeManager().getOriginalName(v);

				if (Parser.isNumber(v))
					continue;

				// ArrayReference lhsAR = new ArrayReference(lhs);
				// ArrayReference lhsExpandedAR = new
				// ArrayReference(expandedLHS);
				// ArrayReference rhsAR = new ArrayReference(v);
				if (lhsExpandedNotInitialized.containsKey(v)
						&& lhsExpandedNotInitialized.get(v).size() > 0) {
					initialized = false;
					// continue;
				}
			}
			if (initialized) {
				if (!equations.get(lhs).isOrderedWithInitialValue()) {
					equations.get(lhs).setOrderedWithInitialValue(initialized);
				}
			}
		}

		// if (!equations.get(lhs).isOrderedWithInitialValue()) {
		// equations.get(lhs).setOrderedWithInitialValue(initialized);
		// }

		return initialized;
	}

	protected boolean containsInitialValueFunction(Equation equation) {
		// want to traverse tree (via tokens)
		for (String token : equation.getTokens()) {
			if (Parser.isFunctionInvocation(token)) {
				FunctionDescription fd = InformationManagers.getInstance()
						.getFunctionManager().getDescription(token);
				if (fd.isSuppliesInitialValue()) {
					return true;
				}
			}
		}
		return false;
	}

	protected String removeValueOf(String var) {
		String retVar = var.replace("valueOf(", "").replaceAll("\"", "")
				.replaceAll("\\)", "");
		return retVar;
	}

	protected void printChain(BufferedWriter bw, int level, String lhs,
			Map<String, HashSet<String>> requires, List<String> alreadyPrinted) {

		if (level > 100) {
			try {
				bw.append("Terminating recursion\n");
				return;
			} catch (IOException e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			}
		}

		StringBuilder blanks = new StringBuilder();
		for (int i = 0; i < level; i++) {
			blanks.append(" ");
		}

		for (String r : sortSet(requires.get(lhs))) {
			if (alreadyPrinted.contains(r)) {
				try {
					bw.append("   ### looping on " + r + "\n");
				} catch (IOException e) {
					// TODO Auto-generated catch block
					e.printStackTrace();
				}
//				 continue;
				return;
			} else {
				try {
					if (InformationManagers.getInstance().getEquationManager().getAnEquationFor(r) != null) {
					bw.append(blanks.toString() + "RHS <" + r + "> "+InformationManagers.getInstance().getEquationManager().getAnEquationFor(r).getVensimEquationOnly()+"\n");
					
					} else {
						log.debug("NULL EQUATION for <"+r+">");
					}
					alreadyPrinted.add(r);
				} catch (IOException e) {
					// TODO Auto-generated catch block
					e.printStackTrace();
				}
				printChain(bw, level + 1, r, requires, alreadyPrinted);
			}
			if (alreadyPrinted.size() > 0)
				alreadyPrinted.remove(alreadyPrinted.size() - 1);
		}

	}
	
	protected void printChain(int level, String lhs,
			Map<String, HashSet<String>> requires, List<String> alreadyPrinted) {

		if (level > 5) {
			
				System.out.println("Terminating recursion\n");
				return;
			
		}

		StringBuilder blanks = new StringBuilder();
		for (int i = 0; i < level; i++) {
			blanks.append(" ");
		}

		for (String r : sortSet(requires.get(lhs))) {
			if (alreadyPrinted.contains(r)) {
					System.out.println("   ### looping on " + r + "\n");
				return;
			} else {
					if (InformationManagers.getInstance().getEquationManager().getAnEquationFor(r) != null) {
					System.out.println(blanks.toString() + "RHS <" + r + "> "+InformationManagers.getInstance().getEquationManager().getAnEquationFor(r).getVensimEquationOnly()+"\n");
					
					} else {
						System.out.println("NULL EQUATION for <"+r+">");
					}
					alreadyPrinted.add(r);
		
				printChain(level + 1, r, requires, alreadyPrinted);
			}
			if (alreadyPrinted.size() > 0)
				alreadyPrinted.remove(alreadyPrinted.size() - 1);
		}

	}

	protected List<String> sortSet(HashSet<String> set) {
		ArrayList<String> al = new ArrayList<String>();
		if (set == null)
			return al;
		Iterator<String> iter = set.iterator();
		while (iter.hasNext())
			al.add(iter.next());

		Collections.sort(al);

		return al;
	}

	protected void printRequires(HashSet<String> set) {

		Iterator<String> iter = set.iterator();
		while (iter.hasNext()) {
			log.debug("           " + iter.next());
		}
		log.debug(">>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>");
	}

//	private boolean nativeMemory() {
//		return memoryImplementation.equalsIgnoreCase("Native");
//	}

	public static void writeMapToFile(String file, Map<String, Set<String>> map) {
		BufferedWriter bw = Utilities.openFileForWriting(file);
		try {
			for (String key : map.keySet()) {
				Set<String> s = map.get(key);
				bw.append(key);
				for (String v : s) {
					bw.append("\t" + v);
				}
				bw.append("\n");
			}
			bw.close();
		} catch (IOException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
	}

	private void print(String s, HashSet<String> hashset) {
		System.out.println(s+" <"+hashset.size()+">");
		for (String req : hashset) {
			System.out.println("   " + req);
		}
	}

	private void print(Set<String> list, Map<String, HashSet<String>> realLhsToExpandedLhsMap, Map<String, HashSet<String>> requiresExpandedLhsExpandedRHS) {
		log.debug("Stuck variables:");
		List<String> al = new ArrayList<String>(list);

		Collections.sort(al);

		for (String s : al) {
			log.debug(s +" "+(equations.get(s) == null ? "NO EQUATION FOUND " : equations.get(s).getVensimEquationOnly()));
			String s2 = s;
			if (ArrayReference.isArrayReference(s2)) {
				s2 = ArrayReference.annotate(s2);
			}
			if (realLhsToExpandedLhsMap.containsKey(s2)) {
				for (String x : realLhsToExpandedLhsMap.get(s2))
					print(x, requiresExpandedLhsExpandedRHS.get(x));
			} else {
				log.debug("How is this possible");
			}

		}
	}
	
	private void printStuckWithFunctions(Set<String> list, Map<String, HashSet<String>> realLhsToExpandedLhsMap, Map<String, HashSet<String>> requiresExpandedLhsExpandedRHS) {
		log.debug("Stuck variables:");
		List<String> al = new ArrayList<String>(list);

		Collections.sort(al);

		for (String s : al) {
			Equation eqn = InformationManagers.getInstance().getEquationManager().getAnEquationFor(s); // equations.get(s);
			if (eqn == null) {
				System.out.println(s +" NO EQUATION FOUND " );
			} else {
				if (eqn.isCallsFunction()) {
					System.out.println("FUNC CALL: "+s +" "+eqn.getVensimEquationOnly());
				} else {
					System.out.println(eqn.getVensimEquationOnly()+"\n");
				}

					String s2 = s;
//					if (ArrayReference.isArrayReference(s2)) {
//						s2 = ArrayReference.annotate(s2);
//					}
					if (realLhsToExpandedLhsMap.containsKey(s2)) {
						for (String x : realLhsToExpandedLhsMap.get(s2)) {
//							print(x, requiresExpandedLhsExpandedRHS.get(x));
						printChain(1, x, requiresExpandedLhsExpandedRHS, new ArrayList<String>());
						}
					} else {
						System.out.println("How is this possible");
					}
				}
			}

		
	}

	private void print(String tag, List<String> list) {
		List<String> al = new ArrayList<String>(list);
		Collections.sort(al);
		
		System.out.println(" ++++++ "+tag+" ++++++");
		for (String s : al)
			System.out.println(s);
	}

	public boolean isCleanEquationOrdering() {
		return cleanEquationOrdering;
	}

	public void setCleanEquationOrdering(boolean cleanEquationOrdering) {
		this.cleanEquationOrdering = cleanEquationOrdering;
	}

}
