package gov.anl.gss.sdm.server.model.translator;


import gov.anl.gss.sdm.server.model.support.ArrayReference;

import java.io.BufferedWriter;
import java.io.IOException;
import java.util.ArrayList;
import java.util.HashSet;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;

public class ActiveInitialProcessor {
	
	public final static String INITIAL_VALUE_MARKING = "__initialValue__";
	
	/*
	 * ACTIVE INITIAL:

(1) get all stocks
(2) generate list of required variable for initialization
(3) take the union of (1) and (2) and grab these from orderedEquations
(4) generate code for all but levels.
(5) extract initial value from stock equation and generate a new equation for this and assign
	to a stock specific value.
(6) modify stock equation tree to insert new variable for initial value
(7) write these equations to a new function.

Need to tell ACTIVE INITIAL that we are either in INITIAL MODE or ACTIVE MODE
Add new argument?

markGenerated
	 */
	
	private Map<String, Equation> equations;
	private List<String> evaluationOrder;
	private List<String> initializationOrder;
	private BufferedWriter bw;
	private List<String> variablesToDeclare;
	private List<String> requiredVariables;
	private List<Equation> equationsWithInitialValue;
	private List<String> initialValueVariables;
	
	private String targetLanguage;
	
	private VariableDependencyProcessor variableDependencyProcessor;
	
	private Map<String, String> lhsToInitialValueVariable =
		new LinkedHashMap<String, String>();
	
	public ActiveInitialProcessor(Map<String, Equation> equations, List<String> evaluationOrder, String targetLanaguage) {
		this.equations = equations;
		this.evaluationOrder = evaluationOrder;
		this.targetLanguage = targetLanaguage;
	}
	
	/**
	 * check to see if there are any ACTIVE INITIAL calls made in the set of equations
	 * @param equations
	 * @return
	 */
	public static boolean requiresProcessing(Map<String, Equation> equations) {
		// check to see if there are any ACTIVE INITIAL calls made in the set of equations
		for(String lhs : equations.keySet()) {
			Equation eqn = equations.get(lhs);
			if (eqn.isActiveInitial())
				return true;
		}
		
		return false;
	}
	
	public static String makeInitialVariable(String variable) {
		if (ArrayReference.isArrayReference(variable)) {
			ArrayReference ar = new ArrayReference(variable);
			return ar.getArrayName()+INITIAL_VALUE_MARKING+ar.getOriginalSubscript();
		} else {
			return variable+INITIAL_VALUE_MARKING;
		}
	}
	
	public void analyze(VariableDependencyProcessor vdp) {
		this.variableDependencyProcessor = vdp;
		// get the list of equations that contain initial values
		equationsWithInitialValue = getEquationsWithInitialValue();
		// for each of these equations, get the variable/constant in the iv position
		initialValueVariables = getInitialValueVariables();
		// for each of these, get the rhs components required
		requiredVariables = getRequiredVariables();
		initializationOrder = getEvalutationOrder();
		
	}
	
	public List<String> getEvalutationOrder() {
		
		variableDependencyProcessor.loadForDependency();
		Map<String, String> expandedLhsToRealLhsMap = variableDependencyProcessor.getExpandedLhsToRealLhsMap();
		Map<String, HashSet<String>> realLhsToExpandedLhsMap = variableDependencyProcessor.getRealLhsToExpandedLhsMap();
		
		List<String> aiOrder = new ArrayList<String>();
		for (String lhs : evaluationOrder) {
			
			if (requiredVariables.contains(lhs)) {
				if (!aiOrder.contains(lhs)) {
//					System.out.println("ADDING TO ORDER(1): "+lhs);
					aiOrder.add(lhs);
				} else {
//					System.out.println("NOT ADDING TO ORDER: "+lhs);
				}
			} else if (expandedLhsToRealLhsMap.containsKey(lhs)) {
				if (requiredVariables.contains(expandedLhsToRealLhsMap.get(lhs))) {
					if (!aiOrder.contains(lhs)) {
//						System.out.println("ADDING TO ORDER(2): "+lhs);
						aiOrder.add(lhs);
					} else {
//						System.out.println("NOT ADDING TO ORDER: "+lhs);
					}
				}
			} else if (realLhsToExpandedLhsMap.containsKey(lhs)) {
				for (String s : realLhsToExpandedLhsMap.get(lhs)) {
					if (requiredVariables.contains(s)) {
						if (!aiOrder.contains(lhs)) {
//							System.out.println("ADDING TO ORDER(3): "+lhs);
							aiOrder.add(lhs);
						} else {
//							System.out.println("NOT ADDING TO ORDER: "+lhs);
						}
					}
				}
			} else {
//				System.out.println("NOT ADDING TO ORDER (else): "+lhs);
			}
		}
		
		for (Equation eqn : equationsWithInitialValue) {
			aiOrder.add(eqn.getLhs());
		}
		return aiOrder;
	}
	
	public List<String> getInitialValueVariables() {
		List<String> al = new ArrayList<String>();
		for (Equation eqn : equationsWithInitialValue) {
			al.addAll(eqn.getFunctionInitialVariables());
			
		}
		
		return al;
	}
	
	public void modifyStockInitialValues() {
		for (Equation eqn : equationsWithInitialValue) {
			// for now, assume that the function with initial value
			// is the first thing on RHS
			Node function = TreeTraversal.getRhs(eqn.getTreeRoot());
			Node last = TreeTraversal.getLastArgument(function);
			
			// make some mods to this node
			last.setToken(makeInitialVariable(eqn.getLhs()));
			
			// make it terminal
			last.setChild(null);
		}
	}
	
	private List<String> getRequiredVariables() {
		List<String> allDepends = new ArrayList<String>();
		for (Equation eq  : equationsWithInitialValue) {
//			System.out.println("\n>>> Req LHS: "+eq.getLhs());
				Set<String> iv = eq.getFunctionInitialVariables();
				for (String v : iv) {
//					System.out.println("### "+v+" ###");
					if (Parser.isNumber(v)) {
						System.out.println("  Req isNumber: "+v);
					} else {
						String orig = InformationManagers.getInstance().getNativeDataTypeManager().getOriginalName(v);
						if (!allDepends.contains(orig)) {
//							System.out.println("  Req orig added: "+v);
							allDepends.add(orig);
						} else {
//							System.out.println("  Req orig NOT added: "+v);
						}
						for (String s : variableDependencyProcessor.getDependencyChain(v)) {
							if (!allDepends.contains(s)) {
//								System.out.println("     Req chain added: "+v);
								allDepends.add(s);
							} else {
//								System.out.println("     Req chain NOT added: "+v);
							}
						}
					}
				}
		}
		
		return allDepends;

	}
	
	public void print() {
	
		
		System.out.println("### equationsWithInitialValue ###");
		for (Equation eqn : equationsWithInitialValue) {
			System.out.println(eqn.getVensimEquationOnly());
		}
		
		System.out.println("### variablesToDeclare ###");
		for (String v : variablesToDeclare) {
			System.out.println(v);
		}
		
		System.out.println("### requiredVariables ###");
		for (String v : requiredVariables) {
			System.out.println(v);
		}
		
		System.out.println("### initializationOrder ###");
		for (String v : initializationOrder) {
			System.out.println(v);
		}
		
		System.out.println("### Code Generation ###");
		
	}
	
	
	public void write(CodeGenerator generator, BufferedWriter bw) {
//		print();
		this.bw = bw;
		writeMethodHeader();
		writeMethodBody(generator);
		
		// now that we've processed the initial value variables
		// modify the code trees to reference these
		
		modifyStockInitialValues();
		
		writeMethodTail();
	}
	
	/**
	 * Returns a list of all stock variables
	 * @return
	 */
	public List<Equation> getEquationsWithInitialValue() {
		List<Equation> al = new ArrayList<Equation>();
		for(String lhs : equations.keySet()) {
			Equation eqn = equations.get(lhs);
			if (eqn.isHasInitialValue())
				al.add(eqn);
		}
		return al;
		
	}
	
	
//	/**
//	 * Returns a list of required aux variables
//	 * @return
//	 */
//	public List<String> getRequiredVariables() {
//		List<String> al = new ArrayList<String>();
//		for(Equation eqn : equationsWithInitialValue) {
//			al.addAll(eqn.getRHSVariables());
//		}
//		return al;
//		
//	}
	
	public void writeMethodBody(CodeGenerator generator) {
		try {
			bw.append("\n/* method body */\n");

			for (String lhs : initializationOrder) {
				
//				System.out.println("wMB: "+lhs);
				
				String realLHS = lhs.replace(INITIAL_VALUE_MARKING, "");
				if (ArrayReference.isArrayReference(realLHS)) {
					realLHS = realLHS.replace("array.", "");
				} else {
					realLHS = realLHS.replace("memory.", "");
				}
				Equation eqn = null;

				eqn = equations.get(realLHS);
				if (eqn == null) {
					eqn = equations.get("array."+realLHS);
				}
				if ( eqn == null) {
					System.out.println("No equation for <"+realLHS+">");
				} else {
//					if (!eqn.isOneTime()) {
//					System.out.println("Gen AI: "+eqn.getVensimEquationOnly());
					// is this equation one for which we are creating a new variable?
					// if not in data structure, just generate as is.
					
						if (!equationsWithInitialValue.contains(eqn)) {
							if (!eqn.isOneTime())
								generator.generateCode(eqn, bw, CodeGenerator.DEBUG_VALUES);
						} else {
							String effectiveLHS = ActiveInitialProcessor.makeInitialVariable(lhs);
							Node functionNode = TreeTraversal.getRhs(eqn.getTreeRoot());
							Node lastArgNode = TreeTraversal.getLastArgument(functionNode);
//							System.out.println("GenRHS: <"+eqn.generateExpression(lastArgNode)+">");
							String generatedEquation;
							if (targetLanguage.equals("Java")) 
							generatedEquation = effectiveLHS+" = "+eqn.generateExpression(lastArgNode);
							else
								generatedEquation = CodeGenerator.scrubC(effectiveLHS+" = "+eqn.generateExpression(lastArgNode));
							
							new Equation(new EquationComponents(generatedEquation+
							"~no units needed~generated for initialization")).writeCode(generator, bw);
						}
				}

			}
		} catch (IOException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
	}
	
	public void writeMethodHeader() {
		try {
			if (targetLanguage.equals("Java")) 
				bw.append("protected ");
			bw.append("void initialize() {\n");
			bw.append("double timeStep = getTIMESTEP();\n");
			bw.append("double time = getINITIALTIME();\n");
			
		} catch (IOException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
	}
	
	public void writeMethodTail() {
		try {
			bw.append("};\n");
		} catch (IOException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
	}

	public BufferedWriter getBw() {
		return bw;
	}

	public void setBw(BufferedWriter bw) {
		this.bw = bw;
	}

	public List<String> getVariablesToDeclare() {
		if (variablesToDeclare == null) {
			variablesToDeclare = new ArrayList<String>();
			for (Equation eqn : this.equationsWithInitialValue) {
				String lhs = eqn.getLhs();
				if (ArrayReference.isArrayReference(lhs)) {
					ArrayReference ar = new ArrayReference(lhs);
					variablesToDeclare.add(ar.getArrayName());
				} else {
					variablesToDeclare.add(lhs);
				}
			}
		}
		return variablesToDeclare;
	}

	public List<String> getInitializationOrder() {
		return initializationOrder;
	}
}
