package gov.anl.gss.sdm.server.model.translator;

import gov.anl.gss.sdm.server.model.support.MutableBoolean;

import java.util.LinkedHashMap;
import java.util.Map;

import org.apache.log4j.Logger;

public class EquationCharacteristics {
static Logger log = Logger.getLogger(EquationCharacteristics.class.getSimpleName());
	
	private static final String LEVEL = "L";
	private static final String SMOOTH = "SM";
	private static final String DELAY = "DE";
	private static final String LEVEL_INITIALIZATION = "LI";
	private static final String INITIAL = "I";
	private static final String CONSTANT = "C";
	private static final String FLOW = "F";
	private static final String AUXILIARY = "A";
	private static final String SUBSCRIPTS = "Sub";
	private static final String DATA = "D";
	private static final String GAME = "G";
	private static final String LOOKUP = "T";
	
	private Equation equation;
	private int variableNumber;
	private Map<String, Boolean> characteristics;

	private boolean level = false;
	private boolean smooth = false;
	private boolean delay = false;
	private boolean levelInitialization = false;
	private boolean initial = false;
	private boolean constant = false;
	private boolean flow = false;
	private boolean auxiliary = false;
	private boolean subscripts = false;
	private boolean data = false;
	private boolean game = false;
	private boolean lookup = false;
	
	private boolean valueRangeSpecified = false;
	
	private boolean undocumented = false;
	private boolean embeddedDataNot01 = false;
	private boolean unitErrorsWarnings = false;
	private boolean notInView = false;
	private boolean incompletelyDefinedSubscript = false;
	private boolean nonmonotonicLookup = false;
	private boolean cascadingLookup = false;
	private boolean nonzeroEndSlopeLookup = false;
	private boolean ifThenElse = false;
	private boolean minMax = false;
	private boolean stepPulse = false;
	private boolean embeddedData01 = false;
	
	private boolean unused = false;
	private boolean supplementary = false;
	private boolean supplementaryUsed = false;
	private boolean complexFormulation = false;
	private boolean complexStock = false;
	
	private MutableBoolean nonZeroLeft = new MutableBoolean(false);
	private MutableBoolean nonZeroRight = new MutableBoolean(false);
	
	private int complexity = 0;
	
	private StringBuilder typeStringBuilder = null;
	
	public EquationCharacteristics() {
		characteristics = new LinkedHashMap<String, Boolean>();
	}
	
	public EquationCharacteristics(Equation equation) {
		this();
		this.equation = equation;
		this.variableNumber = equation.getEquationNumber();
		extractCharacteristics();
	}
	
	public String getType() {
		if (typeStringBuilder == null) {
			typeStringBuilder = new StringBuilder();
			validateType();
			addType(level, LEVEL);
			addType(smooth, SMOOTH);
			addType(delay, DELAY);
			addType(levelInitialization, LEVEL_INITIALIZATION);
			addType(initial, INITIAL);
			addType(constant, CONSTANT);
			addType(flow, FLOW);
			addType(auxiliary, AUXILIARY);
			addType(subscripts, SUBSCRIPTS);
			addType(data, DATA);
			addType(game, GAME);
			addType(lookup, LOOKUP);
		}
		
		if (typeStringBuilder.length() == 0) {
			log.debug("### Can't determine type "+equation.getVensimEquationOnly());
		}

		return typeStringBuilder.toString();
	}
	
	private void validateType() {
		if (level || constant || initial || subscripts || game /*|| smooth*/) {
			auxiliary = false;
			equation.setAuxiliary(false);
		}
	}
	
	private void addType(boolean bval, String marker) {
		if (!bval)
			return;
		if (typeStringBuilder.length() > 0)
			typeStringBuilder.append(",");
		typeStringBuilder.append(marker);
		
	}
	
	private void extractCharacteristics() {
		setLevel(equation.isStock());
		setSmooth(equation.isUsesSmooth());
		setDelay(equation.isUsesDelay());
		setLevelInitialization(equation.isLevelInitialization());
		setInitial(equation.isInitial());
		setConstant(equation.isConstant());
		setSubscripts(equation.isSubscripts());
		setFlow(equation.isFlow());
		setAuxiliary(equation.isAuxiliary());
		setData(equation.isDataByMeta());
		setGame(equation.isGame());
		setLookup(equation.isLookup());
		setUndocumented(equation.isUndocumented());
		setEmbeddedDataNot01(equation.isEmbeddedDataNot01());
		setUnitErrorsWarnings(equation.isUnitErrorsWarnings());
		setNotInView(equation.isNotInView());
		setIncompletelyDefinedSubscript(equation.isIncompletelyDefinedSubscript());
		setNonmonotonicLookup(equation.isNonmonotonicLookup());
		setCascadingLookup(equation.isCascadingLookup());
		setNonzeroEndSlopeLookup(equation.isNonzeroEndSlopeLookup(nonZeroLeft, nonZeroRight));
		setIfThenElse(equation.isUsesIfThenElse());
		setMinMax(equation.isUsesMinMax());
		setStepPulse(equation.isUsesStepPulse());
		setUnused(equation.isUnused());
		setSupplementary(equation.isSupplementary());
		setSupplementaryUsed(equation.isSupplementaryUsed());
		setComplexFormulation(equation.isComplexFormulation());
		setComplexStock(equation.isComplexStock());
		setValueRangeSpecified(equation.isValueRangeSpecified());
		
	}
	
	public boolean getValue(String characteristic) {
			return characteristics.get(characteristic);
	}

	public Equation getEquation() {
		return equation;
	}

	public void setEquation(Equation equation) {
		this.equation = equation;
	}

	public boolean isLevel() {
		return level;
	}

	public void setLevel(boolean level) {
		this.level = level;
		characteristics.put("Level", level);
	}

	public boolean isSmooth() {
		return smooth;
	}

	public void setSmooth(boolean smooth) {
		this.smooth = smooth;
		characteristics.put("Smooth", smooth);
	}

	public boolean isDelay() {
		return delay;
	}

	public void setDelay(boolean delay) {
		this.delay = delay;
		characteristics.put("Delay", delay);
	}

	public boolean isLevelInitialization() {
		return levelInitialization;
	}

	public void setLevelInitialization(boolean levelInitialization) {
		this.levelInitialization = levelInitialization;
		characteristics.put("LevelInitialization", levelInitialization);
	}

	public boolean isInitial() {
		return initial;
	}

	public void setInitial(boolean initial) {
		this.initial = initial;
		characteristics.put("Initial", initial);
	}

	public boolean isConstant() {
		return constant;
		
	}

	public void setConstant(boolean constant) {
		this.constant = constant;
		characteristics.put("Constant", constant);
	}

	public boolean isFlow() {
		return flow;
	}

	public void setFlow(boolean flow) {
		this.flow = flow;
		characteristics.put("Flow", flow);
	}

	public boolean isAuxiliary() {
		return auxiliary;
	}

	public void setAuxiliary(boolean auxiliary) {
		this.auxiliary = auxiliary;
		characteristics.put("Auxiliary", auxiliary);
	}

	public boolean isSubscripts() {
		return subscripts;
	}

	public void setSubscripts(boolean subscripts) {
		this.subscripts = subscripts;
		characteristics.put("Subscripts", subscripts);
	}

	public boolean isData() {
		return data;
	}

	public void setData(boolean data) {
		this.data = data;
		characteristics.put("Data", data);
	}

	public boolean isGame() {
		return game;
	}

	public void setGame(boolean game) {
		this.game = game;
		characteristics.put("Game", game);
	}

	public boolean isLookup() {
		return lookup;
	}

	public void setLookup(boolean lookup) {
		this.lookup = lookup;
		characteristics.put("Lookup", lookup);
	}

	public boolean isUndocumented() {
		return undocumented;
	}

	public void setUndocumented(boolean undocumented) {
		this.undocumented = undocumented;
		characteristics.put("Undocumented", undocumented);
	}

	public boolean isEmbeddedDataNot01() {
		return embeddedDataNot01;
	}

	public void setEmbeddedDataNot01(boolean embeddedDataNot01) {
		this.embeddedDataNot01 = embeddedDataNot01;
		characteristics.put("EmbeddedDataNot01", embeddedDataNot01);
	}

	public boolean isUnitErrorsWarnings() {
		return unitErrorsWarnings;
	}

	public void setUnitErrorsWarnings(boolean unitErrorsWarnings) {
		this.unitErrorsWarnings = unitErrorsWarnings;
		characteristics.put("UnitErrorsWarnings", unitErrorsWarnings);
	}

	public boolean isNotInView() {
		return notInView;
	}

	public void setNotInView(boolean notInView) {
		this.notInView = notInView;
		characteristics.put("NotInView", notInView);
	}

	public boolean isIncompletelyDefinedSubscript() {
		return incompletelyDefinedSubscript;
	}

	public void setIncompletelyDefinedSubscript(boolean incompletelyDefinedSubscript) {
		this.incompletelyDefinedSubscript = incompletelyDefinedSubscript;
		characteristics.put("IncompletelyDefinedSubscript", incompletelyDefinedSubscript);
	}

	public boolean isNonmonotonicLookup() {
		return nonmonotonicLookup;
	}

	public void setNonmonotonicLookup(boolean nonmonotonicLookup) {
		this.nonmonotonicLookup = nonmonotonicLookup;
		characteristics.put("NonmonotonicLookup", nonmonotonicLookup);
	}

	public boolean isCascadingLookup() {
		return cascadingLookup;
	}

	public void setCascadingLookup(boolean cascadingLookup) {
		this.cascadingLookup = cascadingLookup;
		characteristics.put("CascadingLookup", cascadingLookup);
	}

	public boolean isNonzeroEndSlopeLookup() {
		return nonzeroEndSlopeLookup;
	}

	public void setNonzeroEndSlopeLookup(boolean nonzeroEndSlopeLookup) {
		this.nonzeroEndSlopeLookup = nonzeroEndSlopeLookup;
		characteristics.put("NonzeroEndSlopeLookup", nonzeroEndSlopeLookup);
	}

	public boolean isIfThenElse() {
		return ifThenElse;
	}

	public void setIfThenElse(boolean ifThenElse) {
		this.ifThenElse = ifThenElse;
		characteristics.put("IfThenElse", ifThenElse);
	}

	public boolean isMinMax() {
		return minMax;
	}

	public void setMinMax(boolean minMax) {
		this.minMax = minMax;
		characteristics.put("MinMax", minMax);
	}

	public boolean isStepPulse() {
		return stepPulse;
	}

	public void setStepPulse(boolean stepPulse) {
		this.stepPulse = stepPulse;
		characteristics.put("StepPulse", stepPulse);
	}

	public boolean isEmbeddedData01() {
		return embeddedData01;
	}

	public void setEmbeddedData01(boolean embeddedData01) {
		this.embeddedData01 = embeddedData01;
		characteristics.put("EmbeddedData01", embeddedData01);
	}

	public boolean isUnused() {
		return unused;
	}

	public void setUnused(boolean unused) {
		this.unused = unused;
		characteristics.put("Unused", unused);
	}

	public boolean isSupplementary() {
		return supplementary;
	}

	public void setSupplementary(boolean supplementary) {
		this.supplementary = supplementary;
		characteristics.put("Supplementary", supplementary);
	}

	public boolean isSupplementaryUsed() {
		return supplementaryUsed;
	}

	public void setSupplementaryUsed(boolean supplementaryUsed) {
		this.supplementaryUsed = supplementaryUsed;
		characteristics.put("SupplementaryUsed", supplementaryUsed);
	}

	public boolean isComplexFormulation() {
		return complexFormulation;
	}

	public void setComplexFormulation(boolean complexFormulation) {
		this.complexFormulation = complexFormulation;
		characteristics.put("ComplexFormulation", complexFormulation);
	}

	public boolean isComplexStock() {
		return complexStock;
	}

	public void setComplexStock(boolean complexStock) {
		this.complexStock = complexStock;
		characteristics.put("ComplexStock", complexStock);
	}

	public int getVariableNumber() {
		return variableNumber;
	}

	public void setVariableNumber(int variableNumber) {
		this.variableNumber = variableNumber;
	}

	public int getComplexity() {
		return complexity;
	}

	public void setComplexity(int complexity) {
		this.complexity = complexity;
	}

	public MutableBoolean getNonZeroLeft() {
		return nonZeroLeft;
	}

	public MutableBoolean getNonZeroRight() {
		return nonZeroRight;
	}

	public boolean isValueRangeSpecified() {
		return valueRangeSpecified;
	}

	public void setValueRangeSpecified(boolean valueRangeSpecified) {
		this.valueRangeSpecified = valueRangeSpecified;
	}
	
}
