package gov.anl.gss.sdm.server.model.translator;

import gov.anl.gss.sdm.server.model.documentor.MessageGobbler;
import gov.anl.gss.sdm.server.model.documentor.docx.Table;
import gov.anl.gss.sdm.server.model.documentor.docx.reader.DocxReader;
import gov.anl.gss.sdm.server.model.documentor.docx.writer.ReportWriter;
import gov.anl.gss.sdm.server.model.documentor.docx.writer.Reportable;
import gov.anl.gss.sdm.server.model.documentor.xmlwriter.StringConstants;
import gov.anl.gss.sdm.server.model.documentor.xmlwriter.UserSelectedOptions;
import gov.anl.gss.sdm.server.model.support.ArrayReference;
import gov.anl.gss.sdm.server.model.support.Utilities;
import gov.anl.gss.sdm.server.model.support.VensimFormatter;
import gov.anl.gss.sdm.server.model.support.Verify;
import gov.anl.gss.sdm.server.model.tools.MDLProcessor;
import gov.anl.gss.sdm.server.model.tools.SDModel;
import gov.anl.gss.sdm.server.model.xmile.XmileReader;

import java.io.File;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.LinkedHashMap;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.apache.log4j.Logger;

public class ToolTextLanguageTranslator {

	static Logger log = Logger.getLogger(SDM_Doc.class.getSimpleName());
	
	private static final String[] NOT_TO_BE_TRANSLATED = {"TIME STEP", "SAVEPER", "NAREPLACEMENT" , "INITIAL TIME", "FINAL TIME", "TIME"};
	private Set<String> notTranslated = new LinkedHashSet<String>();
	
	private final static String START_TAG = "\t~";
	private final static String EQN_END_TAG = "\t|";
	private final static String EQN_END_TAG2 = START_TAG+"\t"+EQN_END_TAG;
	private final static String GRAPHICS_START = "\\\\\\---/// Sketch information";
	private final static String MACRO_START = ":MACRO:";
	private final static String MACRO_END = ":END OF MACRO:";
	private final static String STARS = "******";
	
	private final static int MAX_COMMENT_LENGTH = 80; // the point at which we continue a single line comment
	
	private final static String SUPPLEMENTARY = ":SUPPLEMENTARY";
	
	private final static String COMMENT_START = "\t~\t";
	
	private final static String LINE_START = "\t\t";
	
	private final static String VARIABLE = "_Variable_";
	private final static String START_COMMENT = "_CommentStart_";
	private final static String END_COMMENT = "_CommentEnd_";
	
	private Map<String, List<String>> variableCommentsMap = new LinkedHashMap<String, List<String>>();
	
	private List<String> mdlContents;

	private String docxFile;
	private String sdInputFile;
	private String sdOutputFile;
	private boolean generateDocx;
	
	private MDLProcessor mdlProcessor;
	private SDModel sdModel;
	
	private boolean xmileOrigin;
	
	private boolean initialize = false;
	
	private NativeDataTypeManager ndtm;
	
	public ToolTextLanguageTranslator() {
		ndtm = InformationManagers.getInstance().getNativeDataTypeManager();
	}

	public ToolTextLanguageTranslator(String mode, String file1, String file2, boolean initialize) {
			this();
			generateDocx = true;
			sdInputFile = file1;
			docxFile = file2;
			this.initialize = initialize;
	}
	
	public ToolTextLanguageTranslator(String mode, String file1, String file2, String file3) {
			this();
			generateDocx = false;
			sdInputFile = file1;
			docxFile = file2;
			sdOutputFile = file3;
	}


	public boolean execute() {
		
		for (String s : NOT_TO_BE_TRANSLATED)
			notTranslated.add(s);

		initialize();

		if (generateDocx)
			return executeGenerateDocx();
		else
			return executeGenerateVensim();
	}

	private boolean executeGenerateDocx() {

		return generateDocx();

	}

	private boolean executeGenerateVensim() {
		
		return generateVensim();
	}

	private void initialize() {
		
		MessageGobbler mf = new MessageGobbler();
		
		String originalModelFile = sdInputFile;
		String modelFile =sdInputFile;
		
		xmileOrigin = false;
		
//		if (Utilities.stringEndsWith(modelFile, "\\.[xX][mM][iI][lL][eE]$")) {
		if (Utilities.isOfXmlOrigin(modelFile)) {
			
			xmileOrigin = true;
			XmileReader xmileReader = new XmileReader(modelFile);
			xmileReader.readAndTranslate();
		}
		
//		modelFile = modelFile.replaceAll("\\.[xX][mM][iI][lL][eE]$",".mdl");
		modelFile = Utilities.getConvertedFileName(modelFile);
		
		MessageGobbler ef = new MessageGobbler();

		String obj = Utilities.makeLegalJavaName(new File(modelFile).getName().replace(".MDL", "").replace(".mdl", "").toUpperCase());
		
		String root = new File(sdInputFile).getParent();

		mdlProcessor = new MDLProcessor(originalModelFile, modelFile, obj, root, mf, ef, 
				new UserSelectedOptions(true));
		mdlProcessor.setVariableNameTranslationMode(true);
		mdlProcessor.process(false);
		sdModel = mdlProcessor.getSDModel();
		
//		printGraphics();

	}
	
	private List<String> getAllVariables() {

		Map<String, Equation> equations = sdModel.getEquations();
		List<String> allLhs = new ArrayList<String>();
		List<String> allSub = new ArrayList<String>();
		for (String lhs : equations.keySet()) {
			if (equations.get(lhs).isSubscripts()) {
				allSub.add("{Subscript} "+lhs);
				for (String v : InformationManagers.getInstance().getNamedSubscriptManager().getValuesFor(lhs)) {
					allSub.add("{Sub Value} "+v);
				}
			} else {
				if (!notTranslated.contains(lhs.toUpperCase()))
					allLhs.add("{Variable} "+lhs);
			}
		}
		// need to add comments from graphics objects
		Map<String, View> graphics = sdModel.getGraphics();
		for (View view : graphics.values()) {
			for (GraphicObject go : view.getGraphicObjects()) {
				if (go.isComment() && go.getAdditionalText() != null) {
					if (!notTranslated.contains(go.getAdditionalText().toUpperCase())) {
						allLhs.add("{Comment} "+go.getAdditionalText());
						System.out.println("TT GO: "+go.getAdditionalText());
					}
				}
			}
		}

		allLhs = sortVariableNames(allLhs);
		if (allSub.size() > 0)
			allLhs.addAll(allSub);
		return allLhs;
		
		

	}
	
	protected boolean generateDocx() {
		
		List<Reportable> reportables = new ArrayList<Reportable>();
		List<List<String>> tableData = new ArrayList<List<String>>();
		
		
		
		Set<String> patternsWrittenToDocx = new LinkedHashSet<String>();
		for (String lhs : getAllVariables()) {
			List<String> al = new ArrayList<String>();
			
			String orig = null;
			String trans = null;
			
			if (ArrayReference.isArrayReference(lhs)) {
				orig = new ArrayReference(lhs).getArrayName().replace(StringConstants.RSSD_LOOKUP, "");
				trans = removeAnnotation(new ArrayReference(lhs).getArrayName());
				if (!patternsWrittenToDocx.contains(orig)) {
					al.add(orig);
					String pat = InformationManagers.getInstance().getVariableNamePatternGenerator().getPatternFor(removeAnnotation(orig));
					if (initialize)
						al.add(trans);
					else
						al.add("");
				}
				
			} else {
				orig = lhs.replace(StringConstants.RSSD_LOOKUP, "");
				trans = removeAnnotation(lhs);
				if (!patternsWrittenToDocx.contains(orig)) {
					al.add(orig);
					String pat = InformationManagers.getInstance().getVariableNamePatternGenerator().getPatternFor(removeAnnotation(orig));
					if (initialize)
						al.add(trans);
					else
						al.add("");
				}
			}
			patternsWrittenToDocx.add(orig);
//			al.add("");
			if (al.size() > 0)
				tableData.add(al);
		}
		String title = "Title";
		String[] columnHeaders = {"Original", "Replacement"};
		reportables.add(new Table(title, columnHeaders, tableData));
		ReportWriter rw = new ReportWriter(reportables, docxFile, false);
		rw.generateDocx();

	
		return true;
	}
	
	private String removeAnnotation(String var) {
		return var.replace(StringConstants.RSSD_LOOKUP, "").replace("{Subscript} ", "").replace("{Sub Value} ", "").
				replace("{Variable} ", "").replace("{Comment} ", "");
	}
	
	
	private List<String> sortVariableNames(List<String> vars) {
		
		Map<String, String> nonQuotedOrigMap = new LinkedHashMap<String, String>();
		Map<String, String> origNonQuotedMap = new LinkedHashMap<String, String>();
		
		for (String var : vars) {
			String noQuotes =  Parser.removeQuotes(var); // var.replace("\"", "");
			nonQuotedOrigMap.put(noQuotes.toUpperCase(), var);
			origNonQuotedMap.put(var, noQuotes.toUpperCase());
		}
		
		List<String> sortedVars = new ArrayList<String>(nonQuotedOrigMap.keySet());
		
		List<String> sorted = new ArrayList<String>();
		
		Collections.sort(sortedVars);
		
		for (String v : sortedVars) {
			sorted.add(nonQuotedOrigMap.get(v));
		}
		
		
		return sorted;
	}
	
	protected boolean generateVensim() {
		// get variable names and comments
		// replace existing comments
		// write vensim files
		
		DocxReader rdr = new DocxReader(docxFile);
		List<String> txt = rdr.getDocText();
		Map<String, String> translation = new HashMap<String, String>();
		
		List<List<List<String>>> tables = convertToTables(txt);
		for (List<List<String>> table : tables) {
			for (List<String> row : table) {
				String orig = InformationManagers.getInstance().getVariableNamePatternGenerator().getPatternFor(removeAnnotation(row.get(0)));
				String trans = row.get(1);
				translation.put(orig, trans);
				System.out.println(orig+" -> "+trans);
			}
		}
		
		Map<String, String> groupText = sdModel.getEquationProcessor().getGroupText();
		
		Map<String, Set<Equation>> groupEqn = new HashMap<String, Set<Equation>>();
		for (String group : groupText.keySet()) {
			groupEqn.put(group, new LinkedHashSet<Equation>());
		}
		for (Equation eqn : sdModel.getEquations().values()) {
			/*
			 * Normally, we would not include any autogenerated equations but the WITH LOOKUP equation
			 * has caused a complication since it is split into two equations...
			 */
			
//			if (eqn.isAutoGenerated()) {
//				continue;
//			}
			
			// first see if it is explicitly marked as not to be included, do not include it
			// doesn't matter if autogen or not
				if (!eqn.isIncludeInTranslation()) {
					continue;
				}
				
//			if (eqn.isIncludeInTranslation()) {
//				
//			}
				
		
			String group = eqn.getGroup();
			if (groupEqn.containsKey(group)) {
//				groupEqn.get(group).add(eqn);
			} else {
				groupText.put(group, "********************************************************"+group+"********************************************************~Autogenerated");
				groupEqn.put(group, new LinkedHashSet<Equation>());
				if (!sdModel.getEquationProcessor().getGroupOrder().contains(group)) {
//					sdModel.getEquationProcessor().getGroupOrder().add(group);
					sdModel.getEquationProcessor().addToHeadOfGroupOrder(group);
				}
				
			}
			groupEqn.get(group).add(eqn);
		}
		
		List<String> vensimLines = new ArrayList<String>();
		
		Set<Equation> equationsWrittenToVensim = new LinkedHashSet<Equation>();
		
		
		vensimLines.add("{UTF-8}");
		for (String group : sdModel.getEquationProcessor().getGroupOrder()) {
			vensimLines.add("********************************************************");
			vensimLines.add(group);
			vensimLines.add("********************************************************~");
			if (!groupText.containsKey(group)) {
				vensimLines.add(VensimFormatter.format("\t"+"Why am i missing group text?"+"\n\t|"));
			} else {
				String[] parts = groupText.get(group).split("~");
				if (parts.length > 1) {
					vensimLines.add(VensimFormatter.format("\t"+parts[1])+"\n\t|"); 
				} else {
					vensimLines.add(VensimFormatter.format("\n\t|"));
				}
			}
			for (Equation eqn : groupEqn.get(group)) {
				if (eqn == null || equationsWrittenToVensim.contains(eqn))
					continue;
				if (InformationManagers.getInstance().getEquationManager().getNumberOfEquationsFor(eqn.getLhs()) == 1) {
					List<String> translated = applyTranslation(eqn, translation);
					String mdlString = asString(translated)+"~"+eqn.getOriginalUnits()+"~"+
					(eqn.getComment() == null ? "" : eqn.getComment())+
					(eqn.isSupplementary() ? "\t~\t:SUPPLEMENTARY" : "" )+
					"|";
					vensimLines.add(VensimFormatter.format(mdlString));
					
				} else {
					int numEqn = InformationManagers.getInstance().getEquationManager().getNumberOfEquationsFor(eqn.getLhs());
					int eqnNum = 1;
					for (Equation e : InformationManagers.getInstance().getEquationManager().getAllEquationsFor(eqn.getLhs())) {
						List<String> translated = applyTranslation(e, translation);
						String mdlString = asString(translated)+(eqnNum++ == numEqn ? "" : "~~|");
						vensimLines.add(VensimFormatter.formatMulti(mdlString));
						equationsWrittenToVensim.add(e);
					}
					vensimLines.add("\t~\t"+eqn.getOriginalUnits()+"\n\t~\t"+(eqn.getComment() == null ? "" : eqn.getComment())+
							(eqn.isSupplementary() ? "\t~\t:SUPPLEMENTARY" : "" )+
							"\t|");
					
				}
				equationsWrittenToVensim.add(eqn);
			}
		}
		vensimLines.add(GRAPHICS_START);
		vensimLines.add("V300  Do not put anything below this section - it will be ignored");

		Map<String, View> graphics = sdModel.getGraphics();
		int v = 0;
		int numV = graphics.size();
		for (View view : graphics.values()) {
			v++;
			vensimLines.add("*"+view.getName());
			vensimLines.add(view.getViewDefaultFont());
			for (GraphicObject go : view.getGraphicObjects()) {
				if (go.isAutoGenerated())
					continue;
				List<String> translated = applyTranslation(go, translation);
				String mdlString = concat(translated);
				vensimLines.add(mdlString);
				if (go.getAdditionalText() != null && go.isComment()) {
					vensimLines.add(go.getAdditionalText());
				}
			}
			if (v < numV)
				vensimLines.add(GRAPHICS_START);
		}
		vensimLines.add("///---\\\\\\");
		for (String xtra : sdModel.getGraphicsProcessor().getExtraStuff()) {
			vensimLines.add(xtra);
		}
		
		Utilities.writeToFileUTF8(sdOutputFile, vensimLines);
		
		// ###########################################################
		
		System.out.println("{UTF-8}");
		for (String group : sdModel.getEquationProcessor().getGroupOrder()) {
			System.out.println("********************************************************");
			System.out.println(group);
			System.out.println("********************************************************~");
			if (!groupText.containsKey(group)) {
				System.out.println(VensimFormatter.format("\t"+"Why am i missing group text?"+"\n\t|"));
			} else {
				String[] parts = groupText.get(group).split("~");
				if (parts.length > 1) {
					System.out.println(VensimFormatter.format("\t"+parts[1])+"\n\t|"); 
				} else {
					System.out.println(VensimFormatter.format("\n\t|"));
				}
			}
			for (Equation eqn : groupEqn.get(group)) {
				//		}
				//		
				//		for (String var : getAllVariables()) {
				//			Equation eqn = sdModel.getEquations().get(var);
				if (eqn == null)
					continue;
				List<String> translated = applyTranslation(eqn, translation);
				//			System.out.println("### "+var);
				//			System.out.println(Parser.tokenize(eqn.getVensimEquationOnly()));
				//			System.out.println(translated);
				//			System.out.println("\n");
				String mdlString = asString(translated)+"~"+eqn.getOriginalUnits()+"~"+eqn.getComment()+"|";
				System.out.println(VensimFormatter.format(mdlString));
				//			System.out.println("\n");
			}
		}
		System.out.println(GRAPHICS_START);
		System.out.println("V300  Do not put anything below this section - it will be ignored");

//		Map<String, View> graphics = sdModel.getGraphics();
		for (View view : graphics.values()) {
			System.out.println("*"+view.getName());
			System.out.println(view.getViewDefaultFont());
			for (GraphicObject go : view.getGraphicObjects()) {
				List<String> translated = applyTranslation(go, translation);
				if (go.isAutoGenerated())
					continue;

				String mdlString = concat(translated);
				System.out.println(mdlString);

				if (go.getAdditionalText() != null) {
					System.out.println(go.getAdditionalText());
				}
			}
			
			System.out.println("///---\\\\\\");
		}
		
		for (String xtra : sdModel.getGraphicsProcessor().getExtraStuff()) {
			System.out.println(xtra);
		}
		
//		print(newMdl);
//		
//		Utilities.writeToFileUTF8(sdOutputFile, newMdl);
		
		return true;
	}
	
	private String asString(List<String> aList) {
		StringBuilder sb = new StringBuilder();
		
		int i = 0;
		for (String s : aList) {
			
//			if (i++ > 0)
//				sb.append(" ");
			sb.append(s);
			
		}
		
		return sb.toString();
	}
	
	private String concat(List<String> aList) {
		StringBuilder sb = new StringBuilder();
		
		for (String s : aList) {
			sb.append(s);
		}
		
		return sb.toString();
	}
	
	private List<String> applyTranslation(Equation eqn, Map<String, String> translation) {
		
		// no quotes have been used in the docx translation file
		
		List<String> orig = Parser.tokenizeMaintainSingleQuotes(eqn.getOriginalVensimEquationOnly());
//		List<String> orig = Parser.tokenize(eqn.getOriginalVensimEquationOnly());
		List<String> translated = new ArrayList<String>();
		for (String o : orig) {
			String pattern = InformationManagers.getInstance().getVariableNamePatternGenerator().getPatternFor(o);
			if (translation.containsKey(pattern)) {
				translated.add(ndtm.quoteIfNecessary(o, translation.get(pattern)));
			} else {
				translated.add(ndtm.quoteIfNecessary(o, o));
			}
		}
		return translated;
	}
	
	private List<String> applyTranslation(GraphicObject go, Map<String, String> translation) {
		List<String> orig = Parser.tokenize(go.getRawObject());
		List<String> translated = new ArrayList<String>();
		for (String o : orig) {
			String pattern = InformationManagers.getInstance().getVariableNamePatternGenerator().getPatternFor(o);
			if (translation.containsKey(pattern)) {
				translated.add(ndtm.quoteIfNecessary(o, translation.get(pattern)));
			} else {
				translated.add(ndtm.quoteIfNecessary(o, o));
			}
		}
		if (go.getAdditionalText() != null) {
			String pattern = InformationManagers.getInstance().getVariableNamePatternGenerator().getPatternFor(go.getAdditionalText());
			if (translation.containsKey(pattern)) {
				go.setAdditionalText(translation.get(pattern));
			}
		}
		return translated;
	}
	
	private List<List<List<String>>> convertToTables(List<String> docText) {
		
		List<List<List<String>>> allTables = new ArrayList<List<List<String>>>();

		int pos = 0;
		// pos always points to next value

		while (pos < docText.size()) {
			String text = docText.get(pos++);
			if (text.startsWith("START OF TABLE")) {
				int numRow = Integer.parseInt(text.split(" = ")[1]);
				int row = 0;
				
				List<List<String>> table = new ArrayList<List<String>>();
				allTables.add(table);
				
				while(row < numRow) {
					text = docText.get(pos++);
					if (text.startsWith("START OF ROW")) {
						int numColProcessed = 0;
						row++;
						int numCol = Integer.parseInt(text.split(" = ")[1]);

						// need to look ahead to see if this is a blank row
						// Note only tables with an index in the first column can be discarded
						int sumLength = 0;
						String cv = docText.get(pos).replace(">>>", "");
						if (Verify.isNumber(cv)) {
							for (int l = 1; l < numCol; l++) {
								int lookAhead = pos + l;
								String colVal = docText.get(lookAhead);
								sumLength += colVal.replace(">>>","").trim().length();
							}
						} else {
							sumLength = 10;
						}

						if (sumLength > 0) {
							List<String> aRow = new ArrayList<String>();
							table.add(aRow);
							while(numColProcessed < numCol) {
								text = docText.get(pos++);
								aRow.add(text.replace(">>>", ""));

								numColProcessed++;
							}
						} else {
							pos += numCol;
						}
					}
				}
			}

		}
		return allTables;
	}
}
