package gov.anl.gss.sdm.server.model.visualization3d;

import gov.anl.gss.sdm.server.model.analysis.graphtheory.ReachabilityMatrix;
import gov.anl.gss.sdm.server.model.analysis.loops.LoopManager;
import gov.anl.gss.sdm.server.model.support.MutableInteger;
import gov.anl.gss.sdm.server.model.translator.InformationManagers;

import java.util.ArrayList;
import java.util.LinkedHashMap;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;


public class Visualization3DData {
	
	private static final String[] distances = {"0", "1", "2", "3", "4", "5"};
	
//	int xyMult = 35;
//	int zMult = 1200;
	
	public static double SPREAD_XY = 40;  // 40, 35
	public static double SPREAD_Z = 2000;  // 2000
	
	// this is in adjusted units
	public static final double OVERLAP_DELTA = 500; // 1000
	
	
	private ConnectionMap connections;
	private Variables variables;
	private String sdmDirectory;
	private String outputDirectory;
	
	private DataManager dataManager;
	private LoopManager loopManager;
	private ReachabilityMatrix reachable;
	
	private Map<Variable, Set<Variable>> stockFlowChain = new LinkedHashMap<Variable, Set<Variable>>();
	
	private Map<Variable, VarToVar> connectivity;
	private Map<Variable, Variable> sameZ = new LinkedHashMap<Variable, Variable>();
	private Map<Variable, List<Variable>> sameZList = new LinkedHashMap<Variable, List<Variable>>();
	private Map<Variable, MutableInteger> miZ = new LinkedHashMap<Variable, MutableInteger>();
	
	private Map<Variable, Set<Variable>> varSameZ = new LinkedHashMap<Variable, Set<Variable>>();
	
	private String mapFile;
	
	private List<MutableInteger> miZList = new ArrayList<MutableInteger>();
	boolean okToProceed;
	
	public Visualization3DData(String sdmDir, String outputDir) {
		
		// we need to use the map file as a source of variable data, not the db
		
		mapFile = sdmDir.replace("_DOC_GRAPH_", ".map");
		
		// "c:/Users/bragen/Desktop/PAT/UDCJS v3aa_DOC_GRAPH_" 
		// "c:/Users/bragen/Desktop/PAT/UDCJS v3aa_3D_"
		
		this.sdmDirectory = sdmDir;
		this.outputDirectory = outputDir;
		
		dataManager = new DataManager();
		
		connections = new ConnectionMap(sdmDirectory+"/structure/tailHeadMap.tab", dataManager);
		
		variables = new Variables(dataManager, sdmDirectory+"/structure/variables.tab",  
				sdmDirectory+"/structure/locations.tab",  
				sdmDirectory+"/runs/baselineHSQLDB");
		
		
		
		okToProceed = variables.isOkToProceed();
		
		if (okToProceed) {

			dataManager.setConnections(connections);
			dataManager.setVariables(variables);
			connections.setVariables();
			
			dataManager.createClouds();
			
			dataManager.fillConnections();

			loopManager = InformationManagers.getInstance().getAnalysisManager().getLoopManager();
			reachable = loopManager.getReachability();

			connectivity = new LinkedHashMap<Variable, VarToVar>();
		}
	}
	
	public void forceDirected() {
		
		VisualizationSpace space = new VisualizationSpace();
		analyzeForceDirected(space);
		
		StockPositioner stockPos = new StockPositioner(space);
		stockPos.position();
		
		AuxiliaryPositioner auxPos = new AuxiliaryPositioner(space);
		auxPos.position();
		
		FlowPositioner flowPos = new FlowPositioner(space);
		flowPos.position();
		
//		dataManager.determineAdjustments(miZ);
//		dataManager.adjustCoordinates();
		
	}
	
	private void analyzeForceDirected(VisualizationSpace space) {
		
		Set<String> views = dataManager.getViews();
		
		Set<Variable> stocks = dataManager.getVariablesOfType(Variables.STOCK);
		Set<Variable> flows = dataManager.getVariablesOfType(Variables.FLOW);
		Set<Variable> auxiliaries = dataManager.getVariablesOfType(Variables.AUXILIARY);
		
		analyzeStocks(stocks);
		analyzeVariableSet(stocks, "STOCKS", stocks, flows, auxiliaries);
		analyzeVariableSet(flows, "FLOWS", stocks, flows, auxiliaries);
		analyzeVariableSet(auxiliaries, "AUXILIARIES", stocks, flows, auxiliaries);
		
		analyzeConnectivity();
	}
	
	private void analyzeConnectivity() {
		
		// private Map<Variable, Set<Variable>> stockFlowChain = new LinkedHashMap<Variable, Set<Variable>>();
		// force chained stocks to be on same list
		
		for (Variable v1: stockFlowChain.keySet()) {
			Set<Variable> same = stockFlowChain.get(v1);
			for (Variable v2 : same) {
				if (!v1.equals(v2)) 
					sameZ.put(v1, v2);
					sameZ.put(v2, v1);
					if (!sameZList.containsKey(v1)) {
						sameZList.put(v1, new ArrayList<Variable>());
					}
					if (!sameZList.get(v1).contains(v2))
						sameZList.get(v1).add(v2);
					
					if (!sameZList.containsKey(v2)) {
						sameZList.put(v2, new ArrayList<Variable>());
					}
					if (!sameZList.get(v2).contains(v1))
						sameZList.get(v2).add(v1);
					
			}
		}
		
		// find variables who have only one output link of length 1
		for (VarToVar v2v : connectivity.values()) {
			if (v2v.dist1() != null) {
//				System.out.println("$$$   " + v2v.getVariable().getName()+" -> "+v2v.dist1().getName());
				sameZ.put(v2v.getVariable(), v2v.dist1());
				if (!sameZList.containsKey(v2v.dist1())) {
					sameZList.put(v2v.dist1(), new ArrayList<Variable>());
				}
				sameZList.get(v2v.dist1()).add(v2v.getVariable());
			}
		}
//		System.out.println("%%% PREMERGE %%%");
		for (Variable var : sameZList.keySet()) {
//			System.out.println(var.getName()+" sameZ "+asListOfNames(sameZList.get(var)));
		}
		
		mergeSameZ();
		
		initializeVarSameZ();
		
		assignZ();
		
	}
	
	private void assignZ() {
		
		int currentLevel = 1;
		
		// process stocks first
		for (Variable v : varSameZ.keySet()) {
			if (!v.isOfType(Variables.STOCK))
				continue;
			if (v.getZ() == null) {
				MutableInteger mi = new MutableInteger(currentLevel++);
				miZList.add(mi);
				for (Variable v2 : varSameZ.get(v)) {
					if (v2.getZ() == null) {
						v2.setZ(mi);
						miZ.put(v2,  mi);
					}
				}
			}
		}
		
		// need a better assignment based on connectivity
		
		for (Variable v : varSameZ.keySet()) {
			if (v.isOfType(Variables.STOCK))
				continue;
			if (v.getZ() == null) {
				MutableInteger mi = new MutableInteger(currentLevel++);
				miZList.add(mi);
				for (Variable v2 : varSameZ.get(v)) {
					if (v2.getZ() == null) {
						v2.setZ(mi);
						miZ.put(v2,  mi);
					}
				}
			}
		}
		
		
		
		for (Variable v : variables.getVariables()) {
			
			int inMax = Integer.MIN_VALUE;
			int outMin = Integer.MAX_VALUE;
			if (miZ.containsKey(v)) {
//				System.out.println(v.getName()+" -> "+ miZ.get(v).value());
			} else {
				
				Set<Connection> ins = v.getInConnections();
				Set<Connection> outs = v.getOutConnections();
//				System.out.println(v.getName());
				for (Connection c : ins) {
					
					if (miZ.containsKey(variables.getVariable(c.getTail()))) {
//						System.out.println("   in "+c.getTail()+" -> "+ miZ.get(variables.getVariable(c.getTail())).value());
						inMax = Math.max(inMax, miZ.get(variables.getVariable(c.getTail())).value());
					}
					else {
//						System.out.println("   in "+c.getTail()+" -> ");
					}
				}
				for (Connection c : outs) {
					if (miZ.containsKey(variables.getVariable(c.getHead()))) {
//						System.out.println("  out "+c.getHead()+" -> "+ miZ.get(variables.getVariable(c.getHead())).value());
						outMin = Math.min(outMin, miZ.get(variables.getVariable(c.getHead())).value());
					} else {
//						System.out.println("  out "+c.getHead()+" -> ");
					}
				}
				
				
				// make choice
				
				int choice;
				
				if (inMax == outMin) {
					choice = inMax;
				} else if (inMax == Integer.MIN_VALUE) {
					choice = outMin;
				} else if (outMin == Integer.MAX_VALUE) {
					choice = inMax;
				} else if (outMin > inMax) {
					choice = inMax+1;
				} else if (outMin < inMax) {
					choice = outMin+1;
				} else {
					choice = outMin;
				}
				
//				System.out.println("choose between "+inMax+" and "+outMin+" choice -> "+choice);
				MutableInteger mi = new MutableInteger(choice);
				v.setZ(mi);
				miZ.put(v, mi);
			}
		}
		
	}
	
	private void initializeVarSameZ() {
		
		for (Variable var : variables.getVariables())
			varSameZ.put(var, new LinkedHashSet<Variable>());
		
		for (Variable var : sameZList.keySet()) {
			Set<Variable> set = asSet(sameZList.get(var));
			set.add(var);
			for (Variable v : set) {
				varSameZ.get(v).addAll(set);
			}
			
		}
	}
	
	private Set<Variable> asSet(List<Variable> list) {
		Set<Variable> set = new LinkedHashSet<Variable>();
		for (Variable var : list)
			set.add(var);
		
		return set;
	}
	
	private void mergeSameZ() {
		
		
		// needs to be recursive
		Set<Variable> alreadyProcessed = new LinkedHashSet<Variable>();
		Set<Variable> toRemove = new LinkedHashSet<Variable>();
		Map<Variable, List<Variable>> toAdd = new LinkedHashMap<Variable, List<Variable>>();

		for (Variable v : sameZList.keySet()) {
			toAdd.put(v, new ArrayList<Variable>());
			if (!alreadyProcessed.contains(v)) {
				alreadyProcessed.add(v);
				for (Variable v2 : sameZList.get(v)) {
					scheduleAdd(v, v2, alreadyProcessed, toRemove, toAdd);
				}
			}
		}

		for (Variable v : toAdd.keySet()) {
			sameZList.get(v).clear();
			sameZList.get(v).addAll(toAdd.get(v));
		}
		
		for (Variable v : toRemove)
			sameZList.remove(v);
	}
	
	private void scheduleAdd(Variable v, Variable v2, Set<Variable> alreadyProcessed, 
			Set<Variable> toRemove, Map<Variable, List<Variable>> toAdd) {
		if (!alreadyProcessed.contains(v2)) {
			alreadyProcessed.add(v2);
			toRemove.add(v2);
			if (!toAdd.get(v).contains(v2))
				toAdd.get(v).add(v2);
			if (sameZList.containsKey(v2)) {
				for (Variable v3 : sameZList.get(v2)) {
					scheduleAdd(v, v3, alreadyProcessed, toRemove, toAdd);	
				}
			}
		}
	}
	
	private String asListOfNames(List<Variable> l) {
		StringBuilder sb = new StringBuilder();
		
		for (Variable v : l) {
			if (sb.length() > 0)
				sb.append(", ");
			sb.append("{"+v.getName()+"}");
		}
		
		return sb.toString();
	}
	
	private void analyzeStocks(Set<Variable> stocks) {

		//	private Map<Variable, Set<Variable>> stockFlowChain = new LinkedHashMap<Variable, Set<Variable>>();

		// this is were we find the chains of stocks and flows so that all stocks linked by flows are on the same z plane
		// we will do this via connections

		for (Variable v1 : stocks) {
			if (!stockFlowChain.containsKey(v1))
				stockFlowChain.put(v1,  new LinkedHashSet<Variable>());
			Set<Connection> v1Connect = connections.getConnectionsEndingAt(v1.getName());
			for (Connection c : v1Connect) {
				if (!c.getTailVariable().isFlow())
					continue;
				stockFlowChain.get(v1).add(v1);
				stockFlowChains(c.getTailVariable(), stockFlowChain.get(v1));
			}
		}

//		// let's line up the stocks and flows in the chains
//
//		for (Variable stock : stockFlowChain.keySet()) {
//			Set<Variable> chain = stockFlowChain.get(stock);
//			Set<Location> stockLocs = dataManager.getLocationsFor(stock.getName());
//			Location stockLoc = null;
//			for (Location l : stockLocs) {
//				stockLoc = l;
//			}
//			// there should be only one
//			boolean horizontal = false;
//			// determine if this chain is horizontal
//			for (Variable v : chain) {
//				if (v.equals(stock))
//					continue;
//				Set<Connection> set = dataManager.getConnectionsBetween(v.getName(), stock.getName());
//				for (Connection c : set) {
//					if (isHorizontal(c)) {
//						// use same y
//						horizontal = true;
//					} else {
//						// use same y
//					}
//				}
//			}
//			// now line up
//			for (Variable v : chain) {
//				for (Location loc : dataManager.getLocationsFor(v.getName())) {
//					if (stockLoc.equals(loc))
//						continue;
//					if (horizontal) {
//						loc.setAdjustedY(stockLoc.getAdjustedY());
//					} else {
//						loc.setAdjustedX(stockLoc.getAdjustedX());
//					}
//				}
//			}
//		}

	}
	
	public void adjustStockFlowChains() {
		// let's line up the stocks and flows in the chains

		for (Variable stock : stockFlowChain.keySet()) {
			Set<Variable> chain = stockFlowChain.get(stock);
			Set<Location> stockLocs = dataManager.getLocationsFor(stock.getName());
			Location stockLoc = null;
			for (Location l : stockLocs) {
				stockLoc = l;
			}
			// there should be only one
			boolean horizontal = false;
			// determine if this chain is horizontal
			for (Variable v : chain) {
				if (v.equals(stock))
					continue;
				Set<Connection> set = dataManager.getConnectionsBetween(v.getName(), stock.getName());
				for (Connection c : set) {
					if (isHorizontal(c)) {
						// use same y
						horizontal = true;
					} else {
						// use same y
					}
				}
			}
			// now line up
			for (Variable v : chain) {
				for (Location loc : dataManager.getLocationsFor(v.getName())) {
					if (stockLoc.equals(loc))
						continue;
					if (horizontal) {
						loc.setAdjustedY(stockLoc.getAdjustedY());
					} else {
						loc.setAdjustedX(stockLoc.getAdjustedX());
					}
				}
			}
		}
	}
	
	public boolean isHorizontal(Connection c) {
		boolean horizontal;
		Location tLoc = dataManager.getOneLocationFor(c.getTail());
		Location hLoc = dataManager.getOneLocationFor(c.getHead());
		
		double xDiff = tLoc.getXd() - hLoc.getXd();
		double yDiff = tLoc.getYd() - hLoc.getYd();
		if (Math.abs(xDiff) > Math.abs(yDiff)) {
			horizontal = true;
		} else {
			horizontal = false;
		}
		
		return horizontal;
	}
	
	private void stockFlowChains(Variable flow, Set<Variable> chainedStocks) {
		chainedStocks.add(flow);
		Set<Connection> flowConnect = connections.getConnectionsStartingAt(flow.getName());
		for (Connection c : flowConnect) {
			if (!c.getHeadVariable().isStock())
				continue;

			if (!chainedStocks.contains(c.getHeadVariable())) {
				chainedStocks.add(c.getHeadVariable());
				Set<Connection> stockConnect = connections.getConnectionsEndingAt(c.getHeadVariable().getName());
				for (Connection c2 : stockConnect) {
					if (c2.getTailVariable().isFlow() && !c2.getTailVariable().equals(c)) {
						stockFlowChains(c2.getHeadVariable(), chainedStocks);
					}
				}
			}

		}
	}
	
	private VarToVar connectivityVariable(Variable v1) {
		if (!connectivity.containsKey(v1))
			connectivity.put(v1, new VarToVar(v1));
		return connectivity.get(v1);
	}
	
	private void analyzeVariableSet(Set<Variable> setToAnalyze, String header, Set<Variable> stocks, Set<Variable> flows, Set<Variable> auxiliaries) {
		
		int min12 = -1;
		
//		System.out.println("######### "+header+" #########");
		
		for (Variable v1 : setToAnalyze) {
			
			VarToVar v2v = connectivityVariable(v1);
			
			System.out.println("### "+v1.getName()+" ###");
			for (Variable v2 : stocks) {
				if (v1.getName().equals(v2.getName()))
					continue;
				if (reachable.isReachable1(v1.getName(), v2.getName())) {
					min12 = reachable.getDistance(v1.getName(), v2.getName());
//					System.out.println(v1.getName()+"->"+v2.getName()+" {"+min12+"} stock");
					v2v.add(v2, Integer.toString(min12));
				}
			}
			for (Variable v2 : flows) {
				if (v1.getName().equals(v2.getName()))
					continue;
				if (reachable.isReachable1(v1.getName(), v2.getName())) {
					min12 = reachable.getDistance(v1.getName(), v2.getName());				
//					System.out.println(v1.getName()+"->"+v2.getName()+" {"+min12+"} flow");
					v2v.add(v2, Integer.toString(min12));
				}
			}
			
			for (Variable v2 : auxiliaries) {
				if (v1.getName().equals(v2.getName()))
					continue;
				if (reachable.isReachable1(v1.getName(), v2.getName())) {
					min12 = reachable.getDistance(v1.getName(), v2.getName());
//					System.out.println(v1.getName()+"->"+v2.getName()+" {"+min12+"} aux");
					v2v.add(v2, Integer.toString(min12));
				}
			}
		}
		
	}	
	
	private void printVariableInfo(Set<Variable> set, String type) {
		System.out.print("########## ");
		System.out.print(type);
		System.out.println(" ##########");
		for (Variable var : set) {
			
			if (type.equals(Variables.STOCK)) {
				System.out.println("### IN FLOWS ###");
				for (Variable v : var.getInFromType(Variables.FLOW)) {
					System.out.println(v.getName()+" "+v.getType());
				}
				System.out.println("### OUT FLOWS ###");
				for (Variable v : var.getOutToType(Variables.FLOW)) {
					System.out.println(v.getName()+" "+v.getType());
				}
			}
			System.out.println(var);
		}
	}
	
	public void writeFormatted() {

		// coordinates were defined in "Vensim" space
		// now adjust for 3d

		//		public static final double SPREAD_XY = 40;  // 40, 35
		//		public static final double SPREAD_Z = 2000;  // 2000

//		for (int s_xy = 10; s_xy < 60; s_xy += 10) {
//			for (int s_z = 1000; s_z <= 2000; s_z += 100) {
//				
//				SPREAD_XY = s_xy;
//				SPREAD_Z = s_z;

				String dir = outputDirectory; // +"/"+s_xy+"_"+s_z;

				dataManager.determineAdjustments(miZ);
				dataManager.adjustVariableCoordinates();
				adjustStockFlowChains();

				connections.verifyConnectiontypes();

				variables.writeVariables(dir+"/"+"variables.csv", miZ);
				variables.writeValues(dir+"/"+"values.csv");

				// other coordinates adjusted. Now it's their turn
				dataManager.adjustConnectionCoordinates();

				connections.computeMidPoints(variables);
				// shouldn't need this any more
				//		connections.handleOverlaps();
				connections.computeRotations(variables);


				connections.writeConnections(variables, dir+"/"+"connections.tab");

//			}
//		}

	}
	
	public static void main(String[] args) {
		
		Visualization3DData v3dd = new Visualization3DData(args[0], args[1]);
		v3dd.writeFormatted();
	}

	public boolean isOkToProceed() {
		return okToProceed;
	}
	
	

}
