package gov.anl.gss.sdm.server.model.xmile;

import gov.anl.gss.sdm.server.model.support.ArrayReference;
import gov.anl.gss.sdm.server.model.translator.Equation;
import gov.anl.gss.sdm.server.model.translator.EquationComponents;
import gov.anl.gss.sdm.server.model.translator.InformationManagers;
import gov.anl.gss.sdm.server.model.translator.NativeDataTypeManager;
import gov.anl.gss.sdm.server.model.translator.Parser;
import gov.anl.gss.sdm.server.model.translator.TreeTraversal;

import java.util.ArrayList;
import java.util.Collections;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.w3c.dom.Element;
import org.w3c.dom.Node;

public abstract class BuildingBlock extends XmileObject {
	
	protected final static String STOCK = "stock";
	protected final static String AUX = "aux";
	protected final static String FLOW = "flow";
	protected final static String ARRAY = "array";
	protected final static String CONNECTOR = "connector";
	
	String modelName;
		
	// required
	protected String type = "undefined";
	protected List<Eqn> eqns = new ArrayList<Eqn>();

	// optional
	protected String access;
	protected String autoexport;
	protected String subscript;
	protected String mathML;
	protected String units;
	protected Doc doc;
	protected Set<XmileElement> elements = new LinkedHashSet<XmileElement>();
	protected List<Dimension> dimensions = new ArrayList<Dimension>();
	protected EventPoster eventposter;
	

	
	protected GraphicalFunction graphicalFunction;
	protected Points points;
	
	protected List<String> to;
	protected String of;
	
	public BuildingBlock (String modelName, Node bbNode) {
		this.modelName = modelName;
		Element e = (Element) bbNode;
		name = new Name(modelName+e.getAttribute("name"));
		
		x = XMLUtilities.getAttribute(bbNode, "x");
		y = XMLUtilities.getAttribute(bbNode, "y");
		extract(bbNode);
	}
	
	protected void extract(Node bbNode) {
		
		if (XMLUtilities.hasAttribute(bbNode, "subscript"))
			subscript = XMLUtilities.getAttribute(bbNode, "subscript");
		else
			subscript = "";
		
		access = XMLUtilities.getAttribute(bbNode, "access");
		
		autoexport = XMLUtilities.getAttribute(bbNode, "auto_export");
		
		if (XMLUtilities.hasElementsWithTag(bbNode, "dimensions")) {
			Element element = XMLUtilities.getElementWithTag((Node) bbNode, "dimensions");
			if (XMLUtilities.hasElementsWithTag(element, "dim")) {
				for (Element el : XMLUtilities.getElementsWithTag((Node) element, "dim")) {
					Dimension dim = new Dimension(el);
					if (dim.isWellFormed()) {
						dimensions.add(dim);
					} else {
						error("bad dimensions");
						
					}
				}
			}
		}
		
		if (XMLUtilities.hasElementsWithTag(bbNode, "element")) {
			for (Element element : XMLUtilities.getElementsWithTag((Node) bbNode, "element")) {
				XmileElement l = new XmileElement(modelName, element);
				if (l.isWellFormed())
					elements.add(l);
				else
					error("Bad element defintion");
			}
			
		}

		if (XMLUtilities.hasElementsWithTag(bbNode, "eqn")) {
			for (Element element : XMLUtilities.getElementsWithTag(bbNode, "eqn")) {
				Eqn r = new Eqn(modelName, element);
				if (r.isWellFormed())
					eqns.add(r);
				else
					error("bad Eqn");
			}
		} else {
			if (access == null)
			error("Missing Eqn");
		}

		if (XMLUtilities.hasElementWithTag(bbNode, "units")) {
			Element e = XMLUtilities.getElementWithTag(bbNode, "units");
			units = XMLUtilities.getNodeValue(e).replace("-", "_");
		}

		if (XMLUtilities.hasElementWithTag(bbNode, "doc")) {
			doc = new Doc(XMLUtilities.getElementWithTag(bbNode, "doc"));
		}

	}
	
	public abstract String getEquation();
	
	protected String getComment() {
		StringBuilder sb = new StringBuilder();
//		sb.append("/*\n");
//		if (doc != null)
//			sb.append(format80(doc.getComment()));
//		sb.append("*/\n\n");
		return sb.toString();
	}
	


	/**
	 * @return the access
	 */
	public String getAccess() {
		return access;
	}
	/**
	 * @return the autoexport
	 */
	public String getAutoexport() {
		return autoexport;
	}
	/**
	 * @return the subscript
	 */
	public String getSubscript() {
		return subscript;
	}
	/**
	 * @return the mathML
	 */
	public String getMathML() {
		return mathML;
	}
	/**
	 * @return the units
	 */
	public String getUnits() {
		return units;
	}
	/**
	 * @return the doc
	 */
	public Doc getDoc() {
		return doc;
	}
	/**
	 * @return the dimensions
	 */
	public List<Dimension> getDimensions() {
		return dimensions;
	}
	/**
	 * @return the eventposter
	 */
	public EventPoster getEventposter() {
		return eventposter;
	}
	/**
	 * @return the graphicalFunction
	 */
	public GraphicalFunction getGraphicalFunction() {
		return graphicalFunction;
	}
	/**
	 * @return the to
	 */
	public List<String> getTo() {
		return to;
	}
	/**
	 * @return the of
	 */
	public String getOf() {
		return of;
	}
	
	protected boolean isGraphicalFunction() {
		return graphicalFunction != null;
	}
	
	public String getFullEquation(String anEquation, Map<String, BuildingBlock> variableInfo) {
		String returnEquation = "";
		NativeDataTypeManager ndtm = InformationManagers.getInstance().getNativeDataTypeManager();
		//		for (String anEquation : getFullEquationAsList(variableInfo)) {
		boolean xlhsHasAR = false;
		boolean elhsHasAR = false;
		boolean xrhsHasAR = false;
		returnEquation = anEquation;

//		System.out.println("getFullEquation: "+anEquation);
		
//		if (anEquation.contains("Age-1"))
//			System.out.println("DEBUG Age-1");

		try {
			EquationComponents ec = new EquationComponents(anEquation);
			Equation eqn = new Equation(ec);
			elhsHasAR = eqn.isHasLHSArrayReference();

			String lhs = eqn.getOriginalLhs();
			
			if (ArrayReference.isArrayReference(lhs)) {
				lhs = ArrayReference.extractArrayName(lhs);
			}
			String lpat = InformationManagers.getInstance().getVariableNamePatternGenerator().getPatternFor(lhs);
			BuildingBlock bbl = null;

			if (variableInfo.containsKey(lpat)) {
				bbl = variableInfo.get(lpat);
				xlhsHasAR = bbl.getDimensions().size() > 0;

				//					if (xlhsHasAR != elhsHasAR)
				//						System.out.println("#######################Mismatch on lhs ar");
			} else {
				//					System.out.println("No match in variableInfo");
			}

			for (String var : eqn.getRHSVariables()) {
				String v = Parser.removeAnnotation(var);
				String vWithArray = Parser.removeAnnotation(var); 
				List<String> subs = new ArrayList<String>();
				ArrayReference ar = null;
				//					System.out.println("RHS var: "+v);
				boolean vIsAR = false;
				boolean isTerm = false;
				boolean containsWildCard = false;
				boolean containsRange = false;
				boolean containsNumericSub = false;
				boolean containsExpression = false;
				int numericSubPos = -1;
				int wildCardPos = -1;
				List<Integer> wildCardPoss = new ArrayList<Integer>();
				List<Integer> numericSubPoss = new ArrayList<Integer>();
				List<Integer> rangePoss = new ArrayList<Integer>();
				List<Integer> expressionPoss = new ArrayList<Integer>();
				int rangePos = -1;
				String range = "";
				String numericSub = "";
				if (ArrayReference.isArrayReference(v)) {
					
					// also need to check for expressions in a subscript location
					// new ranges will need to be created to handle the situation
					vIsAR = true;
					ar = new ArrayReference(v);
					v = ar.getArrayName();
					subs = ar.getSubscripts();
					
					// check for subscript arithmetic
					containsExpression = containsExpression(subs);
					if (containsExpression) {
						expressionPoss = indexOfExpression(subs);
					}
					
					isTerm = !xmileModel.isNamed(subs.get(0));
					containsWildCard = subs.contains("*");
					if (containsWildCard) {
						wildCardPoss = indexOfWildcard(subs);
					}
					
					containsRange = containsRange(subs); // subs.contains("\":\"");
					if (containsRange) {
						rangePoss = indexOfRange(subs); // .indexOf("\":\"");
//						range = subs.get(rangePos);
					}
					
					containsNumericSub = containsNumericSub(subs); // subs.contains("\":\"");
					if (containsNumericSub) {
						numericSubPoss = indexOfNumericSub(subs); // .indexOf("\":\"");
//						numericSub = subs.get(numericSubPos);
					}
					
//					System.out.println(ar+" containsWildCard = "+containsWildCard+" wildCardPos = "+wildCardPos+
//							"containsRange = "+containsRange+" range = "+range);
				}

				String pat = InformationManagers.getInstance().getVariableNamePatternGenerator().getPatternFor(v);
				BuildingBlock bb = null;
				if (variableInfo.containsKey(pat)) {
					bb = variableInfo.get(pat);
					xrhsHasAR = bb.getDimensions().size() > 0;

					// if this variable is dimensioned
					//   if lhs is dimensioned - need some kind of match between subscripts or if no match, then array function
					//   if lhs !dimensioned, then this value needs to collapse to a single value -- probably with a
					//      array function

					if (xrhsHasAR) {
						if (vIsAR) { //  && isTerm) {  if there are subscripts there, assume that they are
							if (containsWildCard) {
								List<String> subers = bb.getLhsArrayReferenceAsList();
								StringBuilder sb = new StringBuilder();
								sb.append("[");
								for (Integer i = 0; i < subers.size(); i++) {
									if (i > 0)
										sb.append(",");
									sb.append(subers.get(i));
									if (wildCardPoss.contains(i)) {
										sb.append("!");
									}
									
								}
								sb.append("]");
								
//								String goober = returnEquation.replaceAll("\\*", "!");
//								System.out.println("goober "+goober);
								
								vWithArray = bb.getNormalizedNameQuoted() + ar.getOriginalSubscript().replace("*", "\\*").replace
										("[", "\\[").replace("]", "\\]");
								String nV = bb.getNormalizedNameQuoted()+sb.toString();
//								System.out.println("nV "+nV);
								
								returnEquation = returnEquation.replaceAll(vWithArray, 
										bb.getNormalizedNameQuoted()+sb.toString()); // "["+xxx+"]");
							} else if (containsRange) {
								for (Integer i : rangePoss) {
									range = subs.get(i);
									String[] parts = range.split("\":\"");
									String sub = bb.getDimensions().get(i).getNormalizedName();
									String rangeDim = sub+"_"+parts[0]+"_"+parts[1];
//									System.out.println("rangeDim = "+rangeDim+"!");
//									System.out.println(rangeDim+":");
									StringBuilder sb = new StringBuilder();
									sb.append(rangeDim+": ");
									int count = 0;
									for (int r = Integer.parseInt(parts[0]); r <= Integer.parseInt(parts[1]); r++) {
										if (count++ > 0)
											sb.append(",");
										sb.append(sub+"_"+r);
									}
									sb.append("\t\t~\t\t~\t\t|");
									xmileModel.addGeneratedSubscriptRange(sb.toString());
									returnEquation = returnEquation.replaceAll(range, rangeDim+"!");
								}
							} else if (containsNumericSub) {
								for (Integer i : numericSubPoss) {
								String sub = bb.getDimensions().get(i).getNormalizedName();
								numericSub = subs.get(i);
								String origSub = numericSub+",";
								String newSub = sub+"_"+numericSub+",";
								
								returnEquation = returnEquation.replaceAll(origSub, newSub);
								}
							} else if (containsExpression) {
								for (Integer i : expressionPoss) {
//									System.out.println("Exp: "+subs.get(i));
									String sub = bb.getDimensions().get(i).getNormalizedName();
									StringBuilder sb = new StringBuilder();
									sb.append("\""+subs.get(i)+"\" : ");
									
									String oper = subs.get(i).replace(sub, "");
									
									Dimension d = xmileModel.getDimensionWithName(sub);
									List<String> elements = d.getElements();
									
//									sb.append(elements.get(elements.size()-1));
									sb.append("\""+sub+"_0\"");
									
									for (int e = 0; e < elements.size()-1; e++) {
										sb.append(", ");
										sb.append(elements.get(e));
									}
									
									sb.append(" -> "+sub);
									
									sb.append("\t\t~\t\t~\t\t|");
									xmileModel.addGeneratedSubscriptForExpression(sb.toString());
									returnEquation = returnEquation.replaceAll(subs.get(i), "\""+subs.get(i)+"\"");
								}
							}
								continue;
							
						}
						if (!elhsHasAR) {

							System.out.println("@@@@ adding range subscript  to function argument: "+
									v+" dim: "+bb.getDimensions().get(0).getNormalizedName());

							// first clean SUM function invocations
							int numArgsProcessed = 0;

							List<gov.anl.gss.sdm.server.model.translator.Node> fcList = TreeTraversal.getFunctionCallsInTree(
									eqn.getTreeRoot(),FunctionTranslations.arrayFunctions());
							for (gov.anl.gss.sdm.server.model.translator.Node funcCall : fcList) {
								List<gov.anl.gss.sdm.server.model.translator.Node> al = TreeTraversal.getFunctionCallArgs(funcCall, 
										FunctionTranslations.arrayFunctions());

								//								returnEquation contains quoted Variables.AUXILIARY generate expression does not quote?

								String argName = ndtm.getOriginalName(al.get(0).getToken());
								if (argName.equals(v)) {
									numArgsProcessed++;
									String exp = gov.anl.gss.sdm.server.model.translator.Node.generateExpression(funcCall).replace("(", "(\"").
											replace(")", "\")").replace("\"\"", "\"");
									String xxx = bb.getDimensions().get(0).getNormalizedName();
									String exp2 = exp.replaceAll("\\)$", "["+xxx+"!])");
									//										System.out.println("Altered Func Call: "+exp2);
									if (!returnEquation.contains(exp)) {
										System.out.println("Cannot find expression in return equation "+exp);
									}
									returnEquation = returnEquation.replace(exp, exp2);
								}
							}
							// if we didn't find a reference to this variable as arg to array function, lets put in a SUM call
							if (numArgsProcessed == 0) {
								String xxx = bb.getDimensions().get(0).getNormalizedName();
								returnEquation = returnEquation.replaceAll(bb.getNormalizedNameQuoted(), "SUM("+bb.getNormalizedNameQuoted()+"["+xxx+"!])");
							}
						} else {
							// both sides are dimensioned
//							System.out.println("@@@@ adding subscript to variable: "+
//									v+" dim: "+bb.getLhsArrayReference());  // bb.getDimensions().get(0).getNormalizedName());
							//							String xxx = bb.getDimensions().get(0).getNormalizedName();
							//							xxx = bbl.getDimensions().get(0).getNormalizedName();
							String subers = bbl.getLhsArrayReference();
							if (returnEquation.contains("\""+bb.getNormalizedName()+"\"")) {
								returnEquation = returnEquation.replaceAll(bb.getNormalizedNameQuoted(), bb.getNormalizedNameQuoted()+subers); // "["+xxx+"]");
							} else {
								returnEquation = returnEquation.replaceAll(bb.getNormalizedNameQuoted(), bb.getNormalizedNameQuoted()+subers); // "["+xxx+"]");
							}
						}

					}  // end dimensioned variable processing
				} else {
					// check if this is a reference to a subscript // mjb July 14, 2021
					// need a more sophisticated method to locate occurence in equation
					if (xmileModel.isNamed(v)) {
//						System.out.println("Subscript used in variable location "+v);
						List<String> tokens = Parser.decompose(returnEquation, true, true);
//						System.out.println("BB SUBR: "+tokens);
						List<Integer> indexOfSubrs = indexOfSubscriptNotInSubscript(tokens, v);
						
						// getSubscriptValue
						
//						for (int k = 0; k < indexOfSubrs.size(); k++) {
//							System.out.println("Subr index: "+indexOfSubrs.get(k)+" token: "+tokens.get(indexOfSubrs.get(k)));
//						}
						
						Collections.sort(indexOfSubrs);
						int iptr = 0;
						
						StringBuilder buf = new StringBuilder();
						// reconstruct equation with edited subscript reference
						int index = indexOfSubrs.get(iptr);
						for (int i = 0; i < tokens.size(); i++) {
							if (i == index) {
//								buf.append("GETSUBSCRIPTVALUE("+"\""+Parser.removeQuotes(tokens.get(i))+"\")");
								buf.append("\""+Parser.removeQuotes(tokens.get(i))+"\"");
							} else {
								buf.append(tokens.get(i));
							}
						}
						
						returnEquation = buf.toString();
						
					} else {
					System.out.println("No match in variableInfo or Subscripts: "+v);
					}
				}
			}

			if (eqn.getFunctionsCalled().size() > 0) {
				//					System.out.println("EQN with func call");
				for (gov.anl.gss.sdm.server.model.translator.Node funcCall : TreeTraversal.getAllButFunctionCallsInTree(eqn.getTreeRoot(), "SUM")) {
					//							System.out.println("Func Call: "+gov.anl.gss.sdm.server.model.translator.Node.generateExpression(funcCall));
					FunctionTranslations.processFunctions(this, eqn);
				}

			}
		} catch (Exception e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
		//		}
//		
//		if (!anEquation.equals(returnEquation)) {
//			System.out.println("Orig: "+anEquation);
//			System.out.println("New: "+returnEquation);
//		}
		return returnEquation+"\t~\t"+(units != null ? units : "")+"\t~\t"+(doc != null ? doc.getComment() : "")+"|";
	}
	
	private boolean containsNumericSub(List<String> subs){
		for (String s : subs) {
			if (Parser.isInteger(s))
				return true;
		}
		return false;
		
	}
		private List<Integer> indexOfNumericSub(List<String> subs) {
			List<Integer> al = new ArrayList<Integer>();
			for (int i = 0; i < subs.size(); i++) {
				if (Parser.isInteger(subs.get(i)))
					al.add(i);
			}
			return al;
		}

	
	private boolean containsRange(List<String> subs){
		for (String s : subs) {
			if (s.contains("\":\""))
				return true;
		}
		return false;
		
	}
	
	private boolean containsExpression(List<String> subs){

		for (String s : subs) {
			List<String> decomposed = Parser.decompose(s, false, true);
			if (decomposed.size() > 1) {
				for (String str : decomposed) {
					if (Parser.isArithmeticOperator(str))
						return true;
				}
			} else {

			}
		}
		return false;

	}
	
	private List<Integer> indexOfExpression(List<String> subs) {
		List<Integer> al = new ArrayList<Integer>();
		for (int i = 0; i < subs.size(); i++) {
			List<String> decomposed = Parser.decompose(subs.get(i), false, true);
			for (String str : decomposed) {
				if (Parser.isArithmeticOperator(str)) {
					al.add(i);
					break;
				}
			}
		}
		return al;
	}
	
		private List<Integer> indexOfRange(List<String> subs) {
			List<Integer> al = new ArrayList<Integer>();
			for (int i = 0; i < subs.size(); i++) {
				if (subs.get(i).contains("\":\""))
					al.add(i);
			}
			return al;
		}
		
		private List<Integer> indexOfSubscriptInSubscript(List<String> equationTokens, String subr) {
			// looking for location of subscript in non substrict location
			boolean subscript = false;
			List<Integer> al = new ArrayList<Integer>();
			for (int i = 0; i < equationTokens.size(); i++) {
				String token = equationTokens.get(i).trim();
				if (token.equals("]")) {
					subscript = false;
				}
				if (token.equals("[")) {
					subscript = true;
				}
				if (subscript && (token.equals(subr)|| token.equals("\""+subr+"\"")))
					al.add(i);
			}
			return al;
		}
		
		private List<Integer> indexOfSubscriptNotInSubscript(List<String> equationTokens, String subr) {
			// looking for location of subscript in non substrict location
			boolean subscript = false;
			List<Integer> al = new ArrayList<Integer>();
			for (int i = 0; i < equationTokens.size(); i++) {
				String token = equationTokens.get(i).trim();
				if (token.equals("]")) {
					subscript = false;
				}
				if (token.equals("[")) {
					subscript = true;
				}
				if (!subscript && (token.equals(subr)|| token.equals("\""+subr+"\"")))
					al.add(i);
			}
			return al;
		}
		
		private List<Integer> indexOfWildcard(List<String> subs) {
			List<Integer> al = new ArrayList<Integer>();
			for (int i = 0; i < subs.size(); i++) {
				if (subs.get(i).contains("*"))
					al.add(i);
			}
			return al;
		}

	public List<String> getFullEquationAsList(Map<String, BuildingBlock> variableInfo) {
		List<String> al = new ArrayList<String>();
		// is this a multi equation?
		int mcnt = 0;
		if (elements.size() > 0) {
			StringBuilder sb = new StringBuilder();
			for (XmileElement xe : elements) {
				//				if (isGraphicalFunction()) {
				if (xe.getGraphicalFunction() != null) {
					if (sb.length() > 0 && mcnt < elements.size()) {
						sb.append("<<MULTI>>");
					}
					mcnt++;
					//					sb.append("\t"+name.getNormalizedNameQuoted()+xe.getMultiGfRHS(name, eqns.get(0)));
					al.add(name.getNormalizedNameQuoted()+xe.getMultiGfRHS(name, eqns.get(0))+"\t~\t"+
							(units != null ? units : "")+"\t~\t"+(doc != null ? doc.getComment() : "")+"|");
					//						al.add(xe.getEquation());
					//					} else {
					////							al.add(getLookupDefinitionEquation());
					//						al.add(this.getGraphicalFunction()+getLookupDefinitionEquation()+"\t~\t"+(units != null ? units : "")+"\t~\t"+(doc != null ? doc.getComment() : "")+"|");
					//					}
				} else {
					if (xe.getEquation() != null) {
						if (sb.length() > 0  && mcnt < elements.size()) {
							sb.append("<<MULTI>>");
						}
						mcnt++;
						List<String> subs = xe.getSubscriptAsList();
						String rhs = xe.getEquation();

						List<Integer> numericSubPoss;
						int numericSubPos;
						String numericSub;

						String eqn = name.getNormalizedNameQuoted()+xe.getMultiEquationRHS()+"\t~\t"+
								(units != null ? units : "")+"\t~\t"+(doc != null ? doc.getComment() : "")+"|";

						boolean containsNumericSub = containsNumericSub(subs);
						if (containsNumericSub) {
							numericSubPoss = indexOfNumericSub(subs); // .indexOf("\":\"");
							for (Integer i : numericSubPoss) {
								numericSubPos = i;
								numericSub = subs.get(numericSubPos);
								String sub = getDimensions().get(numericSubPos).getNormalizedName();

								String origSub = numericSub+",";
								String newSub = sub+"_"+numericSub+",";
								eqn = eqn.replace(origSub, newSub);
							}
							al.add(eqn);
						} else {

							al.add(eqn);
						}



						//						sb.append("\t"+name.getNormalizedNameQuoted()+xe.getMultiEquationRHS());
					} 
				}
			}
			String sbstring = sb.toString().replace("<<MULTI>>", "\t~~|\n");
			// here we add units, comment, etc.
//			al.add(sb.toString()+"\t~\t"+(units != null ? units : "")+"\t~\t"+(doc != null ? doc.getComment() : "")+"|");

		} else {
			if (isGraphicalFunction()) {
				if (subscript == null || subscript.length() == 0) {
					if (eqns.size() > 0) {
						al.add(this.name.getNormalizedNameQuoted()+" = WITH LOOKUP("+eqns.get(0).getEquation()+", "+
								graphicalFunction.getLookupDefinitionEquation(name)+")\n\t~\t"+
								(units != null ? units : "")+"\t~\t"+(doc != null ? doc.getComment() : "")+"|");
					} else {
						al.add(this.name.getNormalizedNameQuoted()+graphicalFunction.getLookupDefinitionEquation(this.name)+"\n\t~\t"+
								(units != null ? units : "")+"\t~\t"+(doc != null ? doc.getComment() : "")+"|");
					}
				} else {
					if (eqns.size() > 0) {
											al.add(this.name.getNormalizedName()+" = WITH LOOKUP("+eqns.get(0).getEquation()+", "+
					graphicalFunction.getLookupDefinitionEquation(name)+")\n\t~\t"+
											(units != null ? units : "")+"\t~\t"+(doc != null ? doc.getComment() : "")+"|");
					} else {
						al.add(this.name.getNormalizedNameQuoted()+"["+subscript+"]"+
								graphicalFunction.getLookupDefinitionEquation(name)+"\n\t~\t"+
								(units != null ? units : "")+"\t~\t"+(doc != null ? doc.getComment() : "")+"|");
					}

				}
				//				
			} else {
//				al.add(this.getEquation()+"\t~\t"+(units != null ? units : "")+"\t~\t"+(doc != null ? doc.getComment() : "")+"|");
				String thisEqn = this.getEquation();
				thisEqn = this.getEquation();
				if (thisEqn.length() > 0)
					al.add(getFullEquation(thisEqn, variableInfo));
			}
		}

		return al;
	}
	
//	public String getLookupDefinitionEquation() {
//		
//		String lookupName = "";
//		StringBuilder sb = new StringBuilder(getComment());
//		if (isGraphicalFunction()) {
//			lookupName = name.getNormalizedName()+"_lookup";
//			sb.append(graphicalFunction.getLookup(lookupName)+")\n");
//		} 
//		return sb.toString();
//	}

	protected void print(Map<String, BuildingBlock> variableInfo) {
		
//		// required
//		private String name;
//		private List<Eqn> eqns = new ArrayList<Eqn>();
//		
//		// optional
//		private String access;
//		private String autoexport;
//		private String subscript;
//		private String mathML;
//		private Units units;
//		private Doc doc;
//		private List<Dimension> dimensions;
//		private EventPoster eventposter;
//		
//		private List<Inflow> inflow = new ArrayList<Inflow>();
//		private List<Outflow> outflow = new ArrayList<Outflow>();
//		
//		private GraphicalFunction graphicalFunction;
//		
//		private List<String> to;
//		private String of;
		
//		System.out.println("//\tBuildingBLock: Type: "+type+"\n");
		
		System.out.println("BuildingBlock: type: "+type);
		if (doc != null)
			doc.print();
		if (units != null)
			System.out.println("Units: "+units);
		for (Dimension dim : dimensions)
			dim.print();
		if (subscript != null && subscript.trim().length() > 0)
			System.out.println("Subscript: "+subscript);
		
		for (String s : getFullEquationAsList(variableInfo)) {
		System.out.println(s);
		}
	}
	
	protected List<String> asVensim(Map<String, BuildingBlock> variableInfo) {
		List<String> equations = new ArrayList<String>();

		for (String s : getFullEquationAsList(variableInfo)) {
				equations.add(s);
		}
		return equations;
	}
	
	protected List<String> getLhsArrayReferenceAsList() {
		List<String> al = new ArrayList<String>();
		if (dimensions.size() == 0) {
			return al;
		} else {
			
			int i = 0;
			for (Dimension dim : dimensions) {
				
				al.add(dim.getName());
			}
			
			return al;
		}
	}
	
	protected String getLhsArrayReference() {
		if (dimensions.size() == 0) {
			return "";
		} else {
			StringBuilder sb = new StringBuilder("[");
			int i = 0;
			for (Dimension dim : dimensions) {
				if (i++ > 0)
					sb.append(", ");
				sb.append(dim.getName());
			}
			sb.append("]");
			return sb.toString();
		}
	}
	
	protected void error(String s) {
		super.error("BuildingBlock : "+name.getAsAppears()+": "+s);
	}

	/**
	 * @return the type
	 */
	public String getType() {
		return type;
	}
	
	public boolean containsGraphicalFunction() {
		return graphicalFunction != null;
	}
	
	private boolean isNonApplyToAllArray(Element element) {
		boolean multi = false;
		if (XMLUtilities.hasElementsWithTag(element, "dimensions") &&
				XMLUtilities.hasElementsWithTag(element, "element"))
			multi = true;
		
		return multi;
	}

	public Points getPoints() {
		return points;
	}

}
