package gov.anl.gss.sdm.server.model.xmile;

import gov.anl.gss.sdm.server.model.documentor.gui.StandaloneGUI;
import gov.anl.gss.sdm.server.model.support.Utilities;
import gov.anl.gss.sdm.server.model.translator.Equation;
import gov.anl.gss.sdm.server.model.translator.EquationComponents;
import gov.anl.gss.sdm.server.model.xmile.utilities.IfThenElseConverter;
import gov.anl.gss.sdm.server.model.xmile.utilities.VensimToXmileFormatter;

import java.io.BufferedWriter;
import java.io.IOException;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Date;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

public class XmileModel extends XmileObject {

	private String version;
	private String level;
	private Header header;
	private Style style;
	private ModelUnits modelUnits;
	private SimSpecs simSpecs;
	private List<Model> models = new ArrayList<Model>();
	private Dimensions dimensions;
	
	private Map<String, List<String>> subscripts = new LinkedHashMap<String, List<String>>();

	private Document document;
	
	private boolean processSimSpecs = true;
	private boolean processModel = true;
	
	private List<String> generatedSubscriptRanges = new ArrayList<String>();
	private List<String> generatedSubscriptExpressions = new ArrayList<String>();
	
	public static int generatedViewNameID;


	public XmileModel(Document document) {
		
		Model.resetNextID();
		generatedViewNameID = 1;
		// normalize text representation
		document.getDocumentElement ().normalize ();

		Element xmileElement = null;

		NodeList nodeList = document.getElementsByTagName("xmile");
		Node aNode = (Element) nodeList.item(0);
		if (aNode.getNodeType() == Node.ELEMENT_NODE) {
			xmileElement = (Element) aNode;
		}

		version = XMLUtilities.getAttribute(xmileElement, "version");
		level = XMLUtilities.getAttribute(xmileElement, "level");


		headerTag(xmileElement);
		styleTag(xmileElement);
		if (processSimSpecs)
			simSpecsTag(xmileElement);
		if (processModel)
			modelTag(xmileElement);
		modelUnitsTag(xmileElement);
		macroTag(xmileElement);
		dataTag(xmileElement);
		behaviorTag(xmileElement);
		dimensionsTag(xmileElement);
		
		buildSubscripts();
		
		disperse(this);

	}
	
	private void buildSubscripts() {
		if (dimensions == null)
			return;
		for (Dimension d : dimensions.getDimensions()) {
			subscripts.put(d.getNormalizedName(), new ArrayList<String>());
			for (String e : d.getElements()) {
				subscripts.get(d.getNormalizedName()).add(e);
			}
		}
	}
	
	public boolean isNamed(String subscript) {
			return subscripts.containsKey(subscript);
	}
	
	public boolean isTermimal(String subscript, String terminal) {
		if (!subscripts.containsKey(subscript)) {
			return false;
		}
		return subscripts.get(subscript).contains(terminal);
	}

	private void dimensionsTag(Element xmileElement) {

		if (!XMLUtilities.hasElementsWithTag(xmileElement, "dimensions")) {
			return;
		} else {
			int elementCount = XMLUtilities.elementCountWithTag((Node) xmileElement, "dimensions");
			if (elementCount != 1) {
				error("ERROR: expecting 1 dimensions tag, found "+elementCount);
				
				return;
			}
		}

		Element element = XMLUtilities.getElementWithTag((Node) xmileElement, "dimensions");
		dimensions = new Dimensions(element);
		if (dimensions.isWellFormed()) {
		} else {
			error("bad dimensions");
			
		}

	}

	private void macroTag(Element xmileElement) {
//		int elementCount = XMLUtilities.elementCountWithTag((Node) xmileElement, "style");
//		if (elementCount == 0)
//			return;
//		if (elementCount > 1) {
//			error("ERROR: expecting 1 style tag, found "+elementCount);
//			
//			return;
//		}
//		//		}
//
//		Element styleElement = XMLUtilities.getElementWithTag((Node) xmileElement, "style");
//		style = new Style((Node) styleElement);
//		if (!style.isWellFormed()) {
//			style = null;
//			error("bad style");
//		}
	}

	private void dataTag(Element xmileElement) {

	}

	private void behaviorTag(Element xmileElement) {

	}

	private void modelUnitsTag(Element xmileElement) {
		int elementCount = XMLUtilities.elementCountWithTag((Node) xmileElement, "model_units");
		if (elementCount == 0)
			return;
		if (elementCount > 1) {
			error("ERROR: expecting 1 model_units tag, found "+elementCount);
			
			return;
		}
		//		}

		Element modelUnitsElement = XMLUtilities.getElementWithTag((Node) xmileElement, "model_units");
		modelUnits = new ModelUnits((Node) modelUnitsElement);
		if (!modelUnits.isWellFormed()) {
			modelUnits = null;
			error("bad modelUnits");
		}
		
//		 <model_units>
//         <unit name="People">
//                 <eqn/>
//                 <alias>person</alias>
//                 <alias>persons</alias>
//         </unit>
//         <unit name="Dimensionless">
//                 <eqn>1</eqn>
//                 <alias>dmnl</alias>
//                 <alias>unitless</alias>
//                 <alias>fraction</alias>
//         </unit>
//         <unit name="Kilometers">
//                 <eqn/>
//                 <alias>km</alias>
//                 <alias>kilometer</alias>
//         </unit>
//         <unit name="Years">
//                 <eqn/>
//                 <alias>yr</alias>
//                 <alias>year</alias>
//         </unit>
//         <unit name="Per Year">
//                 <eqn>1/Years</eqn>
//         </unit>
//         <unit name="lane kilometers">
//                 <eqn>lane-km</eqn>
//                 <alias>lkm</alias>
//         </unit>
//         <unit name="people per lane kilometer">
//                 <eqn>people/lkm</eqn>
//         </unit>
// </model_units>

	}
	

	private void headerTag(Element xmileElement) {
		// verify one header tag
		if (!XMLUtilities.hasElementWithTag(xmileElement, "header")) {
			int elementCount = XMLUtilities.elementCountWithTag((Node) xmileElement, "header");
			error("ERROR: expecting 1 header tag, found "+elementCount);
			
			return;
		}

		Element headerElement = XMLUtilities.getElementWithTag((Node) xmileElement, "header");
		header = new Header((Node) headerElement);
		if (!header.isWellFormed()) {
//			header = null;
			error("bad header");
		}



	}

	private void styleTag(Element xmileElement) {
		// verify 0 or 1 style tag
		//		if (!XMLUtilities.hasElementWithTag((Node) xmileElement, "style")) {
		int elementCount = XMLUtilities.elementCountWithTag((Node) xmileElement, "style");
		if (elementCount == 0)
			return;
		if (elementCount > 1) {
			error("ERROR: expecting 1 style tag, found "+elementCount);
			
			return;
		}
		//		}

		Element styleElement = XMLUtilities.getElementWithTag((Node) xmileElement, "style");
		style = new Style((Node) styleElement);
		if (!style.isWellFormed()) {
			style = null;
			error("bad style");
		}

	}

	private void simSpecsTag(Element xmileElement) {
		// verify one sim_specs tag
		
		if (!XMLUtilities.hasElementWithTag((Node) xmileElement, "sim_specs")) {
			int elementCount = XMLUtilities.elementCountWithTag((Node) xmileElement, "sim_specs");
			error("ERROR: expecting 1 sim_specs tag, found "+elementCount);
			
			return;
		}

		Element simSpecsElement = XMLUtilities.getElementWithTag((Node) xmileElement, "sim_specs");
		simSpecs = new SimSpecs((Node) simSpecsElement);
		if (!simSpecs.isWellFormed()) {
			simSpecs = null;
			error("bad simSpecs");
		}
	}

	private void modelTag(Element xmileElement) {
		// verify one or more models
		if (!XMLUtilities.hasElementsWithTag((Node) xmileElement, "model")) {
			int elementCount = XMLUtilities.elementCountWithTag((Node) xmileElement, "model");
			error("expecting 1 model tag, found "+elementCount);
			
			return;
		}
		
		List<Element> al = XMLUtilities.getElementsWithTag((Node) xmileElement, "model");
		boolean addModelName = al.size() > 1;

		for (Element element : al) {
			Model model = new Model(addModelName, element);
			if (model.isWellFormed()) {
				models.add(model);
			} else {
				error("bad model");
				
			}

		}
	}

	/**
	 * @return the version
	 */
	public String getVersion() {
		return version;
	}

	/**
	 * @return the level
	 */
	public String getLevel() {
		return level;
	}

	/**
	 * @return the header
	 */
	public Header getHeader() {
		return header;
	}

	/**
	 * @return the style
	 */
	public Style getStyle() {
		return style;
	}

	/**
	 * @return the document
	 */
	public Document getDocument() {
		return document;
	}


	/**
	 * @return the simSpecs
	 */
	public SimSpecs getSimSpecs() {
		return simSpecs;
	}

	@Override
	protected void print() {

		System.out.println("XMILE Model:");

		System.out.println("   level: "+level);
		System.out.println("   version: "+version);
		if (header != null)
			header.print();
		if (style != null)
			style.print();
		if (simSpecs != null)
			simSpecs.print();
		if (dimensions != null)
			dimensions.print();
		for (Model model : models)
			model.print();

	}
	
	public void printVensimToConsole() {
		List<String> vensim = asVensim();
		
		System.out.println("{UTF-8}\n");
		System.out.println("********************************************************\n");
		System.out.println("\t"+header.getName()+"\n");
		System.out.println("********************************************************\n");
		System.out.println("\t|");
		
		for (String s : vensim) {
			String eqn = s.replace("\n", " ");
			if (isIfThenElse(eqn)) {
				IfThenElseConverter itec = new IfThenElseConverter(eqn);
//				System.out.println(itec.getOriginalEquation());
				System.out.println(VensimToXmileFormatter.format(itec.getConvertedEquation()));
			} else {
				System.out.println(VensimToXmileFormatter.format(eqn));
			}
		}
	}
	
	public void writeVensimToFile(String file) {
		BufferedWriter bw = Utilities.openFileForWritingUTF8(file);
		List<String> vensim = asVensim();
		try {
			// "I have no value"
			
			bw.append("{UTF-8}\n");
			bw.append("********************************************************\n");
			bw.append("This model translated to Vensim MDL format by SDM-Doc Version "+StandaloneGUI.version);
			bw.append(" "+new SimpleDateFormat("yyyy.MM.dd.HH.mm.ss").format(new Date())+"\n\n");
			bw.append("\t"+header.getName()+"\n");
			bw.append("********************************************************\n");
			bw.append("\t|\n");
			
			for (String s : vensim) {
				String eqn = s.replace("\n", " ");
				
					bw.append(VensimToXmileFormatter.format(eqn)+"\n");

			}
			
			vensim = asVensimViews();
			for (String s : vensim) {
				bw.append(s+"\n");
			}
			vensim = asVensimUnitAliases();
			for (String s : vensim) {
				bw.append(s+"\n");
			}
			
			bw.close();
		} catch (IOException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
	}
	
	public boolean isIfThenElse(String s) {
		if (s.contains("IF") &&
				s.contains("THEN") &&
				s.contains("ELSE")
				) return true;
		else
			return false;
	}
	
	public Dimension getDimensionWithName(String n) {
		return xmileModel.getDimensions().getDimensionWithName(n);
	}
	
	public List<String> asVensim() {
		List<String> equations = new ArrayList<String>();
		equations.addAll(simSpecs.asVensim());

		if (xmileModel.getDimensions() != null) {
			for (Dimension dimension : xmileModel.getDimensions().getDimensions()) {
				EquationComponents ec = new EquationComponents(dimension.getEquation());
				Equation eqn = new Equation(ec);
				equations.add(eqn.getVensimEquation());
			}
		}
		

		for (Model model : models) {
			model.loadConnectors();
		}

		for (Model model : models) {
			equations.addAll(model.asVensim());
		}
		
		for (String s : generatedSubscriptRanges) {
			equations.add(s);
		}
		
		for (String s : generatedSubscriptExpressions) {
			equations.add(s);
		}
		

		return equations;
	}
	
	public void addGeneratedSubscriptRange(String s) {
		if (!generatedSubscriptRanges.contains(s))
			generatedSubscriptRanges.add(s);
	}
	
	public void addGeneratedSubscriptForExpression(String s) {
		if (!generatedSubscriptExpressions.contains(s))
			generatedSubscriptExpressions.add(s);
	}
	
	public List<String> asVensimViews() {
		List<String> equations = new ArrayList<String>();
		for (Model model : models) {
			if (model.getViews() != null) {
				equations.addAll(model.asVensimView());
			}
		}
		equations.add("///---\\\\\\");
		return equations;
	}
	
	public List<String> asVensimUnitAliases() {
		List<String> aliases = new ArrayList<String>();
		if (modelUnits != null) {
			aliases.addAll(modelUnits.asVensim());
		}
		return aliases;
	}

	/**
	 * @return the models
	 */
	public List<Model> getModels() {
		return models;
	}

	/**
	 * @return the dimensions
	 */
	public Dimensions getDimensions() {
		return dimensions;
	}
	
	public void disperse(XmileModel xmileModel) {
		setXmileModel(xmileModel);
		
		if (header != null)
			header.disperse(xmileModel);
		
		if (style != null)
			style.disperse(xmileModel);
		
		if (simSpecs != null)
			simSpecs.disperse(xmileModel);
		
		if (dimensions != null)
			dimensions.disperse(xmileModel);
		
		for (Model model : models)
			model.disperse(xmileModel);
	}
	
	public Model getTopLevelModel() {
		for (Model model : models) {
			if (model.isTopLevel())
				return model;
		}
		return null;
	}

	public Set<Module> getModules() {
		
		// modules are located in the variables section of the top level model
		Model topLevel = getTopLevelModel();
		if (topLevel != null) {
			Variables vars = topLevel.getVariables();
			if (vars != null) {
				Set<Module> mods = vars.getModules();
				return mods;
			} else {
				return null;
			}
		}
		return null;
		
	}
	
	public Connect getGhost(String moduleName, String varName) {
		Connect c = null;
		for (Module m : getModules()) {
			c = m.getConnection(moduleName, varName);
			if (c != null)
				return c;
		}
		return null;
	}

}
